
#include "cluster.h"
#include "fem2d.h"
#include "sparsematrix.h"
#include "supermatrix.h"
#include "h2conversion.h"
#include "blockcluster.h"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <assert.h>
#include <sys/times.h>
#include <time.h>

/* ************************************************************
    Demonstrate the construction of FEM matrices
   ************************************************************ */

int
main()
{
  pclustertree ct;
  pcluster root;
  pgrid2d g2;
  psparsematrix sp;
  psupermatrix M, Minv, Mwork;
  psupermatrix M2;
  double norm,zeit,eps;
  int n, i, k, nmin, maxiter,dofs;
  struct tms buffer;
  clock_t t1, t2;
  double *v, *w;
  char cholesky;
  char buf[80];

  n = 64;
  k = 5;
  eps = 1e-4;
  nmin = 32;

  (void) printf("==============================================\n");
  (void) printf("#     FEM Example                            #\n");
  (void) printf("==============================================\n");
  (void) printf("# Accuracy eps / Rank k = ?\n");
  (void) fgets(buf, 80, stdin);
  (void) sscanf(buf, "%lg", &eps);
  (void) printf("# number of DoF per axis n = ?\n");
  (void) fgets(buf, 80, stdin);
  (void) sscanf(buf, "%d", &n);  
  (void) printf("# [I]nverse or [C]holesky decomposition ?\n"); 
  (void) fgets(buf, 80, stdin);
  remove_newline(buf);
  (void) sscanf(buf, "%c", &cholesky);
  cholesky = tolower(cholesky);
  (void) printf("# Creating triangulation of the unit square\n# using a %d x %d grid\n", n+2, n+2);
  g2 = unitsquare_grid2d(n+1, n+1);
  
  if(eps > 0.999){
    /* fixed rank arithmetic with rank k */
    k = (int) (eps+1e-8);
    eps = 0.0;
    nmin = 32;
    rk_arithmetics_set_adaptive(0);
  }else{
    /* fixed accuracy arithmetic with accuracy eps */
    k = 0;
    if(cholesky=='c'){
      nmin = 32;
    }else{
      nmin = 128;
    }
    rk_arithmetics_set_adaptive(1);
  }
  
  (void) printf("# Creating cluster tree\n");
  ct = buildcluster_grid2d(g2, HLIB_GEOMETRIC, nmin);
  root = ct->root;

  (void) printf("# Building the sparse matrix\n");
  sp = laplace_grid2d(g2, ct);
  
  dofs = root->size;
  (void) printf("# Creating %d x %d supermatrix structure\n",dofs,dofs);

  if(cholesky=='i'){
    if(eps == 0.0){
      M = build_fem_supermatrix_from_cluster(root, root, 2.0, 0, eps, k, 
					     HLIB_FEMDEFAULT);
    }else{
      M = build_fem_supermatrix_from_cluster(root, root, 2.0, nmin, eps, k, 
					     HLIB_FEMDEFAULT);
    }
  }else{
    M = build_fem_supermatrix_from_cluster(root, root, 2.0, nmin, eps, k, 
					   HLIB_FEMSYMM);
  }
  if(cholesky=='i'){
    convertsparse3_supermatrix(sp, M, 0x0, nmin);
  }else{
    convertsparse3sym_supermatrix(sp, M, 0x0, nmin);
  }

  if(dofs <= 50000) {
    (void) printf("# Printing the matrix structure to 'matrix.ps'\n");
    outputsvd_supermatrix(M, "matrix.ps", 0);
  }

  Minv = NULL;
  if(cholesky=='i'){
    (void) printf("#\n# H-Matrix inversion\n");
    (void) printf("# Creating auxiliary supermatrix objects for inversion\n");
    Minv = copystructure_supermatrix(M);
    Mwork = copystructure_supermatrix(M);
    
    (void) printf("# Inverting supermatrix\n");
    t1 = times(&buffer);
    invert_supermatrix(Minv, M, Mwork);
    t2 = times(&buffer);
    zeit = (((double)t2-t1)/CLK_TCK);
    (void) printf("# Time for the inversion was %.2f seconds.\n", zeit);
    (void) printf("  The H-Matrix inverse requires %.2f KB of storage\n",
		  getsize_supermatrix(Minv) / 1024.0);
    if(dofs <= 50000) {
      (void) printf("# Printing the matrix structure to 'matrixi.ps'\n");
      outputsvd_supermatrix(Minv, "matrixi.ps", 0);
    }
    (void) printf("# Computing Inversion error\n");
    norm = norm2prodminusid2_supermatrix(Minv, sp);
    (void) printf("# Result: |Minv * M - I|_2 = %.3e\n", norm);
  }else{
    (void) printf("# Computing the Cholesky factor\n");  
    t1 = times(&buffer);
    choleskydecomposition_supermatrix(M);
    t2 = times(&buffer);
    zeit = (((double)t2-t1)/CLK_TCK);
    (void) printf("# Time for the decomposition was %.2f seconds.\n", zeit);
    (void) printf("  The H-Cholesky factor requires %.2f KB of storage\n",
		  getsize_supermatrix(M) / 1024.0);  
    if(dofs <= 50000) {
      (void) printf("# Printing the matrix structure to 'matrixc.ps'\n");
      outputsvd_supermatrix(M, "matrixc.ps", 0);
    }
    (void) printf("# Computing Inversion error\n");
    norm = norm2cholesky_sparsematrix(M,sp);  
    (void) printf("# Result: |(LL^T)^-1 * M - I|_2 = %.3e\n", norm);
  } 

  v = (double *) malloc((size_t) sizeof(double) * dofs);
  w = (double *) malloc((size_t) sizeof(double) * dofs);

  for(i=0; i<dofs; i++)
    v[i] = (double)(random()%1000)/1000.0;

  t1 = times(&buffer);
  maxiter = 1 + 65536 / dofs;
  for(i=0; i<maxiter; i++)
    if(cholesky=='i'){
      eval_supermatrix(Minv, v, w);
    }else{
      solvecholesky_supermatrix(M,v);
    }
  t2 = times(&buffer);
  zeit = (double)(t2-t1) / (CLK_TCK * maxiter);
  (void) printf("# Time (MVM) is %.4f seconds.\n", zeit);

  (void) printf("#\n# H2-Matrix approximation of the inverse.\n");
  t1 = times(&buffer);
  if(eps==0.0) eps = 1e-5;
  if(cholesky=='i'){
    M2 = buildh2_supermatrix(Minv, root, root, eps, 300, HLIB_EUCLIDEAN_RELATIVE);
  }else{
    M2 = buildh2_supermatrix(M, root, root, eps, 300, HLIB_EUCLIDEAN_RELATIVE);
  }
  t2 = times(&buffer);
  zeit = (((double)t2-t1)/CLK_TCK);
  (void) printf("# Time for the conversion is %.4f seconds.\n", zeit);
  (void) printf("  The H2-Matrix inverse required %.2f KB of storage.\n",
		(getsize_supermatrix(M2) +
		 getsize_clusterbasis(M2->row) +
		 (M2->row != M2->col ? getsize_clusterbasis(M2->col) : 0)) / 1024.0);
  
  (void) printf("# Computing H2-approximation error\n");
  if(cholesky=='i'){
    norm = norm2prodminusid2_supermatrix(M2, sp);
    (void) printf("# Result: |Minv2 * M - I|_2 = %.3e\n", norm);
  }else{
    norm = norm2diff_supermatrix(M2, M) / norm2_supermatrix(M);
    (void) printf("# Result: |L2 - L|/|L| = %.3e\n", norm);
  }
  if(dofs <= 50000) {
    (void) printf("# Printing the matrix structure to 'matrix2.ps'\n");
    outputrank_supermatrix(M2, "matrix2.ps", 1, 0);
  }
  
  t1 = times(&buffer);
  for(i=0; i<maxiter; i++)
    eval_supermatrix(M2,v,w);
  t2 = times(&buffer);
  zeit = (double) (t2-t1) / (CLK_TCK*maxiter);
  (void) printf("# Time (MVM) is %.4f seconds.\n",zeit);
  (void) printf("==============================================\n");
  
  return 0;
}
