
/* ------------------------------------------------------------
   This is the file "aca.c" of the HLib package.

   Copyright 1999 Steffen Boerm and Lars Grasedyck
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#include "aca.h"
#include "basic.h"
#include "lapack.h"

#include <math.h>

static const double ACA_EPS = 1e-16;

static void
compute_column(const double *A, const double *B, int rows, int cols, int k,
	       int j, double *col,
	       int row_off, int col_off,
	       double (*entry)(int row, int col, void *data), void *data)
{
  int i;

  assert(j >= 0);
  assert(j < cols);
  assert(entry != NULL);

  for(i=0; i<rows; i++)
    col[i] = entry(row_off+i, col_off+j, data);

  if(k > 0)
    dgemv_("Not Transposed",
	   &rows, &k,
	   meins_,
	   A, &rows,
	   B+j, &cols,
	   deins_,
	   col, eins_);
}

static void
compute_row(const double *A, const double *B, int rows, int cols, int k,
	    int i, double *row,
	    int row_off, int col_off,
	    double (*entry)(int row, int col, void *data), void *data)
{
  int j;

  assert(i >= 0);
  assert(i < rows);
  assert(entry != NULL);

  for(j=0; j<cols; j++)
    row[j] = entry(row_off+i, col_off+j, data);

  if(k > 0)
    dgemv_("Not Transposed",
	   &cols, &k,
	   meins_,
	   B, &cols,
	   A+i, &rows,
	   deins_,
	   row, eins_);
}

static void
newcompute_row(const double *A, const double *B, int rows, int cols, int k,
	       int i, double *row,
	       int row_off, int col_off,
	       double (*entry)(int row, int col, void *data), void *data,
	       int *col_done)
{
  int j;

  assert(i >= 0);
  assert(i < rows);

  for(j=0; j<cols; j++)
    if(col_done[j]==0)
      row[j] = entry(row_off+i, col_off+j, data);

  if(k > 0)
    dgemv_("Not Transposed",
	   &cols, &k,
	   meins_,
	   B, &cols,
	   A+i, &rows,
	   deins_,
	   row, eins_);
 
  for(j=0; j<cols; j++)
    if(col_done[j]!=0)
      row[j] = 0.0;
 
}

static void
newcompute_column(const double *A, const double *B, int rows, int cols, int k,
		  int j, double *col,
		  int row_off, int col_off,
		  double (*entry)(int row, int col, void *data), void *data,
		  int *row_done)
{
  int i;

  assert(j >= 0);
  assert(j < cols);

  for(i=0; i<rows; i++)
    if(row_done[i]==0)
      col[i] = entry(row_off+i, col_off+j, data);

  if(k > 0)
    dgemv_("Not Transposed",
	   &rows, &k,
	   meins_,
	   A, &rows,
	   B+j, &cols,
	   deins_,
	   col, eins_);

  for(i=0; i<rows; i++)
    if(row_done[i]!=0)
      col[i] = 0.0;
}

int
newaca_fill_block(double *A, double *B, int rows, int cols,
		  int row_off, int col_off,
		  double (*entry)(int row, int col, void *data), void *data,
		  int kmax, double eps, ACAStrategy strategy){
  
  double *col, *col_ref;
  double *row, *row_ref;
  double row_maxval, col_maxval, invmax;
  double blknorm, apxnorm, colnorm,rownorm;
  int i, itmp, j, jtmp, i_ref, j_ref;
  int stop_aca;
  int k;
  int *row_done, *col_done;

  if(kmax<1) return 0;     /* rank 0 approximation */

  row_done = (int *) malloc((size_t) sizeof(int) * rows);
  assert(row_done != NULL);

  col_done = (int *) malloc((size_t) sizeof(int) * cols);
  assert(col_done != NULL);

  k = 0;
  for(i=0; i<rows; i++) row_done[i] = 0;
  for(j=0; j<cols; j++) col_done[j] = 0;

  j_ref = cols / 2;        /* reference column */
  i_ref = rows / 2;        /* reference row */
  col_ref = A + (kmax-1)*rows;
  row_ref = B + (kmax-1)*cols;
  compute_column(A, B, rows, cols, k, j_ref, col_ref, row_off, col_off, entry, data);
  compute_row(A, B, rows, cols, k, i_ref, row_ref, row_off, col_off, entry, data);

  rownorm = dnrm2_(&cols, row_ref, eins_);
  for(i=i_ref; i!=(i_ref+rows-1)%rows && rownorm<ACA_EPS; i=(i+1)%rows){
    compute_row(A, B, rows, cols, k, i, row_ref, row_off, col_off, entry, data);
    rownorm = dnrm2_(&cols, row_ref, eins_);
    if(rownorm<ACA_EPS) row_done[i] = 1;
  }
  if(rownorm<ACA_EPS){
    free(col_done);
    col_done = NULL;
    free(row_done);
    row_done = NULL;

    return 0;
  }else{
    i_ref = i;
  }
  colnorm = dnrm2_(&rows, col_ref, eins_);
  for(j=j_ref; j!=(j_ref+cols-1)%cols && colnorm<ACA_EPS; j=(j+1)%cols){
    compute_column(A, B, rows, cols, k, j, col_ref, row_off, col_off, entry, data);
    colnorm = dnrm2_(&rows, col_ref, eins_);
    if(colnorm<ACA_EPS) col_done[j] = 1;
  }
  if(colnorm<ACA_EPS){
    free(col_done);
    col_done = NULL;
    free(row_done);
    row_done = NULL;

    return 0;
  }else{
    j_ref = j;
  }

  apxnorm = 0.0;
  stop_aca = 0;
  while(k < kmax-1) {

    col = A + k*rows;
    row = B + k*cols;
    
    i = 0; col_maxval = fabs(col_ref[0]);
    for(itmp = 1; itmp < rows; itmp++){
      if(fabs(col_ref[itmp]) > col_maxval) {
	col_maxval = fabs(col_ref[itmp]);
	i = itmp;
      }
    }
    j = 0; row_maxval = fabs(row_ref[0]);
    for(jtmp = 1; jtmp < cols; jtmp++){
      if(fabs(row_ref[jtmp]) > row_maxval) {
	row_maxval = fabs(row_ref[jtmp]);
	j = jtmp;
      }
    }

    if(row_maxval>col_maxval){
      newcompute_column(A, B, rows, cols, k, j, col, row_off, col_off, entry, data,
			row_done);
      i = 0; col_maxval = fabs(col[0]);
      for(itmp = 1; itmp < rows; itmp++){
	if(fabs(col[itmp]) > col_maxval) {
	  col_maxval = fabs(col[itmp]);
	  i = itmp;
	}
      }

      if(col_maxval < ACA_EPS) {
	stop_aca = 1;
      }else{
	newcompute_row(A, B, rows, cols, k, i, row, row_off, col_off, entry, data,
		       col_done);
	invmax = 1.0 / col[i];
	dscal_(&rows, &invmax, col, eins_);

      }
    }else{
      newcompute_row(A, B, rows, cols, k, i, row, row_off, col_off, entry, data,
		     col_done);
      j = 0; row_maxval = fabs(row[0]);
      for(jtmp = 1; jtmp < cols; jtmp++){
	if(fabs(row[jtmp]) > row_maxval) {
	  row_maxval = fabs(row[jtmp]);
	  j = jtmp;
	}
      }

      if(row_maxval < ACA_EPS) {
	stop_aca = 1;
      }else{
	newcompute_column(A, B, rows, cols, k, j, col, row_off, col_off, entry, data,
			  row_done);
	invmax = 1.0 / row[j];
	dscal_(&cols, &invmax, row, eins_);
      }
    }

    row_done[i] = 1;
    col_done[j] = 1;

    if(i!=i_ref){
      invmax = -col[i_ref];
      daxpy_(&cols,&invmax,row,eins_,row_ref,eins_);
      rownorm = dnrm2_(&cols, row_ref, eins_);
    }
    if(i==i_ref || rownorm<ACA_EPS){
      /* new reference row */
      rownorm = 0.0;
      for(i=i_ref; i!=(i_ref+rows-1)%rows && rownorm<ACA_EPS; i=(i+1)%rows){
	if(row_done[i]==0){
	  newcompute_row(A, B, rows, cols, k+1, i, row_ref, row_off, col_off, entry, data,
			 col_done);
	  rownorm = dnrm2_(&cols, row_ref, eins_);
	  if(rownorm<ACA_EPS) row_done[i] = 1;
	}else{
	  rownorm = 0.0;
	}
      }
      if(rownorm<ACA_EPS){
	stop_aca = 1;
      }else{
	i_ref = (i+rows-1)%rows;
      }
    }
    if(stop_aca==0){
      if(j!=j_ref){
	/* update reference column */
	invmax = -row[j_ref];
	daxpy_(&rows,&invmax,col,eins_,col_ref,eins_);
	colnorm = dnrm2_(&rows, col_ref, eins_);
      }
      if(j==j_ref || colnorm<ACA_EPS){
	/* new reference column */
	colnorm = 0.0;
	for(j=j_ref; j!=(j_ref+cols-1)%cols && colnorm<ACA_EPS; j=(j+1)%cols){
	  if(col_done[j]==0){
	    newcompute_column(A, B, rows, cols, k+1, j, col_ref, row_off, col_off, 
			      entry, data, row_done);
	    colnorm = dnrm2_(&rows, col_ref, eins_);
	    if(colnorm<ACA_EPS) col_done[j] = 1;
	  }else{
	    colnorm = 0.0;
	  }
	}
	if(colnorm<ACA_EPS){
	  stop_aca = 1;
	}else{
	  j_ref = (j+cols-1)%cols;
	}
      }
    }

    if(stop_aca==0){
      switch(strategy) {
      default:
      case HLIB_ACA_DEFAULT:

      case HLIB_ACA_RELATIVE2:
	blknorm = (dnrm2_(&rows, col, eins_) *
		   dnrm2_(&cols, row, eins_));
	if(k == 0) {
	  apxnorm = blknorm;
	  stop_aca = 0;
	}
	else
	  stop_aca = (blknorm < apxnorm * eps);
	break;
	
      case HLIB_ACA_ABSOLUTE2:
	blknorm = (dnrm2_(&rows, col, eins_) *
		   dnrm2_(&cols, row, eins_));
	if(k == 0) {
	  apxnorm = blknorm;
	  stop_aca = 0;
	}
	else
	  stop_aca = (blknorm < eps);
	break;
	
      case HLIB_ACA_ABSOLUTE:
	blknorm = (dnrm2_(&rows, col, eins_) *
		   dnrm2_(&cols, row, eins_));
	stop_aca = (blknorm < eps);
	break;
	
      case HLIB_ACA_RELATIVE:
	for(jtmp = 0; jtmp < k; jtmp++)
	  apxnorm += (ddot_(&rows, col, eins_, A+jtmp*rows, eins_) *
		      ddot_(&cols, row, eins_, B+jtmp*cols, eins_));
	blknorm = (ddot_(&rows, col, eins_, col, eins_) *
		   ddot_(&cols, row, eins_, row, eins_));
	stop_aca = (blknorm < apxnorm * eps * eps);
	if(!stop_aca)
	  apxnorm += blknorm;
	break;
      }
    }

    if(stop_aca) break;
    k++;
  }

  free(col_done);
  col_done = NULL;
  free(row_done);
  row_done = NULL;

  return k;
}

int
aca_fill_block(double *A, double *B, int rows, int cols,
	       int row_off, int col_off,
	       double (*entry)(int row, int col, void *data), void *data,
	       int kmax, double eps, ACAStrategy strategy)
{
  double *col;
  double *row;
  double maxval, invmax, blknorm, apxnorm;
  int i, itmp, j, jtmp;
  int tries, stop_aca;
  int k;

  k = 0;
  tries = kmax+cols;
  j = cols / 2;
  apxnorm = 0.0;
  stop_aca = 0;
  while(k < kmax && tries > 0) {
    tries--;

    col = A + k*rows;
    row = B + k*cols;

    compute_column(A, B, rows, cols, k, j, col, row_off, col_off, entry, data);

    maxval = fabs(col[0]);
    i = 0;
    for(itmp = 1; itmp < rows; itmp++)
      if(fabs(col[itmp]) > maxval) {
	maxval = fabs(col[itmp]);
	i = itmp;
      }
	
    if(maxval < ACA_EPS) {
      j = (j+1) % cols;
      continue;
    }

    compute_row(A, B, rows, cols, k, i, row, row_off, col_off, entry, data);

    invmax = 1.0 / col[i];
    dscal_(&rows, &invmax, col, eins_);

    maxval = fabs(row[0]);
    j = 0;
    for(jtmp = 1; jtmp < cols; jtmp++)
      if(fabs(row[jtmp]) > maxval) {
	maxval = fabs(row[jtmp]);
	j = jtmp;
      }
    
    switch(strategy) {
      default:
      case HLIB_ACA_DEFAULT:

      case HLIB_ACA_RELATIVE2:
	blknorm = (dnrm2_(&rows, col, eins_) *
		   dnrm2_(&cols, row, eins_));
	if(k == 0) {
	  apxnorm = blknorm;
	  stop_aca = 0;
	}
	else
	  stop_aca = (blknorm < apxnorm * eps);
	break;

      case HLIB_ACA_ABSOLUTE2:
	blknorm = (dnrm2_(&rows, col, eins_) *
		   dnrm2_(&cols, row, eins_));
	if(k == 0) {
	  apxnorm = blknorm;
	  stop_aca = 0;
	}
	else
	  stop_aca = (blknorm < eps);
	break;

      case HLIB_ACA_ABSOLUTE:
	blknorm = (dnrm2_(&rows, col, eins_) *
		   dnrm2_(&cols, row, eins_));
	stop_aca = (blknorm < eps);
	break;

      case HLIB_ACA_RELATIVE:
	for(jtmp = 0; jtmp < k; jtmp++)
	  apxnorm += (ddot_(&rows, col, eins_, A+jtmp*rows, eins_) *
		      ddot_(&cols, row, eins_, B+jtmp*cols, eins_));
	blknorm = (ddot_(&rows, col, eins_, col, eins_) *
		   ddot_(&cols, row, eins_, row, eins_));
	stop_aca = (blknorm < apxnorm * eps * eps);
	if(!stop_aca)
	  apxnorm += blknorm;
	break;
    }
    

    if(stop_aca)
      break;

    k++;
  }
  
  return k;
}

void
aca2_rkmatrix(double (*entry)(int row, int col, void *data),
	      void *data, double eps,
	      int firstrow, int firstcol,
	      prkmatrix r, ACAStrategy strategy)
{
  assert(r != NULL);

  r->kt = aca_fill_block(r->a, r->b, r->rows, r->cols, firstrow, firstcol,
			 entry, data, strategy, r->k, eps);
}

void
bufaca2_rkmatrix(double (*entry)(int row, int col, void *data),
		 void *data, double eps, int kmax,
		 int firstrow, int firstcol,
		 double **buf, int *bufsize,
		 prkmatrix r, ACAStrategy strategy)
{
  double *rowbuf;
  double *colbuf;
  int rows;
  int cols;
  int k;

  assert(r != NULL);

  rows = r->rows;
  cols = r->cols;

  provide_array((rows+cols)*kmax, buf, bufsize);
  rowbuf = *buf;
  colbuf = *buf + rows * kmax;
  
  k = newaca_fill_block(rowbuf, colbuf, rows, cols, firstrow, firstcol,
			entry, data, kmax, eps, strategy);
  
  reallocstructure_rkmatrix(r, k);
  r->kt = k;
  copy_lapack(k*r->rows, rowbuf, r->a);
  copy_lapack(k*r->cols, colbuf, r->b);
}
