
/* ------------------------------------------------------------
   This is the file "bem3d.c" of the HLib package.

   Copyright 1999 Steffen Boerm and Lars Grasedyck
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#include "bem3d.h"

#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>

#define ASSPT(p,x,y,z) p[0]=(x); p[1]=(y); p[2]=(z)

/* Create an empty grid */

pbemgrid3d
new_bemgrid3d(int vertices, int edges, int triangles)
{
  pbemgrid3d gr;

  gr = (pbemgrid3d) malloc(sizeof(bemgrid3d));
  assert(gr != NULL);

  gr->x = (double (*)[3]) malloc((size_t) sizeof(double[3]) * vertices);
  assert(gr->x != NULL);

  gr->e = (int (*)[2]) malloc((size_t) sizeof(int[2]) * edges);
  assert(gr->e != NULL);

  gr->t = (int (*)[3]) malloc((size_t) sizeof(int[3]) * triangles);
  assert(gr->t != NULL);

  gr->s = (int (*)[3]) malloc((size_t) sizeof(int[3]) * triangles);
  assert(gr->s != NULL);

  gr->gdet = (double *) malloc((size_t) sizeof(double) * triangles);
  assert(gr->gdet != NULL);

  gr->n = (double (*)[3]) malloc((size_t) sizeof(double[3]) * triangles);
  assert(gr->n != NULL);

  gr->vertices = vertices;
  gr->edges = edges;
  gr->triangles = triangles;

  return gr;
}

/* Free the memory used by a grid */

void
del_bemgrid3d(pbemgrid3d gr)
{
  assert(gr != NULL);
  assert(gr->x != NULL);
  assert(gr->e != NULL);
  assert(gr->t != NULL);
  assert(gr->s != NULL);
  assert(gr->gdet != NULL);
  assert(gr->n != NULL);

  free(gr->n);
  free(gr->gdet);
  free(gr->s);
  free(gr->t);
  free(gr->e);
  free(gr->x);
  free(gr);
}

/* Create a grid from a macro grid by splitting */

static void
compare_points(const double *x1, const double *x2,
	       int patch, const char *type, int number)
{
  if(fabs(x1[0]-x2[0]) > 1e-6 ||
     fabs(x1[1]-x2[1]) > 1e-6 ||
     fabs(x1[2]-x2[2]) > 1e-6)
    (void) fprintf(stderr,
		   "Non-matching parametrization in patch %d, %s %d:\n"
		   "(%f %f %f) != (%f %f %f)\n",
		   patch, type, number,
		   x1[0], x1[1], x1[2],
		   x2[0], x2[1], x2[2]);
}

pbemgrid3d
macro2_bemgrid3d(pmacrogrid3d mg, int split, int check_match)
{
  pbemgrid3d gr;
  triangleparametrization3d *pa = mg->pa;
  void **pd = mg->pd;
  int (*t)[3];
  int (*s)[3];
  double xtmp[3];
  int *work, *vwork, *hwork, *dwork;
  int ldwork;
  int tvertices, tedges, ttriangles;
  int vertices, edges, triangles;
  int tr;
  int n0, n1, n2;
  int e0, e1, e2;
  int flip0, flip1, flip2;
  int curr, lastv;
  double ss, tt;
  int i, j;

  assert(split > 0);

  tvertices = (mg->vertices +
	       mg->edges * (split-1) +
	       mg->triangles * (split-2) * (split-1) / 2);
  tedges = (mg->edges * split +
	    3 * mg->triangles * (split-1) * split / 2);
  ttriangles = mg->triangles * split * split;

  gr = new_bemgrid3d(tvertices, tedges, ttriangles);

  t = gr->t;
  s = gr->s;

  ldwork = split+1;

  work = (int *) malloc((size_t) sizeof(int) * ldwork * ldwork);
  assert(work != NULL);

  vwork = (int *) malloc((size_t) sizeof(int) * split * split);
  assert(vwork != NULL);

  hwork = (int *) malloc((size_t) sizeof(int) * split * split);
  assert(hwork != NULL);

  dwork = (int *) malloc((size_t) sizeof(int) * split * split);
  assert(dwork != NULL);

  for(i=0; i<mg->vertices; i++)
    mg->v[i] = -1;
  
  for(i=0; i<mg->edges; i++) {
    mg->e[i] = -1;
    mg->s[i] = -1;
  }

  vertices = 0;
  edges = 0;
  triangles = 0;

  for(tr=0; tr<mg->triangles; tr++) {
    n0 = mg->tv[tr][0];
    n1 = mg->tv[tr][1];
    n2 = mg->tv[tr][2];

    e0 = mg->te[tr][0];
    e1 = mg->te[tr][1];
    e2 = mg->te[tr][2];

    /* Create or re-use vertex 0 */
    if(mg->v[n0] < 0) {
      assert(vertices < tvertices);
      mg->v[n0] = vertices;
      pa[tr](gr->x[vertices], 0.0, 0.0, pd[tr]);
      vertices++;
    }
    else if(check_match) {
      pa[tr](xtmp, 0.0, 0.0, pd[tr]);
      compare_points(xtmp, gr->x[mg->v[n0]], tr, "vertex", n0);
    }

    /* Create or re-use vertex 1 */
    if(mg->v[n1] < 0) {
      assert(vertices < tvertices);
      mg->v[n1] = vertices;
      pa[tr](gr->x[vertices], 1.0, 0.0, pd[tr]);
      vertices++;
    }
    else if(check_match) {
      pa[tr](xtmp, 1.0, 0.0, pd[tr]);
      compare_points(xtmp, gr->x[mg->v[n1]], tr, "vertex", n1);
    }

    /* Create or re-use vertex 2*/
    if(mg->v[n2] < 0) {
      assert(vertices < tvertices);
      mg->v[n2] = vertices;
      pa[tr](gr->x[vertices], 0.0, 1.0, pd[tr]);
      vertices++;
    }
    else if(check_match) {
      pa[tr](xtmp, 0.0, 1.0, pd[tr]);
      compare_points(xtmp, gr->x[mg->v[n2]], tr, "vertex", n2);
    }

    /* Create or re-use vertices on edge 0 */
    flip0 = (mg->ev[e0][0] == n2);
    assert(flip0 ? mg->ev[e0][1] == n1 : mg->ev[e0][1] == n2);
    if(mg->e[e0] < 0) {
      mg->e[e0] = vertices;
      for(i=1; i<split; i++) {
	assert(vertices < tvertices);
	if(flip0)
	  pa[tr](gr->x[vertices],
		 (double) i / split, 1.0 - (double) i / split, pd[tr]);
	else
	  pa[tr](gr->x[vertices],
		 1.0 - (double) i / split, (double) i / split, pd[tr]);
	vertices++;
      }
      assert(vertices - mg->e[e0] == split - 1);
      
      mg->s[e0] = edges;
      lastv = (flip0 ? mg->v[n2] : mg->v[n1]);
      curr = mg->e[e0];
      for(i=1; i<split; i++) {
	assert(edges < tedges);
	gr->e[edges][0] = lastv;
	gr->e[edges][1] = curr;
	edges++;
	lastv = curr;
	curr++;
      }
      assert(edges < tedges);
      gr->e[edges][0] = lastv;
      gr->e[edges][1] = (flip0 ? mg->v[n1] : mg->v[n2]);
      edges++;
      assert(edges - mg->s[e0] == split);
    }
    else if(check_match) {
      for(i=1; i<split; i++) {
	if(flip0)
	  pa[tr](xtmp, (double) i / split, 1.0 - (double) i / split, pd[tr]);
	else
	  pa[tr](xtmp, 1.0 - (double) i / split, (double) i / split, pd[tr]);
	compare_points(xtmp, gr->x[mg->e[e0] + i-1], tr, "edge", e0);
      }
    }

    /* Create or re-use vertices on edge 1 */
    flip1 = (mg->ev[e1][0] == n0);
    assert(flip1 ? mg->ev[e1][1] == n2 : mg->ev[e1][1] == n0);
    if(mg->e[e1] < 0) {
      mg->e[e1] = vertices;
      for(i=1; i<split; i++) {
	assert(vertices < tvertices);
	if(flip1)
	  pa[tr](gr->x[vertices], 0.0, (double) i / split, pd[tr]);
	else
	  pa[tr](gr->x[vertices], 0.0, 1.0 - (double) i / split, pd[tr]);
	vertices++;
      }
      assert(vertices - mg->e[e1] == split - 1);

      mg->s[e1] = edges;
      lastv = (flip1 ? mg->v[n0] : mg->v[n2]);
      curr = mg->e[e1];
      for(i=1; i<split; i++) {
	assert(edges < tedges);
	gr->e[edges][0] = lastv;
	gr->e[edges][1] = curr;
	edges++;
	lastv = curr;
	curr++;
      }
      assert(edges < tedges);
      gr->e[edges][0] = lastv;
      gr->e[edges][1] = (flip1 ? mg->v[n2] : mg->v[n0]);
      edges++;
      assert(edges - mg->s[e1] == split);
    }
    else if(check_match) {
      for(i=1; i<split; i++) {
	if(flip1)
	  pa[tr](xtmp, 0.0, (double) i / split, pd[tr]);
	else
	  pa[tr](xtmp, 0.0, 1.0 - (double) i / split, pd[tr]);
	compare_points(xtmp, gr->x[mg->e[e1] + i-1], tr, "edge", e1);
      }
    }

    /* Create or re-use vertices on edge 2 */
    flip2 = (mg->ev[e2][0] == n1);
    assert(flip2 ? mg->ev[e2][1] == n0 : mg->ev[e2][1] == n1);
    if(mg->e[e2] < 0) {
      mg->e[e2] = vertices;
      for(i=1; i<split; i++) {
	assert(vertices < tvertices);
	if(flip2)
	  pa[tr](gr->x[vertices], 1.0 - (double) i / split, 0.0, pd[tr]);
	else
	  pa[tr](gr->x[vertices], (double) i / split, 0.0, pd[tr]);
	vertices++;
      }
      assert(vertices - mg->e[e2] == split - 1);

      mg->s[e2] = edges;
      lastv = (flip2 ? mg->v[n1] : mg->v[n0]);
      curr = mg->e[e2];
      for(i=1; i<split; i++) {
	assert(edges < tedges);
	gr->e[edges][0] = lastv;
	gr->e[edges][1] = curr;
	edges++;
	lastv = curr;
	curr++;
      }
      assert(edges < tedges);
      gr->e[edges][0] = lastv;
      gr->e[edges][1] = (flip2 ? mg->v[n0] : mg->v[n1]);
      edges++;
      assert(edges - mg->s[e2] == split);
    }
    else if(check_match) {
      for(i=1; i<split; i++) {
	if(flip2)
	  pa[tr](xtmp, 1.0 - (double) i / split, 0.0, pd[tr]);
	else
	  pa[tr](xtmp, (double) i / split, 0.0, pd[tr]);
	compare_points(xtmp, gr->x[mg->e[e2] + i-1], tr, "edge", e2);
      }
    }

    /* Put vertices into an auxiliary array */
    work[0] = mg->v[n0];
    work[split] = mg->v[n1];
    work[split*ldwork] = mg->v[n2];
    for(i=1; i<split; i++) {
      work[(split-i)+i*ldwork] = (flip0 ? mg->e[e0]+(split-1)-i : mg->e[e0]+i-1);
      work[(split-i)*ldwork] = (flip1 ? mg->e[e1]+(split-1)-i : mg->e[e1]+i-1);
      work[i] = (flip2 ? mg->e[e2]+(split-1)-i : mg->e[e2]+i-1);
    }

    /* Put edges into auxiliary arrays */
    for(i=0; i<split; i++) {
      dwork[(split-1-i)+i*split] = (flip0 ? mg->s[e0]+(split-1)-i : mg->s[e0]+i);
      vwork[(split-1-i)*split] = (flip1 ? mg->s[e1]+(split-1)-i : mg->s[e1]+i);
      hwork[i] = (flip2 ? mg->s[e2]+(split-1)-i : mg->s[e2]+i);
    }

    /* Create internal vertices */
    for(i=1; i<split; i++) {
      tt = (double) i / split;
      for(j=1; j<split-i; j++) {
	ss = (double) j / split;

	assert(vertices < tvertices);
	work[j+i*ldwork] = curr = vertices++;

	pa[tr](gr->x[curr], ss, tt, pd[tr]);
      }
    }

    /* Create internal edges */
    for(i=0; i<split-1; i++)
      for(j=0; j<split-1-i; j++) {
	assert(edges < tedges);
	dwork[j+i*split] = curr = edges++;
	gr->e[curr][1] = work[(j+1)+i*ldwork];
	gr->e[curr][0] = work[j+(i+1)*ldwork];
      }
    for(i=0; i<split-1; i++)
      for(j=1; j<split-i; j++) {
	assert(edges < tedges);
	vwork[j+i*split] = curr = edges++;
	gr->e[curr][0] = work[j+(i+1)*ldwork];
	gr->e[curr][1] = work[j+i*ldwork];
      }
    for(i=1; i<split; i++)
      for(j=0; j<split-i; j++) {
	assert(edges < tedges);
	hwork[j+i*split] = curr = edges++;
	gr->e[curr][0] = work[j+i*ldwork];
	gr->e[curr][1] = work[(j+1)+i*ldwork];
      }

    /* Create triangles */
    for(i=0; i<split; i++) {
      for(j=0; j<split-i-1; j++) {
	assert(triangles < ttriangles);
	curr = triangles++;

	t[curr][0] = work[i*ldwork+j];
	t[curr][1] = work[i*ldwork+(j+1)];
	t[curr][2] = work[(i+1)*ldwork+j];

	s[curr][0] = dwork[j+i*split];
	s[curr][1] = vwork[j+i*split];
	s[curr][2] = hwork[j+i*split];

	assert(triangles < ttriangles);
	curr = triangles++;

	t[curr][0] = work[(i+1)*ldwork+(j+1)];
	t[curr][1] = work[(i+1)*ldwork+j];
	t[curr][2] = work[i*ldwork+(j+1)];

	s[curr][0] = dwork[j+i*split];
	s[curr][1] = vwork[(j+1)+i*split];
	s[curr][2] = hwork[j+(i+1)*split];
      }
      assert(triangles < ttriangles);
      curr = triangles++;
    
      t[curr][0] = work[i*ldwork+j];
      t[curr][1] = work[i*ldwork+(j+1)];
      t[curr][2] = work[(i+1)*ldwork+j];

      s[curr][0] = dwork[j+i*split];
      s[curr][1] = vwork[j+i*split];
      s[curr][2] = hwork[j+i*split];
    }
  }

  assert(vertices == tvertices);
  assert(edges = tedges);
  assert(triangles == ttriangles);

  free(dwork);
  free(hwork);
  free(vwork);
  free(work);

  return gr;
}

/* Read a grid from a file */

static void
fetch_line(char *buf, int bufsize, int *line, FILE *in)
{
  (void) fgets(buf, bufsize, in);
  (*line)++;

  while(*buf == '#' || *buf == '\n') {
    (void) fgets(buf, bufsize, in);
    (*line)++;
  }
}

pbemgrid3d
read_bemgrid3d(const char *filename)
{
  FILE *in;
  char buf[80], *r;
  int vertices, edges, triangles;
  int oldversion;

  in = fopen(filename, "r");
  assert(in != NULL);

  r = fgets(buf, 80, in);

  oldversion = (sscanf(buf, "%d %d %d", &vertices, &edges, &triangles) == 3);

  (void) fclose(in);

  return (oldversion ? read1_bemgrid3d(filename) : read2_bemgrid3d(filename));
}

pbemgrid3d
read1_bemgrid3d(const char *filename)
{
  pbemgrid3d gr;
  FILE *in;
  char buf[80];
  double (*x)[3];
  int (*e)[2];
  int (*t)[3];
  int (*s)[3];
  int vertices, edges, triangles;
  int i, line;

  in = fopen(filename, "r");
  assert(in != NULL);

  (void) fgets(buf, 80, in);
  if(sscanf(buf, "%d %d %d", &vertices, &edges, &triangles) != 3) {
    (void) fprintf(stderr,
		   "Cannot process the first line of file \"%s\".\n",
		   filename);
    return NULL;
  }

  gr = new_bemgrid3d(vertices, edges, triangles);

  x = gr->x;
  e = gr->e;
  t = gr->t;
  s = gr->s;

  line = 1;
 
  for(i=0; i<vertices; i++) {
    x[i][0] = x[i][1] = x[i][2] = 0.0;
    (void) fgets(buf, 80, in);
    if(sscanf(buf, "%lf %lf %lf", &x[i][0], &x[i][1], &x[i][2]) != 3) {
      (void) fprintf(stderr,
		     "Line %d of file \"%s\" does not contain a vertex.\n",
		     line, filename);
    }
    line++;
  }

  for(i=0; i<edges; i++) {
    e[i][0] = e[i][1] = 0;
    (void) fgets(buf, 80, in);
    if(sscanf(buf, "%d %d", &e[i][0], &e[i][1]) != 2) {
      (void) fprintf(stderr,
		     "Line %d of file \"%s\" does not contain an edge.\n",
		     line, filename);
    }
    line++;
  }

  for(i=0; i<triangles; i++) {
    t[i][0] = t[i][1] = t[i][2] = 0;
    (void) fgets(buf, 80, in);
    if(sscanf(buf, "%d %d %d %d %d %d",
	      &t[i][0], &t[i][1], &t[i][2],
	      &s[i][0], &s[i][1], &s[i][2]) != 6) {
      (void) fprintf(stderr,
		     "Line %d of file \"%s\" does not contain a triangle.\n",
		     line, filename);
    }
    line++;
  }

  (void) fclose(in);

  return gr;
}

pbemgrid3d
read2_bemgrid3d(const char *filename)
{
  pbemgrid3d gr;
  FILE *in;
  char buf[80];
  double (*x)[3];
  int (*e)[2];
  int (*t)[3];
  int (*s)[3];
  int vertices, edges, triangles;
  double xx, yy, zz;
  int sv, ev;
  int v0, v1, v2, e0, e1, e2;
  int i, line;

  in = fopen(filename, "r");
  assert(in != NULL);
  line = 0;

  fetch_line(buf, 80, &line, in);
  if(sscanf(buf, "nv %d", &vertices) != 1) {
    (void) fprintf(stderr,
		   "Cannot read number of vertices in line %d of file \"%s\".\n",
		   line, filename);
    return NULL;
  }

  fetch_line(buf, 80, &line, in);
  if(sscanf(buf, "ne %d", &edges) != 1) {
    (void) fprintf(stderr,
		   "Cannot read number of edges in line %d of file \"%s\".\n",
		   line, filename);
    return NULL;
  }

  fetch_line(buf, 80, &line, in);
  if(sscanf(buf, "nt %d", &triangles) != 1) {
    (void) fprintf(stderr,
		   "Cannot read number of triangles in line %d of file \"%s\".\n",
		   line, filename);
    return NULL;
  }

  gr = new_bemgrid3d(vertices, edges, triangles);

  x = gr->x;
  e = gr->e;
  t = gr->t;
  s = gr->s;

  do {
    fetch_line(buf, 80, &line, in);

    switch(tolower(*buf)) {
    case 'v':
      xx = yy = zz = 0.0;
      if(sscanf(buf, "v%d x%lf y%lf z%lf", &i, &xx, &yy, &zz) != 4)
	(void) fprintf(stderr,
		       "Line %d of file \"%s\" does not contain a vertex.\n",
		       line, filename);
      else if(i < 0 || i >= vertices) 
	(void) fprintf(stderr,
		       "Vertex number in line %d of file \"%s\" is out of range.\n",
		       line, filename);
      else {
	x[i][0] = xx;
	x[i][1] = yy;
	x[i][2] = zz;
      }
      break;
    case 'e':
      sv = ev = 0;
      if(sscanf(buf, "e%d v%d v%d", &i, &sv, &ev) != 3)
	(void) fprintf(stderr,
		       "Line %d of file \"%s\" does not contain an edge.\n",
		       line, filename);
      else if(i < 0 || i >= edges)
	(void) fprintf(stderr,
		       "Edge number in line %d of file \"%s\" is out of range.\n",
		       line, filename);
      else if(sv < 0 || sv >= vertices)
	(void) fprintf(stderr,
		       "Start number in line %d of file \"%s\" is out of range.\n",
		       line, filename);
      else if(ev < 0 || ev >= vertices)
	(void) fprintf(stderr,
		       "End number in line %d of file \"%s\" is out of range.\n",
		       line, filename);
      else {
	e[i][0] = sv;
	e[i][1] = ev;
      }
      break;
    case 't':
      v0 = v1 = v2 = e0 = e1 = e2 = 0;
      if(sscanf(buf, "t%d v%d v%d v%d e%d e%d e%d", &i, &v0, &v1, &v2, &e0, &e1, &e2) != 7)
	(void) fprintf(stderr,
		       "Line %d of file \"%s\" does not contain a triangle.\n",
		       line, filename);
      else if(i < 0 || i >= triangles)
	(void) fprintf(stderr,
		       "Triangle number in line %d of file \"%s\" is out of range.\n",
		       line, filename);
      else if(v0 < 0 || v0 >= vertices)
	(void) fprintf(stderr,
		       "First vertex in line %d of file \"%s\" is out of range.\n",
		       line, filename);
      else if(v1 < 0 || v1 >= vertices)
	(void) fprintf(stderr,
		       "Second vertex in line %d of file \"%s\" is out of range.\n",
		       line, filename);
      else if(v2 < 0 || v2 >= vertices)
	(void) fprintf(stderr,
		       "Third vertex in line %d of file \"%s\" is out of range.\n",
		       line, filename);
      else if(e0 < 0 || e0 >= edges)
	(void) fprintf(stderr,
		       "First edge in line %d of file \"%s\" is out of range.\n",
		       line, filename);
      else if(e1 < 0 || e1 >= edges)
	(void) fprintf(stderr,
		       "Second edge in line %d of file \"%s\" is out of range.\n",
		       line, filename);
      else if(e2 < 0 || e2 >= edges)
	(void) fprintf(stderr,
		       "Third edge in line %d of file \"%s\" is out of range.\n",
		       line, filename);
      else {
	t[i][0] = v0;
	t[i][1] = v1;
	t[i][2] = v2;
	s[i][0] = e0;
	s[i][1] = e1;
	s[i][2] = e2;
      }
      break;
    default:
      (void) fprintf(stderr,
		     "Unknown line %d in file \"%s\".\n",
		     line, filename);
    }
  }
  while(!feof(in));

  (void) fclose(in);

  return gr;
}

/* Write a grid to a file */

void
write_bemgrid3d(pcbemgrid3d gr, const char *filename)
{
  FILE *out;
  const double (*x)[3] = (const double (*)[3]) gr->x;
  const int (*e)[2] = (const int (*)[2]) gr->e;
  const int (*t)[3] = (const int (*)[3]) gr->t;
  const int (*s)[3] = (const int (*)[3]) gr->s;
  int vertices = gr->vertices;
  int edges = gr->edges;
  int triangles = gr->triangles;
  int i, line;

  out = fopen(filename, "w");
  assert(out != NULL);

  line = 1;
  
  (void) fprintf(out, "%d %d %d\n", vertices, edges, triangles);
  line++;

  for(i=0; i<vertices; i++) {
    (void) fprintf(out, "%g %g %g\n", x[i][0], x[i][1], x[i][2]);
    line++;
  }

  for(i=0; i<edges; i++) {
    (void) fprintf(out, "%d %d\n", e[i][0], e[i][1]);
    line++;
  }

  for(i=0; i<triangles; i++) {
    (void) fprintf(out, "%d %d %d  %d %d %d\n",
		   t[i][0], t[i][1], t[i][2], s[i][0], s[i][1], s[i][2]);
    line++;
  }

  fclose(out);
}

void
write2_bemgrid3d(pcbemgrid3d gr, const char *filename)
{
  FILE *out;
  const double (*x)[3] = (const double (*)[3]) gr->x;
  const int (*e)[2] = (const int (*)[2]) gr->e;
  const int (*t)[3] = (const int (*)[3]) gr->t;
  const int (*s)[3] = (const int (*)[3]) gr->s;
  int vertices = gr->vertices;
  int edges = gr->edges;
  int triangles = gr->triangles;
  int i;

  out = fopen(filename, "w");
  assert(out != NULL);

  (void) fprintf(out,
		 "# HLib BEM3D grid file\n"
		 "# Library version: %s\n"
		 "#\n",
		 VERSION);

  (void) fprintf(out,
		 "# Number of vertices\n"
		 "nv %d\n"
		 "# Number of edges\n"
		 "ne %d\n"
		 "# Number of triangles\n"
		 "nt %d\n"
		 "#\n"
		 "# List of vertices\n"
		 "#\n",
		 vertices, edges, triangles);

  for(i=0; i<vertices; i++)
    (void) fprintf(out,
		   "v %d x %g y %g z %g\n",
		   i, x[i][0], x[i][1], x[i][2]);

  (void) fprintf(out,
		 "#\n"
		 "# List of edges\n"
		 "#\n");

  for(i=0; i<edges; i++)
    (void) fprintf(out,
		   "e %d v %d v %d\n",
		   i, e[i][0], e[i][1]);

  (void) fprintf(out,
		 "#\n"
		 "# List of triangles\n"
		 "#\n");

  for(i=0; i<triangles; i++)
    (void) fprintf(out, "t %d v %d v %d v %d  e %d e %d e %d\n",
		   i,
		   t[i][0], t[i][1], t[i][2],
		   s[i][0], s[i][1], s[i][2]);

  fclose(out);
}

/* Read a grid vector from a file */

double *
readvector_bemgrid3d(const char *filename,
		     pcbemgrid3d gr, pcclustertree ct)
{
  char buf[80], *linebuf;
  const int *iperm = ct->idx2dof;
  double *x;
  FILE *in;
  int n;
  int i, line;

  in = fopen(filename, "r");
  if(in == NULL)
    return NULL;

  line = 0;

  linebuf = fgets(buf, 80, in);
  line++;
  
  if(linebuf == NULL || sscanf(buf, "%d", &n) != 1)
    return NULL;

  if(n != gr->triangles && n != gr->vertices) {
    (void) fclose(in);

    return NULL;
  }

  x = (double *) malloc((size_t) sizeof(double) * n);
  assert(x != NULL);

  for(i=0; i<n; i++) {
    linebuf = fgets(buf, 80, in);
    line++;

    if(linebuf == NULL || sscanf(buf, "%lf", x+iperm[i]) != 1) {
      free(x);
      x = NULL;

      (void) fclose(in);

      return NULL;
    }
  }

  (void) fclose(in);

  return x;
}

/* Write a grid function to a file */

void
writevector_bemgrid3d(const char *filename,
		      int n, const double *x,
		      pcbemgrid3d gr, pcclustertree ct)
{
  FILE *out;
  const int *iperm = ct->idx2dof;
  int i;

  out = fopen(filename, "w");
  assert(out != NULL);

  (void) fprintf(out,
		 "%d\n", n);

  for(i=0; i<n; i++)
    (void) fprintf(out,
		   "%f\n",
		   x[iperm[i]]);

  (void) fclose(out);
}

/* Compute outer normals and Gram determinants for all triangles */

void
prepare_bemgrid3d(pbemgrid3d gr)
{
  const double (*x)[3] = (const double (*)[3]) gr->x;
  const int (*t)[3] = (const int (*)[3]) gr->t;
  double (*n)[3] = gr->n;
  double *gdet = gr->gdet;
  int triangles = gr->triangles;
  double xm[3][3], xd[2][3], g[2][2];
  double norm;
  int i;

  for(i=0; i<triangles; i++) {
    xm[0][0] = x[t[i][0]][0];
    xm[0][1] = x[t[i][0]][1];
    xm[0][2] = x[t[i][0]][2];

    xm[1][0] = x[t[i][1]][0];
    xm[1][1] = x[t[i][1]][1];
    xm[1][2] = x[t[i][1]][2];

    xm[2][0] = x[t[i][2]][0];
    xm[2][1] = x[t[i][2]][1];
    xm[2][2] = x[t[i][2]][2];

    xd[0][0] = xm[1][0] - xm[0][0];
    xd[0][1] = xm[1][1] - xm[0][1];
    xd[0][2] = xm[1][2] - xm[0][2];

    xd[1][0] = xm[2][0] - xm[0][0];
    xd[1][1] = xm[2][1] - xm[0][1];
    xd[1][2] = xm[2][2] - xm[0][2];

    g[0][0] = dsqr(xd[0][0]) + dsqr(xd[0][1]) + dsqr(xd[0][2]);
    g[0][1] = xd[0][0]*xd[1][0] + xd[0][1]*xd[1][1] + xd[0][2]*xd[1][2];
    g[1][0] = g[0][1];
    g[1][1] = dsqr(xd[1][0]) + dsqr(xd[1][1]) + dsqr(xd[1][2]);

    gdet[i] = sqrt(g[0][0]*g[1][1] - g[0][1]*g[1][0]);

    n[i][0] = xd[0][1]*xd[1][2] - xd[0][2]*xd[1][1];
    n[i][1] = xd[0][2]*xd[1][0] - xd[0][0]*xd[1][2];
    n[i][2] = xd[0][0]*xd[1][1] - xd[0][1]*xd[1][0];

    norm = sqrt(dsqr(n[i][0]) + dsqr(n[i][1]) + dsqr(n[i][2]));
    n[i][0] /= norm;
    n[i][1] /= norm;
    n[i][2] /= norm;
  }
}

/* Check whether the grid is closed */

int
closed_bemgrid3d(pcbemgrid3d gr)
{
  const int (*s)[3] = (const int (*)[3]) gr->s;
  int *passed;
  int edges = gr->edges;
  int triangles = gr->triangles;
  int i, j;

  passed = (int *) malloc((size_t) sizeof(int) * edges);
  assert(passed != NULL);

  for(i=0; i<edges; i++)
    passed[i] = 0;

  for(i=0; i<triangles; i++)
    for(j=0; j<3; j++)
      passed[s[i][j]]++;

  for(i=0; i<edges && passed[i] == 2; i++);

  return (i == edges);
}

/* Check whether the grid is closed and oriented */

int
oriented_bemgrid3d(pcbemgrid3d gr)
{
  const int (*e)[2] = (const int (*)[2]) gr->e;
  const int (*t)[3] = (const int (*)[3]) gr->t;
  const int (*s)[3] = (const int (*)[3]) gr->s;
  int *passed;
  int edges = gr->edges;
  int triangles = gr->triangles;
  int i, j;

  passed = (int *) malloc((size_t) sizeof(int) * edges);
  assert(passed != NULL);

  for(i=0; i<edges; i++)
    passed[i] = 0;

  for(i=0; i<triangles; i++)
    for(j=0; j<3; j++)
      if(e[s[i][j]][0] == t[i][(j+1)%3]) {
	assert(e[s[i][j]][0] == t[i][(j+1)%3]);
	assert(e[s[i][j]][1] == t[i][(j+2)%3]);
	passed[s[i][j]]++;
      }
      else {
	assert(e[s[i][j]][0] == t[i][(j+2)%3]);
	assert(e[s[i][j]][1] == t[i][(j+1)%3]);
	passed[s[i][j]]--;
      }
  
  for(i=0; i<edges && passed[i] == 0; i++);

  return (i == edges);
}

/* Create a geometric cluster tree */

pclustertree
buildcluster_bemgrid3d(pcbemgrid3d gr,
		       ClusterStrategy strategy, int leafsize)
{
  pclustertree ct;
  pclusterfactory fac;
  const double (*x)[3] = (const double (*)[3]) gr->x;
  const int (*t)[3] = (const int (*)[3]) gr->t;
  int i, v0, v1, v2;

  fac = new_clusterfactory(gr->triangles, gr->triangles, 3);

  for(i=0; i<gr->triangles; i++) {
    v0 = t[i][0];
    v1 = t[i][1];
    v2 = t[i][2];

    fac->x[i][0] = (x[v0][0] + x[v1][0] + x[v2][0]) / 3.0;
    fac->x[i][1] = (x[v0][1] + x[v1][1] + x[v2][1]) / 3.0;
    fac->x[i][2] = (x[v0][2] + x[v1][2] + x[v2][2]) / 3.0;

    fac->smin[i][0] = dmin(x[v0][0], dmin(x[v1][0], x[v2][0]));
    fac->smin[i][1] = dmin(x[v0][1], dmin(x[v1][1], x[v2][1]));
    fac->smin[i][2] = dmin(x[v0][2], dmin(x[v1][2], x[v2][2]));

    fac->smax[i][0] = dmax(x[v0][0], dmax(x[v1][0], x[v2][0]));
    fac->smax[i][1] = dmax(x[v0][1], dmax(x[v1][1], x[v2][1]));
    fac->smax[i][2] = dmax(x[v0][2], dmax(x[v1][2], x[v2][2]));
  }

  ct = create_clustertree(fac, strategy, leafsize);

  del_clusterfactory(fac);

  return ct;
}

/* Create a geometric cluster tree (vector valued),
   nrdof[i] = number of entries for triangle i */

pclustertree
buildcluster_vec_bemgrid3d(pcbemgrid3d gr,
			   ClusterStrategy strategy, int leafsize,
			   int *nrdof)
{
  pclustertree ct;
  pclusterfactory fac;
  const double (*x)[3] = (const double (*)[3]) gr->x;
  const int (*t)[3] = (const int (*)[3]) gr->t;
  int i, j, k, v0, v1, v2;
  int total;
  
  total = 0;
  for(i=0; i<gr->triangles; i++)
    total += nrdof[i];
  
  fac = new_clusterfactory(total, total, 3);
  
  k = 0;
  for(i=0; i<gr->triangles; i++) {
    v0 = t[i][0];
    v1 = t[i][1];
    v2 = t[i][2];
      
    fac->x[k][0] = (x[v0][0] + x[v1][0] + x[v2][0]) / 3.0;
    fac->x[k][1] = (x[v0][1] + x[v1][1] + x[v2][1]) / 3.0;
    fac->x[k][2] = (x[v0][2] + x[v1][2] + x[v2][2]) / 3.0;
    
    fac->smin[k][0] = dmin(x[v0][0], dmin(x[v1][0], x[v2][0]));
    fac->smin[k][1] = dmin(x[v0][1], dmin(x[v1][1], x[v2][1]));
    fac->smin[k][2] = dmin(x[v0][2], dmin(x[v1][2], x[v2][2]));
    
    fac->smax[k][0] = dmax(x[v0][0], dmax(x[v1][0], x[v2][0]));
    fac->smax[k][1] = dmax(x[v0][1], dmax(x[v1][1], x[v2][1]));
    fac->smax[k][2] = dmax(x[v0][2], dmax(x[v1][2], x[v2][2]));

    k++;

    for(j=1; j<nrdof[i]; j++){
      fac->x[k][0] = fac->x[k-1][0];
      fac->x[k][1] = fac->x[k-1][1];
      fac->x[k][2] = fac->x[k-1][2];
      
      fac->smin[k][0] = fac->smin[k-1][0];
      fac->smin[k][1] = fac->smin[k-1][1];
      fac->smin[k][2] = fac->smin[k-1][2];
      
      fac->smax[k][0] = fac->smax[k-1][0];
      fac->smax[k][1] = fac->smax[k-1][1];
      fac->smax[k][2] = fac->smax[k-1][2];

      k++;
    }
  }

  ct = create_clustertree(fac, strategy, leafsize);

  del_clusterfactory(fac);

  return ct;
}

pclustertree
buildcluster_bemgrid3d_vertices(pcbemgrid3d gr,
				ClusterStrategy strategy, int leafsize)
{
  pclustertree ct;
  pclusterfactory fac;
  const double (*x)[3] = (const double (*)[3]) gr->x;
  const int (*t)[3] = (const int (*)[3]) gr->t;
  double xmin[3], xmax[3];
  int i, j, v0, v1, v2;

  fac = new_clusterfactory(gr->vertices, gr->vertices, 3);

  for(i=0; i<gr->vertices; i++) {
    fac->x[i][0] = fac->smin[i][0] = fac->smax[i][0] = x[i][0];
    fac->x[i][1] = fac->smin[i][1] = fac->smax[i][1] = x[i][1];
    fac->x[i][2] = fac->smin[i][2] = fac->smax[i][2] = x[i][2];
  }

  for(i=0; i<gr->triangles; i++) {
    v0 = t[i][0];
    v1 = t[i][1];
    v2 = t[i][2];

    xmin[0] = dmin(x[v0][0], dmin(x[v1][0], x[v2][0]));
    xmin[1] = dmin(x[v0][1], dmin(x[v1][1], x[v2][1]));
    xmin[2] = dmin(x[v0][2], dmin(x[v1][2], x[v2][2]));

    xmax[0] = dmax(x[v0][0], dmax(x[v1][0], x[v2][0]));
    xmax[1] = dmax(x[v0][1], dmax(x[v1][1], x[v2][1]));
    xmax[2] = dmax(x[v0][2], dmax(x[v1][2], x[v2][2]));

    for(j=0; j<3; j++) {
      v0 = t[i][j];
      assert(0 <= v0 && v0 < gr->vertices);

      fac->smin[v0][0] = dmin(fac->smin[v0][0], xmin[0]);
      fac->smin[v0][1] = dmin(fac->smin[v0][1], xmin[1]);
      fac->smin[v0][2] = dmin(fac->smin[v0][2], xmin[2]);

      fac->smax[v0][0] = dmax(fac->smax[v0][0], xmax[0]);
      fac->smax[v0][1] = dmax(fac->smax[v0][1], xmax[1]);
      fac->smax[v0][2] = dmax(fac->smax[v0][2], xmax[2]);
    }
  }

  ct = create_clustertree(fac, strategy, leafsize);

  del_clusterfactory(fac);

  return ct;
}

/* Create an empty macro grid */

pmacrogrid3d
new_macrogrid3d(int vertices, int edges, int triangles)
{
  pmacrogrid3d mg;
  int i;

  mg = (pmacrogrid3d) malloc(sizeof(macrogrid3d));
  assert(mg);

  mg->v = (int *) malloc((size_t) sizeof(int) * vertices);
  assert(mg->v != NULL);

  mg->e = (int *) malloc((size_t) sizeof(int) * edges);
  assert(mg->e != NULL);

  mg->s = (int *) malloc((size_t) sizeof(int) * edges);
  assert(mg->s != NULL);

  mg->ev = (int (*)[2]) malloc((size_t) sizeof(int [2]) * edges);
  assert(mg->ev != NULL);

  mg->tv = (int (*)[3]) malloc((size_t) sizeof(int [3]) * triangles);
  assert(mg->tv != NULL);

  mg->te = (int (*)[3]) malloc((size_t) sizeof(int [3]) * triangles);
  assert(mg->te != NULL);

  mg->pa = (triangleparametrization3d *) malloc((size_t) sizeof(triangleparametrization3d) * triangles);
  assert(mg->pa != NULL);

  mg->pd = (void **) malloc((size_t) sizeof(void *) * triangles);
  assert(mg->pd != NULL);

  for(i=0; i<triangles; i++)
    mg->pd[i] = NULL;

  mg->vertices = vertices;
  mg->edges = edges;
  mg->triangles = triangles;

  return mg;
}

/* Free the memory used by a macro grid */

void
del_macrogrid3d(pmacrogrid3d mg)
{
  int i;

  assert(mg != NULL);
  assert(mg->v != NULL);
  assert(mg->e != NULL);
  assert(mg->s != NULL);
  assert(mg->ev != NULL);
  assert(mg->tv != NULL);
  assert(mg->te != NULL);
  assert(mg->pa != NULL);
  assert(mg->pd != NULL);

  for(i=0; i<mg->triangles; i++)
    if(mg->pd[i] != NULL)
      free(mg->pd[i]);

  free(mg->pd);
  free(mg->pa);
  free(mg->te);
  free(mg->tv);
  free(mg->ev);
  free(mg->s);
  free(mg->e);
  free(mg->v);
  free(mg);
}

/* Macro grid for an ellipsoid */

typedef struct {
  double rx, ry, rz;
  double p[3][3];
} ellipsoid_triangle_data;

static void
ellipsoid_triangle_parametrization(double *x, double s, double t, void *data)
{
  ellipsoid_triangle_data *etd = (ellipsoid_triangle_data *) data;
  double n;

  x[0] = (1.0-s-t)*etd->p[0][0] + s*etd->p[1][0] + t*etd->p[2][0];
  x[1] = (1.0-s-t)*etd->p[0][1] + s*etd->p[1][1] + t*etd->p[2][1];
  x[2] = (1.0-s-t)*etd->p[0][2] + s*etd->p[1][2] + t*etd->p[2][2];

  n = 1.0 / sqrt(dsqr(x[0]) + dsqr(x[1]) + dsqr(x[2]));

  x[0] *= etd->rx*n;
  x[1] *= etd->ry*n;
  x[2] *= etd->rz*n;
}

pmacrogrid3d
ellipsoid_macrogrid3d(double rx, double ry, double rz)
{
  pmacrogrid3d mg;
  ellipsoid_triangle_data *etd;

  mg = new_macrogrid3d(6, 12, 8);

  /* Equator */

  mg->ev[0][0] = 1; mg->ev[0][1] = 2;
  mg->ev[1][0] = 2; mg->ev[1][1] = 3;
  mg->ev[2][0] = 3; mg->ev[2][1] = 4;
  mg->ev[3][0] = 4; mg->ev[3][1] = 1;

  /* Edges to the north pole */

  mg->ev[4][0] = 0; mg->ev[4][1] = 1;
  mg->ev[5][0] = 0; mg->ev[5][1] = 2;
  mg->ev[6][0] = 0; mg->ev[6][1] = 3;
  mg->ev[7][0] = 0; mg->ev[7][1] = 4;

  /* Edges to the south pole */

  mg->ev[8][0] = 5; mg->ev[8][1] = 1;
  mg->ev[9][0] = 5; mg->ev[9][1] = 4;
  mg->ev[10][0] = 5; mg->ev[10][1] = 3;
  mg->ev[11][0] = 5; mg->ev[11][1] = 2;

  /* Northwestern frontal triangle */

  mg->tv[0][0] = 0; mg->tv[0][1] = 1; mg->tv[0][2] = 2;
  mg->te[0][0] = 0; mg->te[0][1] = 5; mg->te[0][2] = 4;
  mg->pa[0] = ellipsoid_triangle_parametrization;
  mg->pd[0] = etd = (ellipsoid_triangle_data *) malloc(sizeof(ellipsoid_triangle_data));
  etd->rx = rx;
  etd->ry = ry;
  etd->rz = rz;
  ASSPT(etd->p[0], 0.0, 0.0, 1.0);
  ASSPT(etd->p[1], 1.0, 0.0, 0.0);
  ASSPT(etd->p[2], 0.0, 1.0, 0.0);

  /* Northeastern frontal triangle */

  mg->tv[1][0] = 0; mg->tv[1][1] = 2; mg->tv[1][2] = 3;
  mg->te[1][0] = 1; mg->te[1][1] = 6; mg->te[1][2] = 5;
  mg->pa[1] = ellipsoid_triangle_parametrization;
  mg->pd[1] = etd = (ellipsoid_triangle_data *) malloc(sizeof(ellipsoid_triangle_data));
  etd->rx = rx;
  etd->ry = ry;
  etd->rz = rz;
  ASSPT(etd->p[0], 0.0, 0.0, 1.0);
  ASSPT(etd->p[1], 0.0, 1.0, 0.0);
  ASSPT(etd->p[2], -1.0, 0.0, 0.0);

  /* Northwestern rear triangle */

  mg->tv[2][0] = 0; mg->tv[2][1] = 3; mg->tv[2][2] = 4;
  mg->te[2][0] = 2; mg->te[2][1] = 7; mg->te[2][2] = 6;
  mg->pa[2] = ellipsoid_triangle_parametrization;
  mg->pd[2] = etd = (ellipsoid_triangle_data *) malloc(sizeof(ellipsoid_triangle_data));
  etd->rx = rx;
  etd->ry = ry;
  etd->rz = rz;
  ASSPT(etd->p[0], 0.0, 0.0, 1.0);
  ASSPT(etd->p[1], -1.0, 0.0, 0.0);
  ASSPT(etd->p[2], 0.0, -1.0, 0.0);

  /* Northeastern rear triangle */

  mg->tv[3][0] = 0; mg->tv[3][1] = 4; mg->tv[3][2] = 1;
  mg->te[3][0] = 3; mg->te[3][1] = 4; mg->te[3][2] = 7;
  mg->pa[3] = ellipsoid_triangle_parametrization;
  mg->pd[3] = etd = (ellipsoid_triangle_data *) malloc(sizeof(ellipsoid_triangle_data));
  etd->rx = rx;
  etd->ry = ry;
  etd->rz = rz;
  ASSPT(etd->p[0], 0.0, 0.0, 1.0);
  ASSPT(etd->p[1], 0.0, -1.0, 0.0);
  ASSPT(etd->p[2], 1.0, 0.0, 0.0);

  /* Southwestern frontal triangle */

  mg->tv[4][0] = 5; mg->tv[4][1] = 1; mg->tv[4][2] = 4;
  mg->te[4][0] = 3; mg->te[4][1] = 9; mg->te[4][2] = 8;
  mg->pa[4] = ellipsoid_triangle_parametrization;
  mg->pd[4] = etd = (ellipsoid_triangle_data *) malloc(sizeof(ellipsoid_triangle_data));
  etd->rx = rx;
  etd->ry = ry;
  etd->rz = rz;
  ASSPT(etd->p[0], 0.0, 0.0, -1.0);
  ASSPT(etd->p[1], 1.0, 0.0, 0.0);
  ASSPT(etd->p[2], 0.0, -1.0, 0.0);

  /* Southeastern frontal triangle */

  mg->tv[5][0] = 5; mg->tv[5][1] = 4; mg->tv[5][2] = 3;
  mg->te[5][0] = 2; mg->te[5][1] = 10; mg->te[5][2] = 9;
  mg->pa[5] = ellipsoid_triangle_parametrization;
  mg->pd[5] = etd = (ellipsoid_triangle_data *) malloc(sizeof(ellipsoid_triangle_data));
  etd->rx = rx;
  etd->ry = ry;
  etd->rz = rz;
  ASSPT(etd->p[0], 0.0, 0.0, -1.0);
  ASSPT(etd->p[1], 0.0, -1.0, 0.0);
  ASSPT(etd->p[2], -1.0, 0.0, 0.0);

  /* Southwestern rear triangle */

  mg->tv[6][0] = 5; mg->tv[6][1] = 3; mg->tv[6][2] = 2;
  mg->te[6][0] = 1; mg->te[6][1] = 11; mg->te[6][2] = 10;
  mg->pa[6] = ellipsoid_triangle_parametrization;
  mg->pd[6] = etd = (ellipsoid_triangle_data *) malloc(sizeof(ellipsoid_triangle_data));
  etd->rx = rx;
  etd->ry = ry;
  etd->rz = rz;
  ASSPT(etd->p[0], 0.0, 0.0, -1.0);
  ASSPT(etd->p[1], -1.0, 0.0, 0.0);
  ASSPT(etd->p[2], 0.0, 1.0, 0.0);

  /* Southeastern rear triangle */

  mg->tv[7][0] = 5; mg->tv[7][1] = 2; mg->tv[7][2] = 1;
  mg->te[7][0] = 0; mg->te[7][1] = 8; mg->te[7][2] = 11;
  mg->pa[7] = ellipsoid_triangle_parametrization;
  mg->pd[7] = etd = (ellipsoid_triangle_data *) malloc(sizeof(ellipsoid_triangle_data));
  etd->rx = rx;
  etd->ry = ry;
  etd->rz = rz;
  ASSPT(etd->p[0], 0.0, 0.0, -1.0);
  ASSPT(etd->p[1], 0.0, 1.0, 0.0);
  ASSPT(etd->p[2], 1.0, 0.0, 0.0);

  mg->vertices = 6;
  mg->edges = 12;
  mg->triangles = 8;

  return mg;
}

/* Macro grid for a box */

typedef struct {
  double p[3][3];
} box_triangle_data;

static void
box_triangle_parametrization(double *x, double s, double t, void *data)
{
  box_triangle_data *btd = (box_triangle_data *) data;

  x[0] = (1.0-s-t)*btd->p[0][0] + s*btd->p[1][0] + t*btd->p[2][0];
  x[1] = (1.0-s-t)*btd->p[0][1] + s*btd->p[1][1] + t*btd->p[2][1];
  x[2] = (1.0-s-t)*btd->p[0][2] + s*btd->p[1][2] + t*btd->p[2][2];
}

pmacrogrid3d
box_macrogrid3d(double dx, double dy, double dz)
{
  pmacrogrid3d mg;
  box_triangle_data *btd;

  mg = new_macrogrid3d(8, 18, 12);

  /* Front square */

  mg->ev[0][0] = 0; mg->ev[0][1] = 1;
  mg->ev[1][0] = 1; mg->ev[1][1] = 2;
  mg->ev[2][0] = 2; mg->ev[2][1] = 3;
  mg->ev[3][0] = 3; mg->ev[3][1] = 0;

  /* Back square */

  mg->ev[4][0] = 4; mg->ev[4][1] = 5;
  mg->ev[5][0] = 5; mg->ev[5][1] = 6;
  mg->ev[6][0] = 6; mg->ev[6][1] = 7;
  mg->ev[7][0] = 7; mg->ev[7][1] = 4;

  /* Front to back */

  mg->ev[8][0] = 0; mg->ev[8][1] = 5;
  mg->ev[9][0] = 1; mg->ev[9][1] = 4;
  mg->ev[10][0] = 2; mg->ev[10][1] = 7;
  mg->ev[11][0] = 3; mg->ev[11][1] = 6;

  /* Diagonals */

  mg->ev[12][0] = 0; mg->ev[12][1] = 2;
  mg->ev[13][0] = 1; mg->ev[13][1] = 7;
  mg->ev[14][0] = 4; mg->ev[14][1] = 6;
  mg->ev[15][0] = 5; mg->ev[15][1] = 3;
  mg->ev[16][0] = 0; mg->ev[16][1] = 4;
  mg->ev[17][0] = 7; mg->ev[17][1] = 3;

  /* Lower front */

  mg->tv[0][0] = 0; mg->tv[0][1] = 1; mg->tv[0][2] = 2;
  mg->te[0][0] = 1; mg->te[0][1] = 12; mg->te[0][2] = 0;
  mg->pa[0] = box_triangle_parametrization;
  mg->pd[0] = btd = (box_triangle_data *) malloc(sizeof(box_triangle_data));
  ASSPT(btd->p[0], -dx, -dy, dz);
  ASSPT(btd->p[1], dx, -dy, dz);
  ASSPT(btd->p[2], dx, dy, dz);

  /* Upper front */

  mg->tv[1][0] = 2; mg->tv[1][1] = 3; mg->tv[1][2] = 0;
  mg->te[1][0] = 3; mg->te[1][1] = 12; mg->te[1][2] = 2;
  mg->pa[1] = box_triangle_parametrization;
  mg->pd[1] = btd = (box_triangle_data *) malloc(sizeof(box_triangle_data));
  ASSPT(btd->p[0], dx, dy, dz);
  ASSPT(btd->p[1], -dx, dy, dz);
  ASSPT(btd->p[2], -dx, -dy, dz);

  /* Lower right */

  mg->tv[2][0] = 1; mg->tv[2][1] = 4; mg->tv[2][2] = 7;
  mg->te[2][0] = 7; mg->te[2][1] = 13; mg->te[2][2] = 9;
  mg->pa[2] = box_triangle_parametrization;
  mg->pd[2] = btd = (box_triangle_data *) malloc(sizeof(box_triangle_data));
  ASSPT(btd->p[0], dx, -dy, dz);
  ASSPT(btd->p[1], dx, -dy, -dz);
  ASSPT(btd->p[2], dx, dy, -dz);

  /* Upper right */

  mg->tv[3][0] = 1; mg->tv[3][1] = 7; mg->tv[3][2] = 2;
  mg->te[3][0] = 10; mg->te[3][1] = 1; mg->te[3][2] = 13;
  mg->pa[3] = box_triangle_parametrization;
  mg->pd[3] = btd = (box_triangle_data *) malloc(sizeof(box_triangle_data));
  ASSPT(btd->p[0], dx, -dy, dz);
  ASSPT(btd->p[1], dx, dy, -dz);
  ASSPT(btd->p[2], dx, dy, dz);

  /* Lower back */

  mg->tv[4][0] = 5; mg->tv[4][1] = 6; mg->tv[4][2] = 4;
  mg->te[4][0] = 14; mg->te[4][1] = 4; mg->te[4][2] = 5;
  mg->pa[4] = box_triangle_parametrization;
  mg->pd[4] = btd = (box_triangle_data *) malloc(sizeof(box_triangle_data));
  ASSPT(btd->p[0], -dx, -dy, -dz);
  ASSPT(btd->p[1], -dx, dy, -dz);
  ASSPT(btd->p[2], dx, -dy, -dz);

  /* Upper back */

  mg->tv[5][0] = 4; mg->tv[5][1] = 6; mg->tv[5][2] = 7;
  mg->te[5][0] = 6; mg->te[5][1] = 7; mg->te[5][2] = 14;
  mg->pa[5] = box_triangle_parametrization;
  mg->pd[5] = btd = (box_triangle_data *) malloc(sizeof(box_triangle_data));
  ASSPT(btd->p[0], dx, -dy, -dz);
  ASSPT(btd->p[1], -dx, dy, -dz);
  ASSPT(btd->p[2], dx, dy, -dz);

  /* Lower left */

  mg->tv[6][0] = 0; mg->tv[6][1] = 3; mg->tv[6][2] = 5;
  mg->te[6][0] = 15; mg->te[6][1] = 8; mg->te[6][2] = 3;
  mg->pa[6] = box_triangle_parametrization;
  mg->pd[6] = btd = (box_triangle_data *) malloc(sizeof(box_triangle_data));
  ASSPT(btd->p[0], -dx, -dy, dz);
  ASSPT(btd->p[1], -dx, dy, dz);
  ASSPT(btd->p[2], -dx, -dy, -dz);

  /* Upper left */

  mg->tv[7][0] = 5; mg->tv[7][1] = 3; mg->tv[7][2] = 6;
  mg->te[7][0] = 11; mg->te[7][1] = 5; mg->te[7][2] = 15;
  mg->pa[7] = box_triangle_parametrization;
  mg->pd[7] = btd = (box_triangle_data *) malloc(sizeof(box_triangle_data));
  ASSPT(btd->p[0], -dx, -dy, -dz);
  ASSPT(btd->p[1], -dx, dy, dz);
  ASSPT(btd->p[2], -dx, dy, -dz);

  /* Lower top */

  mg->tv[8][0] = 3; mg->tv[8][1] = 2; mg->tv[8][2] = 7;
  mg->te[8][0] = 10; mg->te[8][1] = 17; mg->te[8][2] = 2;
  mg->pa[8] = box_triangle_parametrization;
  mg->pd[8] = btd = (box_triangle_data *) malloc(sizeof(box_triangle_data));
  ASSPT(btd->p[0], -dx, dy, dz);
  ASSPT(btd->p[1], dx, dy, dz);
  ASSPT(btd->p[2], dx, dy, -dz);

  /* Upper top */

  mg->tv[9][0] = 3; mg->tv[9][1] = 7; mg->tv[9][2] = 6;
  mg->te[9][0] = 6; mg->te[9][1] = 11; mg->te[9][2] = 17;
  mg->pa[9] = box_triangle_parametrization;
  mg->pd[9] = btd = (box_triangle_data *) malloc(sizeof(box_triangle_data));
  ASSPT(btd->p[0], -dx, dy, dz);
  ASSPT(btd->p[1], dx, dy, -dz);
  ASSPT(btd->p[2], -dx, dy, -dz);

  /* Lower bottom */

  mg->tv[10][0] = 0; mg->tv[10][1] = 4; mg->tv[10][2] = 1;
  mg->te[10][0] = 9; mg->te[10][1] = 0; mg->te[10][2] = 16;
  mg->pa[10] = box_triangle_parametrization;
  mg->pd[10] = btd = (box_triangle_data *) malloc(sizeof(box_triangle_data));
  ASSPT(btd->p[0], -dx, -dy, dz);
  ASSPT(btd->p[1], dx, -dy, -dz);
  ASSPT(btd->p[2], dx, -dy, dz);

  /* Upper bottom */

  mg->tv[11][0] = 5; mg->tv[11][1] = 4; mg->tv[11][2] = 0;
  mg->te[11][0] = 16; mg->te[11][1] = 8; mg->te[11][2] = 4;
  mg->pa[11] = box_triangle_parametrization;
  mg->pd[11] = btd = (box_triangle_data *) malloc(sizeof(box_triangle_data));
  ASSPT(btd->p[0], -dx, -dy, -dz);
  ASSPT(btd->p[1], dx, -dy, -dz);
  ASSPT(btd->p[2], -dx, -dy, dz);

  return mg;
}
