
/* ------------------------------------------------------------
   This is the file "bem3d.h" of the HLib package.

   Copyright 1999 Steffen Boerm and Lars Grasedyck
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#ifndef BEM3D
#define BEM3D

typedef struct _bemgrid3d bemgrid3d;
typedef bemgrid3d *pbemgrid3d;
typedef const bemgrid3d *pcbemgrid3d;

typedef struct _macrogrid3d macrogrid3d;
typedef macrogrid3d *pmacrogrid3d;
typedef const macrogrid3d *pcmacrogrid3d;

#include "basic.h"
#include "cluster.h"

#include <math.h>
#include <malloc.h>
#include <assert.h>

typedef void (*triangleparametrization3d)(double *x,
					  double s, double t,
					  void *data);

struct _bemgrid3d {
  double (*x)[3];		/* Coordinates of vertices */
  int (*e)[2];			/* Vertices of edges */
  int (*t)[3];			/* Vertices of triangles */
  int (*s)[3];			/* Edges of triangles */

  int vertices;			/* Number of vertices */
  int edges;			/* Number of edges */
  int triangles;		/* Number of triangles */

  double *gdet;			/* Gram determinants */
  double (*n)[3];		/* Outer normal vectors */
};

struct _macrogrid3d {
  int *v;			/* Map abstract vertices to real vertices */
  int *e;			/* Map abstract edges to real vertices */
  int *s;			/* Map abstract edges to real edges */
  int (*ev)[2];			/* Abstract endpoints of abstract edges */
  int (*tv)[3];			/* Abstract vertices of abstract triangles */
  int (*te)[3];			/* Abstract edges of abstract triangles */
  triangleparametrization3d *pa; /* Parametrizations for each patch */
  void **pd;			/* Auxiliary data for each parametrization */

  int vertices;			/* Number of abstract vertices */
  int edges;			/* Number of abstract edges */
  int triangles;		/* Number of abstract triangles */
};

/* Create an empty bemgrid3d object */
pbemgrid3d
new_bemgrid3d(int vertices, int edges, int triangles);

/* Free the memory used by a bemgrid3d object */
void
del_bemgrid3d(pbemgrid3d gr);

/* Create a bemgrid3d object using a macro grid */
pbemgrid3d
macro2_bemgrid3d(pmacrogrid3d mg, int split, int check_match);

/* Create a bemgrid3d object using a file */
pbemgrid3d
read_bemgrid3d(const char *filename);

pbemgrid3d
read1_bemgrid3d(const char *filename);

pbemgrid3d
read2_bemgrid3d(const char *filename);

/* Write a bemgrid3d object to a file */
void
write_bemgrid3d(pcbemgrid3d gr, const char *filename);

void
write2_bemgrid3d(pcbemgrid3d gr, const char *filename);

/* Read a grid function from a file */
double *
readvector_bemgrid3d(const char *filename,
		     pcbemgrid3d gr, pcclustertree ct);

/* Write a grid function to a file */
void
writevector_bemgrid3d(const char *filename,
		      int n, const double *x,
		      pcbemgrid3d gr, pcclustertree ct);

/* Compute outer normal vectors and Gram determinants */
void
prepare_bemgrid3d(pbemgrid3d gr);

/* Check whether a BEM grid is closed */
int
closed_bemgrid3d(pcbemgrid3d gr);

/* Check whether a BEM grid is closed and oriented */
int
oriented_bemgrid3d(pcbemgrid3d gr);

/* Build a cluster tree for the grid defined by a
   bemgrid3d object, initialize perm and iperm arrays */
pclustertree
buildcluster_bemgrid3d(pcbemgrid3d gr,
		       ClusterStrategy strategy, int leafsize);

/* Create a cluster tree (vector valued),
   nrdof[i] = number of entries for triangle i,
   perm,iperm are of length total nr DoF,
*/
pclustertree
buildcluster_vec_bemgrid3d(pcbemgrid3d gr,
			   ClusterStrategy strategy, int leafsize,
			   int *nrdof);

/* Like buildcluster_bemgrid3d, but for piecewise linear basis functions */
pclustertree
buildcluster_bemgrid3d_vertices(pcbemgrid3d gr,
				ClusterStrategy strategy, int leafsize);

/* Create an empty macrogrid3d object */
pmacrogrid3d
new_macrogrid3d(int vertices, int edges, int triangles);

/* Free the memory used by a macrogrid3d object */
void
del_macrogrid3d(pmacrogrid3d mg);

/* Create a macro grid for an ellipsoid */
pmacrogrid3d
ellipsoid_macrogrid3d(double rx, double ry, double rz);

/* Create a macro grid for a box */
pmacrogrid3d
box_macrogrid3d(double dx, double dy, double dz);

#endif
