
/* ------------------------------------------------------------
   This is the file "cluster.c" of the HLib package.

   Copyright 1999 Steffen Boerm and Lars Grasedyck
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#include "basic.h"
#include "cluster.h"
#include "clusterbasis.h"
#include "uniformmatrix.h"

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

struct _qsort_entry {
  double val;
  int idx;
};

const double bbox_eps = 1e-5;

/* ************************************************************
    ClusterFactory
   ************************************************************ */

pclusterfactory
new_clusterfactory(int ndof, int nidx, int d)
{
  clusterfactory *this;
  double **x;
  double **smin;
  double **smax;
  int **perm;
  double *blocks;
  int *iblocks;
  int i;

  this = (clusterfactory *) malloc(sizeof(clusterfactory));
  assert(this != NULL);

  this->x = x = (double **) malloc((size_t) sizeof(double *) * nidx);
  assert(this->x != NULL);

  this->smin = smin = (double **) malloc((size_t) sizeof(double *) * nidx);
  assert(this->smin != NULL);

  this->smax = smax = (double **) malloc((size_t) sizeof(double *) * nidx);
  assert(this->smax != NULL);

  this->perm = perm = (int **) malloc((size_t) sizeof(int *) * d);
  assert(this->perm != NULL);

  this->qe = (struct _qsort_entry *) malloc((size_t) sizeof(struct _qsort_entry) * ndof);
  assert(this->qe != NULL);

  this->tmp = (int *) malloc((size_t) sizeof(int) * 2 * ndof);
  assert(this->tmp != NULL);

  this->vmin = (double *) malloc((size_t) sizeof(double) * d);
  assert(this->vmin != NULL);

  this->vmax = (double *) malloc((size_t) sizeof(double) * d);
  assert(this->vmax != NULL);

  this->blocks = blocks = (double *) malloc((size_t) sizeof(double) * 3 * d * nidx);
  assert(this->blocks != NULL);

  this->iblocks = iblocks = (int *) malloc((size_t) sizeof(int) * d * ndof);
  assert(this->iblocks != NULL);

  for(i=0; i<nidx; i++) {
    x[i] = blocks;
    blocks += d;

    smin[i] = blocks;
    blocks += d;

    smax[i] = blocks;
    blocks += d;
  }

  for(i=0; i<d; i++) {
    perm[i] = iblocks;
    iblocks += ndof;
  }

  this->ndof = ndof;
  this->nidx = nidx;
  this->d = d;

  return this;
}

void
del_clusterfactory(pclusterfactory this)
{
  assert(this != NULL);
  assert(this->x != NULL);
  assert(this->smin != NULL);
  assert(this->smax != NULL);
  assert(this->perm != NULL);
  assert(this->tmp != NULL);
  assert(this->vmin != NULL);
  assert(this->vmax != NULL);
  assert(this->blocks != NULL);
  assert(this->iblocks != NULL);

  free(this->iblocks);
  free(this->blocks);
  free(this->vmax);
  free(this->vmin);
  free(this->tmp);
  free(this->perm);
  free(this->smax);
  free(this->smin);
  free(this->x);
  free(this);
}

static void
update_boundingbox(pcluster tau, const int *index, pclusterfactory factory)
{
  double diam, bdiff;
  int d = tau->d;
  int i, j;

  if(tau->sons > 0) {
    for(j=0; j<d; j++) {
      assert(tau->son[0]->bmin[j] <= tau->son[0]->bmax[j]);
      tau->bmin[j] = tau->son[0]->bmin[j];
      tau->bmax[j] = tau->son[0]->bmax[j];
    }
    
    for(i=1; i<tau->sons; i++)
      for(j=0; j<d; j++) {
	assert(tau->son[i]->bmin[j] <= tau->son[i]->bmax[j]);
	if(tau->son[i]->bmin[j] < tau->bmin[j])
	  tau->bmin[j] = tau->son[i]->bmin[j];
	if(tau->bmax[j] < tau->son[i]->bmax[j])
	  tau->bmax[j] = tau->son[i]->bmax[j];
      }
  }
  else {
    for(j=0; j<d; j++) {
      assert(factory->smin[index[0]][j] <= factory->smax[index[0]][j]);
      tau->bmin[j] = factory->smin[index[0]][j];
      tau->bmax[j] = factory->smax[index[0]][j];
    }
    
    for(i=1; i<tau->size; i++)
      for(j=0; j<d; j++) {
	assert(factory->smin[index[i]][j] <= factory->smax[index[i]][j]);
	if(factory->smin[index[i]][j] < tau->bmin[j])
	  tau->bmin[j] = factory->smin[index[i]][j];
	if(tau->bmax[j] < factory->smax[index[i]][j])
	  tau->bmax[j] = factory->smax[index[i]][j];
      }
  }

  diam = (tau->bmax[0] - tau->bmin[0]) * (tau->bmax[0] - tau->bmin[0]);
  for(j=1; j<d; j++)
    diam += (tau->bmax[j] - tau->bmin[j]) * (tau->bmax[j] - tau->bmin[j]);

  diam = sqrt(diam);
  for(j=0; j<d; j++) {
    bdiff = tau->bmax[j] - tau->bmin[j];
    if(bdiff < bbox_eps * diam) {
      tau->bmax[j] += 0.5*(bbox_eps * diam - bdiff);
      tau->bmin[j] -= 0.5*(bbox_eps * diam - bdiff);
    }
  }
}

/* ************************************************************
    Geometric clustering
   ************************************************************ */

static pcluster
split_geometrically(pclusterfactory factory, int *index,
		    int start, int sz, int leafsize)
{
  pcluster this;
  const double * const *x = (const double * const *) factory->x;
  double *vmin = factory->vmin;
  double *vmax = factory->vmax;
  int d = factory->d;
  double vmid, vdiff;
  int l, r, h;
  int i, j, jmax;

  if(sz <= leafsize)
    this = new_cluster(start, sz, factory->d, 0);
  else {
    for(j=0; j<d; j++) {
      vmin[j] = vmax[j] = x[index[0]][j];
      for(i=1; i<sz; i++)
	if(x[index[i]][j] < vmin[j])
	  vmin[j] = x[index[i]][j];
	else if(vmax[j] < x[index[i]][j])
	  vmax[j] = x[index[i]][j];
    }

    vdiff = vmax[0] - vmin[0]; jmax = 0;
    for(j=1; j<d; j++)
      if(vmax[j] - vmin[j] > 1.1*vdiff) {
	vdiff = vmax[j] - vmin[j]; jmax = j;
      }

    vmid = 0.5 * (vmax[jmax] + vmin[jmax]);
    
    l = 0; r = sz-1;
    while(l < r) {
      while(l < sz && x[index[l]][jmax] <= vmid)
	l++;
      while(r >= 0 && x[index[r]][jmax] > vmid)
	r--;
      
      if(l < r) {
	h = index[l]; index[l] = index[r]; index[r] = h;
      }
    }

    this = new_cluster(start, sz, factory->d, 2);
    this->son[0] = split_geometrically(factory, index, start, l, leafsize);
    this->son[1] = split_geometrically(factory, index+l, start+l, sz-l, leafsize);
  }

  update_boundingbox(this, index, factory);

  return this;
}

static pcluster
do_geometric(pclusterfactory factory,
	     int *index, int n,
	     int leafsize)
{
  assert(n > 0);
  assert(n <= factory->ndof);

  return split_geometrically(factory, index, 0, n, leafsize);
}

/* ************************************************************
    Regular geometric clustering
   ************************************************************ */

static pcluster
split_regularly(pclusterfactory factory, int *index,
		int start, int sz, int jnext, int leafsize)
{
  pcluster this;
  const double * const *x = (const double * const *) factory->x;
  double *vmin = factory->vmin;
  double *vmax = factory->vmax;
  int d = factory->d;
  double vmid, old;
  int l, r, h;

  if(sz <= leafsize)
    this = new_cluster(start, sz, factory->d, 0);
  else {
    vmid = 0.5 * (vmax[jnext] + vmin[jnext]);
    
    l = 0; r = sz-1;
    while(l < r) {
      while(l < sz && x[index[l]][jnext] <= vmid)
	l++;
      while(r >= 0 && x[index[r]][jnext] > vmid)
	r--;
      
      if(l < r) {
	h = index[l]; index[l] = index[r]; index[r] = h;
      }
    }

    if(l == 0) {
      this = new_cluster(start, sz, factory->d, 1);

      old = vmin[jnext];
      vmin[jnext] = vmid;
      this->son[0] = split_regularly(factory, index, start, sz,
				     (jnext+1) % d, leafsize);
      vmin[jnext] = old;
    }
    else if(l == sz) {
      this = new_cluster(start, sz, factory->d, 1);

      old = vmax[jnext];
      vmax[jnext] = vmid;
      this->son[0] = split_regularly(factory, index, start, sz,
				     (jnext+1) % d, leafsize);
      vmax[jnext] = old;
    }
    else {
      this = new_cluster(start, sz, factory->d, 2);

      old = vmax[jnext];
      vmax[jnext] = vmid;
      this->son[0] = split_regularly(factory, index, start, l,
				     (jnext+1) % d, leafsize);
      vmax[jnext] = old;

      old = vmin[jnext];
      vmin[jnext] = vmid;
      this->son[1] = split_regularly(factory, index+l, start+l, sz-l,
				     (jnext+1) % d, leafsize);
      vmin[jnext] = old;
    }
  }

  update_boundingbox(this, index, factory);

  return this;
}

static pcluster
do_regular(pclusterfactory factory,
	   int *index, int n, int leafsize)
{
  const double * const *x = (const double * const *) factory->x;
  double *vmin = factory->vmin;
  double *vmax = factory->vmax;
  int d = factory->d;
  int i, j;

  assert(n > 0);
  assert(n <= factory->ndof);

  for(j=0; j<d; j++) {
    vmin[j] = vmax[j] = x[index[0]][j];
    for(i=1; i<n; i++)
      if(x[index[i]][j] < vmin[j])
	vmin[j] = x[index[i]][j];
      else if(vmax[j] < x[index[i]][j])
	vmax[j] = x[index[i]][j];
  }

  return split_regularly(factory, index, 0, n, 0, leafsize);
}

/* ************************************************************
    Cardinality-balanced clustering
   ************************************************************ */

static int
qsort_compare(const void *x, const void *y)
{
  const struct _qsort_entry *xe = (const struct _qsort_entry *) x;
  const struct _qsort_entry *ye = (const struct _qsort_entry *) y;

  if(xe->val < ye->val)
    return -1;
  else
    if(xe->val > ye->val)
      return 1;

  return 0;
}

static pcluster
split_cardinality(pclusterfactory factory, int *index,
		  int start, int sz, int leafsize)
{
  pcluster this;
  const double * const *x = (const double * const *) factory->x;
  int * const *perm = factory->perm;
  int *tmp = factory->tmp;
  int *tmp2 = factory->tmp + factory->ndof;
  int d = factory->d;
  double vdiff, vmax;
  int l, r;
  int i, j, j2, jmax;

  if(sz <= leafsize) {
    this = new_cluster(start, sz, factory->d, 0);

    (void) memcpy(index, perm[0]+start, (size_t) sizeof(int) * sz);
  }
  else {
    vdiff = x[perm[0][start+sz-1]][0] - x[perm[0][start]][0]; jmax = 0;
    for(j=1; j<d; j++) {
      vmax = x[perm[j][start+sz-1]][j] - x[perm[j][start]][j];
      if(vdiff < vmax) {
	vdiff = vmax; jmax = j;
      }
    }

    for(i=0; i<sz/2; i++)
      tmp[perm[jmax][start+i]] = 1;
    for(; i<sz; i++)
      tmp[perm[jmax][start+i]] = 0;
    
    for(j=1; j<d; j++) {
      j2 = (jmax+j) % d;

      l = 0; r = sz/2;
      for(i=0; i<sz; i++)
	if(tmp[perm[j2][start+i]])
	  tmp2[l++] = perm[j2][start+i];
	else
	  tmp2[r++] = perm[j2][start+i];
      
      assert(l == sz/2);
      assert(r == sz);

      (void) memcpy(perm[j2]+start, tmp2, (size_t) sizeof(int) * sz);
    }
    
    this = new_cluster(start, sz, factory->d, 2);
    this->son[0] = split_cardinality(factory, index, start, sz/2, leafsize);
    this->son[1] = split_cardinality(factory, index+sz/2, start+sz/2, sz-sz/2, leafsize);
  }

  update_boundingbox(this, index, factory);

  return this;
}

static pcluster
do_cardinality(pclusterfactory factory,
	       int *index, int n, int leafsize)
{
  const double * const *x = (const double * const *) factory->x;
  struct _qsort_entry *qe = factory->qe;
  int * const *perm = factory->perm;
  int d = factory->d;
  int i, j;

  assert(n > 0);
  assert(n <= factory->ndof);
  
  for(j=0; j<d; j++) {
    for(i=0; i<n; i++) {
      qe[i].val = x[index[i]][j];
      qe[i].idx = index[i];
    }
    
    qsort(qe, (size_t) n, sizeof(struct _qsort_entry), qsort_compare);

    for(i=0; i<n; i++)
      perm[j][i] = qe[i].idx;
  }

  return split_cardinality(factory, index, 0, n, leafsize);
}

/* ************************************************************
    General clustering routines
   ************************************************************ */

pclustertree
create_clustertree(pclusterfactory factory, ClusterStrategy strategy, int leafsize)
{
  pclustertree ct;
  int n = factory->nidx;
  int *dof2idx, *idx2dof;
  int i;

  assert(n <= factory->ndof);
  assert(n <= factory->nidx);

  ct = new_clustertree(n, n);
  dof2idx = ct->dof2idx;
  idx2dof = ct->idx2dof;

  for(i=0; i<n; i++)
    dof2idx[i] = i;

  switch(strategy) {
  default:
  case HLIB_DEFAULT:
  case HLIB_GEOMETRIC:
    ct->root = do_geometric(factory, dof2idx, n, leafsize);
    break;
  case HLIB_REGULAR:
    ct->root = do_regular(factory, dof2idx, n, leafsize);
    break;
  case HLIB_CARDINALITY:
    ct->root = do_cardinality(factory, dof2idx, n, leafsize);
    break;
  }

  for(i=0; i<n; i++)
    idx2dof[i] = -1;

  for(i=0; i<n; i++)
    idx2dof[dof2idx[i]] = i;

  return ct;
}

pclustertree
create_subclustertree(pclusterfactory factory, const int *index, int n,
		      ClusterStrategy strategy, int leafsize)
{
  pclustertree ct;
  int *dof2idx, *idx2dof;
  int i;

  assert(n <= factory->ndof);

  ct = new_clustertree(n, factory->nidx);
  dof2idx = ct->dof2idx;
  idx2dof = ct->idx2dof;

  for(i=0; i<n; i++) {
    assert(index[i] >= 0);
    assert(index[i] < factory->nidx);
    dof2idx[i] = index[i];
  }

  switch(strategy) {
  default:
  case HLIB_DEFAULT:
  case HLIB_GEOMETRIC:
    ct->root = do_geometric(factory, dof2idx, n, leafsize);
    break;
  case HLIB_REGULAR:
    ct->root = do_regular(factory, dof2idx, n, leafsize);
    break;
  case HLIB_CARDINALITY:
    ct->root = do_cardinality(factory, dof2idx, n, leafsize);
    break;
  }

  for(i=0; i<factory->nidx; i++)
    idx2dof[i] = -1;

  for(i=0; i<n; i++)
    idx2dof[dof2idx[i]] = i;

  return ct;
}

/* ************************************************************
    Cluster Tree
   ************************************************************ */

pclustertree
new_clustertree(int ndof, int nidx)
{
  pclustertree ct;

  assert(ndof >= 0);
  assert(ndof <= nidx);

  ct = (pclustertree) malloc(sizeof(clustertree));
  assert(ct != NULL);

  ct->dof2idx = (int *) malloc((size_t) sizeof(int) * ndof);
  assert(ct->dof2idx != NULL);

  ct->idx2dof = (int *) malloc((size_t) sizeof(int) * nidx);
  assert(ct->idx2dof != NULL);

  ct->root = NULL;

  ct->ndof = ndof;
  ct->nidx = nidx;

  return ct;
}

void
del_clustertree(pclustertree ct)
{
  assert(ct != NULL);
  assert(ct->dof2idx != NULL);
  assert(ct->idx2dof != NULL);

  if(ct->root) {
    del_cluster(ct->root);
    ct->root = NULL;
  }
  
  free(ct->idx2dof);
  ct->idx2dof = NULL;

  free(ct->dof2idx);
  ct->dof2idx = NULL;
}

pclustertree
read_clustertree(const char *filename)
{
  FILE *in;
  char buf[80];
  pclustertree ct;
  pcluster *cluster;
  int *son, *sons;
  int first, size;
  int *dof2idx, *idx2dof;
  int ndof, nidx, clusters;
  int i, j, line;

  in = fopen(filename, "r");
  assert(in != NULL);

  line = 0;

  (void) fgets(buf, 80, in);
  line++;

  if(sscanf(buf, "%d %d %d", &nidx, &ndof, &clusters) != 3) {
    (void) fprintf(stderr,
		   "Cannot process the first line of file \"%s\".\n",
		   filename);
    return NULL;
  }

  ct = new_clustertree(ndof, nidx);
  dof2idx = ct->dof2idx;
  idx2dof = ct->idx2dof;

  cluster = (pcluster *) malloc((size_t) sizeof(pcluster) * clusters);
  assert(cluster != NULL);

  sons = (int *) malloc((size_t) sizeof(int) * (clusters + 1));
  assert(sons != NULL);

  son = (int *) malloc((size_t) sizeof(int) * (clusters - 1));
  assert(son != NULL);

  dof2idx = ct->dof2idx;

  (void) fgets(buf, 80, in);
  line++;

  if(sscanf(buf, "%d", sons) != 1) {
    (void) fprintf(stderr,
		   "Cannot process line %d of file \"%s\".\n",
		   line, filename);
    free(cluster);
    return NULL;
  }

  for(i=0; i<clusters; i++) {
    (void) fgets(buf, 80, in);
    line++;

    if(sscanf(buf, "%d %d %d", sons+i+1, &first, &size) != 3) {
      (void) fprintf(stderr,
		     "Cannot read cluster %d in line %d of file \"%s\".\n",
		     i, line, filename);
      for(j=0; j<i; j++)
	del_cluster(cluster[j]);
      free(son);
      free(sons);
      free(cluster);
      return NULL;
    }

    cluster[i] = new_cluster(first, size, 0, sons[i+1] - sons[i]);
  }

  for(i=0; i<clusters-1; i++) {
    (void) fgets(buf, 80, in);
    line++;

    if(sscanf(buf, "%d", son+i) != 1) {
      (void) fprintf(stderr,
		     "Cannot read son %d in line %d of file \"%s\".\n",
		     i, line, filename);
      for(j=0; j<clusters; j++)
	del_cluster(cluster[j]);
      free(son);
      free(sons);
      free(cluster);
      return NULL;
    }
  }

  for(i=0; i<clusters; i++)
    for(j=sons[i]; j<sons[i+1]; j++)
      cluster[i]->son[j-sons[i]] = cluster[son[j]];

  for(i=0; i<ndof; i++) {
    (void) fgets(buf, 80, in);
    line++;

    if(sscanf(buf, "%d", dof2idx+i) != 1) {
      (void) fprintf(stderr,
		     "Cannot read index %d in line %d of file \"%s\".\n",
		     i, line, filename);
      for(j=0; j<clusters; j++)
	del_cluster(cluster[j]);
      free(son);
      free(sons);
      free(cluster);
      return NULL;
    }

    assert(dof2idx[i] >= 0 && dof2idx[i] < nidx);
  }

  for(i=0; i<nidx; i++)
    idx2dof[i] = -1;

  for(i=0; i<ndof; i++)
    idx2dof[dof2idx[i]] = i;

  ct->root = cluster[0];

  free(son);
  free(sons);
  free(cluster);

  return ct;
}

static void
initialize_widx(pcluster root, int level, int *idx, int *widx,
		int n, int depth)
{
  int i;

  assert(level < depth);
  widx[level * n + root->start] = (*idx)++;

  for(i=0; i<root->sons; i++)
    initialize_widx(root->son[i], level+1, idx, widx, n, depth);
}

static void
write_one_cluster(pccluster root, FILE *out, int *sons)
{
  int i;

  *sons += root->sons;

  (void) fprintf(out,
		 "%d %d %d\n",
		 *sons, root->start, root->size);

  for(i=0; i<root->sons; i++)
    write_one_cluster(root->son[i], out, sons);
}

static void
write_cluster_sons(pccluster root, int level, FILE *out, int *widx, int n)
{
  int i;

  for(i=0; i<root->sons; i++)
    (void) fprintf(out,
		   "%d\n",
		   widx[(level+1) * n + root->son[i]->start]);

  for(i=0; i<root->sons; i++)
    write_cluster_sons(root->son[i], level+1, out, widx, n);
}

void
write_clustertree(pclustertree ct, const char *filename)
{
  FILE *out;
  int *widx;
  int clusters, depth, sons;
  int ndof = ct->ndof;
  int nidx = ct->nidx;
  int i;

  out = fopen(filename, "w");

  depth = depth_cluster(ct->root);
  widx = (int *) malloc((size_t) sizeof(int) * depth * ndof);
  assert(widx != NULL);

  clusters = 0;
  initialize_widx(ct->root, 0, &clusters, widx, ndof, depth);

  (void) fprintf(out,
		 "%d %d %d\n",
		 nidx, ndof, clusters);

  (void) fprintf(out,
		 "0\n");
  sons = 0;
  write_one_cluster(ct->root, out, &sons);

  write_cluster_sons(ct->root, 0, out, widx, ndof);

  for(i=0; i<ndof; i++)
    (void) fprintf(out,
		   "%d\n",
		   ct->dof2idx[i]);

  free(widx);

  (void) fclose(out);
}

/* ************************************************************
    Cluster
   ************************************************************ */    

pcluster
new_cluster(int start, int size, int d, int sons)
{
  pcluster this;
  int i;

  this = (pcluster) malloc(sizeof(cluster));
  assert(this != NULL);

  this->start = start;
  this->size = size;
  this->d = d;
  this->sons = sons;

  if(d > 0) {
    this->bmin = (double *) malloc((size_t) sizeof(double) * d);
    assert(this->bmin != NULL);
    
    this->bmax = (double *) malloc((size_t) sizeof(double) * d);
    assert(this->bmax != NULL);
  }
  else {
    this->bmin = NULL;
    this->bmax = NULL;
  }

  if(sons > 0) {
    this->son = (pcluster *) malloc((size_t) sizeof(pcluster) * sons);
    assert(this->son != NULL);

    for(i=0; i<sons; i++)
      this->son[i] = NULL;
  }
  else
    this->son = NULL;

  this->name = -1;

  return this;
}

void
del_cluster(pcluster this)
{
  int i;

  assert(this != NULL);

  if(this->sons > 0) {
    assert(this->son != NULL);

    for(i=this->sons-1; i>=0; i--) {
      del_cluster(this->son[i]);
      this->son[i] = NULL;
    }

    free(this->son);
    this->son = NULL;
  }
  else
    assert(this->son == NULL);

  if(this->bmax) {
    free(this->bmax);
    this->bmax = NULL;
  }
  if(this->bmin) {
    free(this->bmin);
    this->bmin = NULL;
  }

  free(this);
}

/* ************************************************************
    Compute the diameter of a cluster based on its bounding box
   ************************************************************ */

double
diameter_cluster(pccluster tau)
{
  double sum;
  int i;

  assert(tau->bmax[0] >= tau->bmin[0]);
  sum = (tau->bmax[0] - tau->bmin[0]) * (tau->bmax[0] - tau->bmin[0]);

  for(i=1; i<tau->d; i++) {
    assert(tau->bmax[i] >= tau->bmin[i]);
    sum += (tau->bmax[i] - tau->bmin[i]) * (tau->bmax[i] - tau->bmin[i]);
  }

  return sqrt(sum);
}

/* ************************************************************
    Compute the distance of two clusters
   ************************************************************ */

double
distance_cluster(pccluster tau, pccluster sigma)
{
  double sum;
  int j;

  assert(tau->d == sigma->d);

  sum = 0.0;
  for(j=0; j<tau->d; j++) {
    assert(tau->bmax[j] >= tau->bmin[j]);
    assert(sigma->bmax[j] >= sigma->bmin[j]);

    if(tau->bmax[j] < sigma->bmin[j])
      sum += (sigma->bmin[j] - tau->bmax[j]) * (sigma->bmin[j] - tau->bmax[j]);
    else if(sigma->bmax[j] < tau->bmin[j])
      sum += (tau->bmin[j] - sigma->bmax[j]) * (tau->bmin[j] - sigma->bmax[j]);
  }

  return sqrt(sum);
}

/* ************************************************************
    Create a supermatrix based on the standard admissibility condition
   ************************************************************ */

psupermatrix
build_supermatrix_from_cluster(pccluster row, pccluster col,
			       double eta, int nmin, BlockStrategy strategy)
{
  prkmatrix rk;
  pfullmatrix full;
  psupermatrix super;
  double diam_row, diam_col, dist;
  int admissible;
  int i, j;

  diam_row = diameter_cluster(row);
  diam_col = diameter_cluster(col);
  dist = distance_cluster(row, col);

  admissible = 0;
  switch(strategy & 0xf) {
  default:
  case HLIB_MINADM:
    admissible = (diam_row <= eta*dist || diam_col <= eta*dist);
    break;
  case HLIB_MAXADM:
    admissible = (diam_row <= eta*dist && diam_col <= eta*dist);
    break;
  }

  super = NULL;

  if(admissible) {
    rk = new_rkmatrix(0, row->size, col->size);
    super = new_supermatrix(1, 1, row->size, col->size,
			    (puniformmatrix) 0, rk, (pfullmatrix) 0);
  }
  else
    if(strategy & HLIB_INHOMOGENEOUS) {
      if(row->sons == 0) {
	if(col->sons == 0) {
	  full = new_fullmatrix(row->size, col->size);
	  super = new_supermatrix(1, 1, row->size, col->size,
				  (puniformmatrix) 0, (prkmatrix) 0, full);
	}
	else {
	  super = new_supermatrix(1, col->sons, row->size, col->size,
				  (puniformmatrix) 0, (prkmatrix) 0,
				  (pfullmatrix) 0);
	  for(j=0; j<col->sons; j++)
	    super->s[j] = build_supermatrix_from_cluster(row, col->son[j], eta,
							 nmin, strategy);
	}
      }
      else {
	if(col->sons == 0) {
	  super = new_supermatrix(row->sons, 1, row->size, col->size,
				  (puniformmatrix) 0, (prkmatrix) 0,
				  (pfullmatrix) 0);
	  for(i=0; i<row->sons; i++)
	    super->s[i] = build_supermatrix_from_cluster(row->son[i], col, eta,
							 nmin, strategy);
	}
	else {
	  super = new_supermatrix(row->sons, col->sons, row->size, col->size,
				  (puniformmatrix) 0, (prkmatrix) 0,
				  (pfullmatrix) 0);
	  for(i=0; i<row->sons; i++)
	    for(j=0; j<col->sons; j++)
	      super->s[i+j*row->sons] = 
		build_supermatrix_from_cluster(row->son[i], col->son[j], eta,
					       nmin, strategy);
	}
      }
    }
    else {
      if(row->sons == 0 || col->sons == 0) {
	full = new_fullmatrix(row->size, col->size);
	super = new_supermatrix(1, 1, row->size, col->size,
				(puniformmatrix) 0, (prkmatrix) 0, full);
      }
      else {
	super = new_supermatrix(row->sons, col->sons, row->size, col->size,
				(puniformmatrix) 0, (prkmatrix) 0,
				(pfullmatrix) 0);
	
	for(i=0; i<row->sons; i++)
	  for(j=0; j<col->sons; j++)
	    super->s[i+j*row->sons] = 
	      build_supermatrix_from_cluster(row->son[i], col->son[j], eta,
					     nmin, strategy);
      }
    }
      
  return super;
}

/* *******************************************************************
    Create a supermatrix based on the standard admissibility condition
    where non-diagonal leaves are always admissible
   ******************************************************************* */

psupermatrix
build_fem_supermatrix_from_cluster(pcluster row, pcluster col,
				   double eta, int nmin, 
				   double eps, int k,
				   BuildFemStrategy sym){
  prkmatrix rk;
  pfullmatrix full;
  psupermatrix super;
  double diam_row, diam_col, dist;
  int i, j, weak_adm=0;
  /*
  double mid;
  int dim;
  */

  diam_row = diameter_cluster(row);
  diam_col = diameter_cluster(col);
  dist = distance_cluster(row, col);

  /*
  dim = row->d;
  if(row->d==2 && (row->size<=10000 || col->size<=10000) && k==0){
    weak_adm = 1;
    for(i=0; i<dim; i++){
      mid = 0.5*(row->bmin[i]+row->bmax[i]);
      if(mid > col->bmin[i] && mid < col->bmax[i]) 
	weak_adm=0;
      mid = 0.5*(col->bmin[i]+col->bmax[i]);
      if(mid > row->bmin[i] && mid < row->bmax[i]) 
	weak_adm=0;
    }
  }
  */
  /*
  if(weak_adm==0 && row->d>2 && (row->size<=1024 || col->size<=1024)){
    dim = row->d;
    j = 0;
    for(i=0; i<dim; i++){
      if((0.5*(row->bmin[i]+row->bmax[i]) > col->bmin[i] && 
	  0.5*(row->bmin[i]+row->bmax[i]) < col->bmax[i] )
	 ||
	 (0.5*(col->bmin[i]+col->bmax[i]) > row->bmin[i] && 
	  0.5*(col->bmin[i]+col->bmax[i]) < row->bmax[i]) )
	j++;
    }
    if(j<2) weak_adm = 2; 
  }
  */

  if((diam_row <= eta*dist || diam_col <= eta*dist) 
     || (row->d==2 && row->start != col->start && (row->size<=nmin || col->size<=nmin))
     || (weak_adm>0)){
    if(diam_row <= eta*dist || diam_col <= eta*dist){
      rk = new_rkmatrix(k, row->size, col->size);
    }else{
      rk = new_rkmatrix(3*k, row->size, col->size);
    }
    rk->local_eps = eps;
    super = new_supermatrix(1, 1, row->size, col->size,
			    (puniformmatrix) 0, rk, (pfullmatrix) 0);
  }
  else if(row->sons == 0 || col->sons == 0) {
    full = new_fullmatrix(row->size, col->size);
    super = new_supermatrix(1, 1, row->size, col->size,
			    (puniformmatrix) 0, (prkmatrix) 0, full);
  }
  else {
    super = new_supermatrix(row->sons, col->sons, row->size, col->size,
			    (puniformmatrix) 0, (prkmatrix) 0, (pfullmatrix) 0);
    for(i=0; i<row->sons; i++)
      for(j=0; j<col->sons; j++)
	if(i>=j || row!=col || sym!=HLIB_FEMSYMM){
	  super->s[i+j*row->sons] = 
	    build_fem_supermatrix_from_cluster(row->son[i], col->son[j], 
					      eta, nmin, eps, k, sym);
	}else{
	  rk = new_rkmatrix(0, row->son[i]->size,col->son[j]->size);
	  super->s[i+j*row->sons] = 
	    new_supermatrix(1, 1, row->son[i]->size, col->son[j]->size,
			    (puniformmatrix) 0, rk, (pfullmatrix) 0);
	}
  }
  
  return super;
}

/* ************************************************************
    Create a supermatrix based on the strong admissibility condition
    such that both clusters of inadmissible blocks are leaves and
    admissible blocks are uniform.
   ************************************************************ */

psupermatrix
build_supermatrix_from_cluster2(pclusterbasis row, pclusterbasis col,
				double eta)
{
  puniformmatrix um;
  pfullmatrix full;
  psupermatrix super;
  pccluster c_row, c_col;
  double diam_row, diam_col, dist;
  int i, j;

  c_row = row->t;
  c_col = col->t;

  diam_row = diameter_cluster(c_row);
  diam_col = diameter_cluster(c_col);
  dist = distance_cluster(row->t, col->t);

  if(diam_row <= eta*dist && diam_col <= eta*dist) {
    um = new_uniformmatrix(row, col);
    super = new_supermatrix(1, 1, c_row->size, c_col->size,
			    um, (prkmatrix) 0, (pfullmatrix) 0);
    super->row = row;
    super->col = col;
  }
  else
    if(row->sons == 0)
      if(col->sons == 0) {
	full = new_fullmatrix(c_row->size, c_col->size);
	super = new_supermatrix(1, 1, c_row->size, c_col->size,
				(puniformmatrix) 0, (prkmatrix) 0, full);
	super->row = row;
	super->col = col;
      }
      else {
	super = new_supermatrix(1, col->sons, c_row->size, c_col->size,
				(puniformmatrix) 0, (prkmatrix) 0, (pfullmatrix) 0);
	super->row = row;
	super->col = col;
	for(j=0; j<col->sons; j++)
	  super->s[j] = build_supermatrix_from_cluster2(row, col->son[j], eta);
      }
    else
      if(col->sons == 0) {
	super = new_supermatrix(row->sons, 1, c_row->size, c_col->size,
				(puniformmatrix) 0, (prkmatrix) 0, (pfullmatrix) 0);
	super->row = row;
	super->col = col;
	for(i=0; i<row->sons; i++)
	  super->s[i] = build_supermatrix_from_cluster2(row->son[i], col, eta);
      }
      else {
	super = new_supermatrix(row->sons, col->sons, c_row->size, c_col->size,
				(puniformmatrix) 0, (prkmatrix) 0, (pfullmatrix) 0);
	super->row = row;
	super->col = col;
	for(i=0; i<row->sons; i++)
	  for(j=0; j<col->sons; j++)
	    super->s[i+j*row->sons] = build_supermatrix_from_cluster2(row->son[i], col->son[j], eta);
      }

  return super;
}

psupermatrix
build_supermatrix_from_cluster25(pclusterbasis row, pclusterbasis col,
				 double eta, int nmin, int k)
{
  puniformmatrix um;
  prkmatrix rk;
  pfullmatrix full;
  psupermatrix super;
  pccluster c_row, c_col;
  double diam_row, diam_col, dist;
  int i, j;

  c_row = row->t;
  c_col = col->t;

  diam_row = diameter_cluster(c_row);
  diam_col = diameter_cluster(c_col);
  dist = distance_cluster(row->t, col->t);

  if(diam_row <= eta*dist && diam_col <= eta*dist) {
    if(k*(c_row->size + c_col->size) > nmin){
      um = new_uniformmatrix(row, col);
      super = new_supermatrix(1, 1, c_row->size, c_col->size,
			      um, (prkmatrix) 0, (pfullmatrix) 0);
    }else{
      rk = new_rkmatrix(k, c_row->size, c_col->size);
      super = new_supermatrix(1, 1, c_row->size, c_col->size,
			      (puniformmatrix)0, rk, (pfullmatrix) 0);
    }
    super->row = row;
    super->col = col;
  }
  else
    if(row->sons == 0)
      if(col->sons == 0) {
	full = new_fullmatrix(c_row->size, c_col->size);
	super = new_supermatrix(1, 1, c_row->size, c_col->size,
				(puniformmatrix) 0, (prkmatrix) 0, full);
	super->row = row;
	super->col = col;
      }
      else {
	super = new_supermatrix(1, col->sons, c_row->size, c_col->size,
				(puniformmatrix) 0, (prkmatrix) 0, (pfullmatrix) 0);
	super->row = row;
	super->col = col;
	for(j=0; j<col->sons; j++)
	  super->s[j] = build_supermatrix_from_cluster25(row, col->son[j], eta,
							 nmin,k);
      }
    else
      if(col->sons == 0) {
	super = new_supermatrix(row->sons, 1, c_row->size, c_col->size,
				(puniformmatrix) 0, (prkmatrix) 0, (pfullmatrix) 0);
	super->row = row;
	super->col = col;
	for(i=0; i<row->sons; i++)
	  super->s[i] = build_supermatrix_from_cluster25(row->son[i], col, eta,
							 nmin,k);
      }
      else {
	super = new_supermatrix(row->sons, col->sons, c_row->size, c_col->size,
				(puniformmatrix) 0, (prkmatrix) 0, (pfullmatrix) 0);
	super->row = row;
	super->col = col;
	for(i=0; i<row->sons; i++)
	  for(j=0; j<col->sons; j++)
	    super->s[i+j*row->sons] = 
	      build_supermatrix_from_cluster25(row->son[i], col->son[j], eta,
					       nmin,k);
      }

  return super;
}

/* ************************************************************
    Create a new supermatrix, copying the the structure of a
    given supermatrix, but replacing the original cluster basis
    by a given basis with the same dimension.
   ************************************************************ */

psupermatrix
reuse_supermatrix(psupermatrix src,
		  pclusterbasis row, pclusterbasis col)
{
  psupermatrix trg;
  int i, j;

  assert(src->rows == row->n);
  assert(src->cols == col->n);

  if(src->u) {
    assert(src->row->k == row->k);
    assert(src->row->kt == row->kt);
    assert(src->col->k == col->k);
    assert(src->col->kt == col->kt);

    trg = new_supermatrix(1, 1, src->rows, src->cols,
			  src->u, (prkmatrix) 0, (pfullmatrix) 0);
  }
  else {
    if(src->r)
      trg = new_supermatrix(1, 1, src->rows, src->cols,
			    (puniformmatrix) 0, src->r, (pfullmatrix) 0);
    else {
      if(src->f)
	trg = new_supermatrix(1, 1, src->rows, src->cols,
			      (puniformmatrix) 0, (prkmatrix) 0, src->f);
      else {
	assert(src->s != NULL);

	trg = new_supermatrix(src->block_rows, src->block_cols,
			      src->rows, src->cols,
			      (puniformmatrix) 0, (prkmatrix) 0, (pfullmatrix) 0);

	if(src->block_cols == 1)
	  if(src->block_rows == 1) 
	    trg->s[0] = reuse_supermatrix(src->s[0], row, col);
	  else
	    for(i=0; i<src->block_rows; i++)
	      trg->s[i] = reuse_supermatrix(src->s[i], row->son[i], col);
	else
	  if(src->block_rows == 1)
	    for(j=0; j<src->block_cols; j++)
	      trg->s[j] = reuse_supermatrix(src->s[j], row, col->son[j]);
	  else
	    for(j=0; j<src->block_cols; j++)
	      for(i=0; i<src->block_rows; i++)
		trg->s[i+j*src->block_rows] = reuse_supermatrix(src->s[i+j*src->block_rows], row->son[i], col->son[j]);
      }
    }
  }
  
  trg->row = row;
  trg->col = col;

  return trg;
}

/* ************************************************************
    Create a trivial supermatrix
   ************************************************************ */

psupermatrix
build_full_supermatrix(pccluster row, pccluster col)
{
  pfullmatrix full;
  psupermatrix super;

  full = new_fullmatrix(row->size, col->size);
  super = new_supermatrix(1, 1, row->size, col->size,
			  (puniformmatrix) 0, (prkmatrix) 0, full);

  return super;
}

/* ************************************************************
    Print a cluster tree
   ************************************************************ */

void
print_cluster(pccluster root, const int *index, int indent)
{
  int d = root->d;
  int i;

  assert(d > 0);

  (void) printf("%s(%d %d) {",
		"                                        " + 40 - 2*indent,
		root->start, root->size);
  for(i=0; i<root->size; i++)
    (void) printf(" %d", index[root->start+i]);
  (void) printf(" } [%f, %f]",
		root->bmin[0], root->bmax[0]);
  for(i=1; i<d; i++)
    (void) printf(" x [%f, %f]",
		  root->bmin[i], root->bmax[i]);
  (void) printf("\n");

  for(i=0; i<root->sons; i++)
    (void) print_cluster(root->son[i], index, indent+1);
}

/* ************************************************************
    Determine depth of a cluster tree
   ************************************************************ */

int
depth_cluster(pccluster root)
{
  int d;
  int i;

  d = 0;
  if(root->sons > 0) {
    for(i=0; i<root->sons; i++)
      d = imax(d, depth_cluster(root->son[i]));
  }

  return d+1;
}

/* ************************************************************
    Determine number of nodes in a cluster tree
   ************************************************************ */

int
count_cluster(pccluster root)
{
  int cl;
  int i;

  cl = 1;
  for(i=0; i<root->sons; i++)
    cl += count_cluster(root->son[i]);

  return cl;
}
