
/* ------------------------------------------------------------
   This is the file "clusterbasis.c" of the HLib package.

   Copyright 1999 Steffen Boerm and Lars Grasedyck
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#include "basic.h"
#include "clusterbasis.h"
#include "lapack.h"

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

static const double ORTHO_EPS = 1e-12;

pclusterbasis
new_clusterbasis(pccluster t, int k, int ksons)
{
  pclusterbasis cb;
  int i;
  
  cb = (pclusterbasis) malloc(sizeof(clusterbasis));
  assert(cb != NULL);

  if(k > 0) {
    cb->xt = (double *) malloc((size_t) sizeof(double) * k);
    assert(cb->xt != NULL);
    clear_vector(k, cb->xt);
    
    cb->yt = (double *) malloc((size_t) sizeof(double) * k);
    assert(cb->yt != NULL);
    clear_vector(k, cb->yt);
  }
  else {
    cb->xt = NULL;
    cb->yt = NULL;
  }

  if(t->sons > 0) {
    cb->son = (pclusterbasis *) malloc((size_t) sizeof(pclusterbasis) * t->sons);
    assert(cb->son != NULL);

    cb->T = (double **) malloc((size_t) sizeof(double *) * t->sons);
    assert(cb->T != NULL);

    for(i=0; i<t->sons; i++) {
      cb->son[i] = NULL;

      if(k > 0 && ksons > 0) {
	cb->T[i] = (double *) malloc((size_t) sizeof(double) * k * ksons);
	assert(cb->T[i] != NULL);
	clear_vector(k * ksons, cb->T[i]);
      }
      else
	cb->T[i] = NULL;
    }

    cb->V = NULL;
  }
  else {
    cb->son = NULL;
    cb->T = NULL;

    if(k > 0) {
      cb->V = (double *) malloc((size_t) sizeof(double) * k * t->size);
      assert(cb->V != NULL);
      clear_vector(k * t->size, cb->V);
    }
    else
      cb->V = NULL;
  }

  cb->t = t;
  cb->k = k;
  cb->kt = 0;
  cb->n = t->size;
  cb->sons = t->sons;

  return cb;
}

void
del_clusterbasis(pclusterbasis cb)
{
  int i;

  assert(cb != NULL);

  for(i=0; i<cb->sons; i++) {
    if(cb->T[i]) {
      free(cb->T[i]);
      cb->T[i] = NULL;
    }

    if(cb->son[i]) {
      del_clusterbasis(cb->son[i]);
      cb->son[i] = NULL;
    }
  }

  if(cb->V) {
    free(cb->V);
    cb->V = NULL;
  }

  if(cb->yt) {
    free(cb->yt);
    cb->yt = NULL;
  }

  if(cb->xt) {
    free(cb->xt);
    cb->xt = NULL;
  }

  free(cb);
}

size_t
getsize_clusterbasis(pcclusterbasis cb){
  size_t size;
  int i;

  size = (sizeof(clusterbasis) +
	  (size_t) sizeof(double) * cb->k +
	  (size_t) sizeof(double) * cb->k);

  if(cb->sons > 0) {
    size += ((size_t) sizeof(pclusterbasis) * cb->sons +
	     (size_t) sizeof(double *) * cb->sons);

    for(i=0; i<cb->sons; i++)
      size += ((size_t) sizeof(double) * cb->k * cb->son[i]->k +
	       getsize_clusterbasis(cb->son[i]));
  }
  else
    size += (size_t) sizeof(double) * cb->k * cb->t->size;

  return size;
}

void
reallocstructure_clusterbasis(pclusterbasis cb, int new_k)
{
  double *tmp;
  int i;

  if(new_k > cb->k) {
    tmp = (double *) malloc((size_t) sizeof(double) * new_k);
    assert(tmp != NULL);
    if(cb->xt != NULL) {
      copy_lapack(cb->kt, cb->xt, tmp);
      free(cb->xt);
    }
    cb->xt = tmp;

    tmp = (double *) malloc((size_t) sizeof(double) * new_k);
    assert(tmp != NULL);
    if(cb->yt != NULL) {
      copy_lapack(cb->kt, cb->yt, tmp);
      free(cb->yt);
    }
    cb->yt = tmp;

    if(cb->sons > 0)
      for(i=0; i<cb->sons; i++) {
	tmp = (double *) malloc((size_t) sizeof(double) * new_k * cb->son[i]->k);
	assert(tmp != NULL);

	if(cb->T[i]) {
	  copy_lapack(cb->kt * cb->son[i]->k, cb->T[i], tmp);
	  free(cb->T[i]);
	}
	cb->T[i] = tmp;
      }
    else {
      tmp = (double *) malloc((size_t) sizeof(double) * new_k * cb->t->size);
      assert(tmp != NULL);

      if(cb->V) {
	copy_lapack(cb->kt * cb->t->size, cb->V, tmp);
	free(cb->V);
      }
      cb->V = tmp;
    }

    cb->k = new_k;
  }
}

void
clearx_clusterbasis(pclusterbasis cb)
{
  pclusterbasis *son = cb->son;
  int sons = cb->sons;
  int i;

  for(i=0; i<cb->kt; i++)
    cb->xt[i] = 0.0;

  for(i=0; i<sons; i++)
    clearx_clusterbasis(son[i]);
}

void
cleary_clusterbasis(pclusterbasis cb)
{
  pclusterbasis *son = cb->son;
  int sons = cb->sons;
  int i;

  for(i=0; i<cb->kt; i++)
    cb->yt[i] = 0.0;

  for(i=0; i<sons; i++)
    cleary_clusterbasis(son[i]);
}

void
clear_clusterbasis(pclusterbasis cb)
{
  pclusterbasis *son = cb->son;
  int sons = cb->sons;
  int i;

  for(i=0; i<cb->kt; i++) {
    cb->xt[i] = 0.0;
    cb->yt[i] = 0.0;
  }

  for(i=0; i<sons; i++)
    clear_clusterbasis(son[i]);
}

void
reset_clusterbasis(pclusterbasis cb)
{
  pclusterbasis *son = cb->son;
  int sons = cb->sons;
  int i;
  
  if(sons > 0) {
    if(cb->V != NULL) {
      free(cb->V);
      
      cb->V = NULL;
    }

    for(i=0; i<sons; i++)
      reset_clusterbasis(son[i]);
  }
}

void
uforward_clusterbasis(pclusterbasis cb, const double *x)
{
  pclusterbasis *son = cb->son;
  int sons = cb->sons;
  int xindex;
  int i;

  xindex = 0;
  for(i=0; i<sons; i++) {
    uforward_clusterbasis(son[i], x + xindex);

    xindex += son[i]->n;
  }
  assert(xindex == cb->n);

  if(cb->kt > 0) {
    if(cb->T) {
      clear_vector(cb->kt, cb->xt);
      
      for(i=0; i<sons; i++)
	if(son[i]->kt > 0)
	  addevaltrans_lapack(son[i]->kt, cb->kt, cb->T[i], son[i]->xt, cb->xt);
    }
    else
      evaltrans_lapack(cb->n, cb->kt, cb->V, x, cb->xt);
  }
}

void
forward_clusterbasis(pclusterbasis cb, const double *x)
{
  pclusterbasis *son = cb->son;
  int sons = cb->sons;
  int xindex;
  int i;

  assert(cb->kt <= cb->k);
  assert(cb->kt > 0);

  clear_vector(cb->kt, cb->xt);

  if(sons > 0) {
    xindex = 0;

    for(i=0; i<sons; i++) {
      forward_clusterbasis(son[i], x + xindex);

      assert(son[i]->kt > 0);
      addevaltrans_lapack(son[i]->kt, cb->kt, cb->T[i], son[i]->xt, cb->xt);

      xindex += son[i]->n;
    }
    assert(xindex == cb->n);
  }
  else
    addevaltrans_lapack(cb->n, cb->kt, cb->V, x, cb->xt);
}

void
ubackward_clusterbasis(pclusterbasis cb, double *y)
{
  pclusterbasis *son = cb->son;
  int sons = cb->sons;
  int yindex;
  int i;

  if(cb->kt > 0) {
    if(cb->T) {
      for(i=0; i<sons; i++)
	if(son[i]->kt > 0)
	  addeval_lapack(son[i]->kt, cb->kt, cb->T[i], cb->yt, son[i]->yt);
    }
    else
      addeval_lapack(cb->n, cb->kt, cb->V, cb->yt, y);
  }

  yindex = 0;
  for(i=0; i<sons; i++) {
    ubackward_clusterbasis(son[i], y + yindex);

    yindex += son[i]->n;
  }
}

void
backward_clusterbasis(pclusterbasis cb, double *y)
{
  pclusterbasis *son = cb->son;
  int sons = cb->sons;
  double **T = cb->T;
  double *V = cb->V;
  int yindex;
  int i;

  assert(cb->kt <= cb->k);
  assert(cb->kt > 0);

  if(sons > 0) {
    yindex = 0;

    for(i=0; i<sons; i++) {
      assert(son[i]->kt > 0);
      addeval_lapack(son[i]->kt, cb->kt, T[i], cb->yt, son[i]->yt);

      backward_clusterbasis(son[i], y + yindex);

      yindex += son[i]->n;
    }
    assert(yindex == cb->n);
  }
  else
    addeval_lapack(cb->n, cb->kt, V, cb->yt, y);
}

double
ugetentry_clusterbasis(pclusterbasis cb, int i)
{
  double *V = cb->V;
  double z;

  assert(cb->kt > 0);
  assert(cb->V != NULL);
  assert(i >= 0 && i < cb->n);

  z = ddot_(&cb->kt, V+i, &cb->n, cb->yt, eins_);

  return z;
}

double
getentry_clusterbasis(pclusterbasis cb, int i)
{
  pclusterbasis *son = cb->son;
  int sons = cb->sons;
  double **T = cb->T;
  double *V = cb->V;
  double z;
  int j, f, s;

  assert(i >= 0 && i < cb->n);

  if(sons > 0) {
    s = f = 0;
    for(j=0; j<sons; j++) {
      f = s;
      s += son[j]->n;
      if(i < s)
	break;
    }
    assert(j < sons);

    addeval_lapack(son[j]->kt, cb->kt, T[j], cb->yt, son[j]->yt);
    z = getentry_clusterbasis(son[j], i-f);
  }
  else
    z = ddot_(&cb->kt, V+i, &cb->n, cb->yt, eins_);

  return z;
}

void
buildV_clusterbasis(pclusterbasis cb)
{
  double **T = cb->T;
  pclusterbasis *son = cb->son;
  int sons = cb->sons;
  int i, ii;

  if(cb->V == NULL) {
    assert(cb->k > 0);

    cb->V = (double *) malloc((size_t) sizeof(double) * cb->n * cb->k);
    assert(cb->V != NULL);
  }

  for(i=0; i<cb->n*cb->kt; i++)
    cb->V[i] = 0.0;
  
  ii = 0;
  for(i=0; i<sons; i++) {
    buildV_clusterbasis(son[i]);
    
    assert(ii + son[i]->n <= cb->n);
    
    dgemm_("Transposed", "Transposed",
	   &cb->n, &cb->kt, &(son[i]->n),
	   deins_,
	   cb->son[i]->V, &(son[i]->n),
	   T[i], &(son[i]->kt),
	   dnull_,
	   cb->V + ii, &cb->n);
    
    ii += son[i]->n;
  }
  assert(ii == cb->n);
}

int
checkortho_clusterbasis(pcclusterbasis cb, double tol)
{
  pclusterbasis *son = cb->son;
  int sons = cb->sons;
  int kt = cb->kt;
  double *G, norm;
  int i, j, ortho;

  assert(kt > 0);

  ortho = 1;

  G = (double *) malloc((size_t) sizeof(double) * kt * kt);
  assert(G != NULL);

  clear_vector(kt * kt, G);

  if(sons > 0) {
    for(i=0; i<sons; i++)
      ortho = ortho && checkortho_clusterbasis(son[i], tol);

    for(i=0; i<sons; i++)
      muladdtrans1_lapack(kt, kt, son[i]->kt,
			  cb->T[i], cb->T[i], G);
  }
  else
    multrans1_lapack(kt, kt, cb->n,
		     cb->V, cb->V, G);

  norm = 0.0;
  for(i=0; i<kt; i++)
    for(j=0; j<kt; j++)
      norm += (G[i+j*kt] - (i==j ? 1.0 : 0.0)) * (G[i+j*kt] - (i==j ? 1.0 : 0.0));

  if(norm > tol*tol)
    ortho = 0;

  free(G);
  G = NULL;

  return ortho;
}

static pclusterbasis
oldortho_clusterbasis(pcclusterbasis cb, pclusteroperator *proj,
		      double **tmp, int *tmpsize,
		      double sqeps, TruncationStrategy strategy)
{
  pclusterbasis cb2;
  double *G;
  int ldG;
  double *Y;
  int ldY;
  double *lambda;
  double *work;
  double h, fmax;
  int lwork, mmax, m;
  int k, k2, kt, kt2;
  int info;
  int i, j;

  assert(cb != NULL);

  cb2 = new_clusterbasis(cb->t, 0, 0);
  if(proj != NULL)
    *proj = new_clusteroperator(cb->t, 0, 0);

  k = cb->kt;

  m = 0;
  Y = NULL;
  ldY = k;

  if(cb->sons > 0) {
    mmax = 0;
    for(i=0; i<cb->sons; i++)
      mmax += cb->son[i]->kt;

    Y = (double *) malloc((size_t) sizeof(double) * ldY * mmax);
    assert(Y != NULL);

    for(i=0; i<cb->sons; i++) {
      k2 = cb->son[i]->kt;

      cb2->son[i] = oldortho_clusterbasis(cb->son[i],
					  (proj == NULL ? NULL : &((*proj)->son[i])),
					  tmp, tmpsize,
					  sqeps, strategy);
      kt2 = cb2->son[i]->kt;

      assert(m + kt2 <= mmax);

      dgemm_("Transposed", "Transposed",
	     &k, &kt2, &k2,
	     deins_,
	     cb->T[i], &k2,
	     *tmp, &kt2,
	     dnull_,
	     Y + m*k, &ldY);

      m += kt2;
    }
  }
  else {
    m = cb->n;

    Y = (double *) malloc((size_t) sizeof(double) * ldY * m);
    assert(Y != NULL);

    for(i=0; i<k; i++)
      dcopy_(&m, cb->V + i*m, eins_, Y + i, &ldY);
  }

  lwork = 6 * k;
  ldG = k;
  provide_array(2 * k * k + k + lwork, tmp, tmpsize);
  G = *tmp + k * k;
  lambda = G + k * k;
  work = lambda + k;

  dgemm_("Not Transposed", "Transposed",
	 &k, &k, &m,
	 deins_,
	 Y, &ldY,
	 Y, &ldY,
	 dnull_,
	 G, &ldG);

  dsyev_("Vectors wanted", "Upper half",
	 &k, G, &ldG, lambda, work, &lwork, &info);
  assert(info == 0);

  for(i=0,j=k-1; i<j; i++,j--) {
    swap_lapack(k, G+i*ldG, G+j*ldG);
    h = dmax(0.0, lambda[i]);
    lambda[i] = dmax(0.0, lambda[j]);
    lambda[j] = h;
  }

  switch(strategy) {
  case HLIB_FROBENIUS_ABSOLUTE:
    h = 0.0;
    kt = k;
    while(kt > 1 && h + lambda[kt-1] < sqeps) {
      kt--;
      h += lambda[kt];
    }
    break;
  default:
  case HLIB_FROBENIUS_RELATIVE:
    fmax = 0.0;
    for(i=0; i<k; i++)
      fmax += lambda[i];

    h = 0.0;
    kt = k;
    while(kt > 1 && h + lambda[kt-1] < sqeps * fmax) {
      kt--;
      h += lambda[kt];
    }
    break;
  case HLIB_EUCLIDEAN_RELATIVE:
    kt = 1;
    while(kt < k && lambda[kt] > sqeps*lambda[0])
      kt++;
    break;
  }

  while(kt > 1 && lambda[kt-1] < ORTHO_EPS)
    kt--;

  assert(0 < kt);
  assert(kt <= k);

  reallocstructure_clusterbasis(cb2, kt);
  cb2->kt = kt;

  for(i=0; i<kt; i++) {
    assert(lambda[i] > 0.0);
    scale_lapack(k, G+i*ldG, 1.0 / sqrt(lambda[i]));
  }

  if(cb->sons > 0) {
    clear_vector(kt * k, *tmp);

    mmax = m;
    m = 0;
    for(i=0; i<cb->sons; i++) {
      kt2 = cb2->son[i]->kt;

      dgemm_("Transposed", "Not Transposed",
	     &kt2, &kt, &k,
	     deins_,
	     Y + m*ldY, &ldY,
	     G, &ldG,
	     dnull_,
	     cb2->T[i], &kt2);

      dgemm_("Transposed", "Transposed",
	     &kt, &k, &kt2,
	     deins_,
	     cb2->T[i], &kt2,
	     Y + m*ldY, &ldY,
	     deins_,
	     *tmp, &kt);

      m += kt2;
    }
    assert(m == mmax);
  }
  else {
    dgemm_("Transposed", "Not Transposed",
	   &m, &kt, &k,
	   deins_,
	   Y, &ldY,
	   G, &ldG,
	   dnull_,
	   cb2->V, &m);

    dgemm_("Transposed", "Transposed",
	   &kt, &k, &m,
	   deins_,
	   cb2->V, &m,
	   Y, &ldY,
	   dnull_,
	   *tmp, &kt);
  }

  if(proj) {
    assert(*proj != NULL);

    reallocstructure_clusteroperator(*proj, kt, k);

    copy_lapack(kt * k, *tmp, (*proj)->P);
  }

  free(Y);
  Y = NULL;

  return cb2;
}

pclusterbasis
oldorthogonalize_clusterbasis(pcclusterbasis cb, pclusteroperator *proj,
			      double eps, TruncationStrategy strategy)
{
  pclusterbasis cb2;
  double *tmp;
  int tmpsize;

  tmp = NULL;
  tmpsize = 0;

  cb2 = oldortho_clusterbasis(cb, proj, &tmp, &tmpsize, eps*eps, strategy);

  if(tmp != NULL) {
    free(tmp);
    tmp = NULL;
  }

  return cb2;
}
			   
static pclusterbasis
ortho_clusterbasis(pcclusterbasis cb, pclusteroperator *proj,
		   double **tmp, int *tmpsize,
		   double sqeps, TruncationStrategy strategy)
{
  pclusterbasis cb2;
  double *G;
  int ldG;
  double *Y;
  int ldY;
  double *Z;
  int ldZ;
  double *lambda;
  double *tau;
  double *work;
  double h, fmax;
  int lwork, mmax, m;
  int k, k2, kt, kt2, evpsize;
  int info;
  int i, j;

  assert(cb != NULL);

  cb2 = new_clusterbasis(cb->t, 0, 0);
  if(proj != NULL)
    *proj = new_clusteroperator(cb->t, 0, 0);

  k = cb->kt;

  m = 0;
  Y = NULL;
  ldY = k;

  if(cb->sons > 0) {
    mmax = 0;
    for(i=0; i<cb->sons; i++)
      mmax += cb->son[i]->kt;

    Y = (double *) malloc((size_t) sizeof(double) * ldY * mmax);
    assert(Y != NULL);

    for(i=0; i<cb->sons; i++) {
      k2 = cb->son[i]->kt;

      cb2->son[i] = ortho_clusterbasis(cb->son[i],
				       (proj == NULL ? NULL : &((*proj)->son[i])),
				       tmp, tmpsize,
				       sqeps, strategy);
      kt2 = cb2->son[i]->kt;

      assert(m + kt2 <= mmax);

      dgemm_("Transposed", "Transposed",
	     &k, &kt2, &k2,
	     deins_,
	     cb->T[i], &k2,
	     *tmp, &kt2,
	     dnull_,
	     Y + m*k, &ldY);

      m += kt2;
    }
  }
  else {
    m = cb->n;

    Y = (double *) malloc((size_t) sizeof(double) * ldY * m);
    assert(Y != NULL);

    for(i=0; i<k; i++)
      dcopy_(&m, cb->V + i*m, eins_, Y + i, &ldY);
  }

  lwork = 6 * k;
  ldG = k;
  evpsize = k;
  ldZ = k;
  provide_array(2 * k * k + m * k + 2 * k + lwork, tmp, tmpsize);
  G = *tmp + k * k;
  Z = G + k * k;
  lambda = Z + m * k;
  tau = lambda + k;
  work = tau + k;

  copy_lapack(m * k, Y, Z);

  if(m < k) {
    dgeqrf_(&k, &m, Y, &ldY, tau, work, &lwork, &info);
    assert(info == 0);

    for(i=0; i<m; i++)
      for(j=i+1; j<m; j++)
	Y[j+i*ldY] = 0.0;

    evpsize = m;
  }

  dgemm_("Not Transposed", "Transposed",
	 &evpsize, &evpsize, &m,
	 deins_,
	 Y, &ldY,
	 Y, &ldY,
	 dnull_,
	 G, &ldG);

  dsyev_("Vectors wanted", "Upper half",
	 &evpsize, G, &ldG, lambda, work, &lwork, &info);
  assert(info == 0);

  for(i=0,j=evpsize-1; i<j; i++,j--) {
    swap_lapack(evpsize, G+i*ldG, G+j*ldG);
    h = dmax(0.0, lambda[i]);
    lambda[i] = dmax(0.0, lambda[j]);
    lambda[j] = h;
  }

  switch(strategy) {
  case HLIB_FROBENIUS_ABSOLUTE:
    h = 0.0;
    kt = evpsize;
    while(kt > 1 && h + lambda[kt-1] < sqeps) {
      kt--;
      h += lambda[kt];
    }
    break;
  default:
  case HLIB_FROBENIUS_RELATIVE:
    fmax = 0.0;
    for(i=0; i<evpsize; i++)
      fmax += lambda[i];

    h = 0.0;
    kt = evpsize;
    while(kt > 1 && h + lambda[kt-1] < sqeps * fmax) {
      kt--;
      h += lambda[kt];
    }
    break;
  case HLIB_EUCLIDEAN_RELATIVE:
    kt = 1;
    while(kt < evpsize && lambda[kt] > sqeps*lambda[0])
      kt++;
    break;
  }

  while(kt > 1 && lambda[kt-1] < ORTHO_EPS)
    kt--;

  assert(0 < kt);
  assert(kt <= evpsize);

  reallocstructure_clusterbasis(cb2, kt);
  cb2->kt = kt;

  for(i=0; i<kt; i++) {
    assert(lambda[i] >= ORTHO_EPS);
    scale_lapack(evpsize, G+i*ldG, 1.0 / sqrt(lambda[i]));
  }

  if(cb->sons > 0) {
    clear_vector(kt * k, *tmp);

    mmax = m;
    m = 0;
    for(i=0; i<cb->sons; i++) {
      kt2 = cb2->son[i]->kt;

      dgemm_("Transposed", "Not Transposed",
	     &kt2, &kt, &evpsize,
	     deins_,
	     Y + m*ldY, &ldY,
	     G, &ldG,
	     dnull_,
	     cb2->T[i], &kt2);

      dgemm_("Transposed", "Transposed",
	     &kt, &k, &kt2,
	     deins_,
	     cb2->T[i], &kt2,
	     Z + m*ldZ, &ldZ,
	     deins_,
	     *tmp, &kt);

      m += kt2;
    }
    assert(m == mmax);
  }
  else {
    dgemm_("Transposed", "Not Transposed",
	   &m, &kt, &evpsize,
	   deins_,
	   Y, &ldY,
	   G, &ldG,
	   dnull_,
	   cb2->V, &m);

    dgemm_("Transposed", "Transposed",
	   &kt, &k, &m,
	   deins_,
	   cb2->V, &m,
	   Z, &ldZ,
	   dnull_,
	   *tmp, &kt);
  }

  if(proj) {
    assert(*proj != NULL);

    reallocstructure_clusteroperator(*proj, kt, k);

    copy_lapack(kt * k, *tmp, (*proj)->P);
  }

  free(Y);
  Y = NULL;

  return cb2;
}

pclusterbasis
orthogonalize_clusterbasis(pcclusterbasis cb, pclusteroperator *proj,
			   double eps, TruncationStrategy strategy)
{
  pclusterbasis cb2;
  double *tmp;
  int tmpsize;

  tmp = NULL;
  tmpsize = 0;

  cb2 = ortho_clusterbasis(cb, proj, &tmp, &tmpsize, eps*eps, strategy);

  if(tmp != NULL) {
    free(tmp);
    tmp = NULL;
  }

  return cb2;
}
			   
pclusteroperator
new_clusteroperator(pccluster t, int kr, int kc)
{
  pclusteroperator co;
  int sons = t->sons;
  int i;

  co = (pclusteroperator) malloc(sizeof(clusteroperator));
  assert(co != NULL);

  if(kr > 0 && kc > 0) {
    co->P = (double *) malloc((size_t) sizeof(double) * kr * kc);
    assert(co->P != NULL);
  }
  else
    co->P = NULL;

  if(sons > 0) {
    co->son = (pclusteroperator *) malloc((size_t) sizeof(pclusteroperator) * sons);
    assert(co->son != NULL);

    for(i=0; i<sons; i++)
      co->son[i] = NULL;
  }
  else
    co->son = NULL;

  co->t = t;
  co->kr = kr;
  co->ktr = 0;
  co->kc = kc;
  co->ktc = 0;
  co->sons = sons;

  return co;
}

void
del_clusteroperator(pclusteroperator co)
{
  int i;

  assert(co != NULL);

  if(co->sons > 0) {
    assert(co->son != NULL);

    for(i=0; i<co->sons; i++)
      if(co->son[i]) {
	del_clusteroperator(co->son[i]);
	co->son[i] = NULL;
      }
    
    free(co->son);
    co->son = NULL;
  }

  if(co->kr > 0 && co->kc > 0) {
    assert(co->P != NULL);
    free(co->P);
    co->P = NULL;
  }

  free(co);
}

size_t
getsize_clusteroperator(pcclusteroperator co)
{
  size_t sz;
  int i;

  sz = (size_t) (sizeof(clusteroperator) +
		 sizeof(double) * co->kr * co->kc);

  if(co->sons > 0) {
    sz += (size_t) sizeof(pclusteroperator) * co->sons;

    for(i=0; i<co->sons; i++)
      sz += getsize_clusteroperator(co->son[i]);
  }

  return sz;
}

void
reallocstructure_clusteroperator(pclusteroperator co, int kr, int kc)
{
  double *tmp;
  int i;

  if(co->kr < kr || co->kc < kc) {
    tmp = (double *) malloc((size_t) sizeof(double) * kr * kc);
    assert(tmp != NULL);

    clear_vector(kr * kc, tmp);

    if(co->P) {
      for(i=0; i<co->ktc; i++)
	dcopy_(&co->ktr, co->P+i*co->ktr, eins_, tmp+i*co->ktr, eins_);
      free(co->P);
      co->P = NULL;
    }

    co->P = tmp;

    co->kr = kr;
    co->kc = kc;
  }
}

static pclusteroperator
compute_product(pcclusterbasis row, pcclusterbasis col, double **p, int *psize)
{
  pclusteroperator co;
  int i;

  assert(col->sons == row->sons);
  assert(col->t == row->t);

  co = new_clusteroperator(col->t, row->kt, col->kt);
  co->ktr = row->kt;
  co->ktc = col->kt;

  if(row->kt > 0 && col->kt > 0) {
    assert(co->P != NULL);

    if(co->sons > 0) {
      assert(co->son != NULL);
      assert(row->T != NULL);
      assert(col->T != NULL);

      clear_vector(co->ktr * co->ktc, co->P);

      for(i=0; i<co->sons; i++) {
	co->son[i] = compute_product(row->son[i], col->son[i], p, psize);

	provide_array(row->son[i]->kt * col->kt, p, psize);

	mul_lapack(row->son[i]->kt, col->kt, col->son[i]->kt,
		   co->son[i]->P, col->T[i], *p);
	muladdtrans1_lapack(row->kt, col->kt, row->son[i]->kt,
			    row->T[i], *p, co->P);
      }
    }
    else {
      assert(co->son == NULL);
      assert(row->V != NULL);
      assert(col->V != NULL);

      multrans1_lapack(row->kt, col->kt, row->t->size,
		       row->V, col->V, co->P);
    }
  }
  return co;
}

pclusteroperator
newproduct_clusteroperator(pcclusterbasis row, pcclusterbasis col)
{
  pclusteroperator co;
  double *p;
  int psize;

  p = NULL;
  psize = 0;

  co = compute_product(row, col, &p, &psize);

  if(p) {
    free(p);
    p = NULL;
  }

  return co;
}
