
/* ------------------------------------------------------------
   This is the file "fullmatrix.c" of the HLib package.

   Copyright 1999 Lars Grasedyck and Steffen Boerm
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#include "basic.h"
#include "fullmatrix.h"
#include "lapack.h"
#include "math.h"
#include "rkmatrix.h"
#include "supermatrix.h"

#include <assert.h>

#define FULL_SVD_EPS 1.0e-16
#define FULL_SVD_EPS2 1.0e-32
#define FULL_SVD_EPS3 1.0e-64

long int allocated_fullmatrices;
long int current_fullmatrices;
double full_abseps;
double full_releps;

double fullmatrix_e_flops;
double fullmatrix_a_flops;
double fullmatrix_m_flops;
double fullmatrix_c_flops;
double fullmatrix_i_flops;
double fullmatrix_r_flops;
double fullmatrix_nrsvd;

void
init_fullmatrices(int verbosity){
  if(verbosity>0)
    printf("Initializing FullMatrices.\n");
  allocated_fullmatrices=0;
  current_fullmatrices=0;
  fullmatrix_e_flops = 0.0;
  fullmatrix_a_flops = 0.0;
  fullmatrix_m_flops = 0.0;
  fullmatrix_c_flops = 0.0;
  fullmatrix_i_flops = 0.0;
  fullmatrix_r_flops = 0.0;
  fullmatrix_nrsvd = 0;
  full_abseps = FULL_SVD_EPS2;
  full_releps = FULL_SVD_EPS;
}

void
reportflops_fullmatrix(){
  printf("Fullmatrix Eval flops = %g\n", fullmatrix_e_flops);
  printf("Fullmatrix Add  flops = %g\n", fullmatrix_a_flops);
  printf("Fullmatrix Mul  flops = %g\n", fullmatrix_m_flops);
  printf("Fullmatrix #Conv = %g\n", fullmatrix_nrsvd);
  printf("Fullmatrix Conv flops = %g\n", fullmatrix_c_flops);
  printf("Fullmatrix Inv  flops = %g\n", fullmatrix_i_flops);
  printf("Fullmatrix Rest flops = %g\n", fullmatrix_r_flops);
}

void
resetflops_fullmatrix(){
  fullmatrix_e_flops = 0.0;
  fullmatrix_a_flops = 0.0;
  fullmatrix_m_flops = 0.0;
  fullmatrix_c_flops = 0.0;
  fullmatrix_i_flops = 0.0;
  fullmatrix_r_flops = 0.0;
  fullmatrix_nrsvd = 0;
}

void
finish_fullmatrices(int verbosity){
  if(verbosity>0){
    printf("Finishing FullMatrices is ");
    if(current_fullmatrices!=0){
      printf("not successful:\n");
      printf(" Still In Use : %ld\n",current_fullmatrices);
    }else{
      printf("successful:\n");
    }
    printf(" Allocated : %ld\n",allocated_fullmatrices);
  }
}

pfullmatrix
new_fullmatrix(int rows_new, int cols_new){
  long int i;
  size_t size;
  pfullmatrix f;
  f = (pfullmatrix) malloc(sizeof(fullmatrix));
  assert(f != 0x0);
  f->rows = rows_new;
  f->cols = cols_new;
  size = (size_t)rows_new * (size_t)cols_new;
  f->e = (double*) malloc(size*sizeof(double));
  assert(f->e != 0x0);
  for(i=0; i<size; i++) f->e[i] = 0.0;
  allocated_fullmatrices++;
  current_fullmatrices++;
  return f;
}

void
setabseps_fullmatrix(double abseps){
  full_abseps = abseps;
}


void
setreleps_fullmatrix(double releps){
  full_releps = releps;
}

size_t
size_fullmatrix(int rows_new, int cols_new){
  size_t size = 0;
  size += sizeof(fullmatrix);
  size += rows_new*cols_new*sizeof(double);
  return size;
}

size_t
getsize_fullmatrix(pfullmatrix f){
  size_t size = 0;
  size += sizeof(fullmatrix);
  size += sizeof(double) * f->rows * f->cols;
  return size;
}

pfullmatrix
copy_fullmatrix(pfullmatrix f){
  int size;
  pfullmatrix fc;
  fc = new_fullmatrix(f->rows,f->cols);
  size = f->rows * f->cols;
  dcopy_(&size,f->e,eins_,fc->e,eins_);
  return fc;
}

pfullmatrix
copystructure_fullmatrix(pfullmatrix f){
  pfullmatrix fc;
  fc = new_fullmatrix(f->rows,f->cols);
  return fc;
}

void
copycontent_fullmatrix(pfullmatrix fsource, pfullmatrix ftarget){
  int size;
  assert(fsource!=0x0);
  assert(ftarget!=0x0);
  assert(fsource->e!=0x0);
  assert(ftarget->e!=0x0);
  assert(fsource->rows == ftarget->rows);
  assert(fsource->cols == ftarget->cols);
  size = fsource->rows * fsource->cols;
  dcopy_(&size,fsource->e,eins_,ftarget->e,eins_);
}

void
del_fullmatrix(pfullmatrix f){
  assert(f!=0x0);
  assert(f->e!=0x0);
  free(f->e);
  free(f);
  f = 0x0;
  current_fullmatrices--;
}

void
init_fullmatrix(pfullmatrix f, double v){
  int i, size;
  assert(f!=0x0);
  assert(f->e!=0x0);
  size = f->rows * f->cols;
  for(i=0; i<size; i++) f->e[i] = 0.0;
}

void 
eval_fullmatrix(pfullmatrix f, double* v, double* w){
  assert(f!=0x0);
  assert(f->e!=0x0);
  assert(v!=0x0);
  assert(w!=0x0);
  dgemv_(ntrans,&f->rows,&f->cols,deins_,f->e,&f->rows,v,eins_,dnull_,w,eins_);
  fullmatrix_e_flops += 2.0 * f->rows * f->cols;
}

void 
addeval_fullmatrix(pfullmatrix f, double* v, double* w){
  assert(f!=0x0);
  assert(f->e!=0x0);
  assert(v!=0x0);
  assert(w!=0x0);
  dgemv_(ntrans,&f->rows,&f->cols,deins_,f->e,&f->rows,v,eins_,deins_,w,eins_);
  fullmatrix_e_flops += 2.0 * f->rows * f->cols;
}

void 
addscaleeval_fullmatrix(pfullmatrix f, double alpha, double *v, double *w){
  assert(f!=0x0);
  assert(f->e!=0x0);
  assert(v!=0x0);
  assert(w!=0x0);
  dgemv_(ntrans, &f->rows, &f->cols,
	 &alpha, f->e, &f->rows, v, eins_,
	 deins_, w, eins_);
  fullmatrix_e_flops += 2.0 * f->rows * f->cols + f->rows;
}

void 
evaltrans_fullmatrix(pfullmatrix f, double* v, double* w){
  assert(f!=0x0);
  assert(f->e!=0x0);
  assert(v!=0x0);
  assert(w!=0x0);
  dgemv_(ttrans,&f->rows,&f->cols,deins_,f->e,&f->rows,v,eins_,dnull_,w,eins_);
  fullmatrix_e_flops += 2.0 * f->rows * f->cols;
}

void 
addevaltrans_fullmatrix(pfullmatrix f, double* v, double* w){
  assert(f!=0x0);
  assert(f->e!=0x0);
  assert(v!=0x0);
  assert(w!=0x0);
  dgemv_(ttrans,&f->rows,&f->cols,deins_,f->e,&f->rows,v,eins_,deins_,w,eins_);
  fullmatrix_e_flops += 2.0 * f->rows * f->cols;
}

void 
subeval_fullmatrix(pfullmatrix f, double* v, double* w){
  assert(f!=0x0);
  assert(f->e!=0x0);
  assert(v!=0x0);
  assert(w!=0x0);
  dgemv_(ntrans,&f->rows,&f->cols,meins_,f->e,&f->rows,v,eins_,deins_,w,eins_);
  fullmatrix_e_flops += 2.0 * f->rows * f->cols; 
}

void 
subevaltrans_fullmatrix(pfullmatrix f, double* v, double* w){
  assert(f!=0x0);
  assert(f->e!=0x0);
  assert(v!=0x0);
  assert(w!=0x0);
  dgemv_(ttrans,&f->rows,&f->cols,meins_,f->e,&f->rows,v,eins_,deins_,w,eins_);
  fullmatrix_e_flops += 2.0 * f->rows * f->cols; 
}

void
generaleval_fullmatrix(pfullmatrix f, int uplo, int trans, int add,
		       int diag, double alpha, double *v, double *w){
  double target_scale = 0.0;
  char u='L',t='N',d='N';
  double *v2;
  
  if(uplo==2) u='U';
  if(trans==1) t='T';
  if(diag==1) d='U';
  if(add==1) target_scale = 1.0;
  if(uplo==0){
    dgemv_(&t,&f->rows,&f->cols,&alpha,f->e,&f->rows,
	   v,eins_,&target_scale,w,eins_);
  }else{
    if(add==0){
      copy_lapack(f->rows,v,w);
      scale_lapack(f->rows,w,alpha);
      dtrmv_(&u,&t,&d,&f->rows,f->e,&f->rows,w,eins_);
    }else{
      v2 = (double*) malloc(f->rows * sizeof(double));
      copy_lapack(f->rows,v,v2);
      dtrmv_(&u,&t,&d,&f->rows,f->e,&f->rows,v2,eins_);
      daxpy_(&f->rows,&alpha,v2,eins_,w,eins_);
      free(v2);
    }
  }
  fullmatrix_e_flops += 2.0 * f->rows * f->cols; 
}

void
lowersolve_fullmatrix(pfullmatrix f, double *v)
{
  assert(f != 0x0);
  assert(f->e != 0x0);
  assert(f->rows == f->cols);
  assert(v != 0x0);
  lowersolve_lapack(f->rows, f->e, v);
}

void
uppersolve_fullmatrix(pfullmatrix f, double *v)
{
  assert(f != 0x0);
  assert(f->e != 0x0);
  assert(f->rows == f->cols);
  assert(v != 0x0);
  uppersolve_lapack(f->rows, f->e, v);
}

double
getentry_fullmatrix(pfullmatrix f, int row, int col){
  return f->e[row + col*f->rows];
}

void
setentry_fullmatrix(pfullmatrix f, int row, int col, double value){
  assert(row>=0 && col>=0);
  assert(f->rows>row && f->cols>col);
  f->e[row + col*f->rows] = value;
}


void
scale_fullmatrix(pfullmatrix f, double v){
  int size;
  assert(f!=0x0);
  assert(f->e!=0x0);
  size = f->rows*f->cols;
  dscal_(&size,&v,f->e,eins_);
  fullmatrix_r_flops += size;
}

double
normoo_fullmatrix(pfullmatrix f){
  int i,rows,cols;
  double diff,diffmax=0.0;
  assert(f!=0x0);
  assert(f->e!=0x0);
  rows = f->rows;
  cols = f->cols;
  for(i=0; i<rows; i++){
    diff = dasum_(&cols,&f->e[i],&rows);
    if(diff>diffmax) diffmax = diff;
  }
  return diffmax;
}

void
add_fullmatrix(pfullmatrix target, pfullmatrix addend1, pfullmatrix addend2){
  int size;
  assert(target != 0x0);
  assert(target->e != 0x0);
  assert(addend1 != 0x0);
  assert(addend2 != 0x0);
  assert(addend1->e != 0x0);
  assert(addend2->e != 0x0);
  assert(target->rows == addend1->rows);
  assert(target->rows == addend2->rows);
  assert(target->cols == addend1->cols);
  assert(target->cols == addend2->cols);
  size = target->rows * target->cols;
  dcopy_(&size,addend1->e,eins_,target->e,eins_);
  daxpy_(&size,deins_,addend2->e,eins_,target->e,eins_);
  fullmatrix_a_flops += size;
}

void
addfull2_fullmatrix(pfullmatrix target, pfullmatrix f){
  int size;
  assert(target != 0x0);
  assert(target->e != 0x0);
  assert(f != 0x0);
  assert(f->e != 0x0);
  assert(target->rows == f->rows);
  assert(target->cols == f->cols);
  size = target->rows * target->cols;
  daxpy_(&size,deins_,f->e,eins_,target->e,eins_);
  fullmatrix_a_flops += size;
}

void
addfull2_rkmatrix(pfullmatrix f, prkmatrix rt, int  row_offset, int col_offset){
  int rows = rt->rows;
  int cols = rt->cols;
  int size = rows*cols;
  int minsize;
  int i,j;
  double *u,*v,*s,*work,*tmp;
  int lwork;
  int info;
  double norm;
  pfullmatrix ff;

  minsize = rows; if(cols<rows) minsize=cols;
  /* compute f+rt in fullmatrix representation */
  tmp = (double*) malloc(size*sizeof(double));
  assert(tmp!=0x0);
  if(row_offset==0 && col_offset==0 && f->rows==rows && f->cols==cols){
    dcopy_(&size,f->e,eins_,tmp,eins_);
  }else{
    for(i=0; i < cols; i++)
     dcopy_(&rows,&f->e[(i+col_offset)*f->rows + row_offset],eins_,&tmp[i*rows],eins_);
  }
  if(rt->kt>0){
    dgemm_(ntrans,ttrans,&rows,&cols,&rt->kt,deins_,
	   rt->a,&rows,rt->b,&cols,deins_,
	   tmp,&rows);
    fullmatrix_c_flops += 2.0 * size * rt->kt;
  }

  info = 0;
  /* count nonzero rows without squaring entries */
  for(i=0; i<rows; i++){
    norm = 0.0;
    for(j=0; j<cols; j++) norm += fabs(tmp[i+j*rows]);
    if(norm <= 0.9*FULL_SVD_EPS){
      info = 1;
      i = rows;
    }
  }
    /* count nonzero columns without squaring entries */
  if(info==0){
    for(j=0; j<cols; j++){
      norm = 0.0;
      for(i=0; i<rows; i++) norm += fabs(tmp[i+j*rows]);
      if(norm <= 0.9*FULL_SVD_EPS2){
	info = 1;
	j = cols;
      }
    }
  }
  if(info==1){
    /* eliminate zero rows or columns */
    ff = (pfullmatrix) malloc(sizeof(fullmatrix));
    ff->rows = rows;
    ff->cols = cols;
    ff->e = tmp;
    fastfull2_rkmatrix(ff,rt);
    free(ff);
  }else{
    fullmatrix_nrsvd++;
    lwork = 8*(rows+cols); 
    work = (double*) malloc((lwork + rows*minsize + cols*minsize + minsize)
			    *sizeof(double));
    assert(work!=0x0);
    u = &work[lwork];
    v = &work[lwork + rows*minsize];
    s = &work[lwork + rows*minsize + cols*minsize];

    dgesvd_("S","S",&rows,&cols,tmp,&rows,s,u,&rows,v,&minsize,
	    work,&lwork,&info);
    fullmatrix_c_flops += 60 * size * minsize;
    if(info!=0){
      printf("add_full2rkmatrix: info(dgesvd)=%d\n",info);
      exit(0);
    }

    if(rk_arithmetics_adaptive()==1 && rt->local_eps>=0){
      j=0;
      for(i=0; i < rows && i < cols; i++){
	if(s[i] > (rt->local_eps * s[0]) && s[i]>full_abseps) j++;
      }
      if(j>rt->k){
	reallocstructure_rkmatrix(rt,j);
      }
    }
    rt->kt=0;
    size = rows*rt->k; for(i=0; i<size; i++) rt->a[i]=0.0;
    size = cols*rt->k; for(i=0; i<size; i++) rt->b[i]=0.0;
    for(i=0; i < rt->k && i < minsize && s[i]>full_releps*s[0] &&
	  s[i]>full_abseps; i++){
      norm = sqrt(s[i]);
      daxpy_(&rows,&norm,&u[i*rows],eins_,&rt->a[i*rows],eins_);
      daxpy_(&cols,&norm,&v[i],&minsize,&rt->b[i*cols],eins_);
      rt->kt++;
      fullmatrix_c_flops += 2.0 * (rows+cols);
    }
    free(work);
  }
  free(tmp);
}

void
fastfull2_rkmatrix(pfullmatrix f, prkmatrix rt){
  int rows,cols,size;
  int minsize;
  int i,j;
  prkmatrix r_tmp;
  int nz_rows=0;
  int nz_cols=0;
  int *nz_row;
  int *nz_col;
  double norm;

  assert(rt != 0x0);
  assert(f != 0x0);
  assert(f->rows == rt->rows);
  assert(f->cols == rt->cols);
  rows = rt->rows;
  cols = rt->cols;
  size = rows*cols;
  
  minsize = rows; if(cols<rows) minsize=cols;
  nz_row = (int*) malloc((rows+cols)*sizeof(int));
  assert(nz_row != 0x0);
  nz_col = &nz_row[rows];
  
  /* count nonzero rows without squaring entries */
  for(i=0; i<rows; i++){
    norm = 0.0;
    nz_row[i] = 0;
    for(j=0; j<cols; j++) norm += fabs(f->e[i+j*rows]);
    if(norm > FULL_SVD_EPS){
      nz_rows++;
      nz_row[i] = 1;
    }
  }
  /* count nonzero columns without squaring entries */
  for(j=0; j<cols; j++){
    norm = 0.0;
    nz_col[j] = 0;
    for(i=0; i<rows; i++) norm += fabs(f->e[i+j*rows]);
    if(norm > FULL_SVD_EPS){
      nz_cols++;
      nz_col[j] = 1;
    }
  }
  minsize = nz_rows; if(nz_cols<minsize) minsize = nz_cols;

  if(minsize==0){
    rt->kt=0;
    return;
  }
    
  /* just copy the minnm nonzero cols/rows */
  if(minsize <= rt->k){
    rt->kt = minsize;
    if(nz_rows<nz_cols){
      for(j=0,i=0; i<rows; i++){
	if(nz_row[i]==1){
	  dcopy_(&cols,&f->e[i],&rows,&rt->b[j*cols],eins_);
	  dscal_(&rows,dnull_,&rt->a[j*rows],eins_);
	  rt->a[i+rows*j]=1.0;
	  j++;
	}
      }
    }else{
      for(j=0,i=0; i<cols; i++){
	if(nz_col[i]==1){
	  dcopy_(&rows,&f->e[i*rows],eins_,&rt->a[j*rows],eins_);
	  dscal_(&cols,dnull_,&rt->b[j*cols],eins_);
	  rt->b[i+cols*j]=1.0;
	  j++;
	}
      }
    }
  }else{
  /* call the formatted R(k) truncation */
    r_tmp = new_rkmatrix(minsize,rows,cols);
    r_tmp->kt = minsize;
    if(nz_rows < nz_cols){
      for(j=0,i=0; i<rows; i++){
	if(nz_row[i]==1){
	  dcopy_(&cols,&f->e[i],&rows,&r_tmp->b[j*cols],eins_);
	  dscal_(&rows,dnull_,&r_tmp->a[j*rows],eins_);
	  r_tmp->a[i+rows*j]=1.0;
	  j++;
	}
      }
    }else{
      for(j=0,i=0; i<cols; i++){
	if(nz_col[i]==1){
	  dcopy_(&rows,&f->e[i*rows],eins_,&r_tmp->a[j*rows],eins_);
	  dscal_(&cols,dnull_,&r_tmp->b[j*cols],eins_);
	  r_tmp->b[i+cols*j]=1.0;
	  j++;
	}
      }
    }
    rt->kt = 0;
    addpart2_rkmatrix(rt,0,0,r_tmp,0,0,rows,cols);
    del_rkmatrix(r_tmp);
    
  }
  
  free(nz_row);
}

void  
addfullpart2_rkmatrix(pfullmatrix f, prkmatrix r, int row_offset, int col_offset){
  int i;
  pfullmatrix tmp; 
  assert(r != 0x0);
  tmp = new_fullmatrix(r->rows,r->cols); 
  for(i=0; i<r->cols; i++)
    dcopy_(&r->rows,&f->e[row_offset+(i+col_offset)*f->rows],eins_,
	   &tmp->e[i*r->rows],eins_);
  addfull2_rkmatrix(tmp,r,0,0);
  del_fullmatrix(tmp);
}

void
mul_fullmatrix(pfullmatrix fc, pfullmatrix fa, pfullmatrix fb){
  assert(fc!=0x0);
  assert(fc->e!=0x0);
  assert(fa!=0x0);
  assert(fa->e!=0x0);
  assert(fb!=0x0);
  assert(fb->e!=0x0);
  assert(fc->rows == fa->rows);
  assert(fa->cols == fb->rows);
  assert(fb->cols == fc->cols);
  if(fc->rows>0 && fc->cols>0 && fa->cols>0)
    dgemm_(ntrans,ntrans,&fc->rows,&fc->cols,&fa->cols,deins_,fa->e,
	   &fc->rows,fb->e,&fb->rows,dnull_,fc->e,&fc->rows);
  fullmatrix_m_flops += 2.0 * fa->rows * fa->cols * fb->cols;
}

void
muladd_fullmatrix(pfullmatrix fc, pfullmatrix fa, pfullmatrix fb){
  assert(fc!=0x0);
  assert(fc->e!=0x0);
  assert(fa!=0x0);
  assert(fa->e!=0x0);
  assert(fb!=0x0);
  assert(fb->e!=0x0);
  assert(fc->rows == fa->rows);
  assert(fa->cols == fb->rows);
  assert(fb->cols == fc->cols);
  if(fc->rows>0 && fc->cols>0 && fa->cols>0)
    dgemm_(ntrans,ntrans,&fc->rows,&fc->cols,&fa->cols,deins_,fa->e,
	   &fc->rows,fb->e,&fb->rows,deins_,fc->e,&fc->rows);
  fullmatrix_m_flops += 2.0 * fa->rows * fa->cols * fb->cols;
}

pfullmatrix
getmul_fullmatrix(psupermatrix a, psupermatrix b){
  pfullmatrix fa,fb,tmp=0x0;
  psupermatrix stmp;
  int i,rows,cols,size;
  double *v, *w;
  assert(a!=0x0);
  assert(b!=0x0);
  assert(a->cols == b->rows);
  
  fa = a->f;
  fb = b->f;
  rows = a->rows;
  size = a->cols;
  cols = b->cols;

  if(fa!=0x0 && fb!=0x0){
    assert(fa->e!=0x0);
    assert(fb->e!=0x0);
    tmp = new_fullmatrix(rows,cols);
    if(rows>0 && cols>0 && a->cols>0)
      dgemm_(ntrans,ntrans,&rows,&cols,&size,deins_,fa->e,&rows,fb->e,&size,
	     dnull_,tmp->e,&rows);
    fullmatrix_m_flops += 2.0 * rows * size * cols;
  }else{
    if(fa!=0x0 && fb==0x0){
      v = (double*) malloc(size*sizeof(double)); 
      assert(v!=0x0);
      w = (double*) malloc(cols*sizeof(double)); 
      assert(w!=0x0);
      tmp = new_fullmatrix(rows,cols);
      for(i=0; i < rows; i++){
	dcopy_(&size,&fa->e[i],&rows,v,eins_);
	evaltrans_supermatrix(b,v,w);
	dcopy_(&cols,w,eins_,&tmp->e[i],&rows);
      }
      free(v);
      free(w);
    }
    if(fa==0x0 && fb!=0x0){
      assert(fb->e!=0x0);
      tmp = new_fullmatrix(rows,cols);
      for(i=0; i < cols; i++){
	eval_supermatrix(a,&fb->e[i*size],&tmp->e[i*rows]);
      }
    }
    if(fa==0x0 && fb==0x0){
      if(rows<cols){
	fa = new_fullmatrix(rows,size);
	stmp = new_supermatrix(1,1,rows,size,0x0,0x0,fa);
	convertsuper2_fullmatrix(fa,a);
	tmp = getmul_fullmatrix(stmp,b);
	del_supermatrix(stmp);
      }else{
	fb = new_fullmatrix(size,cols);
	stmp = new_supermatrix(1,1,size,cols,0x0,0x0,fb);
	convertsuper2_fullmatrix(fb,b);
	tmp = getmul_fullmatrix(a,stmp);
	del_supermatrix(stmp);
      }
    }
  }
  return tmp;
}

pfullmatrix
getgeneralmul_fullmatrix(psupermatrix a, psupermatrix b, 
		       int transpose_a, int transpose_b){
  pfullmatrix fa,fb,tmp=0x0;
  psupermatrix stmp;
  int i,rows,cols,size;
  double *v, *w;
  assert(a!=0x0);
  assert(b!=0x0);
  

  fa = a->f;
  fb = b->f;

  if(transpose_a==0){
    rows = a->rows;
    size = a->cols;  
  }else{
    rows = a->cols;
    size = a->rows;  
  }

  if(transpose_b==0){
    cols = b->cols;
  }else{
    cols = b->rows;
  }

  if(fa!=0x0 && fb!=0x0){
    assert(fa->e!=0x0);
    assert(fb->e!=0x0);
    tmp = new_fullmatrix(rows,cols);
    if(transpose_a==0){
      if(transpose_b==0){
	mul_lapack(rows,cols,size,fa->e,fb->e,tmp->e);
      }else{
	multrans2_lapack(rows,cols,size,fa->e,fb->e,tmp->e);
      }
    }else{
      if(transpose_b==0){
	multrans1_lapack(rows,cols,size,fa->e,fb->e,tmp->e);
      }else{
	multrans12_lapack(rows,cols,size,fa->e,fb->e,tmp->e);
      }
    }
    /*fullmatrix_m_flops += 2.0 * rows * size * cols;*/
  }else{
    if(fa!=0x0 && fb==0x0){
      v = (double*) malloc(size*sizeof(double)); 
      assert(v!=0x0);
      w = (double*) malloc(cols*sizeof(double)); 
      assert(w!=0x0);
      tmp = new_fullmatrix(rows,cols);
      for(i=0; i<rows; i++){
	if(transpose_a==0){
	  dcopy_(&size,&fa->e[i],&rows,v,eins_);
	}else{
	  dcopy_(&size,&fa->e[i*size],eins_,v,eins_);
	}
	if(transpose_b==0){
	  evaltrans_supermatrix(b,v,w);
	}else{
	  eval_supermatrix(b,v,w);
	}
	dcopy_(&cols,w,eins_,&tmp->e[i],&rows);
      }
      free(v);
      free(w);
    }
    if(fa==0x0 && fb!=0x0){
      assert(fb->e!=0x0);
      tmp = new_fullmatrix(rows,cols);
      v = (double*) malloc(size*sizeof(double)); 
      assert(v!=0x0);
      for(i=0; i < cols; i++){
	if(transpose_b==0){
	  dcopy_(&size,&fb->e[i*fb->rows],eins_,v,eins_);
	}else{
	  dcopy_(&size,&fb->e[i],&fb->rows,v,eins_);
	}
	if(transpose_a==0){
	  eval_supermatrix(a,v,&tmp->e[i*rows]);
	}else{
	  evaltrans_supermatrix(a,v,&tmp->e[i*rows]);
	}
      }
      free(v);
  
    }
    if(fa==0x0 && fb==0x0){
      if(rows<cols){
	fa = new_fullmatrix(a->rows,a->cols);
	stmp = new_supermatrix(1,1,fa->rows,fa->cols,0x0,0x0,fa);
	convertsuper2_fullmatrix(fa,a);
	tmp = getgeneralmul_fullmatrix(stmp,b,transpose_a,transpose_b);
	del_supermatrix(stmp);
      }else{
	fb = new_fullmatrix(b->rows,b->cols);
	stmp = new_supermatrix(1,1,fb->rows,fb->cols,0x0,0x0,fb);
	convertsuper2_fullmatrix(fb,b);
	tmp = getgeneralmul_fullmatrix(a,stmp,transpose_a,transpose_b);
	del_supermatrix(stmp);
      }
    }
  }
  return tmp;
}

void
invert_fullmatrix(pfullmatrix fi, pfullmatrix f, pfullmatrix fw){
  int rows,cols,size,info;
  int *ipiv;
  double *tmp;

  assert(f != 0x0);
  assert(fi != 0x0);
  assert(fi->rows == fi->cols);
  assert(f->rows == f->cols);
  assert(fi->rows == f->rows);
  rows = f->rows;
  cols = f->cols;
  size = rows*cols;
  ipiv = (int*) malloc(rows*sizeof(int));
  assert(ipiv!=0x0);
  assert(f->e != 0x0);
  assert(fi->e != 0x0);
  assert(fw->e != 0x0);
  dcopy_(&size,f->e,eins_,fi->e,eins_);
  dgetrf_(&rows,&cols,fi->e,&rows,ipiv,&info);
  if(info!=0){
    printf("invert_fullmatrix: info(dgetrf)=%d\n",info);
    exit(0);
  }
  if(fw!=0x0 && fw->rows==f->rows && fw->cols==f->cols){
    dgetri_(&rows,fi->e,&rows,ipiv,fw->e,&size,&info);
    if(info!=0){
      printf("invert_fullmatrix: info(dgetri,1)=%d\n",info);
      exit(0);
    }
  }else{
    tmp = (double*) malloc(size*sizeof(double));
    assert(tmp != 0x0);
    dgetri_(&rows,fi->e,&rows,ipiv,tmp,&size,&info);
    if(info!=0){
      printf("invert_fullmatrix: info(dgetri,2)=%d\n",info);
      exit(0);
    }
    free(tmp);
  }
  free(ipiv);
  fullmatrix_i_flops += rows * cols * rows;
}

void
outputleaf_fullmatrix(pfullmatrix f, FILE* datei, pblockcluster bc,
		      int row_offset, int col_offset, 
		      int row_max, int col_max, 
		      int rank, int fill, int border, double tol){
  int rows,cols,size;
  double norm;
  prkmatrix r;
  assert(f != 0x0);
  rows = f->rows;
  cols = f->cols;
  size = rows; if(size>cols) size=cols;
  if(fill > 1){
    r = new_rkmatrix(size,rows,cols);
    addfull2_rkmatrix(f,r,0,0);  
    outputleaf_rkmatrix(r,datei,bc,row_offset,col_offset,row_max,col_max,
			rank,fill,border,tol);
    del_rkmatrix(r);
  }else{
    norm = normoo_fullmatrix(f);
    if(norm<1e-16) return;
    if(fill>0){
      /* fill red(fr) inside */
      fprintf(datei,"%.3f %.3f m\n",col_offset+0.0*cols,
	      row_max-rows-row_offset+0.0*rows);
      fprintf(datei,"%.3f %.3f l\n",col_offset+0.0*cols,
	      row_max-rows-row_offset+1.0*rows);
      fprintf(datei,"%.3f %.3f l\n",col_offset+1.0*cols,
	      row_max-rows-row_offset+1.0*rows);
      fprintf(datei,"%.3f %.3f l\n",col_offset+1.0*cols,
	      row_max-rows-row_offset+0.0*rows);
      fprintf(datei,"%.3f %.3f l\n",col_offset+0.0*cols,
	      row_max-rows-row_offset+0.0*rows);
      fprintf(datei,"fi\n");
      fprintf(datei,"cs\n");
    }
    if(border>0){
      /* black(bl) border */
      fprintf(datei,"bl\n");
      fprintf(datei,"%d %d m\n",col_offset,     row_max-row_offset-rows);
      fprintf(datei,"%d %d l\n",col_offset,     row_max-row_offset);
      fprintf(datei,"%d %d l\n",col_offset+cols,row_max-row_offset);
      fprintf(datei,"%d %d l\n",col_offset+cols,row_max-row_offset-rows);
      fprintf(datei,"%d %d l\n",col_offset,     row_max-row_offset-rows);
      fprintf(datei,"cs\n");
    }
    if(rank > 0){
      /* print min(rows,cols) */
      fprintf(datei,"bl\n");
      fprintf(datei,"hf %.1f sf\n",size*0.5);
      fprintf(datei,"%.1f %.1f m (%d) show\n",col_offset+cols*0.2,
	      row_max-rows-row_offset+rows*0.1,size);
    }
  }  
}

