
/* ------------------------------------------------------------
   This is the file "h2conversion.c" of the HLib package.

   Copyright 1999 Steffen Boerm and Lars Grasedyck
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#include "h2conversion.h"

/* Display local singular values for each cluster
#define DEBUG_LOCAL_SVD
*/

#include <math.h>
#ifdef DEBUG_LOCAL_SVD
#include <stdio.h>
#endif

typedef struct _convblock convblock;
typedef convblock *pconvblock;

typedef struct _conversion conversion;
typedef conversion *pconversion;

struct _convblock {
  psupermatrix original;
  pcclusteroperator gram;
  
  double *a;
  double *gb;
  int k;

  int start;
  int size;

  pconvblock next;
};

struct _conversion {
  double *evp;
  int evpsize;
  double *tmp;
  int tmpsize;

  int kmax;
  double sqeps;
  int transposed;
  TruncationStrategy strategy;
};

static void
del_block(pconvblock b)
{
  assert(b != NULL);
  assert(b->a != NULL);
  assert(b->gb != NULL);

  free(b->gb);
  b->gb = NULL;

  free(b->a);
  b->a = NULL;

  free(b);
}

static void
splitrow_supermatrix(psupermatrix s, int sstart,
		     pconvblock *super, pconvblock *rk,
		     int start, int size)
{
  pconvblock b;
  int i, j, off;
  
  assert(sstart <= start);
  assert(start + size <= sstart + s->rows);
  assert(s->row == NULL && s->col == NULL);

  if(s->rows > size || (s->s && s->block_rows == 1)) {
    assert(s->s != NULL);

    off = sstart;
    for(i=0; i<s->block_rows; i++) {
      if(start + size <= off + s->s[i]->rows)
	break;
      off += s->s[i]->rows;
    }

    assert(i < s->block_rows);

    for(j=0; j<s->block_cols; j++)
      splitrow_supermatrix(s->s[i+j*s->block_rows], off,
			   super, rk, start, size);
  }
  else {
    assert(sstart == start);

    if(s->s) {
      b = (pconvblock) malloc(sizeof(convblock));
      assert(b != NULL);

      b->original = s;
      b->gram = NULL;
      b->a = NULL;
      b->gb = NULL;
      b->k = 0;
      b->start = sstart;
      b->size = 0;

      b->next = *super;
      *super = b;
    }
    else if(s->r) {
      if(s->r->kt > 0) {
	b = (pconvblock) malloc(sizeof(convblock));
	assert(b != NULL);
	
	b->original = NULL;
	b->gram = NULL;
	b->a = (double *) malloc((size_t) sizeof(double) * s->r->rows * s->r->kt);
	assert(b->a != NULL);
	b->gb = (double *) malloc((size_t) sizeof(double) * s->r->kt * s->r->kt);
	assert(b->gb != NULL);
	b->k = s->r->kt;
	b->start = sstart;
	b->size = s->r->rows;
	
	b->next = *rk;
	*rk = b;
	
	copy_lapack(s->r->rows * s->r->kt, s->r->a, b->a);
	multrans1_lapack(s->r->kt, s->r->kt, s->r->cols, s->r->b, s->r->b, b->gb);
      }
    }
    else
      assert(s->f != NULL);
  }
}

static void
splitcol_supermatrix(psupermatrix s, int sstart,
		     pconvblock *super, pconvblock *rk,
		     int start, int size)
{
  pconvblock b;
  int i, j, off;
  
  assert(sstart <= start);
  assert(start + size <= sstart + s->cols);
  assert(s->row == NULL && s->col == NULL);

  if(s->cols > size || (s->s && s->block_cols == 1)) {
    assert(s->s != NULL);

    off = sstart;
    for(i=0; i<s->block_cols; i++) {
      if(start + size <= off + s->s[i*s->block_rows]->cols)
	break;
      off += s->s[i]->cols;
    }

    assert(i < s->block_cols);

    for(j=0; j<s->block_rows; j++)
      splitcol_supermatrix(s->s[j+i*s->block_rows], off,
			   super, rk, start, size);
  }
  else {
    assert(sstart == start);

    if(s->s) {
      b = (pconvblock) malloc(sizeof(convblock));
      assert(b != NULL);

      b->original = s;
      b->gram = NULL;
      b->a = NULL;
      b->gb = NULL;
      b->k = 0;
      b->start = sstart;
      b->size = 0;

      b->next = *super;
      *super = b;
    }
    else if(s->r) {
      if(s->r->kt > 0) {
	b = (pconvblock) malloc(sizeof(convblock));
	assert(b != NULL);
	
	b->original = NULL;
	b->gram = NULL;
	b->a = (double *) malloc((size_t) sizeof(double) * s->r->cols * s->r->kt);
	assert(b->a != NULL);
	b->gb = (double *) malloc((size_t) sizeof(double) * s->r->kt * s->r->kt);
	assert(b->gb != NULL);
	b->k = s->r->kt;
	b->start = sstart;
	b->size = s->r->cols;
	
	b->next = *rk;
	*rk = b;
	
	copy_lapack(s->r->cols * s->r->kt, s->r->b, b->a);
	multrans1_lapack(s->r->kt, s->r->kt, s->r->rows, s->r->a, s->r->a, b->gb);
      }
    }
    else
      assert(s->f != NULL);
  }
}

static pclusterbasis
build_basis(pccluster t, pconvblock super, pconvblock rk, pconversion aux)
{
  pclusterbasis cb;
  double *G;
  double *lambda;
  double *work;
  int start = t->start;
  int size = t->size;
  int kmax = aux->kmax;
  double sqeps = aux->sqeps;
  int lwork, info;
  double h, fmax;
  pconvblock super1;
  pconvblock rk1;
  psupermatrix s;
  pconvblock b;
  int m_tau, kt;
  int i, j, off, off2;

  /* Create new clusterbasis */
  cb = new_clusterbasis(t, 0, 0);

  /* Add new entries to block list */
  super1 = NULL;
  rk1 = rk;
  b = super;
  while(b) {
    s = b->original;

    if(aux->transposed)
      splitcol_supermatrix(s, b->start, &super1, &rk1,
			   t->start, t->size);
    else
      splitrow_supermatrix(s, b->start, &super1, &rk1,
			   t->start, t->size);

    b = b->next;
  }

  /* Determine size of EVP and build sons bases if necessary */
  m_tau = 0;
  if(t->sons > 0)
    for(i=0; i<t->sons; i++) {
      cb->son[i] = build_basis(t->son[i], super1, rk1, aux);
      
      m_tau += cb->son[i]->kt;
    }
  else
    m_tau = size;

  /* Provide storage for EVP */
  provide_array(m_tau * (m_tau+7), &aux->evp, &aux->evpsize);
  G = aux->evp;
  lambda = aux->evp + m_tau * m_tau;
  work = aux->evp + m_tau * (m_tau+1);
  lwork = 6*m_tau;

  /* Initialize Gram matrix */
  clear_vector(m_tau * m_tau, G);

  if(t->sons > 0) {
    /* Build transformed Gram matrix */
    b = rk1;
    while(b) {
      provide_array(m_tau * b->k, &aux->tmp, &aux->tmpsize);
      
      off = start;
      off2 = 0;
      for(i=0; i<t->sons; i++) {
	dgemm_("Not Transposed", "Not Transposed",
	       &cb->son[i]->kt, &b->k, &b->k,
	       deins_,
	       b->a + off - b->start, &b->size,
	       b->gb, &b->k,
	       dnull_,
	       aux->tmp + off2, &m_tau);
	off += t->son[i]->size;
	off2 += cb->son[i]->kt;
      }
      assert(off == start + size);
      assert(off2 == m_tau);

      off = start;
      off2 = 0;
      for(i=0; i<t->sons; i++) {
	dgemm_("Not Transposed", "Transposed",
	       &m_tau, &cb->son[i]->kt, &b->k,
	       deins_,
	       aux->tmp, &m_tau,
	       b->a + off - b->start, &b->size,
	       deins_,
	       G + off2 * m_tau, &m_tau);
	off += t->son[i]->size;
	off2 += cb->son[i]->kt;
      }

      b = b->next;
    }
  }
  else {
    /* Build Gram matrix */
    b = rk1;
    while(b) {
      provide_array(m_tau * b->k, &aux->tmp, &aux->tmpsize);

      dgemm_("Not Transposed", "Not Transposed",
	     &m_tau, &b->k, &b->k,
	     deins_,
	     b->a + start - b->start, &b->size,
	     b->gb, &b->k,
	     dnull_,
	     aux->tmp, &m_tau);

      dgemm_("Not Transposed", "Transposed",
	     &m_tau, &m_tau, &b->k,
	     deins_,
	     aux->tmp, &m_tau,
	     b->a + start - b->start, &b->size,
	     deins_,
	     G, &m_tau);

      b = b->next;
    }
  }

  /* Solve eigenvalue problem */
  dsyev_("Vectors", "Upper", &m_tau, G, &m_tau,
	 lambda, work, &lwork, &info);
  assert(info == 0);

  /* Sort eigenvalues and eigenvectors */
  for(i=0,j=m_tau-1; i<j; i++,j--) {
    swap_lapack(m_tau, G+i*m_tau, G+j*m_tau);
    h = lambda[i];
    lambda[i] = lambda[j];
    lambda[j] = h;
  }

  /* Determine appropriate rank */
  switch(aux->strategy) {
  case HLIB_FROBENIUS_ABSOLUTE:
    h = 0.0;
    kt = m_tau;
    while(kt > 1 && h + fabs(lambda[kt-1]) < sqeps) {
      kt--;
      h += fabs(lambda[kt]);
    }
    break;
  default:
  case HLIB_FROBENIUS_RELATIVE:
    fmax = 0.0;
    for(i=0; i<m_tau; i++)
      fmax += fabs(lambda[i]);

    h = 0.0;
    kt = m_tau;
    while(kt > 1 && h + fabs(lambda[kt-1]) < sqeps * fmax) {
      kt--;
      h += fabs(lambda[kt]);
    }
    break;
  case HLIB_EUCLIDEAN_RELATIVE:
    kt = 1;
    while(kt < m_tau && lambda[kt] > sqeps*lambda[0])
      kt++;
    break;
  }

  if(kt > kmax)
    kt = kmax;

  reallocstructure_clusterbasis(cb, kt);
  cb->kt = kt;

#ifdef DEBUG_LOCAL_SVD
  printf("%d:", kt);
  for(i=0; i<m_tau; i++)
    printf(" %g", lambda[i]);
  printf("\n");
#endif

  /* Copy eigenvectors to cluster basis matrices */
  if(t->sons > 0) {
    off = 0;
    for(i=0; i<t->sons; i++) {
      for(j=0; j<kt; j++)
	copy_lapack(cb->son[i]->kt,
		    G + off + j*m_tau,
		    cb->T[i] + j*cb->son[i]->kt);
      
      off += cb->son[i]->kt;
    }
    assert(off == m_tau);
  }
  else
    copy_lapack(kt * m_tau, G, cb->V);

  /* Apply projection to rank-k-matrices */
  b = rk;
  while(b) {
    clear_vector(kt * b->k, aux->tmp);

    if(t->sons > 0) {
      off = start;
      for(i=0; i<t->sons; i++) {
	dgemm_("Transposed", "Not Transposed",
	       &kt, &b->k, &cb->son[i]->kt,
	       deins_,
	       cb->T[i], &cb->son[i]->kt,
	       b->a + off - b->start, &b->size,
	       deins_,
	       aux->tmp, &kt);
	off += t->son[i]->size;
      }
      assert(off == start + size);
    }
    else
      dgemm_("Transposed", "Not Transposed",
	     &kt, &b->k, &size,
	     deins_,
	     cb->V, &size,
	     b->a + start - b->start, &b->size,
	     deins_,
	     aux->tmp, &kt);

    for(i=0; i<b->k; i++)
      copy_lapack(kt, aux->tmp + i*kt, b->a + start - b->start + i*b->size);
    
    b = b->next;
  }

  /* Free auxiliary memory */
  while(super1) {
    b = super1->next;
    free(super1);
    super1 = b;
  }

  while(rk1 != rk) {
    assert(rk1 != NULL);

    b = rk1->next;
    del_block(rk1);
    rk1 = b;
  }

  return cb;
}

static void
splitrow2_supermatrix(psupermatrix s, pcclusteroperator gram, int sstart,
		      pconvblock *super, pconvblock *rk, double *G, int ldG,
		      int start, int size, pcclusterbasis cb, pconversion aux)
{
  pconvblock b;
  pcclusteroperator gram1;
  int i, j, off;
  
  assert(sstart <= start);
  assert(start + size <= sstart + s->rows);
  assert(gram != NULL);
  assert(cb != NULL);
  assert(s->row != NULL && s->col != NULL);
  assert(s->col->t == gram->t);
  assert(s->col->kt == gram->ktc && s->col->kt == gram->ktr);

  if(s->rows > size ||
     (s->s && s->block_rows == 1 && s->s[0]->row == cb)) {
    off = sstart;
    for(i=0; i<s->block_rows; i++) {
      if(start + size <= off + s->s[i]->rows)
	break;
      off += s->s[i]->rows;
    }

    assert(i < s->block_rows);

    for(j=0; j<s->block_cols; j++) {
      gram1 = gram;
      if(s->s[i+j*s->block_rows]->col->t != gram->t) {
	assert(j < gram->sons);
	gram1 = gram->son[j];
      }

      splitrow2_supermatrix(s->s[i+j*s->block_rows], gram1, off,
			    super, rk, G, ldG, start, size, cb, aux);
    }
  }
  else {
    assert(sstart == start);

    if(s->s) {
      b = (pconvblock) malloc(sizeof(convblock));
      assert(b != NULL);

      b->original = s;
      b->gram = gram;
      b->a = NULL;
      b->gb = NULL;
      b->k = 0;
      b->start = sstart;
      b->size = 0;

      b->next = *super;
      *super = b;
    }
    else if(s->r) {
      if(s->r->kt > 0) {
	b = (pconvblock) malloc(sizeof(convblock));
	assert(b != NULL);
	
	b->original = NULL;
	b->gram = NULL;
	b->a = (double *) malloc((size_t) sizeof(double) * s->r->rows * s->r->kt);
	assert(b->a != NULL);
	b->gb = (double *) malloc((size_t) sizeof(double) * s->r->kt * s->r->kt);
	assert(b->gb != NULL);
	b->k = s->r->kt;
	b->start = sstart;
	b->size = s->r->rows;
	
	b->next = *rk;
	*rk = b;
	
	copy_lapack(s->r->rows * s->r->kt, s->r->a, b->a);
	multrans1_lapack(s->r->kt, s->r->kt, s->r->cols, s->r->b, s->r->b, b->gb);
      }
    }
    else if(s->u) {
      assert(s->row == s->u->row);
      assert(s->col == s->u->col);
      assert(s->u->row == cb);
      assert(s->u->col->t == gram->t);
      assert(s->u->ktc == gram->ktc);
      assert(s->u->ktc == gram->ktr);
      assert(G != NULL);
      assert(cb->kt <= ldG);

      provide_array(gram->ktr * cb->kt, &aux->tmp, &aux->tmpsize);

      dgemm_("Not Transposed", "Transposed",
	     &gram->ktr, &cb->kt, &gram->ktr,
	     deins_,
	     gram->P, &gram->ktr,
	     s->u->S, &s->u->row->kt,
	     dnull_,
	     aux->tmp, &gram->ktr);
      dgemm_("Not Transposed", "Not Transposed",
	     &cb->kt, &cb->kt, &gram->ktr,
	     deins_,
	     s->u->S, &s->u->row->kt,
	     aux->tmp, &gram->ktr,
	     deins_,
	     G, &ldG);
    }
    else
      assert(s->f != NULL);
  }
}

static void
splitcol2_supermatrix(psupermatrix s, pcclusteroperator gram, int sstart,
		      pconvblock *super, pconvblock *rk, double *G, int ldG,
		      int start, int size, pcclusterbasis cb, pconversion aux)
{
  pconvblock b;
  pcclusteroperator gram1;
  int i, j, off;
  
  assert(sstart <= start);
  assert(start + size <= sstart + s->cols);
  assert(gram != NULL);
  assert(cb != NULL);
  assert(s->row != NULL && s->col != NULL);
  assert(s->row->t == gram->t);
  assert(s->row->kt == gram->ktc && s->row->kt == gram->ktr);

  if(s->cols > size ||
     (s->s && s->block_cols == 1 && s->s[0]->col == cb)) {
    off = sstart;
    for(i=0; i<s->block_cols; i++) {
      if(start + size <= off + s->s[i*s->block_rows]->cols)
	break;
      off += s->s[i*s->block_rows]->cols;
    }

    assert(i < s->block_cols);

    for(j=0; j<s->block_rows; j++) {
      gram1 = gram;
      if(s->s[j+i*s->block_rows]->row->t != gram->t) {
	assert(j < gram->sons);
	gram1 = gram->son[j];
      }

      splitcol2_supermatrix(s->s[j+i*s->block_rows], gram1, off,
			    super, rk, G, ldG, start, size, cb, aux);
    }
  }
  else {
    assert(sstart == start);

    if(s->s) {
      b = (pconvblock) malloc(sizeof(convblock));
      assert(b != NULL);

      b->original = s;
      b->gram = gram;
      b->a = NULL;
      b->gb = NULL;
      b->k = 0;
      b->start = sstart;
      b->size = 0;

      b->next = *super;
      *super = b;
    }
    else if(s->r) {
      if(s->r->kt > 0) {
	b = (pconvblock) malloc(sizeof(convblock));
	assert(b != NULL);
	
	b->original = NULL;
	b->gram = NULL;
	b->a = (double *) malloc((size_t) sizeof(double) * s->r->cols * s->r->kt);
	assert(b->a != NULL);
	b->gb = (double *) malloc((size_t) sizeof(double) * s->r->kt * s->r->kt);
	assert(b->gb != NULL);
	b->k = s->r->kt;
	b->start = sstart;
	b->size = s->r->cols;
	
	b->next = *rk;
	*rk = b;
	
	copy_lapack(s->r->cols * s->r->kt, s->r->b, b->a);
	multrans1_lapack(s->r->kt, s->r->kt, s->r->rows, s->r->a, s->r->a, b->gb);
      }
    }
    else if(s->u) {
      assert(s->row == s->u->row);
      assert(s->col == s->u->col);
      assert(s->u->col == cb);
      assert(s->u->row->t == gram->t);
      assert(s->u->ktr == gram->ktr);
      assert(s->u->ktr == gram->ktc);
      assert(G != NULL);
      assert(cb->kt <= ldG);

      provide_array(gram->ktr * cb->kt, &aux->tmp, &aux->tmpsize);

      dgemm_("Not Transposed", "Not Transposed",
	     &gram->ktr, &cb->kt, &gram->ktr,
	     deins_,
	     gram->P, &gram->ktr,
	     s->u->S, &s->u->ktr,
	     dnull_,
	     aux->tmp, &gram->ktr);
      dgemm_("Transposed", "Not Transposed",
	     &cb->kt, &cb->kt, &gram->ktr,
	     deins_,
	     s->u->S, &s->u->ktr,
	     aux->tmp, &gram->ktr,
	     deins_,
	     G, &ldG);
    }
    else
      assert(s->f != NULL);
  }
}

static pclusterbasis
build2_basis(pccluster t, pcclusterbasis ocb,
	     pconvblock super, pconvblock rk,
	     double *G1, int ldG1, pconversion aux)
{
  pclusterbasis cb;
  double *G, *G2, *X;
  double *lambda;
  double *work;
  int start = t->start;
  int size = t->size;
  int kmax = aux->kmax;
  int ldG2;
  double sqeps = aux->sqeps;
  int lwork, info;
  double h, fmax;
  pconvblock super1;
  pconvblock rk1;
  psupermatrix s;
  pcclusteroperator gram;
  pconvblock b;
  int m_tau, m_max, kt;
  int i, j, off, off2;

  assert(ocb != NULL);
  assert(G1 != NULL);
  assert(ocb->t == t);

  /* Create new clusterbasis */
  cb = new_clusterbasis(t, 0, 0);

  /* Add new entries to block list */
  super1 = NULL;
  rk1 = rk;
  b = super;
  while(b) {
    s = b->original;
    gram = b->gram;

    if(aux->transposed)
      splitcol2_supermatrix(s, gram, b->start, &super1, &rk1, G1, ldG1,
			    t->start, t->size, ocb, aux);
    else
      splitrow2_supermatrix(s, gram, b->start, &super1, &rk1, G1, ldG1,
			    t->start, t->size, ocb, aux);

    b = b->next;
  }

  /* Determine size of EVP and build son bases if necessary */
  m_tau = 0;
  X = NULL;
  if(t->sons > 0) {
    assert(ocb->sons == t->sons);

    m_max = kmax * t->sons;

    X = (double *) malloc((size_t) sizeof(double) * m_max * ocb->kt);
    assert(X != NULL);

    for(i=0; i<t->sons; i++) {
      ldG2 = imax(kmax, ocb->son[i]->kt);
      G2 = (double *) malloc((size_t) sizeof(double) * ldG2 * ldG2);
      assert(G2 != NULL);
      
      assert(ocb->kt <= ldG1);
      
      provide_array(ldG2 * ldG1, &aux->tmp, &aux->tmpsize);
      dgemm_("Not Transposed", "Transposed",
	     &ocb->kt, &ocb->son[i]->kt, &ocb->kt,
	     deins_,
	     G1, &ldG1,
	     ocb->T[i], &ocb->son[i]->kt,
	     dnull_,
	     aux->tmp, &ldG1);
      dgemm_("Not Transposed", "Not Transposed",
	     &ocb->son[i]->kt, &ocb->son[i]->kt, &ocb->kt,
	     deins_,
	     ocb->T[i], &ocb->son[i]->kt,
	     aux->tmp, &ldG1,
	     dnull_,
	     G2, &ldG2);
      
      cb->son[i] = build2_basis(t->son[i], ocb->son[i],
				super1, rk1, G2, ldG2, aux);

      assert(m_tau + cb->son[i]->kt <= m_max);
      dgemm_("Transposed", "Transposed",
	     &ocb->kt, &cb->son[i]->kt, &ocb->son[i]->kt,
	     deins_,
	     ocb->T[i], &ocb->son[i]->kt,
	     G2, &ldG2,
	     dnull_,
	     X + ocb->kt * m_tau, &ocb->kt);
      
      free(G2);
      G2 = NULL;
      
      m_tau += cb->son[i]->kt;
    }
  }
  else
    m_tau = size;

  /* Provide storage for EVP */
  provide_array(m_tau * (m_tau+7), &aux->evp, &aux->evpsize);
  G = aux->evp;
  lambda = aux->evp + m_tau * m_tau;
  work = aux->evp + m_tau * (m_tau+1);
  lwork = 6*m_tau;

  /* Initialize Gram matrix */
  if(t->sons == 0) {
    assert(ocb->V != NULL);
    assert(ocb->n == m_tau);
    assert(ocb->kt <= ldG1);
    
    provide_array(m_tau * ocb->kt, &aux->tmp, &aux->tmpsize);
    dgemm_("Not Transposed", "Transposed",
	   &ocb->kt, &m_tau, &ocb->kt,
	   deins_,
	   G1, &ldG1,
	   ocb->V, &m_tau,
	   dnull_,
	   aux->tmp, &ocb->kt);
    dgemm_("Not Transposed", "Not Transposed",
	   &m_tau, &m_tau, &ocb->kt,
	   deins_,
	   ocb->V, &m_tau,
	   aux->tmp, &ocb->kt,
	   dnull_,
	   G, &m_tau);
  }
  else {
    assert(X != NULL);
    assert(ocb->kt <= ldG1);
    
    provide_array(m_tau * ocb->kt, &aux->tmp, &aux->tmpsize);
    dgemm_("Not Transposed", "Not Transposed",
	   &ocb->kt, &m_tau, &ocb->kt,
	   deins_,
	   G1, &ldG1,
	   X, &ocb->kt,
	   dnull_,
	   aux->tmp, &ocb->kt);
    dgemm_("Transposed", "Not Transposed",
	   &m_tau, &m_tau, &ocb->kt,
	   deins_,
	   X, &ocb->kt,
	   aux->tmp, &ocb->kt,
	   dnull_,
	   G, &m_tau);
  }

  if(t->sons > 0) {
    /* Build transformed Gram matrix */
    b = rk1;
    while(b) {
      provide_array(m_tau * b->k, &aux->tmp, &aux->tmpsize);
      
      off = start;
      off2 = 0;
      for(i=0; i<t->sons; i++) {
	dgemm_("Not Transposed", "Not Transposed",
	       &cb->son[i]->kt, &b->k, &b->k,
	       deins_,
	       b->a + off - b->start, &b->size,
	       b->gb, &b->k,
	       dnull_,
	       aux->tmp + off2, &m_tau);
	off += t->son[i]->size;
	off2 += cb->son[i]->kt;
      }
      assert(off == start + size);
      assert(off2 == m_tau);

      off = start;
      off2 = 0;
      for(i=0; i<t->sons; i++) {
	dgemm_("Not Transposed", "Transposed",
	       &m_tau, &cb->son[i]->kt, &b->k,
	       deins_,
	       aux->tmp, &m_tau,
	       b->a + off - b->start, &b->size,
	       deins_,
	       G + off2 * m_tau, &m_tau);
	off += t->son[i]->size;
	off2 += cb->son[i]->kt;
      }
      assert(off == start + size);
      assert(off2 == m_tau);

      b = b->next;
    }
  }
  else {
    /* Build Gram matrix */
    b = rk1;
    while(b) {
      provide_array(m_tau * b->k, &aux->tmp, &aux->tmpsize);

      dgemm_("Not Transposed", "Not Transposed",
	     &m_tau, &b->k, &b->k,
	     deins_,
	     b->a + start - b->start, &b->size,
	     b->gb, &b->k,
	     dnull_,
	     aux->tmp, &m_tau);

      dgemm_("Not Transposed", "Transposed",
	     &m_tau, &m_tau, &b->k,
	     deins_,
	     aux->tmp, &m_tau,
	     b->a + start - b->start, &b->size,
	     deins_,
	     G, &m_tau);

      b = b->next;
    }
  }

  /* Solve eigenvalue problem */
  dsyev_("Vectors", "Upper", &m_tau, G, &m_tau,
	 lambda, work, &lwork, &info);
  assert(info == 0);

  /* Sort eigenvalues and eigenvectors */
  for(i=0,j=m_tau-1; i<j; i++,j--) {
    swap_lapack(m_tau, G+i*m_tau, G+j*m_tau);
    h = lambda[i];
    lambda[i] = lambda[j];
    lambda[j] = h;
  }

  /* Determine appropriate rank */
  switch(aux->strategy) {
  case HLIB_FROBENIUS_ABSOLUTE:
    h = 0.0;
    kt = m_tau;
    while(kt > 1 && h + fabs(lambda[kt-1]) < sqeps) {
      kt--;
      h += fabs(lambda[kt]);
    }
    break;
  default:
  case HLIB_FROBENIUS_RELATIVE:
    fmax = 0.0;
    for(i=0; i<m_tau; i++)
      fmax += fabs(lambda[i]);

    h = 0.0;
    kt = m_tau;
    while(kt > 1 && h + fabs(lambda[kt-1]) < sqeps * fmax) {
      kt--;
      h += fabs(lambda[kt]);
    }
    break;
  case HLIB_EUCLIDEAN_RELATIVE:
    kt = 1;
    while(kt < m_tau && lambda[kt] > sqeps*lambda[0])
      kt++;
    break;
  }

  if(kt > kmax)
    kt = kmax;

  reallocstructure_clusterbasis(cb, kt);
  cb->kt = kt;

#ifdef DEBUG_LOCAL_SVD
  printf("%d:", kt);
  for(i=0; i<m_tau; i++)
    printf(" %g", lambda[i]);
  printf("\n");
#endif

  /* Copy eigenvectors to cluster basis matrices */
  if(t->sons > 0) {
    off = 0;
    for(i=0; i<t->sons; i++) {
      for(j=0; j<kt; j++)
	copy_lapack(cb->son[i]->kt,
		    G + off + j*m_tau,
		    cb->T[i] + j*cb->son[i]->kt);
      
      off += cb->son[i]->kt;
    }
    assert(off == m_tau);
  }
  else
    copy_lapack(kt * m_tau, G, cb->V);

  /* Compute cluster basis transformation */
  if(ocb->sons > 0) {
    assert(X != NULL);
    assert(cb->kt <= ldG1);
    
    dgemm_("Transposed", "Transposed",
	   &cb->kt, &ocb->kt, &m_tau,
	   deins_,
	   G, &m_tau,
	   X, &ocb->kt,
	   dnull_,
	   G1, &ldG1);
    
    free(X);
    X = NULL;
  }
  else {
    assert(ocb->V != NULL);
    assert(cb->kt <= ldG1);
    
    dgemm_("Transposed", "Not Transposed",
	   &cb->kt, &ocb->kt, &m_tau,
	   deins_,
	   G, &m_tau,
	   ocb->V, &m_tau,
	   dnull_,
	   G1, &ldG1);
  }

  /* Apply projection to rank-k-matrices */
  b = rk;
  while(b) {
    clear_vector(kt * b->k, aux->tmp);

    if(t->sons > 0) {
      off = start;
      for(i=0; i<t->sons; i++) {
	dgemm_("Transposed", "Not Transposed",
	       &kt, &b->k, &cb->son[i]->kt,
	       deins_,
	       cb->T[i], &cb->son[i]->kt,
	       b->a + off - b->start, &b->size,
	       deins_,
	       aux->tmp, &kt);
	off += t->son[i]->size;
      }
      assert(off == start + size);
    }
    else
      dgemm_("Transposed", "Not Transposed",
	     &kt, &b->k, &size,
	     deins_,
	     cb->V, &size,
	     b->a + start - b->start, &b->size,
	     deins_,
	     aux->tmp, &kt);

    for(i=0; i<b->k; i++)
      copy_lapack(kt, aux->tmp + i*kt, b->a + start - b->start + i*b->size);
    
    b = b->next;
  }

  /* Free auxiliary memory */
  while(super1) {
    b = super1->next;
    free(super1);
    super1 = b;
  }

  while(rk1 != rk) {
    assert(rk1 != NULL);

    b = rk1->next;
    del_block(rk1);
    rk1 = b;
  }

  return cb;
}

pclusterbasis
buildrow2_supermatrix(psupermatrix s, pccluster ct,
		      double eps, int kmax, TruncationStrategy strategy)
{
  pclusterbasis cb;
  pclusteroperator gram;
  double *G1;
  int ldG1;
  conversion aux;
  convblock super;

  aux.evp = NULL;
  aux.evpsize = 0;
  aux.tmp = NULL;
  aux.tmpsize = 0;
  aux.kmax = kmax;
  aux.sqeps = (strategy == HLIB_FROBENIUS_ABSOLUTE ?
	       eps * eps / count_cluster(ct) :
	       eps * eps);
  aux.strategy = strategy;
  aux.transposed = 0;

  super.original = s;
  super.a = NULL;
  super.gb = NULL;
  super.k = 0;
  super.start = 0;
  super.size = 0;
  super.next = NULL;

  if(s->row != NULL) {
    assert(s->col != NULL);

    gram = newproduct_clusteroperator(s->col, s->col);
    super.gram = gram;

    ldG1 = imax(kmax, s->row->kt);
    G1 = (double *) malloc((size_t) sizeof(double) * ldG1 * ldG1);
    assert(G1 != NULL);
    clear_vector(ldG1 * ldG1, G1);

    cb = build2_basis(ct, s->row, &super, NULL, G1, ldG1, &aux);

    free(G1);
    G1= NULL;

    del_clusteroperator(gram);
    gram = NULL;
  }
  else {
    assert(s->col == NULL);

    super.gram = NULL;

    cb = build_basis(ct, &super, NULL, &aux);
  }

  if(aux.evp) {
    free(aux.evp);
    aux.evp = NULL;
  }

  if(aux.tmp) {
    free(aux.tmp);
    aux.tmp = NULL;
  }

  return cb;
}

pclusterbasis
buildcol2_supermatrix(psupermatrix s, pccluster ct,
		      double eps, int kmax, TruncationStrategy strategy)
{
  pclusterbasis cb;
  pclusteroperator gram;
  double *G1;
  int ldG1;
  conversion aux;
  convblock super;

  aux.evp = NULL;
  aux.evpsize = 0;
  aux.tmp = NULL;
  aux.tmpsize = 0;
  aux.kmax = kmax;
  aux.sqeps = (strategy == HLIB_FROBENIUS_ABSOLUTE ?
	       eps * eps / count_cluster(ct) :
	       eps * eps);
  aux.strategy = strategy;
  aux.transposed = 1;

  super.original = s;
  super.a = NULL;
  super.gb = NULL;
  super.k = 0;
  super.start = 0;
  super.size = 0;
  super.next = NULL;

  if(s->col != NULL) {
    assert(s->row != NULL);

    gram = newproduct_clusteroperator(s->row, s->row);
    super.gram = gram;

    ldG1 = imax(kmax, s->col->kt);
    G1 = (double *) malloc((size_t) sizeof(double) * ldG1 * ldG1);
    assert(G1 != NULL);
    clear_vector(ldG1 * ldG1, G1);

    cb = build2_basis(ct, s->col, &super, NULL, G1, ldG1, &aux);

    free(G1);
    G1 = NULL;

    del_clusteroperator(gram);
    gram = NULL;
  }
  else {
    assert(s->row == NULL);

    super.gram = NULL;

    cb = build_basis(ct, &super, NULL, &aux);
  }

  if(aux.evp) {
    free(aux.evp);
    aux.evp = NULL;
  }

  if(aux.tmp) {
    free(aux.tmp);
    aux.tmp = NULL;
  }

  return cb;
}

static psupermatrix
project2bases(psupermatrix s, int mixed,
	      pclusterbasis row, pclusterbasis col)
{
  psupermatrix s2;
  pfullmatrix f;
  prkmatrix r;
  puniformmatrix u;
  int block_rows = s->block_rows;
  int block_cols = s->block_cols;
  int i, j;

  assert(s->rows == row->n);
  assert(s->cols == col->n);
  assert(s->u == NULL);

  s2 = NULL;

  while(row->sons == 1)
    row = row->son[0];

  while(col->sons == 1)
    col = col->son[0];

  if(s->s != NULL) {
    s2 = new_supermatrix(block_rows, block_cols, s->rows, s->cols,
			 (puniformmatrix) 0, (prkmatrix) 0, (pfullmatrix) 0);
    s2->row = row;
    s2->col = col;

    if(row->sons > 0) {
      if(col->sons > 0) {
	assert(block_rows == row->sons);
	assert(block_cols == col->sons);
	
	for(i=0; i<block_cols; i++)
	  for(j=0; j<block_rows; j++)
	    s2->s[j+i*block_rows] = project2bases(s->s[j+i*block_rows], mixed,
						  row->son[j], col->son[i]);
      }
      else {
	assert(block_rows == row->sons);
	assert(block_cols == 1);

	for(j=0; j<block_rows; j++)
	  s2->s[j] = project2bases(s->s[j], mixed,
				   row->son[j], col);
      }
    }
    else {
      if(col->sons > 0) {
	assert(block_rows == 1);
	assert(block_cols == col->sons);

	for(j=0; j<block_cols; j++)
	  s2->s[j] = project2bases(s->s[j], mixed,
				   row, col->son[j]);
      }
      else {
	assert(block_rows == 1);
	assert(block_cols == 1);

	s2->s[0] = project2bases(s->s[0], mixed,
				 row, col);
      }
    }
  }
  else if(s->f != NULL) {
    f = copy_fullmatrix(s->f);
    s2 = new_supermatrix(1, 1, s->f->rows, s->f->cols,
			 (puniformmatrix) 0, (prkmatrix) 0, f);
    s2->row = row;
    s2->col = col;
  }
  else {
    assert(s->r != NULL);

    if(mixed && s->r->kt * (s->rows + s->cols) < row->kt * col->kt) {
      r = copy_rkmatrix(s->r);
      s2 = new_supermatrix(1, 1, s->r->rows, s->r->cols,
			   (puniformmatrix) 0, r, (pfullmatrix) 0);
      s2->row = row;
      s2->col = col;
    }
    else {
      u = new_uniformmatrix(row, col);
      convertrk2_uniformmatrix(s->r, u);
      s2 = new_supermatrix(1, 1, s->r->rows, s->r->cols,
			   u, (prkmatrix) 0, (pfullmatrix) 0);
      s2->row = row;
      s2->col = col;
    }
  }

  return s2;
}

static psupermatrix
project2bases2(psupermatrix s, int mixed,
	       pclusterbasis row, pclusterbasis col,
	       pcclusteroperator torow, pcclusteroperator tocol,
	       double **tmp, int *tmpsize)
{
  psupermatrix s2;
  pfullmatrix f;
  prkmatrix r;
  puniformmatrix u;
  int block_rows = s->block_rows;
  int block_cols = s->block_cols;
  int i, j;

  assert(s->rows == torow->t->size);
  assert(s->rows == row->n);
  assert(torow->t == row->t);
  assert(torow->sons == row->sons);
  assert(s->cols == tocol->t->size);
  assert(s->cols == col->n);
  assert(tocol->t == col->t);
  assert(tocol->sons == col->sons);

  while(row->sons == 1 && s->row->t != row->t) {
    row = row->son[0];
    torow = torow->son[0];
  }

  while(col->sons == 1 && s->col->t != col->t) {
    col = col->son[0];
    tocol = tocol->son[0];
  }

  assert(s->row->t == row->t);
  assert(s->col->t == col->t);

  s2 = NULL;

  if(s->s != NULL) {
    s2 = new_supermatrix(block_rows, block_cols, s->rows, s->cols,
			 (puniformmatrix) 0, (prkmatrix) 0, (pfullmatrix) 0);
    s2->row = row;
    s2->col = col;

    if(row->sons > 0) {
      if(col->sons > 0) {
	assert(block_rows == row->sons);
	assert(block_cols == col->sons);
	
	for(i=0; i<block_cols; i++)
	  for(j=0; j<block_rows; j++)
	    s2->s[j+i*block_rows] = project2bases2(s->s[j+i*block_rows], mixed,
						   row->son[j], col->son[i],
						   torow->son[j], tocol->son[i],
						   tmp, tmpsize);
      }
      else {
	assert(block_rows == row->sons);
	assert(block_cols == 1);

	for(j=0; j<block_rows; j++)
	  s2->s[j] = project2bases2(s->s[j], mixed,
				    row->son[j], col,
				    torow->son[j], tocol,
				    tmp, tmpsize);
      }
    }
    else {
      if(col->sons > 0) {
	assert(block_rows == 1);
	assert(block_cols == col->sons);

	for(j=0; j<block_cols; j++)
	  s2->s[j] = project2bases2(s->s[j], mixed,
				    row, col->son[j],
				    torow, tocol->son[j],
				    tmp, tmpsize);
      }
      else {
	assert(block_rows == 1);
	assert(block_cols == 1);

	s2->s[0] = project2bases2(s->s[0], mixed,
				  row, col,
				  torow, tocol,
				  tmp, tmpsize);
      }
    }
  }
  else if(s->f != NULL) {
    f = copy_fullmatrix(s->f);
    s2 = new_supermatrix(1, 1, s->f->rows, s->f->cols,
			 (puniformmatrix) 0, (prkmatrix) 0, f);
    s2->row = row;
    s2->col = row;
  }
  else if(s->r != NULL) {
    if(mixed && s->r->kt * (s->rows + s->cols) < row->kt * col->kt) {
      r = copy_rkmatrix(s->r);
      s2 = new_supermatrix(1, 1, s->r->rows, s->r->cols,
			   (puniformmatrix) 0, r, (pfullmatrix) 0);
      s2->row = row;
      s2->col = row;
    }
    else {
      u = new_uniformmatrix(row, col);
      convertrk2_uniformmatrix(s->r, u);
      s2 = new_supermatrix(1, 1, s->r->rows, s->r->cols,
			   u, (prkmatrix) 0, (pfullmatrix) 0);
      s2->row = row;
      s2->col = row;
    }
  }
  else {
    assert(s->u != NULL);
    assert(s->u->row == s->row);
    assert(s->u->col == s->col);

    u = new_uniformmatrix(row, col);

    provide_array(s->row->kt * col->kt, tmp, tmpsize);
    multrans2_lapack(s->row->kt, col->kt, s->col->kt,
		     s->u->S, tocol->P, *tmp);
    mul_lapack(row->kt, col->kt, s->row->kt,
	       torow->P, *tmp, u->S);

    s2 = new_supermatrix(1, 1, s->rows, s->cols,
			 u, (prkmatrix) 0, (pfullmatrix) 0);
    s2->row = row;
    s2->col = col;
  }

  return s2;
}

psupermatrix
project2bases_supermatrix(psupermatrix s, int mixed,
			  pclusterbasis row, pclusteroperator rproj,
			  pclusterbasis col, pclusteroperator cproj)
{
  psupermatrix s2;
  pclusteroperator torow;
  pclusteroperator tocol;
  double *tmp;
  int tmpsize;

  s2 = NULL;

  if(s->row || s->col) {
    assert(s->row != NULL);
    assert(s->col != NULL);

    torow = (rproj ? rproj : newproduct_clusteroperator(row, s->row));
    tocol = (cproj ? cproj : newproduct_clusteroperator(col, s->col));

    tmp = NULL;
    tmpsize = 0;

    s2 = project2bases2(s, mixed, row, col, torow, tocol, &tmp, &tmpsize);

    if(tmp)
      free(tmp);
    tmp = NULL;

    if(cproj == NULL) {
      del_clusteroperator(tocol);
      tocol = NULL;
    }
    if(rproj == NULL) {
      del_clusteroperator(torow);
      torow = NULL;
    }
  }
  else
    s2 = project2bases(s, mixed, row, col);

  return s2;
}

psupermatrix
buildh2_supermatrix(psupermatrix s, pccluster row, pccluster col,
		    double eps, int kmax, TruncationStrategy strategy)
{
  pclusterbasis rb, cb;

  rb = buildrow2_supermatrix(s, row, sqrt(0.5) * eps, kmax, strategy);

  cb = buildcol2_supermatrix(s, col, sqrt(0.5) * eps, kmax, strategy);

  return project2bases_supermatrix(s, 0, rb, NULL, cb, NULL);
}

psupermatrix
buildh2symm_supermatrix(psupermatrix s, pccluster ct,
			double eps, int kmax, TruncationStrategy strategy)
{
  pclusterbasis rb;

  rb = buildrow2_supermatrix(s, ct, sqrt(0.5) * eps, kmax, strategy);

  return project2bases_supermatrix(s, 0, rb, NULL, rb, NULL);
}

psupermatrix
orthogonalize_supermatrix(psupermatrix s,
			  double eps, TruncationStrategy strategy)
{
  psupermatrix snew;
  pclusterbasis rb;
  pclusterbasis cb;
  pclusteroperator ro;
  pclusteroperator co;

  assert(s->row != NULL);
  assert(s->col != NULL);

  rb = orthogonalize_clusterbasis(s->row, &ro, sqrt(0.5) * eps, strategy);
  cb = orthogonalize_clusterbasis(s->col, &co, sqrt(0.5) * eps, strategy);

  snew = project2bases_supermatrix(s, 0, rb, ro, cb, co);

  del_clusteroperator(co);
  del_clusteroperator(ro);

  return snew;
}
