
/* ------------------------------------------------------------
   This is the file "h2virtual.c" of the HLib package.

   Copyright 1999 Steffen Boerm and Lars Grasedyck
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#include "h2virtual.h"

/* Display local singular values for each cluster
#define DEBUG_LOCAL_SVD
*/

/* Talk about the steps in the recompression procedure */
#define VERBOSE_RECOMPRESSION
/**/

#include <math.h>
#ifdef DEBUG_LOCAL_SVD
#include <stdio.h>
#endif

static const double RECOMP_EPS = 1e-32;
static const double VIRTORTHO_EPS = 1e-12;

typedef struct _h2block h2block;
typedef h2block *ph2block;

struct _h2block {
  pblockcluster blk;
  pclusteroperator ww;
  pclusterbasis orb;
  pclusteroperator orbp;
  pclusterbasis ocb;
  pclusteroperator ocbp;

  double *S;
  int kr;
  int kc;

  psupermatrix M;
  pclusterbasis row;
  pclusteroperator rowproj;

  ph2block next;
};

/* ============================================================

   Virtual H^2 recompression

   ============================================================ */

ph2recompression
new_h2recompression()
{
  ph2recompression h2r;

  h2r = (ph2recompression) malloc(sizeof(h2recompression));
  assert(h2r != NULL);

  h2r->Q = NULL;
  h2r->Qsize = 0;
  h2r->tmp = NULL;
  h2r->tmpsize = 0;

  h2r->transposed = 0;

  h2r->strategy = HLIB_FROBENIUS_RELATIVE;
  h2r->kmax = 16;
  h2r->eps = 1e-4;

  h2r->buildN = NULL;
  h2r->buildS = NULL;
  h2r->buildVrow = NULL;
  h2r->buildTrow = NULL;
  h2r->getrankrow = NULL;
  h2r->buildVcol = NULL;
  h2r->buildTcol = NULL;
  h2r->getrankcol = NULL;
  h2r->data = NULL;
  h2r->disp = NULL;
  h2r->dispdata = NULL;

  return h2r;
}

void
del_h2recompression(ph2recompression h2r)
{
  assert(h2r != NULL);

  h2r->data = NULL;
  h2r->getrankcol = NULL;
  h2r->buildTcol = NULL;
  h2r->buildVcol = NULL;
  h2r->getrankrow = NULL;
  h2r->buildTrow = NULL;
  h2r->buildVrow = NULL;
  h2r->buildS = NULL;
  h2r->buildN = NULL;

  free(h2r);
}

/* ************************************************************
    Cluster basis products for virtual bases
   ************************************************************ */

static pclusteroperator
compute_product(pccluster t, ph2recompression h2r, double **p, int *psize)
{
  pclusteroperator co;
  double *T;
  int ldT;
  double *V;
  int ldV;
  int k, k2;
  int i;

  k = (h2r->transposed ?
       h2r->getrankcol(t, h2r->data) :
       h2r->getrankrow(t, h2r->data));
  assert(k > 0);

  co = new_clusteroperator(t, k, k);
  co->ktr = k;
  co->ktc = k;

  if(k > 0) {
    assert(co->P != NULL);

    if(co->sons > 0) {
      assert(co->son != NULL);

      clear_vector(k * k, co->P);

      for(i=0; i<co->sons; i++) {
	k2 = (h2r->transposed ?
	      h2r->getrankcol(t->son[i], h2r->data) :
	      h2r->getrankrow(t->son[i], h2r->data));
	assert(k2 > 0);

	co->son[i] = compute_product(t->son[i], h2r, p, psize);

	provide_array(2 * k * k2, p, psize);
	T = (*p) + k * k2;
	ldT = k2;

	if(h2r->transposed)
	  h2r->buildTcol(t->son[i], t, h2r->data, T, ldT);
	else
	  h2r->buildTrow(t->son[i], t, h2r->data, T, ldT);

	mul_lapack(k2, k, k2, co->son[i]->P, T, *p);
	muladdtrans1_lapack(k, k, k2, T, *p, co->P);
      }
    }
    else {
      assert(co->son == NULL);

      provide_array(k * t->size, p, psize);
      V = *p;
      ldV = t->size;
      
      if(h2r->transposed)
	h2r->buildVcol(t, h2r->data, V, ldV);
      else
	h2r->buildVrow(t, h2r->data, V, ldV);

      multrans1_lapack(k, k, t->size, V, V, co->P);
    }
  }

  h2r->done++;
  if(h2r->disp)
    h2r->disp(1, h2r->done, h2r->total, h2r->dispdata);

  return co;
}

pclusteroperator
newproductvirtual_clusteroperator(pccluster t, ph2recompression h2r)
{
  pclusteroperator co;
  double *p;
  int psize;

  p = NULL;
  psize = 0;

  co = compute_product(t, h2r, &p, &psize);

  if(p) {
    free(p);
    p = NULL;
  }

  return co;
}

/* ************************************************************
    Recompression of virtual H^2 matrices
   ************************************************************ */

static void
splitrow_recomp(pblockcluster blk, pclusteroperator ww,
		ph2block *active, ph2block *passive,
		pccluster t, ph2recompression h2r)
{
  ph2block b;
  pcluster t0;
  int kr, kc;
  int i, j;

  assert(blk->col == ww->t);

  if(blk->row != t || (blk->rowsons == 1 && blk->son[0]->row == t)) {
    assert(blk->rowsons > 0);
    assert(blk->colsons > 0);
    assert(blk->colsons == 1 || blk->colsons == ww->sons);

    for(i=0; i<blk->rowsons; i++) {
      t0 = blk->son[i]->row;
      if(t0->start <= t->start && t->start + t->size <= t0->start + t0->size)
	break;
    }
    assert(i < blk->rowsons);

    if(blk->colsons > 1)
      for(j=0; j<blk->colsons; j++)
	splitrow_recomp(blk->son[i+j*blk->rowsons], ww->son[j],
			active, passive, t, h2r);
    else
      splitrow_recomp(blk->son[i], ww,
		      active, passive, t, h2r);
  }
  else {
    assert(blk->row == t);
    assert(blk->col == ww->t);
    assert(blk->adm == 0 || blk->adm == 3);

    switch(blk->adm) {
    case 0:
      if(blk->rowsons > 0 && blk->colsons > 0) {
	b = (ph2block) malloc(sizeof(h2block));
	assert(b != NULL);
	
	b->blk = blk;
	b->ww = ww;
	b->S = NULL;
	b->kr = 0;
	b->kc = 0;
	b->M = NULL;
	b->rowproj = NULL;
	b->next = *passive;
	*passive = b;
      }
      break;
    case 3:
      b = (ph2block) malloc(sizeof(h2block));
      assert(b != NULL);

      kr = h2r->getrankrow(blk->row, h2r->data);
      kc = h2r->getrankcol(blk->col, h2r->data);

      b->blk = blk;
      b->ww = ww;
      b->S = (double *) malloc((size_t) sizeof(double) * kr * kc);
      assert(b->S != NULL);
      b->kr = kr;
      b->kc = kc;
      b->M = NULL;
      b->rowproj = NULL;
      h2r->buildS(blk->row, blk->col, h2r->data, b->S, kr);
      b->next = *active;
      *active = b;
      break;
    default:
      break;
    }
  }
}

static void
splitcol_recomp(pblockcluster blk, pclusteroperator ww,
		pclusterbasis row, pclusterbasis col, pclusteroperator proj,
		psupermatrix *M, ph2block *active, ph2block *passive,
		pccluster t, ph2recompression h2r)
{
  ph2block b;
  pcluster t0;
  puniformmatrix u;
  pfullmatrix f;
  int kr, kc;
  int i, j;

  assert(M != NULL);
  assert(blk->row == row->t);
  assert(blk->row == ww->t);
  assert(blk->row == proj->t);

  if(blk->col != t || (blk->colsons == 1 && blk->son[0]->col == t)) {
    assert(blk->rowsons > 0);
    assert(blk->colsons > 0);
    assert(blk->rowsons == 1 || blk->rowsons == row->sons);
    assert(blk->rowsons == 1 || blk->rowsons == proj->sons);
    assert(blk->rowsons == 1 || blk->rowsons == ww->sons);

    if(*M == NULL) {
      *M = new_supermatrix(blk->rowsons, blk->colsons,
			   blk->row->size, blk->col->size,
			   NULL, NULL, NULL);

      for(i=0; i<(*M)->block_rows * (*M)->block_cols; i++)
	(*M)->s[i] = NULL;
    }

    assert((*M)->block_rows == blk->rowsons);
    assert((*M)->block_cols == blk->colsons);
    assert((*M)->rows == blk->row->size);
    assert((*M)->cols == blk->col->size);

    for(i=0; i<blk->colsons; i++) {
      t0 = blk->son[i*blk->rowsons]->col;
      if(t0->start <= t->start && t->start + t->size <= t0->start + t0->size)
	break;
    }
    assert(i < blk->colsons);

    if(blk->rowsons > 1)
      for(j=0; j<blk->rowsons; j++)
	splitcol_recomp(blk->son[j+i*blk->rowsons], ww->son[j],
			row->son[j], col, proj->son[j],
			&(*M)->s[j+i*(*M)->block_rows], active, passive, t, h2r);
    else {
      assert(blk->rowsons == 1);
      splitcol_recomp(blk->son[i], ww,
		      row, col, proj,
		      &(*M)->s[i], active, passive, t, h2r);
    }
  }
  else {
    assert(blk->col == t);
    assert(blk->row == ww->t);
    assert(blk->adm == 0 || blk->adm == 3);

    switch(blk->adm) {
    case 0:
      if(blk->rowsons > 0 && blk->colsons > 0) {
	b = (ph2block) malloc(sizeof(h2block));
	assert(b != NULL);

	if(*M == NULL) {
	  *M = new_supermatrix(blk->rowsons, blk->colsons,
			       blk->row->size, blk->col->size,
			       NULL, NULL, NULL);
	  (*M)->row = row;
	  (*M)->col = col;

	  for(i=0; i<(*M)->block_rows * (*M)->block_cols; i++)
	    (*M)->s[i] = NULL;
	}
	assert((*M)->s != NULL);
	
	b->blk = blk;
	b->ww = ww;
	b->S = NULL;
	b->kr = 0;
	b->kc = 0;
	b->M = *M;
	b->row = row;
	b->rowproj = proj;
	b->next = *passive;
	*passive = b;
      }
      else {
	if(*M == NULL) {
	  f = new_fullmatrix(blk->row->size, blk->col->size);
	  *M = new_supermatrix(1, 1, f->rows, f->cols,
			       NULL, NULL, f);
	  (*M)->row = row;
	  (*M)->col = col;
	  h2r->buildN(blk->row, blk->col, h2r->data, f->e, f->rows);
	}
	assert((*M)->f != NULL);
      }
      break;
    case 3:
      b = (ph2block) malloc(sizeof(h2block));
      assert(b != NULL);

      if(*M == NULL) {
	u = new_uniformmatrix(NULL, NULL);
	*M = new_supermatrix(1, 1, blk->row->size, blk->col->size,
			     u, NULL, NULL);
      }
      assert((*M)->u != NULL);

      kr = h2r->getrankrow(blk->row, h2r->data);
      kc = h2r->getrankcol(blk->col, h2r->data);

      b->blk = blk;
      b->ww = ww;
      b->S = (double *) malloc((size_t) sizeof(double) * kr * kc);
      assert(b->S != NULL);
      h2r->buildS(blk->row, blk->col, h2r->data, b->S, kr);
      b->kr = kr;
      b->kc = kc;
      b->M = *M;
      b->row = row;
      b->rowproj = proj;
      b->next = *active;
      *active = b;
      break;
    default:
      break;
    }
  }
}

static pclusterbasis
build_basis_h2(pccluster t, pclusteroperator *proj,
	       ph2block active, ph2block passive,
	       double *G, int ldG, ph2recompression h2r)
{
  pclusterbasis cb;
  ph2block active2, passive2;
  ph2block b, bnext;
  double *T;
  int ldT;
  double *G2;
  int ldG2;
  double *X;
  int ldX;
  double *lambda, *work;
  double h, fmax, sqeps;
  int k, k2, smax, kt, kmax;
  int m, off;
  int ldwork, info;
  int i, j;

  assert(t != NULL);
  assert(G != NULL);
  assert(h2r != NULL);

  /* Get rank of original cluster basis */

  k = (h2r->transposed ?
       h2r->getrankcol(t, h2r->data) :
       h2r->getrankrow(t, h2r->data));

  assert(k == ldG);

  /* Create new cluster basis and corresponding projection */

  cb = new_clusterbasis(t, 0, 0);

  if(proj)
    *proj = new_clusteroperator(t, 0, 0);

  /* Determine active blocks */

  active2 = active;
  passive2 = NULL;
  for(b=passive; b!=NULL; b=b->next)
    if(h2r->transposed)
      splitcol_recomp(b->blk, b->ww, b->row, cb, b->rowproj, &b->M, &active2, &passive2, t, h2r);
    else
      splitrow_recomp(b->blk, b->ww, &active2, &passive2, t, h2r);

  assert(t->sons > 0 || passive2 == NULL);

  /* Update Gram matrix */

  for(b=active2; b!=active; b=b->next) {
    assert(b != NULL);
    if(h2r->transposed) {
      assert(b->ww->ktr == b->kr);
      assert(b->ww->ktc == b->kr);
      assert(b->kc == k);

      provide_array(b->kr * b->kc, &h2r->tmp, &h2r->tmpsize);
      dgemm_("Not Transposed", "Not Transposed",
	     &b->kr, &b->kc, &b->kr,
	     deins_,
	     b->ww->P, &b->kr,
	     b->S, &b->kr,
	     dnull_,
	     h2r->tmp, &b->kr);

      assert(ldG == b->kc);
      dgemm_("Transposed", "Not Transposed",
	     &b->kc, &b->kc, &b->kr,
	     deins_,
	     b->S, &b->kr,
	     h2r->tmp, &b->kr,
	     deins_,
	     G, &ldG);
    }
    else {
      assert(b->ww->ktr == b->kc);
      assert(b->ww->ktc == b->kc);
      assert(b->kr == k);

      provide_array(b->kc * b->kr, &h2r->tmp, &h2r->tmpsize);
      dgemm_("Not Transposed", "Transposed",
	     &b->kc, &b->kr, &b->kc,
	     deins_,
	     b->ww->P, &b->kc,
	     b->S, &b->kr,
	     dnull_,
	     h2r->tmp, &b->kc);

      assert(ldG == b->kr);
      dgemm_("Not Transposed", "Not Transposed",
	     &b->kr, &b->kr, &b->kc,
	     deins_,
	     b->S, &b->kr,
	     h2r->tmp, &b->kc,
	     deins_,
	     G, &ldG);
    }
  }

  /* Find expansion coefficients */

  m = 0;
  X = NULL;
  if(t->sons > 0) {
    X = (double *) malloc((size_t) sizeof(double) * t->sons * h2r->kmax * k);
    assert(X != NULL);
    ldX = k;

    smax = 0;
    if(h2r->transposed)
      for(i=0; i<t->sons; i++)
	smax = imax(smax, h2r->getrankcol(t->son[i], h2r->data));
    else
      for(i=0; i<t->sons; i++)
	smax = imax(smax, h2r->getrankrow(t->son[i], h2r->data));

    G2 = (double *) malloc((size_t) sizeof(double) * smax * smax);
    assert(G2 != NULL);

    T = (double *) malloc((size_t) sizeof(double) * smax * k);
    assert(T != NULL);
    
    for(i=0; i<t->sons; i++) {
      k2 = (h2r->transposed ?
	    h2r->getrankcol(t->son[i], h2r->data) :
	    h2r->getrankrow(t->son[i], h2r->data));
      assert(k2 <= smax);

      ldG2 = k2;
      ldT = k2;

      if(h2r->transposed)
	h2r->buildTcol(t->son[i], t, h2r->data, T, ldT);
      else
	h2r->buildTrow(t->son[i], t, h2r->data, T, ldT);

      provide_array(k * k2, &h2r->tmp, &h2r->tmpsize);

      dgemm_("Not Transposed", "Transposed",
	     &k, &k2, &k,
	     deins_,
	     G, &ldG,
	     T, &ldT,
	     dnull_,
	     h2r->tmp, &k);
      
      dgemm_("Not Transposed", "Not Transposed",
	     &k2, &k2, &k,
	     deins_,
	     T, &ldT,
	     h2r->tmp, &k,
	     dnull_,
	     G2, &ldG2);

      cb->son[i] = build_basis_h2(t->son[i], (proj ? &(*proj)->son[i] : NULL),
				  active2, passive2, G2, ldG2, h2r);

      dgemm_("Transposed", "Transposed",
	     &k, &cb->son[i]->kt, &k2,
	     deins_,
	     T, &ldT,
	     G2, &ldG2,
	     dnull_,
	     X + m*ldX, &ldX);

      m += cb->son[i]->kt;
    }

    free(T);
    T = NULL;

    free(G2);
    G2 = NULL;
  }
  else {
    m = t->size;

    X = (double *) malloc((size_t) sizeof(double) * m * k);
    assert(X != NULL);
    ldX = k;

    provide_array(m * k, &h2r->tmp, &h2r->tmpsize);

    if(h2r->transposed)
      h2r->buildVcol(t, h2r->data, h2r->tmp, m);
    else
      h2r->buildVrow(t, h2r->data, h2r->tmp, m);

    for(i=0; i<m; i++)
      dcopy_(&k, h2r->tmp + i, &m, X + i*k, eins_);
  }

  /* Expand Gram matrix */

  provide_array(m * k, &h2r->tmp, &h2r->tmpsize);
  dgemm_("Not Transposed", "Not Transposed",
	 &k, &m, &k,
	 deins_,
	 G, &ldG,
	 X, &ldX,
	 dnull_,
	 h2r->tmp, &k);

  provide_array(m * m, &h2r->Q, &h2r->Qsize);
  dgemm_("Transposed", "Not Transposed",
	 &m, &m, &k,
	 deins_,
	 X, &ldX,
	 h2r->tmp, &k,
	 dnull_,
	 h2r->Q, &m);

  /* Solve eigenvalue problem */

  ldwork = 8 * m;
  provide_array(ldwork + m, &h2r->tmp, &h2r->tmpsize);
  work = h2r->tmp;
  lambda = h2r->tmp + ldwork;

  dsyev_("Vectors Wanted", "Upper half given",
	 &m, h2r->Q, &m, lambda, work, &ldwork, &info);
  assert(info == 0);
	 
  for(i=0,j=m-1; i<j; i++,j--) {
    swap_lapack(m, h2r->Q+i*m, h2r->Q+j*m);
    h = lambda[i];
    lambda[i] = lambda[j];
    lambda[j] = h;
  }

  /* Determine rank of new cluster basis */

  sqeps = h2r->sqeps;
  kmax = h2r->kmax;

  switch(h2r->strategy) {
  case HLIB_FROBENIUS_ABSOLUTE:
    h = 0.0;
    kt = m;
    while(kt > 1 && h + fabs(lambda[kt-1]) < sqeps) {
      kt--;
      h += fabs(lambda[kt]);
    }
    break;
  default:
  case HLIB_FROBENIUS_RELATIVE:
    fmax = 0.0;
    for(i=0; i<m; i++)
      fmax += fabs(lambda[i]);

    if(fmax < RECOMP_EPS)
      kt = 1;
    else {
      h = 0.0;
      kt = m;
      while(kt > 1 && h + fabs(lambda[kt-1]) < sqeps * fmax) {
	kt--;
	h += fabs(lambda[kt]);
      }
    }
    break;
  case HLIB_EUCLIDEAN_RELATIVE:
    kt = 1;
    while(kt < m && lambda[kt] > sqeps*lambda[0])
      kt++;
    break;
  }

  if(kt > kmax)
    kt = kmax;

  assert(kt > 0);

  reallocstructure_clusterbasis(cb, kt);
  cb->kt = kt;

#ifdef DEBUG_LOCAL_SVD
  printf("%d:", kt);
  for(i=0; i<m; i++)
    printf(" %g", lambda[i]);
  printf("\n");
#endif

  /* Extract transfer or leaf matrices from eigenvectors */

  if(t->sons > 0) {
    off = 0;
    for(i=0; i<t->sons; i++) {
      for(j=0; j<kt; j++)
	copy_lapack(cb->son[i]->kt,
		    h2r->Q + off + j*m,
		    cb->T[i] + j*cb->son[i]->kt);
      
      off += cb->son[i]->kt;
    }
    assert(off == m);
  }
  else
    copy_lapack(kt * m, h2r->Q, cb->V);

  assert(X != NULL);

  /* Find cluster basis projection */

  dgemm_("Transposed", "Transposed",
	 &kt, &k, &m,
	 deins_,
	 h2r->Q, &m,
	 X, &k,
	 dnull_,
	 G, &ldG);

  if(proj) {
    reallocstructure_clusteroperator(*proj, kt, k);
    for(i=0; i<k; i++)
      dcopy_(&kt, G+i*ldG, eins_, (*proj)->P+i*kt, eins_);
    (*proj)->ktr = kt;
    (*proj)->ktc = k;
  }
  
  free(X);
  X = NULL;

  /* Retire active blocks */

  b = active2;
  while(b != active) {
    assert(b != NULL);

    bnext = b->next;

    if(h2r->transposed) {
      assert(b->M != NULL);
      assert(b->row != NULL);
      assert(b->rowproj != NULL);

      b->M->row = b->row;
      b->M->col = cb;
      reallocstructure_uniformmatrix(b->M->u, b->row, cb);
    
      assert(b->rowproj->ktc == b->kr);
      assert(b->rowproj->ktr == b->row->kt);

      provide_array(b->kr * kt, &h2r->tmp, &h2r->tmpsize);

      dgemm_("Not Transposed", "Transposed",
	     &b->kr, &kt, &k,
	     deins_,
	     b->S, &b->kr,
	     G, &ldG,
	     dnull_,
	     h2r->tmp, &b->kr);
      dgemm_("Not Transposed", "Not Transposed",
	     &b->row->kt, &kt, &b->kr,
	     deins_,
	     b->rowproj->P, &b->row->kt,
	     h2r->tmp, &b->kr,
	     dnull_,
	     b->M->u->S, &b->row->kt);
    }

    free(b->S);
    b->S = NULL;

    free(b);
    b = bnext;
  }

  /* Retire passive blocks */

  b = passive2;
  while(b != NULL) {
    bnext = b->next;

    free(b);
    b = bnext;
  }

  h2r->done++;
  if(h2r->disp)
    h2r->disp(1, h2r->done, h2r->total, h2r->dispdata);

  return cb;
}

psupermatrix
oldvirtual2_supermatrix(pccluster row, pccluster col,
			ph2recompression h2r, pblockcluster blk)
{
  pclusterbasis rb, cb;
  pclusteroperator ww, rowproj;
  h2block b;
  double *G;
  int ldG;

  assert(h2r != NULL);
  assert(blk->rowsons > 0);
  assert(blk->colsons > 0);

#ifdef VERBOSE_RECOMPRESSION
  (void) printf("  Building column cluster basis product\n");
#endif

  h2r->transposed = 1;

  if(h2r->disp) {
    h2r->total = count_cluster(row);
    h2r->done = 0;
    h2r->disp(0, h2r->done, h2r->total, h2r->dispdata);
  }

  ww = newproductvirtual_clusteroperator(col, h2r);

  if(h2r->disp)
    h2r->disp(2, h2r->done, h2r->total, h2r->dispdata);

#ifdef VERBOSE_RECOMPRESSION
  (void) printf("  Building row cluster basis\n");
#endif

  h2r->transposed = 0;
  h2r->sqeps = (h2r->strategy == HLIB_FROBENIUS_ABSOLUTE ?
		dsqr(h2r->eps) / count_cluster(row) / 2.0 :
		dsqr(h2r->eps) / 2.0);

  b.blk = blk;
  b.ww = ww;
  b.S = NULL;
  b.kr = 0;
  b.kc = 0;
  b.M = NULL;
  b.row = NULL;
  b.rowproj = NULL;
  b.next = NULL;

  ldG = h2r->getrankrow(row, h2r->data);
  G = (double *) malloc((size_t) sizeof(double) * ldG * ldG);
  assert(G != NULL);
  clear_vector(ldG * ldG, G);

  if(h2r->disp) {
    h2r->total = count_cluster(row);
    h2r->done = 0;
    h2r->disp(0, h2r->done, h2r->total, h2r->dispdata);
  }

  rowproj = NULL;
  rb = build_basis_h2(row, &rowproj, NULL, &b, G, ldG, h2r);

  if(h2r->disp)
    h2r->disp(2, h2r->done, h2r->total, h2r->dispdata);

  free(G);
  G = NULL;

  del_clusteroperator(ww);

#ifdef VERBOSE_RECOMPRESSION
  (void) printf("  Building row cluster basis product\n");
#endif

  h2r->transposed = 0;
  
  if(h2r->disp) {
    h2r->total = count_cluster(row);
    h2r->done = 0;
    h2r->disp(0, h2r->done, h2r->total, h2r->dispdata);
  }

  ww = newproductvirtual_clusteroperator(row, h2r);

  if(h2r->disp)
    h2r->disp(2, h2r->done, h2r->total, h2r->dispdata);

#ifdef VERBOSE_RECOMPRESSION
  (void) printf("  Building column cluster basis and matrix\n");
#endif

  h2r->transposed = 1;
  h2r->sqeps = (h2r->strategy == HLIB_FROBENIUS_ABSOLUTE ?
		dsqr(h2r->eps) / count_cluster(col) / 2.0 :
		dsqr(h2r->eps) / 2.0);

  b.ww = ww;
  b.M = NULL;
  b.row = rb;
  b.rowproj = rowproj;

  ldG = h2r->getrankcol(col, h2r->data);
  G = (double *) malloc((size_t) sizeof(double) * ldG * ldG);
  assert(G != NULL);
  clear_vector(ldG * ldG, G);

  if(h2r->disp) {
    h2r->total = count_cluster(col);
    h2r->done = 0;
    h2r->disp(0, h2r->done, h2r->total, h2r->dispdata);
  }

  cb = build_basis_h2(col, NULL, NULL, &b, G, ldG, h2r);

  if(h2r->disp)
    h2r->disp(2, h2r->done, h2r->total, h2r->dispdata);

  free(G);
  G = NULL;

#ifdef VERBOSE_RECOMPRESSION
  (void) printf("  Temporary storage requirements: %.1f MB\n",
		(getsize_clusteroperator(rowproj) +
		 getsize_clusteroperator(ww)) / 1024.0 / 1024.0);
#endif

  del_clusteroperator(rowproj);
  del_clusteroperator(ww);

  return b.M;
}

/* ************************************************************
    Build orthogonalized cluster basis
   ************************************************************ */

static void
orthovirtual_clusterbasis(pccluster t,
			  pclusterbasis *cb, pclusteroperator *proj,
			  ph2recompression h2r)
{
  double sqeps = h2r->sqeps;
  double **tmp = &h2r->tmp;
  int *tmpsize = &h2r->tmpsize;
  double **tmp2 = &h2r->Q;
  int *tmp2size = &h2r->Qsize;
  TruncationStrategy strategy = h2r->strategy;
  double *G;
  int ldG;
  double *Y;
  int ldY;
  double *Z;
  int ldZ;
  double *lambda;
  double *tau;
  double *work;
  double h, fmax;
  int lwork, mmax, m;
  int k, k2, kt, kt2, evpsize;
  int info;
  int i, j;

  assert(cb != NULL);
  assert(proj != NULL);

  *cb = new_clusterbasis(t, 0, 0);
  *proj = new_clusteroperator(t, 0, 0);

  k = (h2r->transposed ?
       h2r->getrankcol(t, h2r->data) :
       h2r->getrankrow(t, h2r->data));

  mmax = t->size;
  if(t->sons > 0) {
    mmax = 0;
    for(i=0; i<t->sons; i++)
      mmax += (h2r->transposed ?
	       h2r->getrankcol(t->son[i], h2r->data) :
	       h2r->getrankrow(t->son[i], h2r->data));
  }

  Y = (double *) malloc((size_t) sizeof(double) * mmax * k);
  assert(Y != NULL);
  ldY = k;

  if(t->sons > 0) {
    m = 0;

    for(i=0; i<t->sons; i++) {
      k2 = (h2r->transposed ?
	    h2r->getrankcol(t->son[i], h2r->data) :
	    h2r->getrankrow(t->son[i], h2r->data));

      orthovirtual_clusterbasis(t->son[i],
				&((*cb)->son[i]), &((*proj)->son[i]),
				h2r);
      kt2 = (*cb)->son[i]->kt;

      provide_array(k * k2, tmp2, tmp2size);
      if(h2r->transposed)
	h2r->buildTcol(t->son[i], t, h2r->data, *tmp2, k2);
      else
	h2r->buildTrow(t->son[i], t, h2r->data, *tmp2, k2);

      assert(m + kt2 <= mmax);

      dgemm_("Transposed", "Transposed",
	     &k, &kt2, &k2,
	     deins_,
	     *tmp2, &k2,
	     *tmp, &kt2,
	     dnull_,
	     Y + m*k, &ldY);

      m += kt2;
    }
  }
  else {
    m = t->size;

    provide_array(m * k, tmp2, tmp2size);
    if(h2r->transposed)
      h2r->buildVcol(t, h2r->data, *tmp2, m);
    else
      h2r->buildVrow(t, h2r->data, *tmp2, m);

    for(i=0; i<k; i++)
      dcopy_(&m, *tmp2 + i*m, eins_, Y + i, &ldY);
  }

  lwork = 6 * k;
  evpsize = k;
  ldG = k;
  ldZ = k;
  provide_array(2 * k * k + mmax * k + 2 * k + lwork, tmp, tmpsize);
  G = *tmp + k * k;
  Z = G + k * k;
  lambda = Z + mmax * k;
  tau = lambda + k;
  work = tau + k;

  copy_lapack(m * k, Y, Z);

  if(m < k) {
    dgeqrf_(&k, &m, Y, &ldY, tau, work, &lwork, &info);
    assert(info == 0);

    for(i=0; i<m; i++)
      for(j=i+1; j<m; j++)
	Y[j+i*ldY] = 0.0;

    evpsize = m;
  }

  dgemm_("Not Transposed", "Transposed",
	 &evpsize, &evpsize, &m,
	 deins_,
	 Y, &ldY,
	 Y, &ldY,
	 dnull_,
	 G, &ldG);

  dsyev_("Vectors wanted", "Upper half",
	 &evpsize, G, &ldG, lambda, work, &lwork, &info);
  assert(info == 0);

  for(i=0,j=evpsize-1; i<j; i++,j--) {
    swap_lapack(evpsize, G+i*ldG, G+j*ldG);
    h = dmax(0.0, lambda[i]);
    lambda[i] = dmax(0.0, lambda[j]);
    lambda[j] = h;
  }

  switch(strategy) {
  case HLIB_FROBENIUS_ABSOLUTE:
    h = 0.0;
    kt = evpsize;
    while(kt > 1 && h + lambda[kt-1] < sqeps) {
      kt--;
      h += lambda[kt];
    }
    break;
  default:
  case HLIB_FROBENIUS_RELATIVE:
    fmax = 0.0;
    for(i=0; i<evpsize; i++)
      fmax += lambda[i];

    h = 0.0;
    kt = evpsize;
    while(kt > 1 && h + lambda[kt-1] < sqeps * fmax) {
      kt--;
      h += lambda[kt];
    }
    break;
  case HLIB_EUCLIDEAN_RELATIVE:
    kt = 1;
    while(kt < evpsize && lambda[kt] > sqeps*lambda[0])
      kt++;
    break;
  }

  while(kt > 1 && lambda[kt-1] < VIRTORTHO_EPS)
    kt--;

  assert(0 < kt);
  assert(kt <= evpsize);

  reallocstructure_clusterbasis(*cb, kt);
  (*cb)->kt = kt;

  for(i=0; i<kt; i++) {
    assert(lambda[i] >= VIRTORTHO_EPS);
    scale_lapack(evpsize, G+i*ldG, 1.0 / sqrt(lambda[i]));
  }

  if(t->sons > 0) {
    clear_vector(kt * k, *tmp);

    mmax = m;
    m = 0;
    for(i=0; i<t->sons; i++) {
      kt2 = (*cb)->son[i]->kt;

      dgemm_("Transposed", "Not Transposed",
	     &kt2, &kt, &evpsize,
	     deins_,
	     Y + m*ldY, &ldY,
	     G, &ldG,
	     dnull_,
	     (*cb)->T[i], &kt2);

      dgemm_("Transposed", "Transposed",
	     &kt, &k, &kt2,
	     deins_,
	     (*cb)->T[i], &kt2,
	     Z + m*ldZ, &ldZ,
	     deins_,
	     *tmp, &kt);

      m += kt2;
    }
    assert(m == mmax);
  }
  else {
    dgemm_("Transposed", "Not Transposed",
	   &m, &kt, &evpsize,
	   deins_,
	   Y, &ldY,
	   G, &ldG,
	   dnull_,
	   (*cb)->V, &m);

    dgemm_("Transposed", "Transposed",
	   &kt, &k, &m,
	   deins_,
	   (*cb)->V, &m,
	   Z, &ldZ,
	   dnull_,
	   *tmp, &kt);
  }

  reallocstructure_clusteroperator(*proj, kt, k);
  copy_lapack(kt * k, *tmp, (*proj)->P);

  h2r->done++;
  if(h2r->disp)
    h2r->disp(1, h2r->done, h2r->total, h2r->dispdata);

  free(Y);
  Y = NULL;
}

pclusterbasis
virtual2row_clusterbasis(pccluster t, pclusteroperator *proj,
			 ph2recompression h2r)
{
  pclusterbasis cb;

  h2r->transposed = 0;
  h2r->sqeps = (h2r->strategy == HLIB_FROBENIUS_ABSOLUTE ?
		0.5 * dsqr(h2r->eps) / count_cluster(t) :
		0.5 * dsqr(h2r->eps));

  h2r->done = 0;
  h2r->total = count_cluster(t);
  if(h2r->disp)
    h2r->disp(0, 0, h2r->total, h2r->dispdata);

  orthovirtual_clusterbasis(t, &cb, proj, h2r);

  if(h2r->disp)
    h2r->disp(2, h2r->done, h2r->total, h2r->dispdata);

  return cb;
}
			   
pclusterbasis
virtual2col_clusterbasis(pccluster t, pclusteroperator *proj,
			 ph2recompression h2r)
{
  pclusterbasis cb;

  h2r->transposed = 1;
  h2r->sqeps = (h2r->strategy == HLIB_FROBENIUS_ABSOLUTE ?
		0.5 * dsqr(h2r->eps) / count_cluster(t) :
		0.5 * dsqr(h2r->eps));

  h2r->done = 0;
  h2r->total = count_cluster(t);
  if(h2r->disp)
    h2r->disp(0, 0, h2r->total, h2r->dispdata);

  orthovirtual_clusterbasis(t, &cb, proj, h2r);

  if(h2r->disp)
    h2r->disp(2, h2r->done, h2r->total, h2r->dispdata);

  return cb;
}

/* ************************************************************
    Build projected H^2-matrix
   ************************************************************ */

static psupermatrix
projectbuild_supermatrix(pblockcluster blk,
			 pclusterbasis rb, pclusteroperator rproj,
			 pclusterbasis cb, pclusteroperator cproj,
			 ph2recompression h2r)
{
  psupermatrix s;
  pfullmatrix f;
  puniformmatrix u;
  pclusterbasis rb2, cb2;
  pclusteroperator rproj2, cproj2;
  double *S, *S2;
  int krow, kcol;
  int i, j;

  assert(blk->row == rb->t);
  assert(blk->col == cb->t);
  assert(blk->row == rproj->t);
  assert(blk->col == cproj->t);
  assert(blk->adm == 0 || blk->adm == 3);

  s = NULL;
  switch(blk->adm) {
    case 0:
      if(blk->rowsons > 0 && blk->colsons > 0) {
	s = new_supermatrix(blk->rowsons, blk->colsons,
			    blk->row->size, blk->col->size,
			    NULL, NULL, NULL);
	for(i=0; i<blk->colsons; i++) {
	  cb2 = cb;
	  cproj2 = cproj;
	  if(blk->son[i*blk->rowsons]->col != cb2->t) {
	    assert(i < cb->sons);
	    assert(i < cproj->sons);
	    cb2 = cb->son[i];
	    cproj2 = cproj->son[i];
	  }
	  for(j=0; j<blk->rowsons; j++) {
	    rb2 = rb;
	    rproj2 = rproj;
	    if(blk->son[j+i*blk->rowsons]->row != rb2->t) {
	      assert(j < rb->sons);
	      assert(j < rproj->sons);
	      rb2 = rb->son[j];
	      rproj2 = rproj->son[j];
	    }
	    s->s[j+i*blk->rowsons] = projectbuild_supermatrix(blk->son[j+i*blk->rowsons],
							      rb2, rproj2,
							      cb2, cproj2,
							      h2r);
	  }
	}
      }
      else {
	f = new_fullmatrix(blk->row->size, blk->col->size);
	s = new_supermatrix(1, 1, blk->row->size, blk->col->size,
			    NULL, NULL, f);

	h2r->buildN(blk->row, blk->col, h2r->data, f->e, blk->row->size);
      }
      break;

    default:
    case 3:
      u = new_uniformmatrix(rb, cb);
      s = new_supermatrix(1, 1, blk->row->size, blk->col->size,
			  u, NULL, NULL);

      krow = h2r->getrankrow(rb->t, h2r->data);
      kcol = h2r->getrankcol(cb->t, h2r->data);

      provide_array(krow * (kcol + cb->kt), &h2r->tmp, &h2r->tmpsize);
      S = h2r->tmp;
      S2 = h2r->tmp + krow * kcol;

      h2r->buildS(blk->row, blk->col, h2r->data, S, krow);

      dgemm_("Not Transposed", "Transposed",
	     &krow, &cb->kt, &kcol,
	     deins_,
	     S, &krow,
	     cproj->P, &cb->kt,
	     dnull_,
	     S2, &krow);

      dgemm_("Not Transposed", "Not Transposed",
	     &rb->kt, &cb->kt, &krow,
	     deins_,
	     rproj->P, &rb->kt,
	     S2, &krow,
	     dnull_,
	     u->S, &rb->kt);

      break;
  }

  assert(s != NULL);
  s->row = rb;
  s->col = cb;

  h2r->done++;
  if(h2r->disp)
    h2r->disp(1, h2r->done, h2r->total, h2r->dispdata);

  return s;
}

psupermatrix
virtualortho_supermatrix(pccluster row, pccluster col,
			 ph2recompression h2r, pblockcluster blk)
{
  psupermatrix s;
  pclusterbasis rb, cb;
  pclusteroperator rproj, cproj;

  rproj = NULL;
  rb = virtual2row_clusterbasis(row, &rproj, h2r);

  cproj = NULL;
  cb = virtual2col_clusterbasis(col, &cproj, h2r);

  h2r->done = 0;
  h2r->total = getnrnodes_blockcluster(blk);
  if(h2r->disp)
    h2r->disp(0, 0, h2r->total, h2r->dispdata);

  s = projectbuild_supermatrix(blk, rb, rproj, cb, cproj, h2r);

  if(h2r->disp)
    h2r->disp(2, h2r->done, h2r->total, h2r->dispdata);

  del_clusteroperator(cproj);
  del_clusteroperator(rproj);

  return s;
}
			   
/* ************************************************************
    Recompression of pre-orthogonalized virtual H^2 matrices
   ************************************************************ */

static void
splitrow_pcomp(pblockcluster blk, pccluster t,
	       ph2block *active, ph2block *passive,
	       pclusterbasis orb, pclusteroperator orbp,
	       pclusterbasis ocb, pclusteroperator ocbp,
	       ph2recompression h2r)
{
  ph2block b;
  pcluster t0;
  pclusterbasis orb2;
  pclusteroperator orbp2;
  int kr, kc;
  int okr, okc;
  int i, j;

  assert(blk->row == orb->t);
  assert(blk->row == orbp->t);
  assert(blk->col == ocb->t);
  assert(blk->col == ocbp->t);

  if(blk->row != t || (blk->rowsons == 1 && blk->son[0]->row == t)) {
    assert(blk->rowsons > 0);
    assert(blk->colsons > 0);

    t0 = NULL;
    for(i=0; i<blk->rowsons; i++) {
      t0 = blk->son[i]->row;
      if(t0->start <= t->start && t->start + t->size <= t0->start + t0->size)
	break;
    }
    assert(i < blk->rowsons);

    orb2 = orb;
    if(orb2->t != t0) {
      assert(i < orb->sons);
      orb2 = orb->son[i];
    }

    orbp2 = orbp;
    if(orbp2->t != t0) {
      assert(i < orbp->sons);
      orbp2 = orbp->son[i];
    }

    if(blk->colsons > 1)
      for(j=0; j<blk->colsons; j++)
	splitrow_pcomp(blk->son[i+j*blk->rowsons], t,
		       active, passive,
		       orb2, orbp2,
		       ocb->son[j], ocbp->son[j], h2r);
    else
      splitrow_pcomp(blk->son[i], t,
		     active, passive,
		     orb2, orbp2,
		     ocb, ocbp, h2r);
  }
  else {
    assert(blk->row == t);
    assert(blk->adm == 0 || blk->adm == 3);

    switch(blk->adm) {
    case 0:
      if(blk->rowsons > 0 && blk->colsons > 0) {
	b = (ph2block) malloc(sizeof(h2block));
	assert(b != NULL);
	
	b->blk = blk;
	b->orb = orb;
	b->orbp = orbp;
	b->ocb = ocb;
	b->ocbp = ocbp;
	b->S = NULL;
	b->kr = 0;
	b->kc = 0;
	b->M = NULL;
	b->rowproj = NULL;
	b->next = *passive;
	*passive = b;
      }
      break;
    case 3:
      b = (ph2block) malloc(sizeof(h2block));
      assert(b != NULL);

      kr = h2r->getrankrow(blk->row, h2r->data);
      kc = h2r->getrankcol(blk->col, h2r->data);
      okr = orb->kt;
      okc = ocb->kt;

      b->blk = blk;
      b->orb = orb;
      b->orbp = orbp;
      b->ocb = ocb;
      b->ocbp = ocbp;
      b->S = (double *) malloc((size_t) sizeof(double) * okr * okc);
      assert(b->S != NULL);
      b->kr = okr;
      b->kc = okc;
      b->M = NULL;
      b->rowproj = NULL;
      b->next = *active;
      *active = b;

      provide_array(kr * (kc + okc), &h2r->tmp, &h2r->tmpsize);
      h2r->buildS(blk->row, blk->col, h2r->data, h2r->tmp, kr);

      dgemm_("Not Transposed", "Transposed",
	     &kr, &okc, &kc,
	     deins_,
	     h2r->tmp, &kr,
	     ocbp->P, &okc,
	     dnull_,
	     h2r->tmp + kr * kc, &kr);

      dgemm_("Not Transposed", "Not Transposed",
	     &okr, &okc, &kr,
	     deins_,
	     orbp->P, &okr,
	     h2r->tmp + kr * kc, &kr,
	     dnull_,
	     b->S, &okr);
      break;
    default:
      break;
    }
  }
}

static void
splitcol_pcomp(pblockcluster blk, pccluster t,
	       pclusterbasis row, pclusterbasis col, pclusteroperator proj,
	       psupermatrix *M,
	       ph2block *active, ph2block *passive,
	       pclusterbasis orb, pclusteroperator orbp,
	       pclusterbasis ocb, pclusteroperator ocbp,
	       ph2recompression h2r)
{
  ph2block b;
  pcluster t0;
  pclusterbasis ocb2;
  pclusteroperator ocbp2;
  puniformmatrix u;
  pfullmatrix f;
  int kr, kc;
  int okr, okc;
  int i, j;

  assert(M != NULL);
  assert(blk->col == ocb->t);
  assert(blk->col == ocbp->t);
  assert(blk->row == orb->t);
  assert(blk->row == orbp->t);
  assert(blk->row == row->t);
  assert(blk->row == proj->t);

  if(blk->col != t || (blk->colsons == 1 && blk->son[0]->col == t)) {
    assert(blk->rowsons > 0);
    assert(blk->colsons > 0);
    assert(blk->rowsons == 1 || blk->rowsons == row->sons);
    assert(blk->rowsons == 1 || blk->rowsons == proj->sons);

    if(*M == NULL) {
      *M = new_supermatrix(blk->rowsons, blk->colsons,
			   blk->row->size, blk->col->size,
			   NULL, NULL, NULL);

      for(i=0; i<(*M)->block_rows * (*M)->block_cols; i++)
	(*M)->s[i] = NULL;
    }

    assert((*M)->block_rows == blk->rowsons);
    assert((*M)->block_cols == blk->colsons);
    assert((*M)->rows == blk->row->size);
    assert((*M)->cols == blk->col->size);

    t0 = NULL;
    for(i=0; i<blk->colsons; i++) {
      t0 = blk->son[i*blk->rowsons]->col;
      if(t0->start <= t->start && t->start + t->size <= t0->start + t0->size)
	break;
    }
    assert(i < blk->colsons);

    ocb2 = ocb;
    if(ocb2->t != t0) {
      assert(i < ocb->sons);
      ocb2 = ocb->son[i];
    }

    ocbp2 = ocbp;
    if(ocbp2->t != t0) {
      assert(i < ocbp->sons);
      ocbp2 = ocbp->son[i];
    }

    if(blk->rowsons > 1)
      for(j=0; j<blk->rowsons; j++)
	splitcol_pcomp(blk->son[j+i*blk->rowsons], t,
		       row->son[j], col, proj->son[j],
		       &(*M)->s[j+i*(*M)->block_rows], active, passive,
		       orb->son[j], orbp->son[j],
		       ocb2, ocbp2, h2r);
    else {
      assert(blk->rowsons == 1);
      splitcol_pcomp(blk->son[i], t,
		     row, col, proj,
		     &(*M)->s[i], active, passive,
		     orb, orbp,
		     ocb2, ocbp2, h2r);
    }
  }
  else {
    assert(blk->col == ocb->t);
    assert(blk->col == ocbp->t);
    assert(blk->col == t);
    assert(blk->adm == 0 || blk->adm == 3);

    switch(blk->adm) {
    case 0:
      if(blk->rowsons > 0 && blk->colsons > 0) {
	b = (ph2block) malloc(sizeof(h2block));
	assert(b != NULL);

	if(*M == NULL) {
	  *M = new_supermatrix(blk->rowsons, blk->colsons,
			       blk->row->size, blk->col->size,
			       NULL, NULL, NULL);
	  (*M)->row = row;
	  (*M)->col = col;

	  for(i=0; i<(*M)->block_rows * (*M)->block_cols; i++)
	    (*M)->s[i] = NULL;
	}
	assert((*M)->s != NULL);
	
	b->blk = blk;
	b->orb = orb;
	b->orbp = orbp;
	b->ocb = ocb;
	b->ocbp = ocbp;
	b->S = NULL;
	b->kr = 0;
	b->kc = 0;
	b->M = *M;
	b->row = row;
	b->rowproj = proj;
	b->next = *passive;
	*passive = b;
      }
      else {
	if(*M == NULL) {
	  f = new_fullmatrix(blk->row->size, blk->col->size);
	  *M = new_supermatrix(1, 1, f->rows, f->cols,
			       NULL, NULL, f);
	  (*M)->row = row;
	  (*M)->col = col;
	  h2r->buildN(blk->row, blk->col, h2r->data, f->e, f->rows);
	}
	assert((*M)->f != NULL);
      }
      break;
    case 3:
      b = (ph2block) malloc(sizeof(h2block));
      assert(b != NULL);

      if(*M == NULL) {
	u = new_uniformmatrix(NULL, NULL);
	*M = new_supermatrix(1, 1, blk->row->size, blk->col->size,
			     u, NULL, NULL);
      }
      assert((*M)->u != NULL);

      kr = h2r->getrankrow(blk->row, h2r->data);
      kc = h2r->getrankcol(blk->col, h2r->data);
      okr = orb->kt;
      okc = ocb->kt;

      b->blk = blk;
      b->orb = orb;
      b->orbp = orbp;
      b->ocb = ocb;
      b->ocbp = ocbp;
      b->S = (double *) malloc((size_t) sizeof(double) * okr * okc);
      assert(b->S != NULL);
      b->kr = okr;
      b->kc = okc;
      b->M = *M;
      b->row = row;
      b->rowproj = proj;
      b->next = *active;
      *active = b;

      provide_array(kr * (kc + okc), &h2r->tmp, &h2r->tmpsize);
      h2r->buildS(blk->row, blk->col, h2r->data, h2r->tmp, kr);

      dgemm_("Not Transposed", "Transposed",
	     &kr, &okc, &kc,
	     deins_,
	     h2r->tmp, &kr,
	     ocbp->P, &okc,
	     dnull_,
	     h2r->tmp + kr * kc, &kr);

      dgemm_("Not Transposed", "Not Transposed",
	     &okr, &okc, &kr,
	     deins_,
	     orbp->P, &okr,
	     h2r->tmp + kr * kc, &kr,
	     dnull_,
	     b->S, &okr);
      break;
    default:
      break;
    }
  }
}

static pclusterbasis
build_pbasis_h2(pccluster t, pclusteroperator *proj,
		ph2block active, ph2block passive,
		pclusterbasis ocb, pclusteroperator ocbp,
		double *G, int ldG, ph2recompression h2r)
{
  pclusterbasis cb;
  ph2block active2, passive2;
  ph2block b, bnext;
  double *G2;
  int ldG2;
  double *X;
  int ldX;
  double *lambda, *work;
  double h, fmax, sqeps;
  int k, k2, smax, kt, kmax;
  int m, off;
  int ldwork, info;
  int i, j;

  assert(t != NULL);
  assert(G != NULL);
  assert(h2r != NULL);

  /* Get rank of original cluster basis */

  k = ocb->kt;

  assert(k == ldG);

  /* Create new cluster basis and corresponding projection */

  cb = new_clusterbasis(t, 0, 0);

  if(proj)
    *proj = new_clusteroperator(t, 0, 0);

  /* Determine active blocks */

  active2 = active;
  passive2 = NULL;
  for(b=passive; b!=NULL; b=b->next)
    if(h2r->transposed)
      splitcol_pcomp(b->blk, t, b->row, cb, b->rowproj, &b->M, &active2, &passive2, b->orb, b->orbp, b->ocb, b->ocbp, h2r);
    else
      splitrow_pcomp(b->blk, t, &active2, &passive2, b->orb, b->orbp, b->ocb, b->ocbp, h2r);

  assert(t->sons > 0 || passive2 == NULL);

  /* Update Gram matrix */

  for(b=active2; b!=active; b=b->next) {
    assert(b != NULL);
    if(h2r->transposed) {
      assert(b->kc == k);

      assert(ldG == b->kc);
      dgemm_("Transposed", "Not Transposed",
	     &b->kc, &b->kc, &b->kr,
	     deins_,
	     b->S, &b->kr,
	     b->S, &b->kr,
	     deins_,
	     G, &ldG);
    }
    else {
      assert(b->kr == k);

      assert(ldG == b->kr);
      dgemm_("Not Transposed", "Transposed",
	     &b->kr, &b->kr, &b->kc,
	     deins_,
	     b->S, &b->kr,
	     b->S, &b->kr,
	     deins_,
	     G, &ldG);
    }
  }

  /* Find expansion coefficients */

  m = 0;
  X = NULL;
  if(t->sons > 0) {
    X = (double *) malloc((size_t) sizeof(double) * t->sons * h2r->kmax * k);
    assert(X != NULL);
    ldX = k;

    smax = 0;
    for(i=0; i<t->sons; i++)
      smax = imax(smax, ocb->son[i]->kt);

    G2 = (double *) malloc((size_t) sizeof(double) * smax * smax);
    assert(G2 != NULL);

    for(i=0; i<t->sons; i++) {
      k2 = ocb->son[i]->kt;
      assert(k2 <= smax);

      ldG2 = k2;

      provide_array(k * k2, &h2r->tmp, &h2r->tmpsize);

      dgemm_("Not Transposed", "Transposed",
	     &k, &k2, &k,
	     deins_,
	     G, &ldG,
	     ocb->T[i], &k2,
	     dnull_,
	     h2r->tmp, &k);
      
      dgemm_("Not Transposed", "Not Transposed",
	     &k2, &k2, &k,
	     deins_,
	     ocb->T[i], &k2,
	     h2r->tmp, &k,
	     dnull_,
	     G2, &ldG2);

      cb->son[i] = build_pbasis_h2(t->son[i], (proj ? &(*proj)->son[i] : NULL),
				   active2, passive2, ocb->son[i], ocbp->son[i],
				   G2, ldG2, h2r);

      dgemm_("Transposed", "Transposed",
	     &k, &cb->son[i]->kt, &k2,
	     deins_,
	     ocb->T[i], &k2,
	     G2, &ldG2,
	     dnull_,
	     X + m*ldX, &ldX);

      m += cb->son[i]->kt;
    }

    free(G2);
    G2 = NULL;
  }
  else {
    m = t->size;

    X = (double *) malloc((size_t) sizeof(double) * m * k);
    assert(X != NULL);
    ldX = k;

    for(i=0; i<m; i++)
      dcopy_(&k, ocb->V + i, &m, X + i*k, eins_);
  }

  /* Expand Gram matrix */

  provide_array(m * k, &h2r->tmp, &h2r->tmpsize);
  dgemm_("Not Transposed", "Not Transposed",
	 &k, &m, &k,
	 deins_,
	 G, &ldG,
	 X, &ldX,
	 dnull_,
	 h2r->tmp, &k);

  provide_array(m * m, &h2r->Q, &h2r->Qsize);
  dgemm_("Transposed", "Not Transposed",
	 &m, &m, &k,
	 deins_,
	 X, &ldX,
	 h2r->tmp, &k,
	 dnull_,
	 h2r->Q, &m);

  /* Solve eigenvalue problem */

  ldwork = 8 * m;
  provide_array(ldwork + m, &h2r->tmp, &h2r->tmpsize);
  work = h2r->tmp;
  lambda = h2r->tmp + ldwork;

  dsyev_("Vectors Wanted", "Upper half given",
	 &m, h2r->Q, &m, lambda, work, &ldwork, &info);
  assert(info == 0);
	 
  for(i=0,j=m-1; i<j; i++,j--) {
    swap_lapack(m, h2r->Q+i*m, h2r->Q+j*m);
    h = lambda[i];
    lambda[i] = lambda[j];
    lambda[j] = h;
  }

  /* Determine rank of new cluster basis */

  sqeps = h2r->sqeps;
  kmax = h2r->kmax;

  switch(h2r->strategy) {
  case HLIB_FROBENIUS_ABSOLUTE:
    h = 0.0;
    kt = m;
    while(kt > 1 && h + fabs(lambda[kt-1]) < sqeps) {
      kt--;
      h += fabs(lambda[kt]);
    }
    break;
  default:
  case HLIB_FROBENIUS_RELATIVE:
    fmax = 0.0;
    for(i=0; i<m; i++)
      fmax += fabs(lambda[i]);

    if(fmax < RECOMP_EPS)
      kt = 1;
    else {
      h = 0.0;
      kt = m;
      while(kt > 1 && h + fabs(lambda[kt-1]) < sqeps * fmax) {
	kt--;
	h += fabs(lambda[kt]);
      }
    }
    break;
  case HLIB_EUCLIDEAN_RELATIVE:
    kt = 1;
    while(kt < m && lambda[kt] > sqeps*lambda[0])
      kt++;
    break;
  }

  if(kt > kmax)
    kt = kmax;

  assert(kt > 0);

  reallocstructure_clusterbasis(cb, kt);
  cb->kt = kt;

#ifdef DEBUG_LOCAL_SVD
  printf("%d:", kt);
  for(i=0; i<m; i++)
    printf(" %g", lambda[i]);
  printf("\n");
#endif

  /* Extract transfer or leaf matrices from eigenvectors */

  if(t->sons > 0) {
    off = 0;
    for(i=0; i<t->sons; i++) {
      for(j=0; j<kt; j++)
	copy_lapack(cb->son[i]->kt,
		    h2r->Q + off + j*m,
		    cb->T[i] + j*cb->son[i]->kt);
      
      off += cb->son[i]->kt;
    }
    assert(off == m);
  }
  else
    copy_lapack(kt * m, h2r->Q, cb->V);

  assert(X != NULL);

  /* Find cluster basis projection */

  dgemm_("Transposed", "Transposed",
	 &kt, &k, &m,
	 deins_,
	 h2r->Q, &m,
	 X, &k,
	 dnull_,
	 G, &ldG);

  if(proj) {
    reallocstructure_clusteroperator(*proj, kt, k);
    for(i=0; i<k; i++)
      dcopy_(&kt, G+i*ldG, eins_, (*proj)->P+i*kt, eins_);
    (*proj)->ktr = kt;
    (*proj)->ktc = k;
  }
  
  free(X);
  X = NULL;

  /* Retire active blocks */

  b = active2;
  while(b != active) {
    assert(b != NULL);

    bnext = b->next;

    if(h2r->transposed) {
      assert(b->M != NULL);
      assert(b->row != NULL);
      assert(b->rowproj != NULL);

      b->M->row = b->row;
      b->M->col = cb;
      reallocstructure_uniformmatrix(b->M->u, b->row, cb);
    
      assert(b->rowproj->ktc == b->kr);
      assert(b->rowproj->ktr == b->row->kt);

      provide_array(b->kr * kt, &h2r->tmp, &h2r->tmpsize);

      dgemm_("Not Transposed", "Transposed",
	     &b->kr, &kt, &k,
	     deins_,
	     b->S, &b->kr,
	     G, &ldG,
	     dnull_,
	     h2r->tmp, &b->kr);
      dgemm_("Not Transposed", "Not Transposed",
	     &b->row->kt, &kt, &b->kr,
	     deins_,
	     b->rowproj->P, &b->row->kt,
	     h2r->tmp, &b->kr,
	     dnull_,
	     b->M->u->S, &b->row->kt);
    }

    free(b->S);
    b->S = NULL;

    free(b);
    b = bnext;
  }

  /* Retire passive blocks */

  b = passive2;
  while(b != NULL) {
    bnext = b->next;

    free(b);
    b = bnext;
  }

  h2r->done++;
  if(h2r->disp)
    h2r->disp(1, h2r->done, h2r->total, h2r->dispdata);

  return cb;
}

psupermatrix
virtual2_supermatrix(pccluster row, pccluster col,
		     ph2recompression h2r, pblockcluster blk)
{
  pclusterbasis rb, cb;
  pclusterbasis orb, ocb;
  pclusteroperator orbp, ocbp;
  pclusteroperator rowproj;
  h2block b;
  double *G;
  int ldG;

  assert(h2r != NULL);
  assert(blk->rowsons > 0);
  assert(blk->colsons > 0);

#ifdef VERBOSE_RECOMPRESSION
  (void) printf("  Orthogonalizing row cluster basis\n");
#endif

  orbp = NULL;
  orb = virtual2row_clusterbasis(row, &orbp, h2r);

#ifdef VERBOSE_RECOMPRESSION
  (void) printf("  Orthogonalizing column cluster basis\n");
#endif

  ocbp = NULL;
  ocb = virtual2col_clusterbasis(col, &ocbp, h2r);

#ifdef VERBOSE_RECOMPRESSION
  (void) printf("  Building row cluster basis\n");
#endif

  h2r->transposed = 0;
  h2r->sqeps = (h2r->strategy == HLIB_FROBENIUS_ABSOLUTE ?
		dsqr(h2r->eps) / count_cluster(row) / 2.0 :
		dsqr(h2r->eps) / 2.0);

  b.blk = blk;
  b.ww = NULL;
  b.orb = orb;
  b.orbp = orbp;
  b.ocb = ocb;
  b.ocbp = ocbp;
  b.S = NULL;
  b.kr = 0;
  b.kc = 0;
  b.M = NULL;
  b.row = NULL;
  b.rowproj = NULL;
  b.next = NULL;

  ldG = orb->kt;
  G = (double *) malloc((size_t) sizeof(double) * ldG * ldG);
  assert(G != NULL);
  clear_vector(ldG * ldG, G);

  if(h2r->disp) {
    h2r->total = count_cluster(row);
    h2r->done = 0;
    h2r->disp(0, h2r->done, h2r->total, h2r->dispdata);
  }

  rowproj = NULL;
  rb = build_pbasis_h2(row, &rowproj, NULL, &b, orb, orbp, G, ldG, h2r);

  if(h2r->disp)
    h2r->disp(2, h2r->done, h2r->total, h2r->dispdata);

  free(G);
  G = NULL;

#ifdef VERBOSE_RECOMPRESSION
  (void) printf("  Building column cluster basis and matrix\n");
#endif

  h2r->transposed = 1;
  h2r->sqeps = (h2r->strategy == HLIB_FROBENIUS_ABSOLUTE ?
		dsqr(h2r->eps) / count_cluster(col) / 2.0 :
		dsqr(h2r->eps) / 2.0);

  b.orb = orb;
  b.orbp = orbp;
  b.ocb = ocb;
  b.ocbp = ocbp;
  b.M = NULL;
  b.row = rb;
  b.rowproj = rowproj;

  ldG = ocb->kt;
  G = (double *) malloc((size_t) sizeof(double) * ldG * ldG);
  assert(G != NULL);
  clear_vector(ldG * ldG, G);

  if(h2r->disp) {
    h2r->total = count_cluster(col);
    h2r->done = 0;
    h2r->disp(0, h2r->done, h2r->total, h2r->dispdata);
  }

  cb = build_pbasis_h2(col, NULL, NULL, &b, ocb, ocbp, G, ldG, h2r);

  if(h2r->disp)
    h2r->disp(2, h2r->done, h2r->total, h2r->dispdata);

  free(G);
  G = NULL;

#ifdef VERBOSE_RECOMPRESSION
  (void) printf("  Temporary storage requirements: %.1f MB\n",
		(getsize_clusterbasis(orb) +
		 getsize_clusteroperator(orbp) +
		 getsize_clusterbasis(ocb) +
		 getsize_clusteroperator(ocbp) +
		 getsize_clusteroperator(rowproj)) / 1048576.0);
#endif

  del_clusteroperator(rowproj);

  del_clusteroperator(orbp);
  orbp = NULL;
  del_clusterbasis(orb);
  orb = NULL;

  del_clusteroperator(ocbp);
  ocbp = NULL;
  del_clusterbasis(ocb);
  ocb = NULL;

  return b.M;
}

/* ============================================================

   Virtual ACA recompression

   ============================================================ */

typedef struct _acablock acablock;
typedef acablock *pacablock;

typedef struct _acaconv acaconv;
typedef acaconv *pacaconv;

struct _acablock {
  psupermatrix original;
  
  double *a;
  double *b;
  double *gb;
  int k;

  int rstart;
  int cstart;
  int start;
  int size;

  pacablock next;
};

struct _acaconv {
  double *evp;
  int evpsize;
  double *tmp;
  int tmpsize;

  double (*entry)(int row, int col, void *data);
  void *data;
  ACAStrategy acastrategy;
  int acakmax;
  double acaeps;

  TruncationStrategy strategy;
  int kmax;
  double sqeps;
  int transposed;
};

static void
del_acablock(pacablock b)
{
  assert(b != NULL);
  assert(b->a != NULL);
  assert(b->b != NULL);
  assert(b->gb != NULL);

  free(b->gb);
  b->gb = NULL;

  free(b->b);
  b->b = NULL;

  free(b->a);
  b->a = NULL;

  free(b);
}

static void
split_row(psupermatrix s, int rstart, int cstart,
	  pacablock *super, pacablock *rk, pacaconv aux,
	  int start, int size)
{
  pacablock b;
  double *A;
  double *B;
  int kt;
  int roff, coff;
  int i, j;

  assert(rstart <= start);
  assert(start + size <= rstart + s->rows);

  if(s->rows > size || (s->s && s->block_rows == 1)) {
    assert(s->s != NULL);

    roff = rstart;
    for(i=0; i<s->block_rows; i++) {
      if(start + size <= roff + s->s[i]->rows)
	break;
      roff += s->s[i]->rows;
    }

    assert(i < s->block_rows);
    
    coff = cstart;
    for(j=0; j<s->block_cols; j++) {
      split_row(s->s[i+j*s->block_rows], roff, coff, super, rk, aux, start, size);
      coff += s->s[i+j*s->block_rows]->cols;
    }
    assert(coff == cstart + s->cols);
  }
  else {
    assert(rstart == start);
    assert(s->rows == size);

    if(s->s) {
      b = (pacablock) malloc(sizeof(acablock));
      assert(b != NULL);

      b->original = s;
      b->a = NULL;
      b->b = NULL;
      b->gb = NULL;
      b->k = 0;
      b->rstart = rstart;
      b->cstart = cstart;
      b->start = rstart;
      b->size = s->rows;

      b->next = *super;
      *super = b;
    }
    else if(s->r) {
      s->row = NULL;

      provide_array((s->rows + s->cols) * aux->acakmax, &aux->tmp, &aux->tmpsize);
      A = aux->tmp;
      B = aux->tmp + s->rows * aux->acakmax;

      kt = aca_fill_block(A, B, s->rows, s->cols, rstart, cstart,
			  aux->entry, aux->data, aux->acastrategy,
			  aux->acakmax, aux->acaeps);

      b = (pacablock) malloc(sizeof(acablock));
      assert(b != NULL);

      b->original = s;
      b->a = (double *) malloc((size_t) sizeof(double) * s->rows * kt);
      assert(b->a != NULL);
      b->b = (double *) malloc((size_t) sizeof(double) * s->cols * kt);
      assert(b->b != NULL);
      b->gb = (double *) malloc((size_t) sizeof(double) * kt * kt);
      assert(b->gb != NULL);
      b->k = kt;
      b->rstart = rstart;
      b->cstart = cstart;
      b->start = rstart;
      b->size = s->rows;

      b->next = *rk;
      *rk = b;

      copy_lapack(s->rows * kt, A, b->a);
      copy_lapack(s->cols * kt, B, b->b);
      multrans1_lapack(kt, kt, s->cols, B, B, b->gb);
    }
    else
      assert(s->f != NULL);
  }
}

static void
retire_inadmissible_row(pacablock b, pclusterbasis cb)
{
  psupermatrix s;

  s = b->original;
  assert(s->s != NULL);

  if(s->row == NULL)
    s->row = cb;
}

static void
retire_row(pacablock b, pclusterbasis cb, pacaconv aux)
{
  psupermatrix s;

  s = b->original;
  assert(s->r != NULL);
  assert(cb->n == s->rows);

  if(s->row == NULL)
    s->row = cb;
}

static void
split_col(psupermatrix s, int rstart, int cstart,
	  pacablock *super, pacablock *rk, pacaconv aux,
	  int start, int size)
{
  pacablock b;
  double *A, *B;
  int kt;
  int roff, coff;
  int i, j;
  
  assert(cstart <= start);
  assert(start + size <= cstart + s->cols);

  if(s->cols > size || (s->s && s->block_cols == 1)) {
    assert(s->s != NULL);

    coff = cstart;
    for(i=0; i<s->block_cols; i++) {
      if(start + size <= coff + s->s[i*s->block_rows]->cols)
	break;
      coff += s->s[i*s->block_rows]->cols;
    }

    assert(i < s->block_cols);

    roff = rstart;
    for(j=0; j<s->block_rows; j++) {
      split_col(s->s[j+i*s->block_rows], roff, coff, super, rk, aux, start, size);
      roff += s->s[j+i*s->block_rows]->rows;
    }
    assert(roff == rstart + s->rows);
  }
  else {
    assert(cstart == start);
    assert(s->cols == size);

    if(s->s) {
      b = (pacablock) malloc(sizeof(acablock));
      assert(b != NULL);

      b->original = s;
      b->a = NULL;
      b->gb = NULL;
      b->k = 0;
      b->rstart = rstart;
      b->cstart = cstart;
      b->start = cstart;
      b->size = s->cols;

      b->next = *super;
      *super = b;
    }
    else if(s->r) {
      s->col = NULL;
      
      provide_array((s->rows + s->cols) * aux->acakmax, &aux->tmp, &aux->tmpsize);
      A = aux->tmp;
      B = aux->tmp + s->rows * aux->acakmax;
      kt = aca_fill_block(A, B, s->rows, s->cols, rstart, cstart,
			  aux->entry, aux->data, aux->acastrategy,
			  aux->acakmax, aux->acaeps);

      b = (pacablock) malloc(sizeof(acablock));
      assert(b != NULL);

      b->original = s;
      b->a = (double *) malloc((size_t) sizeof(double) * s->cols * kt);
      assert(b->a != NULL);
      b->b = (double *) malloc((size_t) sizeof(double) * s->rows * kt);
      assert(b->b != NULL);
      b->gb = (double *) malloc((size_t) sizeof(double) * kt * kt);
      assert(b->gb != NULL);
      b->k = kt;
      b->rstart = rstart;
      b->cstart = cstart;
      b->start = cstart;
      b->size = s->cols;

      b->next = *rk;
      *rk = b;

      copy_lapack(s->cols * kt, B, b->a);
      copy_lapack(s->rows * kt, A, b->b);
      multrans1_lapack(kt, kt, s->r->rows, A, A, b->gb);
    }
    else {
      assert(s->f != NULL);
      assert(s->f->rows == s->rows);
      assert(s->f->cols == s->cols);

      for(i=0; i<s->cols; i++)
	for(j=0; j<s->rows; j++)
	  s->f->e[j+i*s->rows] = aux->entry(j+rstart, i+cstart, aux->data);
    }
  }
}

static void
retire_inadmissible_col(pacablock b, pclusterbasis cb)
{
  psupermatrix s;

  s = b->original;
  assert(s->s != NULL);

  if(s->col == NULL)
    s->col = cb;
}

static void
retire_col(pacablock b, pclusterbasis cb, pacaconv aux)
{
  psupermatrix s;
  puniformmatrix u;
  int i;

  s = b->original;
  assert(s->r != NULL);
  assert(cb->n == s->cols);

  if(s->col == NULL) {
    assert(s->r != NULL);
    assert(s->u == NULL);

    assert(s->row != NULL);
    assert(s->cols == b->size);
    assert(b->k > 0);

    s->col = cb;
    
    del_rkmatrix(s->r);
    s->r = NULL;
    
    u = new_uniformmatrix(s->row, s->col);
    s->u = u;

    assert(u->ktr == s->row->kt);
    assert(u->ktc == s->col->kt);
    
    provide_array(u->ktr * b->k, &aux->tmp, &aux->tmpsize);
    
    for(i=0; i<b->k; i++) {
      forward_clusterbasis(u->row, b->b + i*s->rows);
      copy_lapack(u->ktr, u->row->xt, aux->tmp + i*u->ktr);
    }

    dgemm_("Not Transposed", "Transposed",
	   &u->ktr, &u->ktc, &b->k,
	   deins_,
	   aux->tmp, &u->ktr,
	   b->a, &b->size,
	   dnull_,
	   u->S, &u->ktr);
  }
}

static pclusterbasis
build_basis_aca(pccluster t, pacablock super, pacablock rk, pacaconv aux)
{
  pclusterbasis cb;
  double *G;
  double *lambda;
  double *work;
  int start = t->start;
  int size = t->size;
  int kmax = aux->kmax;
  double sqeps = aux->sqeps;
  int lwork, info;
  double h;
  pacablock super1;
  pacablock rk1;
  psupermatrix s;
  pacablock b;
  int m_tau, kt;
  int i, j, off, off2;

  /* Create new clusterbasis */
  cb = new_clusterbasis(t, 0, 0);

  /* Add new entries to block list */
  super1 = NULL;
  rk1 = rk;
  b = super;
  while(b) {
    s = b->original;

    if(aux->transposed)
      split_col(s, b->rstart, b->cstart, &super1, &rk1, aux, start, size);
    else
      split_row(s, b->rstart, b->cstart, &super1, &rk1, aux, start, size);

    b = b->next;
  }

  /* Determine size of EVP and build sons bases if necessary */
  m_tau = 0;
  if(t->sons > 0)
    for(i=0; i<t->sons; i++) {
      cb->son[i] = build_basis_aca(t->son[i], super1, rk1, aux);
      
      m_tau += cb->son[i]->kt;
    }
  else
    m_tau = size;

  /* Provide storage for EVP */
  provide_array(m_tau * (m_tau+7), &aux->evp, &aux->evpsize);
  G = aux->evp;
  lambda = aux->evp + m_tau * m_tau;
  work = aux->evp + m_tau * (m_tau+1);
  lwork = 6*m_tau;

  /* Initialize Gram matrix */
  clear_vector(m_tau * m_tau, G);

  if(t->sons > 0) {
    /* Build transformed Gram matrix */
    b = rk1;
    while(b) {
      assert(b->start <= start);
      assert(start + size <= b->start + b->size);

      provide_array(m_tau * b->k, &aux->tmp, &aux->tmpsize);
      
      off = start;
      off2 = 0;
      for(i=0; i<t->sons; i++) {
	dgemm_("Not Transposed", "Not Transposed",
	       &cb->son[i]->kt, &b->k, &b->k,
	       deins_,
	       b->a + off - b->start, &b->size,
	       b->gb, &b->k,
	       dnull_,
	       aux->tmp + off2, &m_tau);
	off += t->son[i]->size;
	off2 += cb->son[i]->kt;
      }
      assert(off == start + size);
      assert(off2 == m_tau);

      off = start;
      off2 = 0;
      for(i=0; i<t->sons; i++) {
	dgemm_("Not Transposed", "Transposed",
	       &m_tau, &cb->son[i]->kt, &b->k,
	       deins_,
	       aux->tmp, &m_tau,
	       b->a + off - b->start, &b->size,
	       deins_,
	       G + off2 * m_tau, &m_tau);
	off += t->son[i]->size;
	off2 += cb->son[i]->kt;
      }

      b = b->next;
    }
  }
  else {
    /* Build Gram matrix */
    b = rk1;
    while(b) {
      assert(b->start <= start);
      assert(start + size <= b->start + b->size);

      provide_array(m_tau * b->k, &aux->tmp, &aux->tmpsize);

      dgemm_("Not Transposed", "Not Transposed",
	     &m_tau, &b->k, &b->k,
	     deins_,
	     b->a + start - b->start, &b->size,
	     b->gb, &b->k,
	     dnull_,
	     aux->tmp, &m_tau);

      dgemm_("Not Transposed", "Transposed",
	     &m_tau, &m_tau, &b->k,
	     deins_,
	     aux->tmp, &m_tau,
	     b->a + start - b->start, &b->size,
	     deins_,
	     G, &m_tau);

      b = b->next;
    }
  }

  /* Solve eigenvalue problem */
  dsyev_("Vectors", "Upper", &m_tau, G, &m_tau,
	 lambda, work, &lwork, &info);
  assert(info == 0);

  /* Sort eigenvalues and eigenvectors */
  for(i=0,j=m_tau-1; i<j; i++,j--) {
    swap_lapack(m_tau, G+i*m_tau, G+j*m_tau);
    h = lambda[i];
    lambda[i] = lambda[j];
    lambda[j] = h;
  }

  /* Determine appropriate rank */
  switch(aux->strategy) {
  case HLIB_FROBENIUS_ABSOLUTE:
    h = 0.0;
    kt = m_tau;
    while(kt > 1 && h + fabs(lambda[kt-1]) < sqeps) {
      kt--;
      h += fabs(lambda[kt]);
    }
    break;
  default:
  case HLIB_EUCLIDEAN_RELATIVE:
    kt = 1;
    while(kt < m_tau && lambda[kt] > sqeps*lambda[0])
      kt++;
    break;
  }

  if(kt > kmax)
    kt = kmax;

  reallocstructure_clusterbasis(cb, kt);
  cb->kt = kt;

#ifdef DEBUG_LOCAL_SVD
  printf("%d:", kt);
  for(i=0; i<m_tau; i++)
    printf(" %g", lambda[i]);
  printf("\n");
#endif

  /* Copy eigenvectors to cluster basis matrices */
  if(t->sons > 0) {
    off = 0;
    for(i=0; i<t->sons; i++) {
      for(j=0; j<kt; j++)
	copy_lapack(cb->son[i]->kt,
		    G + off + j*m_tau,
		    cb->T[i] + j*cb->son[i]->kt);
      
      off += cb->son[i]->kt;
    }
    assert(off == m_tau);
  }
  else
    copy_lapack(kt * m_tau, G, cb->V);

  /* Apply projection to rank-k-matrices */
  if(t->sons > 0) {
    b = rk1;
    while(b) {
      assert(b->start <= start);
      assert(start + size <= b->start + b->size);

      assert(kt * b->k <= aux->tmpsize);
      clear_vector(kt * b->k, aux->tmp);

      off = start;
      for(i=0; i<t->sons; i++) {
	dgemm_("Transposed", "Not Transposed",
	       &kt, &b->k, &cb->son[i]->kt,
	       deins_,
	       cb->T[i], &cb->son[i]->kt,
	       b->a + off - b->start, &b->size,
	       deins_,
	       aux->tmp, &kt);

	off += t->son[i]->size;
      }
      assert(off == start + size);

      for(i=0; i<b->k; i++)
	copy_lapack(kt, aux->tmp + i*kt, b->a + start - b->start + i*b->size);

      b = b->next;
    }
  }
  else {
    b = rk1;
    while(b) {
      assert(b->start <= start);
      assert(start + size <= b->start + b->size);

      assert(kt * b->k <= aux->tmpsize);

      dgemm_("Transposed", "Not Transposed",
	     &kt, &b->k, &size,
	     deins_,
	     cb->V, &size,
	     b->a + start - b->start, &b->size,
	     dnull_,
	     aux->tmp, &kt);

      for(i=0; i<b->k; i++)
	copy_lapack(kt, aux->tmp + i*kt, b->a + start - b->start + i*b->size);

      b = b->next;
    }
  }

  /* Retire inadmissible blocks */
  while(super1) {
    if(aux->transposed)
      retire_inadmissible_col(super1, cb);
    else
      retire_inadmissible_row(super1, cb);

    b = super1->next;
    free(super1);
    super1 = b;
  }

  /* Retire admissible blocks */
  while(rk1 != rk) {
    assert(rk1 != NULL);
    assert(rk1->start == start);
    assert(rk1->size == size);

    if(aux->transposed)
      retire_col(rk1, cb, aux);
    else
      retire_row(rk1, cb, aux);

    b = rk1->next;
    del_acablock(rk1);
    rk1 = b;
  }

  return cb;
}

void
aca2fill_supermatrix(psupermatrix s, pccluster rct, pccluster cct,
		     double eps, int kmax, TruncationStrategy strategy,
		     double acaeps, int acakmax, ACAStrategy acastrategy,
		     double (*entry)(int row, int col, void *data), void *data)
{
  pclusterbasis rb, cb;
  acaconv aux;
  acablock super;

  aux.evp = NULL;
  aux.evpsize = 0;
  aux.tmp = NULL;
  aux.tmpsize = 0;
  aux.kmax = kmax;
  aux.strategy = strategy;
  aux.acakmax = acakmax;
  aux.acaeps = acaeps;
  aux.acastrategy = acastrategy;
  aux.entry = entry;
  aux.data = data;

  super.original = s;
  super.a = NULL;
  super.gb = NULL;
  super.k = 0;
  super.rstart = 0;
  super.cstart = 0;
  super.start = 0;
  super.size = 0;
  super.next = NULL;

#ifdef VERBOSE_RECOMPRESSION
  (void) printf("  Building row cluster basis\n");
#endif

  aux.transposed = 0;
  aux.sqeps = (strategy == HLIB_FROBENIUS_ABSOLUTE ?
	       eps * eps / count_cluster(rct) :
	       eps * eps);
  rb = build_basis_aca(rct, &super, NULL, &aux);

#ifdef VERBOSE_RECOMPRESSION
  (void) printf("  Building column cluster basis and matrix\n");
#endif

  aux.transposed = 1;
  aux.sqeps = (strategy == HLIB_FROBENIUS_ABSOLUTE ?
	       eps * eps / count_cluster(cct) :
	       eps * eps);
  cb = build_basis_aca(cct, &super, NULL, &aux);

  if(aux.evp) {
    free(aux.evp);
    aux.evp = NULL;
  }

  if(aux.tmp) {
    free(aux.tmp);
    aux.tmp = NULL;
  }
}
