
/* ------------------------------------------------------------
   This is the file "h2virtual.h" of the HLib package.

   Copyright 1999 Steffen Boerm and Lars Grasedyck
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#ifndef H2Virtual
#define H2Virtual

typedef struct _h2recompression h2recompression;
typedef h2recompression *ph2recompression;
typedef const h2recompression *pch2recompression;

#include "aca.h"
#include "blockcluster.h"
#include "supermatrix.h"

/* ============================================================
   Creating H^2 matrices directly (see the module surfacebemfactory)
   is very fast, but requires also a large amount of memory. In
   BEM applications, we typically do not need all expansion functions,
   so we can try to eliminate redundant ones (see the module
   h2conversion). In order to conserve memory, it is desirable to
   apply this recompression without storing the entire matrix.
   This can be done in two different fashions: We can keep the
   original cluster bases and create the coefficient matrices only
   temporarily, or we can even "virtualize" the cluster bases.
   In both cases, we need a way of creating the matrices
   "just in time", so we introduce the class "h2recompression"
   that contains all necessary callback functions.
   ============================================================ */

struct _h2recompression {
  TruncationStrategy strategy;	/* Rank truncation strategy */
  int kmax;			/* Maximal cluster rank */
  double eps;			/* Truncation tolerance */

  void (*buildN)(pccluster row, pccluster col, /* Nearfield matrices */
		 void *data, double *N, int ldN);
  void (*buildS)(pccluster row, pccluster col, /* Coefficient matrices */
		 void *data, double *S, int ldS);

  int (*getrankrow)(pccluster row, void *data);	/* Row basis rank */
  void (*buildVrow)(pccluster tau, /* Row leaf basis matrices */
		    void *data, double *V, int ldV);
  void (*buildTrow)(pccluster trg, pccluster src, /* Row transfer matrices */
		    void *data, double *T, int ldT);

  int (*getrankcol)(pccluster col, void *data); /* Column basis rank */
  void (*buildVcol)(pccluster tau, /* Column leaf basis matrices */
		    void *data, double *V, int ldV);
  void (*buildTcol)(pccluster trg, pccluster src, /* Column transfer matrices */
		    void *data, double *T, int ldT);
  void *data;			/* Data for callback functions */

  void (*disp)(int mode, int done, int total, void *dispdata); /* Progress reports */
  void *dispdata;		/* Data for disp callback */

  double *tmp;			/* Internal: Auxiliary storage */
  int tmpsize;
  double *Q;			/* Internal: Transformed Gram matrix / Projection */
  int Qsize;

  double sqeps;			/* Internal: Local tolerance */

  int transposed;		/* Internal: Transpose virtual matrix? */

  int done;			/* Internal: Number of completed clusters */
  int total;			/* Internal: Number of clusters to treat */
};

/* Create an empty h2recompression object */
ph2recompression
new_h2recompression();

/* Delete an h2recompression object */
void
del_h2recompression(ph2recompression h2r);

/* Compute a cluster basis product for virtual cluster bases
   t: Root cluster
   h2r: Description of the virtual original matrix */
pclusteroperator
newproductvirtual_clusteroperator(pccluster t, ph2recompression h2r);

/* Build a recompressed H^2-matrix
   row: Row root cluster
   col: Column root cluster
   h2r: Description of the virtual original H^2-matrix
   blk: Description of the desired block cluster tree */
psupermatrix
oldvirtual2_supermatrix(pccluster row, pccluster col,
			ph2recompression h2r, pblockcluster blk);

/* Build an orthogonal row cluster basis
   t: Root cluster
   proj: Pointer to cluster basis projection (!= NULL)
   eps: Desired precision
   strategy: Strategy for the truncation,
   h2r: Description of the virtual original H^2 matrix */
pclusterbasis
virtual2row_clusterbasis(pccluster t, pclusteroperator *proj,
			 ph2recompression h2r);
			   
/* Build an orthogonal column cluster basis
   t: Root cluster
   proj: Pointer to cluster basis projection (!= NULL)
   eps: Desired precision
   strategy: Strategy for the truncation,
   h2r: Description of the virtual H^2 matrix */
pclusterbasis
virtual2col_clusterbasis(pccluster t, pclusteroperator *proj,
			 ph2recompression h2r);

/* Build an orthogonalized H^2-matrix
   row: Row root cluster
   col: Column root cluster
   h2r: Description of the virtual original H^2-matrix
   blk: Description of the desired block cluster tree */
psupermatrix
virtualortho_supermatrix(pccluster row, pccluster col,
			 ph2recompression h2r, pblockcluster blk);

/* Build a recompressed H^2-matrix using pre-orthogonalization
   row: Row root cluster
   col: Column root cluster
   h2r: Description of the virtual original H^2-matrix
   blk: Description of the desired block cluster tree */
psupermatrix
virtual2_supermatrix(pccluster row, pccluster col,
		     ph2recompression h2r, pblockcluster blk);

/* This routine takes a supermatrix containing only fullmatrix and rkmatrix
   blocks. For each rkmatrix block, the ACA algorithm is used to determine
   a virtual low-rank approximation, and this approximation is used to
   find row and column cluster bases. Then all rkmatrix blocks are approximated
   by uniformmatrix blocks.
   The data that may be contained in the original rkmatrix blocks is not
   used, so it is perfectly acceptable for them to represent rank 0 matrices.
   In short: This routine takes the general structure of a supermatrix and
   computes an ACA/H^2 approximation of an operator.
   s: Original supermatrix, overwritten with H^2 matrix
   rct: Row cluster tree
   cct: Column cluster tree
   eps: Tolerance for H^2 compression
   kmax: Maximal rank for cluster bases
   strategy: Truncation strategy for H^2 compression
   acaeps: Tolerance for ACA algorithm
   acakmax: Maximal local rank for virtual ACA blocks
   acastrategy: Truncation strategy for ACA algorithm
   entry: Callback function for computing a matrix entry
   data: Additional data for callback function "entry" */
void
aca2fill_supermatrix(psupermatrix s, pccluster rct, pccluster cct,
		     double eps, int kmax, TruncationStrategy strategy,
		     double acaeps, int acakmax, ACAStrategy acastrategy,
		     double (*entry)(int row, int col, void *data), void *data);

#endif
