
/* ------------------------------------------------------------
   This is the file "net3d.c" of the HLib package.

   Copyright 1999 Lars Grasedyck and Steffen Boerm
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#include "basic.h"
#include "net3d.h"

#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

/*********************************  3D *********************************/

/* Create an empty net3d */
pnet3d
new_net3d(int vertices){
  pnet3d n3;
  
  n3 = (pnet3d) malloc(sizeof(net3d));
  assert(n3!=0x0);
  
  if(vertices>0){
    n3->p = (double (*)[3]) malloc((size_t) sizeof(double [3]) * vertices);
    assert(n3->p!=0x0);
    
    n3->perm = (int *) malloc((size_t) sizeof(int) * vertices);
    assert(n3->perm!=0x0);
    
    n3->iperm = (int *) malloc((size_t) sizeof(int) * vertices);
    assert(n3->iperm!=0x0);
  }else{
    n3->p = 0x0;
    n3->perm = 0x0;
    n3->iperm = 0x0;
  }

  n3->vertices = vertices;
  return n3;
}

pnet3d
buildunitcube_net3d(int verticesperaxis){
  pnet3d n3;
  int x,y,z;
  int i;
  
  n3 = new_net3d(verticesperaxis*verticesperaxis*verticesperaxis);
  for(x=0; x<verticesperaxis; x++){
    for(y=0; y<verticesperaxis; y++){
      for(z=0; z<verticesperaxis; z++){
	i = x + y*verticesperaxis + z*verticesperaxis*verticesperaxis;
	n3->p[i][0] = ((double)x)/((double)verticesperaxis+1.0); 
	n3->p[i][1] = ((double)y)/((double)verticesperaxis+1.0); 
	n3->p[i][2] = ((double)z)/((double)verticesperaxis+1.0); 
	n3->perm[i] = i;
	n3->iperm[i] = i;
      }
    }
  }

  return n3;
}

psparsematrix
buildsparsematrixunitcube_net3d(pnet3d n3, double conv){
  psparsematrix sp;
  psparsefactory fac;
  int i,j,n,x,y,z,x2,y2,z2;
  double coeff,h;
  const int *iperm = n3->iperm;
  
  n = (int) (0.5+pow((double)n3->vertices,1.0/3.0));
  fac = new_sparsefactory(n*n*n,n*n*n);
  h = 1.0/((double)n+1.0);

  for(x=0; x<n; x++){
    for(y=0; y<n; y++){
      for(z=0; z<n; z++){
	i = x + y*n + z*n*n;
	for(x2=x-1; x2<=x+1; x2++){
	  for(y2=y-1; y2<=y+1; y2++){
	    for(z2=z-1; z2<=z+1; z2++){
	      if(x2>=0 && x2<n && y2>=0 && y2<n && z2>=0 && z2<n){
		j = x2 + y2*n +z2*n*n;
		if(i==j || fabs(x-x2)+fabs(y-y2)+fabs(z-z2)<1.5){
		  addnz_sparsefactory(fac, iperm[i], iperm[j]);
		}
	      }
	    }
	  }
	}
      } 
    } 
  } 
  

  sp = new_sparsematrix(fac);
  for(x=0; x<n; x++){
    for(y=0; y<n; y++){
      for(z=0; z<n; z++){
	i = x + y*n + z*n*n;
	for(x2=x-1; x2<=x+1; x2++){
	  for(y2=y-1; y2<=y+1; y2++){
	    for(z2=z-1; z2<=z+1; z2++){
	      if(x2>=0 && x2<n && y2>=0 && y2<n && z2>=0 && z2<n){
		j = x2 + y2*n +z2*n*n;
		coeff = 0.0;
		if(i==j){
		  coeff = 6.0*h - conv*h*h;
		  addcoeff_sparsematrix(sp, iperm[i], iperm[j], coeff);
		}else{
		  if(fabs(x-x2)+fabs(y-y2)+fabs(z-z2)<1.5){
		    if(y==y2 && z==z2 && x2==x+1){
		      coeff = -1.0*h + conv*h*h;
		    }else{
		      coeff = -1.0*h;
		    }
		    addcoeff_sparsematrix(sp, iperm[i], iperm[j], coeff);
		  }
		}
	      }
	    }
	  }
	}
      } 
    } 
  } 
  del_sparsefactory(fac);

  return sp;
}

/* Free the memory used by a net3d object */
void
del_net3d(pnet3d n3){
  if(n3->p!=0x0) free(n3->p);
  if(n3->perm!=0x0) free(n3->perm);
  if(n3->iperm!=0x0) free(n3->iperm);
  free(n3);
}

/* Build a geometric regular cluster tree for a net3d */
pclustertree
buildcluster_net3d(pnet3d n3, ClusterStrategy strategy, int leafsize){
  pclustertree ct;
  pclusterfactory fac;
  const double (*p)[3] = (const double (*)[3]) n3->p;
  int n;
  int i, j;
  double gmax[3], gmin[3], vol;

  /* degrees of freedom */
  n = n3->vertices;

  fac = new_clusterfactory(n, n, 3);

  /* Initialize x, smin and smax (bounding box for support)
     WARNING: These heuristics apply only to volume geometries! */

  for(j=0; j<3; j++){
    gmin[j] = p[0][j];
    gmax[j] = p[0][j];
  }
  for(i=1; i<n; i++) {
    for(j=0; j<3; j++) {
      if(p[i][j] < gmin[j])
	gmin[j] = p[i][j];
      if(p[i][j] > gmax[j])
	gmax[j] = p[i][j];
    }
  }
  vol = (gmax[0] - gmin[0]) * (gmax[1] - gmin[1]) * (gmax[2] - gmin[2]);
  vol = pow(vol / n, 1.0/3.0);

  j = 0;
  for(i=0; i<n; i++)
    for(j=0; j<3; j++) {
      fac->x[i][j] = p[i][j];
      fac->smin[i][j] = p[i][j] - 0.5*vol;
      fac->smax[i][j] = p[i][j] + 0.5*vol;
    }

  /* Create the cluster tree */
  ct = create_clustertree(fac, strategy, leafsize);

  /* Copy the idx2dof and dof2idx arrays */
  (void) memcpy(n3->perm, ct->dof2idx, (size_t) sizeof(int) * ct->ndof);
  (void) memcpy(n3->iperm, ct->idx2dof, (size_t) sizeof(int) * ct->nidx);
  
  del_clusterfactory(fac);

  return ct;
}

/*********************************  2D *********************************/

/* Create an empty net2d */
pnet2d
new_net2d(int vertices){
  pnet2d n2;
  
  n2 = (pnet2d) malloc(sizeof(net2d));
  assert(n2!=0x0);
  
  if(vertices>0){
    n2->p = (double (*)[2]) malloc((size_t) sizeof(double [2]) * vertices);
    assert(n2->p!=0x0);
    
    n2->perm = (int *) malloc((size_t) sizeof(int) * vertices);
    assert(n2->perm!=0x0);
    
    n2->iperm = (int *) malloc((size_t) sizeof(int) * vertices);
    assert(n2->iperm!=0x0);
  }else{
    n2->p = 0x0;
    n2->perm = 0x0;
    n2->iperm = 0x0;
  }

  n2->vertices = vertices;
  return n2;
}

pnet2d
buildunitsquare_net2d(int verticesperaxis){
  pnet2d n2;
  int x,y;
  int i;
  
  n2 = new_net2d(verticesperaxis*verticesperaxis);
  for(x=0; x<verticesperaxis; x++){
    for(y=0; y<verticesperaxis; y++){
      i = x + y*verticesperaxis;
      n2->p[i][0] = ((double)x)/((double)verticesperaxis+1.0); 
      n2->p[i][1] = ((double)y)/((double)verticesperaxis+1.0); 
      n2->perm[i] = i;
      n2->iperm[i] = i;
    }
  }
  return n2;
}

psparsematrix
buildsparsematrixunitsquare_net2d(pnet2d n2, double conv){
  psparsematrix sp;
  psparsefactory fac;
  int i,j,n,x,y,x2,y2;
  double coeff,h;
  double xc,yc;
  const int *iperm = n2->iperm;
  
  n = (int) (0.5+sqrt((double)n2->vertices));
  fac = new_sparsefactory(n*n,n*n);
  h = 1.0/((double)n+1.0);
  
  for(x=0; x<n; x++){
    for(y=0; y<n; y++){
      i = x + y*n;
      for(x2=x-1; x2<=x+1; x2++){
	for(y2=y-1; y2<=y+1; y2++){
	  if(x2>=0 && x2<n && y2>=0 && y2<n){
	    j = x2 + y2*n;
	    if(i==j || fabs(x-x2)+fabs(y-y2)<1.5){
	      addnz_sparsefactory(fac, iperm[i], iperm[j]);
	    }
	  }
	}
      } 
    } 
  } 
  

  sp = new_sparsematrix(fac);
  for(x=0; x<n; x++){
    for(y=0; y<n; y++){
      i = x + y*n;
      for(x2=x-1; x2<=x+1; x2++){
	for(y2=y-1; y2<=y+1; y2++){
	  if(x2>=0 && x2<n && y2>=0 && y2<n){
	    j = x2 + y2*n;
	    coeff = 0.0;
	    xc = 1.0;
	    yc = 0.0;
	    
	    if(i==j){
	      coeff = 4.0-conv*h;
	      addcoeff_sparsematrix(sp, iperm[i], iperm[j], coeff);
	    }else{
	      if(fabs(x-x2)+fabs(y-y2)<1.5){
		if(y==y2 && x2==x+1){
		  if(xc>0)
		    coeff = -1.0 + xc*conv*h; 
		  else
		    coeff = -1.0;
		}
		if(y==y2 && x2==x-1){
		  if(xc<0) coeff = -1.0 - xc*conv*h; 
		  else
		    coeff = -1.0;
		}
		if(y2==y+1 && x2==x){
		  if(yc>0) coeff = -1.0 + yc*conv*h; 
		  else
		    coeff = -1.0;
		}
		if(y2==y-1 && x2==x){
		  if(yc<0) coeff = -1.0 - yc*conv*h; 
		  else
		    coeff = -1.0;
		}
		addcoeff_sparsematrix(sp, iperm[i], iperm[j], coeff);
	      }
	    }
	  }
	}
      } 
    } 
  } 
  del_sparsefactory(fac);

  return sp;
}

/* Free the memory used by a net2d object */
void
del_net2d(pnet2d n2){
  if(n2->p!=0x0) free(n2->p);
  if(n2->perm!=0x0) free(n2->perm);
  if(n2->iperm!=0x0) free(n2->iperm);
  free(n2);
}

/* Build a geometric regular cluster tree for a net2d */
pclustertree
buildcluster_net2d(pnet2d n2, ClusterStrategy strategy, int leafsize){
  pclustertree ct;
  pclusterfactory fac;
  const double (*p)[2] = (const double (*)[2]) n2->p;
  int n;
  int i, j;
  double gmax[2], gmin[2], vol;

  /* degrees of freedom */
  n = n2->vertices;

  fac = new_clusterfactory(n, n, 2);

  /* Initialize x, smin and smax (bounding box for support)
     WARNING: These heuristics apply only to plane geometries! */

  for(j=0; j<2; j++){
    gmin[j] = p[0][j];
    gmax[j] = p[0][j];
  }
  for(i=1; i<n; i++)
    for(j=0; j<2; j++) {
      if(p[i][j] < gmin[j])
	gmin[j] = p[i][j];
      if(p[i][j] > gmax[j])
	gmax[j] = p[i][j];
    }

  vol = (gmax[0] - gmin[0]) * (gmax[1] - gmin[1]);
  vol = sqrt(vol / n);

  j = 0;
  for(i=0; i<n; i++)
    for(j=0; j<2; j++) {
      fac->x[i][j] = p[i][j];
      fac->smin[i][j] = p[i][j]-vol;
      fac->smax[i][j] = p[i][j]+vol;
    }
  
  /* Create the cluster tree */
  ct = create_clustertree(fac, strategy, leafsize);
  
  del_clusterfactory(fac);

  /* Copy the idx2dof and dof2idx arrays */
  (void) memcpy(n2->perm, ct->dof2idx, (size_t) sizeof(int) * ct->ndof);
  (void) memcpy(n2->iperm, ct->idx2dof, (size_t) sizeof(int) * ct->nidx);
  
  return ct;
}
