
/* ------------------------------------------------------------
   This is the file "quadrature2d.h" of the HLib package.

   Copyright 1999 Steffen Boerm and Lars Grasedyck
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#ifndef Quadrature2D
#define Quadrature2D

struct _quadrature2d;
typedef struct _quadrature2d quadrature2d;
typedef quadrature2d *pquadrature2d;
typedef const quadrature2d *pcquadrature2d;

#include "basic.h"

struct _quadrature2d {
  double (*x_identical)[2];	/* Quadrature for identical triangles */
  double (*y_identical)[2];
  double *w_identical;
  int n_identical;

  double (*x_edge)[2];		/* Quadrature for a common edge */
  double (*y_edge)[2];
  double *w_edge;
  int n_edge;

  double (*x_vertex)[2];	/* Quadrature for a common vertex */
  double (*y_vertex)[2];
  double *w_vertex;
  int n_vertex;

  double (*x_distant)[2];	/* Quadrature for distant triangles */
  double (*y_distant)[2];
  double *w_distant;
  int n_distant;
};

/* Compute Duffy/Gauss quadrature rules
   q: Order of the quadrature
   qq: Will be overwritten with the number of quadrature points
   xg: Will be set to an array containing the quadrature points
   wq: Will be set to an array containing the quadrature weights
       It is up to the user to free xg and wg after he is done with them. */
void
build_triangle_gauss(int q, int *qq, double (**xg)[2], double **wg);

/* A quadrature2D object contains all the information required for the
   hierarchical quadrature of singular homogeneous or logarithmic
   kernel functions on triangles T x S.
   Homogeneity means that there is an exponent a of of homogeneity
   such that k(s x, s y) = s^a k(x, y) holds for all x,y and s>0.
   Four cases have to be handled: The triangles can have up to
   three vertices in common, implying that there are three different
   singular cases and one regular case. */

/* Create a quadrature2d object for the symmetric case (k(x,y)=k(x,y))
   order: Order of the quadrature (1=constant, 2=linear, etc.)
   homogeneity: Exponent of homogeneity */
pquadrature2d
new_quadrature2d(int order, double homogeneity);

/* Create a quadrature2D object for the antisymmetric case (k(x,y)=-k(y,x))
   order: Order of the quadrature (1=constant, 2=linear, etc.)
   homogeneity: Exponent of homogeneity */
pquadrature2d
new_quadrature2d_antisymm(int order, double homogeneity);

/* Create a quadrature2d object for Stefan Sauter's quadrature
   order: Order of the quadrature (1=constant, 2=linear, etc.) */
pquadrature2d
new_quadrature2d_sauter(int order);

/* Free the memory used by a quadrature2d object */
void
del_quadrature2d(pquadrature2d q2);

/* Print the quadrature rules described by a quadrature2d object */
void
print_quadrature2d(pcquadrature2d q2);

/* Find which of the four cases mentioned above holds. The functions returns
   the number of vertices the triangles have in common, arrays giving the
   correspondence of these endpoints, and the correct quadrature parameters.
   t: Names of the vertices of the first triangle
   s: Names of the vertices of the second triangle
   tp: Array of size 3, will be overwritten with a permutation of {0, 1, 2}
   sp: Array of size 3, will be overwritten with a permutation of {0, 1, 2}
       If the return value is 1, 2 or 3, we have t[tp[0]] == s[sp[0]].
       If the return value is 2 or 3, we have t[tp[1]] == s[sp[1]].
       If the return value is 3, we have t[tp[2]] == s[sp[2]].
   x: Will be set to point to the appropriate quadrature points for x
   y: Will be set to point to the appropriate quadrature points for y
   w: Will be set to point to the appropriate quadrature weights
   n: Number of quadrature points */
int
select_quadrature2d(pcquadrature2d q2,
		    const int *t, const int *s,
		    int *tp, int *sp,
		    const double (**x)[2], const double (**y)[2],
		    const double **w, int *n);

/* Create Duffy/Gauss quadrature points with integrated mapped linear
   basis functions */
void
build_triangle_gauss_l(int q, int *qq, double (**xg)[2], double **wg);

/* Create Sauter quadrature points with integrated mapped linear basis
   functions */
pquadrature2d
new_quadrature2d_sauter_ll(int order);

#endif
