
/* ------------------------------------------------------------
   This is the file "rkmatrix.c" of the HLib package.

   Copyright 1999 Lars Grasedyck and Steffen Boerm
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#include "basic.h"
#include "fullmatrix.h"
#include "lapack.h"
#include "rkmatrix.h"
#include "supermatrix.h"

#include <assert.h>
#include <math.h>
#include <stdio.h>

/* Are we going to use the realloc() function? */
#define USE_REALLOC

#define RK_SVD_EPS 1.0e-16
#define RK_SVD_EPS2 1.0e-32
#define RK_SVD_EPS3 1.0e-64

int RK_ADAPTIVE;
long int allocated_rkmatrices;
long int current_rkmatrices;
double rk_abseps;

double rk_svd_flops;
double rk_mvm_flops;
double rk_flops;
double rk_nrsvd;

void
init_rkmatrices(int verbosity){
  if(verbosity>0)
    printf("Initializing RkMatrices.\n");
  allocated_rkmatrices=0;
  current_rkmatrices=0;
  RK_ADAPTIVE = 0;
  rk_svd_flops = 0.0;
  rk_mvm_flops = 0.0;
  rk_flops = 0.0;
  rk_nrsvd = 0;
  rk_abseps = RK_SVD_EPS2;
}

void
finish_rkmatrices(int verbosity){
  if(verbosity>0){
    printf("Finishing RkMatrices is ");
    if(current_rkmatrices!=0){
      printf("not successful:\n");
      printf(" Still In Use : %ld\n",current_rkmatrices);
    }else{
      printf("successful:\n");
    }
    printf(" Allocated : %ld\n",allocated_rkmatrices);
  }
}

void
reportflops_rkmatrix(){
  printf("R(k)-Matrix #SVD = %g\n",rk_nrsvd);
  printf("R(k)-Matrix SVD flops = %g\n",rk_svd_flops);
  printf("R(k)-Matrix MVM flops = %g\n",rk_mvm_flops);
  printf("R(k)-Matrix     flops = %g\n",rk_flops);
}

void
resetflops_rkmatrix(){
  rk_svd_flops = 0.0;
  rk_mvm_flops = 0.0;
  rk_flops = 0.0;
  rk_nrsvd = 0;
}

int
rk_arithmetics_adaptive(){
  return RK_ADAPTIVE;
}

void
rk_arithmetics_set_adaptive(int ad){
  RK_ADAPTIVE = ad;
}

void
rk_arithmetics_set_abseps(double abseps){
  rk_abseps = abseps;
}

prkmatrix
new_rkmatrix(int k_new, int rows_new, int cols_new){
  int i;
  prkmatrix r;
  r = (prkmatrix) malloc(sizeof(rkmatrix));
  assert(r!= 0x0);
  r->k = k_new;
  r->kt = 0;
  r->rows = rows_new;
  r->cols = cols_new;
  r->a = 0x0;
  r->b = 0x0;
  r->local_eps = 1e-8; /* sqrt(RK_SVD_EPS);*/
  if(k_new>0){
    r->a = (double*) malloc(k_new*rows_new*sizeof(double));
    assert(r->a != 0x0);
    for(i=0; i<rows_new*k_new; i++) r->a[i] = 0.0;
    r->b = (double*) malloc(k_new*cols_new*sizeof(double));
    assert(r->b != 0x0);
    for(i=0; i<cols_new*k_new; i++) r->b[i] = 0.0;
  }
  allocated_rkmatrices++;
  current_rkmatrices++;
  return r;
}

size_t
getsize_rkmatrix(prkmatrix r){
  size_t size = 0;
  size += sizeof(rkmatrix);
  size += sizeof(double) * r->k * r->rows;
  size += sizeof(double) * r->k * r->cols;
  return size;
}

size_t
size_rkmatrix(int k_new, int rows_new, int cols_new){
  size_t size = 0;
  size += sizeof(rkmatrix);
  size += sizeof(double) * k_new * rows_new;
  size += sizeof(double) * k_new * cols_new;
  return size;
}

prkmatrix
copy_rkmatrix(prkmatrix r){
  int size;
  prkmatrix rc;
  assert(r != 0x0);
  rc = new_rkmatrix(r->k,r->rows,r->cols);
  assert(rc != 0x0);
  rc->kt = r->kt;
  size = r->rows * r->k;
  if(size>0) dcopy_(&size,r->a,eins_,rc->a,eins_);
  size = r->cols * r->k;
  if(size>0) dcopy_(&size,r->b,eins_,rc->b,eins_);
  rc->local_eps = r->local_eps;
  return rc;
}

prkmatrix
extendedcopy_rkmatrix(prkmatrix r, int newrank){
  int size;
  prkmatrix rc;
  int rc_k = newrank;
  assert(r != 0x0);
  if(rc_k<r->k) rc_k = r->k;  
  rc = new_rkmatrix(rc_k,r->rows,r->cols);
  assert(rc != 0x0);
  rc->kt = r->kt;
  size = r->rows * r->kt;
  if(size>0) dcopy_(&size,r->a,eins_,rc->a,eins_);
  size = r->cols * r->kt;
  if(size>0) dcopy_(&size,r->b,eins_,rc->b,eins_);
  rc->local_eps = r->local_eps;
  return rc;
}

prkmatrix
copystructure_rkmatrix(prkmatrix r){
  prkmatrix rc;
  rc = new_rkmatrix(r->k,r->rows,r->cols);
  assert(rc != 0x0);
  rc->local_eps = r->local_eps;
  rc->kt = 0;
  return rc;
}

void
reallocstructure_rkmatrix(prkmatrix r, int new_rank){
  int i;
#ifndef USE_REALLOC
  double *tmp;
#endif

  assert(r != 0x0);

  if(new_rank > r->k){
    if(r->a != 0x0){
#ifdef USE_REALLOC
      r->a = (double*) realloc(r->a,new_rank*r->rows*sizeof(double));
#else
      tmp = (double*) malloc(new_rank*r->rows*sizeof(double));
      i = r->k*r->rows;
      if(i>0) dcopy_(&i,r->a,eins_,tmp,eins_);
      free(r->a);
      r->a = tmp;
#endif
    }else{
      r->a = (double*) malloc(new_rank*r->rows*sizeof(double));
    }
    assert(r->a != 0x0);
    if(r->b != 0x0){
#ifdef USE_REALLOC
      r->b = (double*) realloc(r->b,new_rank*r->cols*sizeof(double));
#else
      tmp = (double*) malloc(new_rank*r->cols*sizeof(double));
      i = r->k*r->cols;
      if(i>0) dcopy_(&i,r->b,eins_,tmp,eins_);
      free(r->b);
      r->b = tmp;
#endif
    }else{
      r->b = (double*) malloc(new_rank*r->cols*sizeof(double));
    }
    assert(r->b != 0x0);
    
    for(i=r->k*r->rows; i<new_rank*r->rows; i++) r->a[i]=0.0;
    for(i=r->k*r->cols; i<new_rank*r->cols; i++) r->b[i]=0.0;
    r->k = new_rank;
  }
}

void
copycontent_rkmatrix(prkmatrix rsource, prkmatrix rtarget){
  int size;
  assert(rsource != 0x0);
  assert(rtarget != 0x0);
  size = rsource->rows * rsource->k;
  if(rtarget->k < rsource->k){
    reallocstructure_rkmatrix(rtarget, rsource->k);
  }
  if(size>0) dcopy_(&size,rsource->a,eins_,rtarget->a,eins_);
  size = rsource->cols * rsource->k;
  if(size>0) dcopy_(&size,rsource->b,eins_,rtarget->b,eins_);
  rtarget->kt = rsource->kt;
  rtarget->local_eps = rsource->local_eps;
}

void
del_rkmatrix(prkmatrix r){
  assert(r != 0x0);
  if(r->k > 0){
    free(r->a);
    free(r->b);
  }
  free(r);
  current_rkmatrices--;
}

void 
eval_rkmatrix(prkmatrix r, double* v, double* w){
  /* w = r*v */
  double tmp;
  int i;
  assert(r != 0x0);
  assert(v != 0x0);
  assert(w != 0x0);

  for(i=0; i<r->rows; i++) w[i] = 0.0;
  for(i=0; i<r->kt; i++){
    tmp = ddot_(&r->cols,v,eins_,&(r->b[i*r->cols]),eins_);
    daxpy_(&r->rows,&tmp,&(r->a[i*r->rows]),eins_,w,eins_);
  }

  rk_mvm_flops += 2.0*r->kt*(r->rows + r->cols);
}

void 
addeval_rkmatrix(prkmatrix r, double* v, double* w){
  /* w = w + r*v */
  double tmp;
  int i;
  assert(r != 0x0);
  assert(v != 0x0);
  assert(w != 0x0);

  for(i=0; i<r->kt; i++){
    tmp = ddot_(&r->cols,v,eins_,&(r->b[i*r->cols]),eins_);
    daxpy_(&r->rows,&tmp,&(r->a[i*r->rows]),eins_,w,eins_);
  }

  rk_mvm_flops += 2.0*r->kt*(r->rows + r->cols);
}

void 
addscaleeval_rkmatrix(prkmatrix r, double alpha, double *v, double *w) {
  /* w = w + alpha*r*v */
  double tmp;
  int i;
  assert(r != 0x0);
  assert(v != 0x0);
  assert(w != 0x0);

  for(i=0; i<r->kt; i++){
    tmp = alpha * ddot_(&r->cols,v,eins_,&(r->b[i*r->cols]),eins_);
    daxpy_(&r->rows,&tmp,&(r->a[i*r->rows]),eins_,w,eins_);
  }

  rk_mvm_flops += 2.0*r->kt*(r->rows + r->cols) + r->kt;
}

void 
evaltrans_rkmatrix(prkmatrix r, double* v, double* w){
  /* w = r'*v */
  double tmp;
  int i;
  assert(r != 0x0);
  assert(v != 0x0);
  assert(w != 0x0);

  for(i=0; i<r->cols; i++) w[i] = 0.0;
  for(i=0; i<r->kt; i++){
    tmp = ddot_(&r->rows,v,eins_,&(r->a[i*r->rows]),eins_);
    daxpy_(&r->cols,&tmp,&(r->b[i*r->cols]),eins_,w,eins_);
  }
  
  rk_mvm_flops += 2.0*r->kt*(r->rows + r->cols);
}

void 
addevaltrans_rkmatrix(prkmatrix r, double* v, double* w){
  /* w = w + r'*v */
  double tmp;
  int i;
  assert(r != 0x0);
  assert(v != 0x0);
  assert(w != 0x0);

  for(i=0; i<r->kt; i++){
    tmp = ddot_(&r->rows,v,eins_,&(r->a[i*r->rows]),eins_);
    daxpy_(&r->cols,&tmp,&(r->b[i*r->cols]),eins_,w,eins_);
  }
  
  rk_mvm_flops += 2.0*r->kt*(r->rows + r->cols);
}

void 
subeval_rkmatrix(prkmatrix r, double* v, double* w){
  /* w = w + r*v */
  double tmp, *x;
  int i;
  assert(r != 0x0);
  assert(v != 0x0);
  assert(w != 0x0);

  if(r->kt>15){
    x = (double*) malloc(r->kt*sizeof(double));
    dgemv_(ttrans,&r->cols,&r->kt,deins_,r->b,&r->cols,v,eins_,dnull_,x,eins_);
    dgemv_(ntrans,&r->rows,&r->kt,meins_,r->a,&r->rows,x,eins_,deins_,w,eins_);
    free(x);
  }else{
    for(i=0; i<r->kt; i++){
      tmp = -ddot_(&r->cols,v,eins_,&(r->b[i*r->cols]),eins_);
      daxpy_(&r->rows,&tmp,&(r->a[i*r->rows]),eins_,w,eins_);
    }
  }
  rk_mvm_flops += 2.0*r->kt*(r->rows + r->cols);
}

void 
subevaltrans_rkmatrix(prkmatrix r, double* v, double* w){
  /* w = w + r'*v */
  double tmp, *x;
  int i;
  assert(r != 0x0);
  assert(v != 0x0);
  assert(w != 0x0);

  if(r->kt>15){
    x = (double*) malloc(r->kt*sizeof(double));
    dgemv_(ttrans,&r->rows,&r->kt,deins_,r->a,&r->rows,v,eins_,dnull_,x,eins_);
    dgemv_(ntrans,&r->cols,&r->kt,meins_,r->b,&r->cols,x,eins_,deins_,w,eins_);
    free(x);
  }else{
    for(i=0; i<r->kt; i++){
      tmp = -ddot_(&r->rows,v,eins_,&(r->a[i*r->rows]),eins_);
      daxpy_(&r->cols,&tmp,&(r->b[i*r->cols]),eins_,w,eins_);
    }
  }
  rk_mvm_flops += 2.0*r->kt*(r->rows + r->cols); 
}

void
generaleval_rkmatrix(prkmatrix r, int uplo, int trans, int add,
             double alpha, double *v, double *w){
  /* w = (w) +alpha * r(tran,uplo) * v */
  double tmp;
  int i;
  assert(r != 0x0);
  assert(v != 0x0);
  assert(w != 0x0);
  assert(uplo==0);

  if(add==0){
    if(trans==0){
      scale_lapack(r->rows,w,0.0);
    }else{
      scale_lapack(r->cols,w,0.0);
    }
  }
  for(i=0; i<r->kt; i++){
    if(trans==0){
      tmp = alpha*ddot_(&r->cols,v,eins_,&(r->b[i*r->cols]),eins_);
      daxpy_(&r->rows,&tmp,&(r->a[i*r->rows]),eins_,w,eins_);
    }else{
      tmp = alpha*ddot_(&r->rows,v,eins_,&(r->a[i*r->rows]),eins_);
      daxpy_(&r->cols,&tmp,&(r->b[i*r->cols]),eins_,w,eins_);
    }
  }
  rk_mvm_flops += 2.0*r->kt*(r->rows + r->cols);  
}

double
getentry_rkmatrix(prkmatrix r, int row, int col){
  int i;
  double entry=0.0;

  for(i=0; i<r->kt; i++){
    entry += r->a[row + i*r->rows] * r->b[col + i*r->cols];
  }
  return entry;
}

void
setentry_rkmatrix(prkmatrix r, int row, int col, double value){
  int i;
  prkmatrix rk_r[1];
  int rk_row_offset[1];
  int rk_col_offset[1];
  assert(row>=0 && col>=0);
  assert(r->rows>row && r->cols>col);
  if(r->kt < r->k){
    for(i=0; i<r->rows; i++) r->a[r->kt*r->rows +i] = 0.0;
    r->a[r->kt*r->rows +row] = 1.0;
    for(i=0; i<r->cols; i++) r->b[r->kt*r->cols +i] = 0.0;
    r->b[r->kt*r->cols +col] = value;
    r->kt++;
  }else{
    rk_row_offset[0] = 0;
    rk_col_offset[0] = 0;
    rk_r[0] = new_rkmatrix(1,r->rows,r->cols);
    rk_r[0]->a[row] = 1.0;
    rk_r[0]->b[col] = value;
    rk_r[0]->kt = 1;
    addparts2_rkmatrix(r, 1, rk_row_offset, rk_col_offset, rk_r);
    del_rkmatrix(rk_r[0]);
  }
}

void
scale_rkmatrix(prkmatrix r, double v){
  int size;
  assert(r != 0x0);
  assert(v != 0x0);

  size = r->rows * r->kt;
  dscal_(&size,&v,r->a,eins_);
  rk_flops += size;
}

prkmatrix
getmul_rkmatrix(psupermatrix a, psupermatrix b){
  prkmatrix ra, rb, rab, rmin=0x0;
  int i,j,rows,cols,size,k=0;
  assert(a != 0x0);
  assert(b != 0x0);

  ra = a->r;
  rb = b->r;
  assert(ra!=0x0 || rb!=0x0);

  rows = a->rows;
  cols = b->cols;
  size = a->cols;  

  if(ra!=0x0 && rb==0x0) rmin = ra;
  if(ra==0x0 && rb!=0x0) rmin = rb;
  if(ra!=0x0 && rb!=0x0){
    if(ra->kt < rb->kt){
      rmin = ra;
    }else{
      rmin = rb;
    }
  }
  k = rmin->kt; 
  rab = new_rkmatrix(k,rows,cols);
  if(rk_arithmetics_adaptive()==1){
    rab->local_eps = rmin->local_eps;
  }
  rab->kt = k;
  if(k==0) return rab;

  if(rmin==ra){
    j=0; 
    for(i=0; i<k; i++){
      evaltrans_supermatrix(b,&ra->b[i*size],&rab->b[j*cols]);
      if((dnrm2_(&rows,&ra->a[i*rows],eins_) + 
      dnrm2_(&cols,&rab->b[j*cols],eins_)) > RK_SVD_EPS2){
    dcopy_(&rows,&ra->a[i*rows],eins_,&rab->a[j*rows],eins_);
    j++;
      }
    }
  }else{
    j=0; 
    for(i=0; i<k; i++){
      eval_supermatrix(a,&rb->a[i*size],&rab->a[j*rows]);
      if((dnrm2_(&rows,&rab->a[j*rows],eins_) + 
      dnrm2_(&cols,&rb->b[i*cols],eins_)) > RK_SVD_EPS2){
    dcopy_(&cols,&rb->b[i*cols],eins_,&rab->b[j*cols],eins_);
    j++;
      }
    }
  }
  return rab;
}

void
mul_rkmatrix(prkmatrix r, psupermatrix a, psupermatrix b){
  prkmatrix rab;
  assert(r != 0x0);
  assert(a != 0x0);
  assert(b != 0x0);
  assert(a->r!=0x0 || b->r!=0x0);

  rab = getmul_rkmatrix(a,b);
  convertrk2_rkmatrix(rab,r,0,0);
  del_rkmatrix(rab);
}

void 
add_rkmatrix(prkmatrix target, const prkmatrix r1, const prkmatrix r2){
  int* rk_row_offset;
  int* rk_col_offset;
  prkmatrix* rk_r;
  assert(target != 0x0);
  assert(r1 != 0x0);
  assert(r2 != 0x0);
  
  target->kt=0;
  if(r2->kt+r1->kt == 0) return;
  
  rk_row_offset = (int*) malloc(4*sizeof(int));
  assert(rk_row_offset != 0x0);
  rk_col_offset = &rk_row_offset[2];
  rk_r = (prkmatrix*) malloc(2*sizeof(prkmatrix));
  assert(rk_r != 0x0);
  rk_r[0] = r1;
  rk_r[1] = r2;
  rk_row_offset[0] = 0;
  rk_col_offset[0] = 0;
  rk_row_offset[1] = 0;
  rk_col_offset[1] = 0;
  addparts2_rkmatrix(target,2,rk_row_offset,rk_col_offset,rk_r);
  free(rk_row_offset);
  free(rk_r);
}

void  
addpart2_rkmatrix(prkmatrix r1, int row_offset1, int col_offset1, 
          const prkmatrix r2, int row_offset2, int col_offset2, 
          int rows, int cols){
  int i;
  int* rk_row_offset;
  int* rk_col_offset;
  prkmatrix* rk_r;
  assert(r1 != 0x0);
  assert(r2 != 0x0);
  
  if(r2->kt == 0) return;
  
  rk_row_offset = (int*) malloc(2*sizeof(int));
  assert(rk_row_offset != 0x0);
  rk_col_offset = &rk_row_offset[1];
  rk_r = (prkmatrix*) malloc(sizeof(prkmatrix));
  assert(rk_r != 0x0);
  rk_r[0] = new_rkmatrix(r2->kt,rows,cols);
  rk_r[0]->kt = r2->kt;
  rk_r[0]->local_eps = r2->local_eps;

  for(i=0; i<r2->kt; i++){
    dcopy_(&rows,&r2->a[row_offset2+i*r2->rows],eins_,&rk_r[0]->a[i*rows],eins_);
    dcopy_(&cols,&r2->b[col_offset2+i*r2->cols],eins_,&rk_r[0]->b[i*cols],eins_);
  }
  rk_row_offset[0] = row_offset1;
  rk_col_offset[0] = col_offset1;
  addparts2_rkmatrix(r1,1,rk_row_offset,rk_col_offset,rk_r);
  del_rkmatrix(rk_r[0]);
  free(rk_row_offset);
  free(rk_r);
}

void 
addparts2_rkmatrix(prkmatrix rt, int nr, int* rk_row_offset, int* rk_col_offset, 
           prkmatrix* rk_r){
  double *atmp;
  double *btmp;
  int rows = rt->rows;
  int cols = rt->cols;
  int kmax;
  int i,j,k;
  double diff;
  double* qr_work;
  double* tau1;
  double* tau2;
  double* sigma;
  double* rarb;
  double* u;
  double* v;
  pfullmatrix f;
  int lwork;
  int info;
  int minsize,maxsize;
  
  assert(rt != 0x0);
  assert(rk_row_offset != 0x0);
  assert(rk_col_offset != 0x0);
  assert(rk_r != 0x0);

  /*
  if(rows>60 || cols>60 || k<8){
    addpartsgram2_rkmatrix(rt, nr, rk_row_offset, rk_col_offset, rk_r);
    return;
  }
  */

  minsize = rows; if(cols < rows) minsize = cols;
  maxsize = rows; if(cols > rows) maxsize = cols;
  kmax=rt->kt;
  for(i=0; i<nr; i++){
    assert(rk_r[i] != 0x0);
    kmax += rk_r[i]->kt;
  }

  if(kmax==0){
    rt->kt=0;
    return;
  }  
  
  if(kmax > rows || kmax > cols){
    f = new_fullmatrix(rows,cols);
    for(i=0; i<nr; i++){
      if(rk_r[i]->kt>0)
    dgemm_(ntrans,ttrans,&rk_r[i]->rows,&rk_r[i]->cols,&rk_r[i]->kt,
           deins_,rk_r[i]->a,&rk_r[i]->rows,
           rk_r[i]->b,&rk_r[i]->cols,
           deins_,&f->e[rk_row_offset[i]+rk_col_offset[i]*f->rows],&f->rows);
      rk_svd_flops += 2.0 * rk_r[i]->rows * rk_r[i]->cols * rk_r[i]->kt;
    }
    addfull2_rkmatrix(f,rt,0,0);
    del_fullmatrix(f);
    return;
  }

  if(kmax <= rt->k){
    for(i=(rt->kt)*rows; i<rows*kmax; i++) rt->a[i] = 0.0;
    for(i=(rt->kt)*cols; i<cols*kmax; i++) rt->b[i] = 0.0;
    k = rt->kt;
    for(j=0; j<nr; j++){
      for(i=0; i<rk_r[j]->kt; i++){
    dcopy_(&rk_r[j]->rows,
           &rk_r[j]->a[i*rk_r[j]->rows],eins_,
           &rt->a[rk_row_offset[j]+(k+i)*rows],eins_);
    dcopy_(&rk_r[j]->cols,
           &rk_r[j]->b[i*rk_r[j]->cols],eins_,
           &rt->b[rk_col_offset[j]+(k+i)*cols],eins_);
      }
      k += rk_r[j]->kt;
    }
    rt->kt = kmax;
    return;
  }

  atmp = (double*) malloc((rows+cols)*kmax*sizeof(double));
  assert(atmp != 0x0);
  btmp = &atmp[rows*kmax];

  for(i=0; i<rows*kmax; i++) atmp[i] = 0.0;
  for(i=0; i<cols*kmax; i++) btmp[i] = 0.0;
  for(i=0; i<rt->kt; i++){
    dcopy_(&rows,&rt->a[i*rows],eins_,&atmp[i*rows],eins_);
    dcopy_(&cols,&rt->b[i*cols],eins_,&btmp[i*cols],eins_);
  }
  k = rt->kt;
  for(j=0; j<nr; j++){
    for(i=0; i<rk_r[j]->kt; i++){
      dcopy_(&rk_r[j]->rows,
         &rk_r[j]->a[i*rk_r[j]->rows],eins_,
         &atmp[rk_row_offset[j]+(k+i)*rows],eins_);
      dcopy_(&rk_r[j]->cols,
         &rk_r[j]->b[i*rk_r[j]->cols],eins_,
         &btmp[rk_col_offset[j]+(k+i)*cols],eins_);
    }
    k += rk_r[j]->kt;
  }

  /* eliminate zero vectorproducts */
  j=0;
  for(i=0; i<k; i++){
    diff = dnrm2_(&rows,&atmp[i*rows],eins_)*dnrm2_(&cols,&btmp[i*cols],eins_);
    if(diff>RK_SVD_EPS2 && diff>rk_abseps){
      if(j<i){
    dcopy_(&rows,&atmp[i*rows],eins_,&atmp[j*rows],eins_);
    dcopy_(&cols,&btmp[i*cols],eins_,&btmp[j*cols],eins_);
      }
      j++;
    }
  }
  k = j;
  kmax = j;
  
  if(kmax <= rt->k){
    for(i=0; i<kmax; i++){
      dcopy_(&rows,&atmp[i*rows],eins_,&rt->a[i*rows],eins_);
      dcopy_(&cols,&btmp[i*cols],eins_,&rt->b[i*cols],eins_);
    }
    rt->kt = kmax;
    free(atmp);
    return;
  }
  

  lwork = 8*minsize + maxsize;
  qr_work = (double*) malloc((lwork+rows+cols+minsize+3*kmax*kmax)*sizeof(double)); 
  assert(qr_work!=0x0);
  tau1 = &qr_work[lwork];
  tau2 = &qr_work[lwork+rows];
  sigma = &qr_work[lwork+rows+cols];
  rarb = &qr_work[lwork+rows+cols+minsize];
  u = &qr_work[lwork+rows+cols+minsize+kmax*kmax];
  v = &qr_work[lwork+rows+cols+minsize+2*kmax*kmax];
  
  rk_nrsvd++;

  dgeqrf_(&rows,&kmax,atmp,&rows,tau1,qr_work,&lwork,&info);
  rk_svd_flops += 2 * kmax * kmax * rows;
  if(info!=0){
    printf("addparts2rkmatrix: info(dgeqrf,a)=%d\n",info);
    exit(0);
  }
  dgeqrf_(&cols,&kmax,btmp,&cols,tau2,qr_work,&lwork,&info);
  rk_svd_flops += 2 * kmax * kmax * cols;
  if(info!=0){
    printf("addparts2rkmatrix: info(dgeqrf,b)=%d\n",info);
    exit(0);
  }
  
  for(i=0; i<kmax*kmax; i++) rarb[i] = 0.0;
  for(i=0; i<kmax; i++){
    for(j=0; j<kmax; j++){
      for(k=0; k<kmax; k++){
    if(k>=i && k>=j) rarb[i+kmax*j] += atmp[i+k*rows]*btmp[j + k*cols];
      }
    }
  }
  dorgqr_(&rows,&kmax,&kmax,atmp,&rows,tau1,qr_work,&lwork,&info);
  if(info!=0){
    printf("addparts2rkmatrix: info(dorgqr,a)=%d\n",info);
    exit(0);
  }
  dorgqr_(&cols,&kmax,&kmax,btmp,&cols,tau2,qr_work,&lwork,&info);
  if(info!=0){
    printf("addparts2rkmatrix: info(dorgqr,b)=%d\n",info);
    exit(0);
  }
  
  diff = 0.0;
  for(i=0; i<kmax*kmax; i++){
    if(fabs(rarb[i]) < RK_SVD_EPS2) rarb[i] = 0.0;  
    diff += fabs(rarb[i]);
  }
  
  if(diff<RK_SVD_EPS2 || diff<rk_abseps){
    rt->kt = 0;
  }else{
    dgesvd_("A","A",&kmax,&kmax,rarb,&kmax,sigma,u,&kmax,v,&kmax,
        qr_work,&lwork,&info);
    rk_svd_flops += 60 * kmax * kmax * kmax;
    
    if(info!=0){
      printf("addparts2rkmatrix: info(dgesvd)=%d\n",info);
      exit(0);
    }
    for(i=0; i<kmax; i++){
      dscal_(&kmax,&sigma[i],&u[i*kmax],eins_);
      dcopy_(&kmax,&v[i],&kmax,&rarb[i*kmax],eins_);
    }
    if(rk_arithmetics_adaptive()==1 && rt->local_eps>=0){
      for(k=0; k<kmax && sigma[k] > (rt->local_eps*sigma[0]) && 
        k<rows && k<cols && sigma[k]>rk_abseps; k++);
      if(k>rt->k) reallocstructure_rkmatrix(rt,k);
    }else{
      for(k=0; k<kmax && k<rt->k && k<rows && k<cols && 
        sigma[k]>RK_SVD_EPS*sigma[0] && sigma[k]>rk_abseps; k++);
    }

    rt->kt = k;

    if(k>0){
      dgemm_(ntrans,ntrans,&rows,&k,&kmax,deins_,atmp,&rows,u,&kmax,dnull_,
         rt->a,&rows);
      dgemm_(ntrans,ntrans,&cols,&k,&kmax,deins_,btmp,&cols,rarb,&kmax,dnull_,
         rt->b,&cols);
      rk_svd_flops += 2 * (rows+cols) * k * k;
    }
  }    

  free(qr_work);
  free(atmp);
}

void 
addpartsgram2_rkmatrix(prkmatrix rt, int nr, 
               int* rk_row_offset, int* rk_col_offset, 
               prkmatrix* rk_r){
  double *ata, *qa;
  double *btb, *qb;
  double *s, *sigma;
  int i,j,ii,jj,k,nu1,nu2,ktotal,k1,k2;
  int rows = rt->rows,cols = rt->cols;
  double *wa, *wb, *v, *work;
  int lwork, info=0;
  int ka, kb;
  pfullmatrix f;
  
  k = rt->kt;
  ktotal = k;
  for(i=0; i<nr; i++) ktotal += rk_r[i]->kt;
  
  if(ktotal==0){
    rt->kt = 0;
    return;
  }
  
  if(ktotal > rows || ktotal > cols){
    f = new_fullmatrix(rows,cols);
    for(i=0; i<nr; i++){
      if(rk_r[i]->kt>0)
    dgemm_(ntrans,ttrans,&rk_r[i]->rows,&rk_r[i]->cols,&rk_r[i]->kt,
           deins_,rk_r[i]->a,&rk_r[i]->rows,
           rk_r[i]->b,&rk_r[i]->cols,
           deins_,&f->e[rk_row_offset[i]+rk_col_offset[i]*f->rows],&f->rows);
    }
    addfull2_rkmatrix(f,rt,0,0);
    del_fullmatrix(f);
    return;
  }

  if(ktotal <= rt->k){
    for(i=(rt->kt)*rows; i<rows*ktotal; i++) rt->a[i] = 0.0;
    for(i=(rt->kt)*cols; i<cols*ktotal; i++) rt->b[i] = 0.0;
    k = rt->kt;
    for(j=0; j<nr; j++){
      for(i=0; i<rk_r[j]->kt; i++){
    dcopy_(&rk_r[j]->rows,
           &rk_r[j]->a[i*rk_r[j]->rows],eins_,
           &rt->a[rk_row_offset[j]+(k+i)*rows],eins_);
    dcopy_(&rk_r[j]->cols,
           &rk_r[j]->b[i*rk_r[j]->cols],eins_,
           &rt->b[rk_col_offset[j]+(k+i)*cols],eins_);
      }
      k += rk_r[j]->kt;
    }
    rt->kt = ktotal;
    return;
  }
  
  lwork = (4+ktotal+rows+cols)*ktotal;
  work = (double*) malloc((6*ktotal*ktotal + 3*ktotal + lwork)*sizeof(double));
  ata = &work[lwork];
  btb = &work[lwork +   ktotal*ktotal];
  s   = &work[lwork + 2*ktotal*ktotal];
  v   = &work[lwork + 3*ktotal*ktotal];
  wb  = &work[lwork + 4*ktotal*ktotal];
  wa  = &work[lwork + 4*ktotal*ktotal + ktotal];
  sigma  = &work[lwork + 4*ktotal*ktotal + 2*ktotal];
  qa = &work[lwork + 4*ktotal*ktotal + 3*ktotal];
  qb = &work[lwork + 5*ktotal*ktotal + 3*ktotal];
  
  

  for(nu1=0; nu1<rt->kt; nu1++){
    for(nu2=0; nu2<rt->kt; nu2++){
      if(nu1<=nu2){
    ata[nu1+nu2*ktotal] = ddot_(&rows,&rt->a[nu1*rows],eins_,
                    &rt->a[nu2*rows],eins_);
      }
    }
    k2 = rt->kt;
    for(j=0; j<nr; j++){
      for(nu2=0; nu2<rk_r[j]->kt; nu2++){
    ii = nu1;
    jj = k2+nu2;
    if(ii<=jj){
      ata[ii + jj*ktotal] = ddot_(&rk_r[j]->rows,
                      &rt->a[nu1*rows + rk_row_offset[j]],eins_,
                      &rk_r[j]->a[nu2*rk_r[j]->rows],eins_);
    }
      }
      k2 += rk_r[j]->kt;
    }
  }

  k1 = rt->kt;
  for(i=0; i<nr; i++){
    k2 = rt->kt;
    for(j=0; j<nr; j++){
      for(nu1=0; nu1<rk_r[i]->kt; nu1++){
    for(nu2=0; nu2<rk_r[j]->kt; nu2++){
      ii = k1+nu1;
      jj = k2+nu2;
      if(ii<=jj){
        if(rk_r[i]->rows+rk_row_offset[i]>rk_row_offset[j] &&
           rk_r[j]->rows+rk_row_offset[j]>rk_row_offset[i]){
          ata[ii + jj*ktotal] = 
        ddot_(&rk_r[i]->rows,&rk_r[i]->a[nu1*rk_r[i]->rows],
              eins_,
              &rk_r[j]->a[nu2*rk_r[j]->rows],eins_);
        }else{
          ata[ii + jj*ktotal] = 0.0;
        }
      }
    }
      }
      k2 += rk_r[j]->kt;
    }
    k1 += rk_r[i]->kt;
  }

  for(nu1=0; nu1<rt->kt; nu1++){
    for(nu2=0; nu2<rt->kt; nu2++){
      if(nu1<=nu2){
    btb[nu1 + nu2*ktotal] = ddot_(&cols,&rt->b[nu1*cols],eins_,
                    &rt->b[nu2*cols],eins_);
      }
    }
    k2 = rt->kt;
    for(j=0; j<nr; j++){
      for(nu2=0; nu2<rk_r[j]->kt; nu2++){
    ii = nu1;
    jj = k2+nu2;
    if(ii<=jj){
      btb[ii + jj*ktotal] = ddot_(&rk_r[j]->cols,
                      &rt->b[nu1*cols + rk_col_offset[j]],eins_,
                      &rk_r[j]->b[nu2*rk_r[j]->cols],eins_);
    }
      }
      k2 += rk_r[j]->kt;
    }
  }

  k1 = rt->kt;
  for(i=0; i<nr; i++){
    k2 = rt->kt;
    for(j=0; j<nr; j++){
      for(nu1=0; nu1<rk_r[i]->kt; nu1++){
    for(nu2=0; nu2<rk_r[j]->kt; nu2++){
      ii = k1+nu1;
      jj = k2+nu2;
      if(ii<=jj){
        if(rk_r[i]->cols+rk_col_offset[i]>rk_col_offset[j] &&
           rk_r[j]->cols+rk_col_offset[j]>rk_col_offset[i]){
          btb[ii + jj*ktotal] = 
        ddot_(&rk_r[i]->cols,&rk_r[i]->b[nu1*rk_r[i]->cols],
              eins_,
              &rk_r[j]->b[nu2*rk_r[j]->cols],eins_);
        }else{
          btb[ii + jj*ktotal] = 0.0;
        }
      }
    }
      }
      k2 += rk_r[j]->kt;
    }
    k1 += rk_r[i]->kt;
  }
  /*
  for(i=0; i<ktotal; i++){
    if(ata[i+i*ktotal]<1e-10){
      for(j=i; j<ktotal; j++) ata[i+j*ktotal] = 0.0;
      for(j=0; j<i; j++) ata[j+i*ktotal] = 0.0;
    }
    if(btb[i+i*ktotal]<1e-10){
      for(j=i; j<ktotal; j++) btb[i+j*ktotal] = 0.0;
      for(j=0; j<i; j++) btb[j+i*ktotal] = 0.0;
    }
  }
  */

  dsyev_("V","U",&ktotal,btb,&ktotal,wb,work,&lwork,&info);
  if(info!=0){
    printf(" warning: info(dgeev b[%d x %d, %d])=%d\n",rows,cols,ktotal,info);
  }
  kb = ktotal;
  
  for(i=0; i<ktotal; i++)
    if(wb[i]<1e-15) kb--;
  
  if(kb==0){
    rt->kt = 0;
    free(work);
    return;
  }

  dsyev_("V","U",&ktotal,ata,&ktotal,wa,work,&lwork,&info);
  if(info!=0){
    printf(" warning: info(dgeev a[%d x %d, %d])=%d\n",rows,cols,ktotal,info);
  }
  ka = ktotal;
  
  for(i=0; i<ktotal; i++)
    if(wa[i]<1e-15) ka--;
  
  if(ka==0){
    rt->kt = 0;
    free(work);
    return;
  }

  copy_lapack(ka*ktotal,&ata[(ktotal-ka)*ktotal],qa);
  copy_lapack(kb*ktotal,&btb[(ktotal-kb)*ktotal],qb);

  for(i=ktotal-ka; i<ktotal; i++){
    wa[i] = sqrt(wa[i]);
    dscal_(&ktotal,&wa[i],&ata[i*ktotal],eins_);
  }
  for(i=ktotal-kb; i<ktotal; i++){
    wb[i] = sqrt(wb[i]);
    dscal_(&ktotal,&wb[i],&btb[i*ktotal],eins_);
  }
  for(i=ktotal-ka; i<ktotal; i++){
    wa[i] = 1.0/wa[i];
    dscal_(&ktotal,&wa[i],&qa[(i-(ktotal-ka))*ktotal],eins_);
  }
  for(i=ktotal-kb; i<ktotal; i++){
    wb[i] = 1.0/wb[i];
    dscal_(&ktotal,&wb[i],&qb[(i-(ktotal-kb))*ktotal],eins_);
  }

  dgemm_("T","N",&ka,&kb,&ktotal,deins_,
     &ata[(ktotal-ka)*ktotal],&ktotal,&btb[(ktotal-kb)*ktotal],&ktotal,
     dnull_,s,&ka);
  
  dgesvd_("A","A",&ka,&kb,s,&ka,sigma,ata,&ka,btb,&kb,
      work,&lwork,&info);
  if(info!=0){
    printf(" warning: info(dgesvd [%d x %d, %d])=%d\n",rows,cols,ktotal,info);
  }
  
  if(rk_arithmetics_adaptive()==1 && rt->local_eps>=0){
    for(k=0; k<ka && k<kb && sigma[k] > (rt->local_eps*sigma[0]) 
      && sigma[k]>rk_abseps; k++);
    if(k>rt->k) reallocstructure_rkmatrix(rt,k);
  }else{
    for(k=0; k<ka && k<kb && k<rt->k && 
      sigma[k]>RK_SVD_EPS*sigma[0] && sigma[k]>rk_abseps; k++);
  }
  
  if(k==0){
    rt->kt = 0;
    free(work);
    return;
  }
  
  for(i=0; i<k; i++){
    sigma[i] = sqrt(sigma[i]);
    dscal_(&ka,&sigma[i],&ata[i*ka],eins_);
    dscal_(&kb,&sigma[i],&btb[i],&kb);
  }
  
  mul_lapack(ktotal,k,ka,qa,ata,s);
  /* rt */
  if(rt->kt>0){
    copy_lapack(rows*rt->kt,rt->a,work);
    dgemm_("N","N",&rows,&k,&rt->kt,deins_,
       work,&rows,s,&ktotal,dnull_,rt->a,&rows);
  }else{
    for(i=0; i<rows*k; i++) rt->a[i] = 0.0;
  }
  /* rk_r */
  nu1 = rt->kt;
  for(i=0; i<nr; i++){
    if(rk_r[i]->kt>0){
      dgemm_("N","N",&rk_r[i]->rows,&k,&rk_r[i]->kt,deins_,
         rk_r[i]->a,&rk_r[i]->rows,
         &s[nu1],&ktotal,deins_,&rt->a[rk_row_offset[i]],&rows);
      nu1 += rk_r[i]->kt;
    }
  }
  
  dgemm_("N","T",&ktotal,&k,&kb,deins_,qb,&ktotal,btb,&kb,dnull_,s,&ktotal);
  /* rt */
  if(rt->kt>0){
    copy_lapack(cols*rt->kt,rt->b,work);
    dgemm_("N","N",&cols,&k,&rt->kt,deins_,
       work,&cols,s,&ktotal,dnull_,rt->b,&cols);
  }else{
    for(i=0; i<cols*k; i++) rt->b[i] = 0.0;
  }

  /* rk_r */
  nu1 = rt->kt;
  for(i=0; i<nr; i++){
    if(rk_r[i]->kt>0){
      dgemm_("N","N",&rk_r[i]->cols,&k,&rk_r[i]->kt,deins_,
         rk_r[i]->b,&rk_r[i]->cols,
         &s[nu1],&ktotal,deins_,&rt->b[rk_col_offset[i]],&cols);
      nu1 += rk_r[i]->kt;
    }
  }

  rt->kt = k;
  
  free(work);
}

void
addprod2_rkmatrix(prkmatrix r, psupermatrix a, psupermatrix b, 
          int row_offset, int col_offset){
  int i,j,k,block_rows,block_cols,block_size,rows,cols,size;
  int* rk_row_offset;
  int* rk_col_offset;
  int* rk_row_offset2;
  int* rk_col_offset2;
  prkmatrix* rk_row;
  prkmatrix* rk_col;
  prkmatrix rtmp;
  psupermatrix stmp;
  prkmatrix r2=0x0;
  pfullmatrix f;
  
  assert(r != 0x0);
  assert(a != 0x0);
  assert(b != 0x0);

  block_rows = a->block_rows;
  block_size = a->block_cols;
  block_cols = b->block_cols;
  if(a->s!=0x0 && b->s!=0x0){
    /* subdivision of a and b, convert subblocks to rk, then add all subblocks */
    assert(b->block_rows == block_size);
    rk_row_offset = (int*) 
      malloc((block_rows+block_cols+block_cols+block_rows) * sizeof(int));
    assert(rk_row_offset != 0x0);
    rk_row_offset2 = &rk_row_offset[block_rows];
    rk_col_offset  = &rk_row_offset[block_rows+block_cols];
    rk_col_offset2 = &rk_row_offset[block_rows+block_cols+block_cols];
    rk_row = (prkmatrix*) malloc((block_rows+block_cols)*sizeof(prkmatrix));
    assert(rk_row != 0x0);
    rk_col = &rk_row[block_rows];
    
    rk_col_offset[0] = col_offset;
    for(j=0; j<block_cols; j++){
      if(j>0) rk_col_offset[j] = rk_col_offset[j-1] + b->s[(j-1)*block_size]->cols;
      rk_row_offset2[j] = 0;
    }
    for(i=0; i<block_rows; i++){
      rk_row[i] = new_rkmatrix(r->k,a->s[i]->rows,b->cols);
      if(rk_arithmetics_adaptive()==1) rk_row[i]->local_eps = r->local_eps;
      for(j=0; j<block_cols; j++){
    rk_col[j] = new_rkmatrix(r->k,a->s[i]->rows,b->s[j*block_size]->cols);
    if(rk_arithmetics_adaptive()==1){
      rk_col[j]->local_eps = r->local_eps;
    }
    for(k=0; k<block_size; k++){
      addprod2_rkmatrix(rk_col[j],a->s[i+k*block_rows],b->s[k+j*block_size],0,0);
    }
      }
      addparts2_rkmatrix(rk_row[i],block_cols,rk_row_offset2,rk_col_offset,rk_col);
      for(j=0; j<block_cols; j++) del_rkmatrix(rk_col[j]);
    }
    rk_row_offset[0] = row_offset;
    for(i=0; i<block_rows; i++){
      if(i>0) rk_row_offset[i] = rk_row_offset[i-1] + a->s[i-1]->rows;
      rk_col_offset2[i] = col_offset;
    }
    addparts2_rkmatrix(r,block_rows,rk_row_offset,rk_col_offset2,rk_row);
    for(i=0; i<block_rows; i++) del_rkmatrix(rk_row[i]);
    free(rk_row_offset);
    free(rk_row);
  }else{
    if(a->r!=0x0 || b->r!=0x0){
      /* one of a and b is rkmatrix -> trivial */
      r2 = getmul_rkmatrix(a,b);
      addpart2_rkmatrix(r,row_offset,col_offset,r2,0,0,r2->rows,r2->cols);
      del_rkmatrix(r2);
    }else{
      if(a->f!=0x0 || b->f!=0x0){
    rows = a->rows;
    cols = b->cols;
    size = a->cols; 
    /* both are fullmatrices but a->cols can be very small
       so that we consider a*b as an rkmatrix factorisation */
    if(size<rows && size<cols){
      if(rows<cols || (b->f==0x0 && a->f!=0x0)){
        rtmp = new_rkmatrix(size,rows,size);
        stmp = new_supermatrix(1,1,rows,size,0x0,rtmp,0x0);
        addfull2_rkmatrix(a->f,rtmp,0,0);
        r2 = getmul_rkmatrix(stmp,b);
        del_supermatrix(stmp);
        addpart2_rkmatrix(r,row_offset,col_offset,r2,0,0,r2->rows,r2->cols);
        del_rkmatrix(r2);
      }else{
        rtmp = new_rkmatrix(size,size,cols);
        stmp = new_supermatrix(1,1,size,cols,0x0,rtmp,0x0);
        addfull2_rkmatrix(b->f,rtmp,0,0);
        r2 = getmul_rkmatrix(a,stmp);
        del_supermatrix(stmp);
        addpart2_rkmatrix(r,row_offset,col_offset,r2,0,0,r2->rows,r2->cols);
        del_rkmatrix(r2);
      }
    }else{
      f = getmul_fullmatrix(a,b);
      r2 = new_rkmatrix(r->k,rows,cols);
      if(rk_arithmetics_adaptive()==1){
        r2->local_eps  = r->local_eps;
      }
      addfull2_rkmatrix(f,r2,0,0);
      addpart2_rkmatrix(r,row_offset,col_offset,r2,0,0,r2->rows,r2->cols);
      del_rkmatrix(r2);
      del_fullmatrix(f);
    }
      }else{
    assert(a->u!=0x0 || b->u!=0x0);
    /* one of the factors is a uniform matrix */
    r2 = getmul_uniformmatrix(a,b);
    addpart2_rkmatrix(r,row_offset,col_offset,r2,0,0,r2->rows,r2->cols);
    del_rkmatrix(r2);
      }
    } 
  }
}

prkmatrix
getgeneralmul_rkmatrix(psupermatrix a, psupermatrix b, 
               int transpose_a, int transpose_b){
  prkmatrix ra, rb, rab, rmin=0x0;
  int i,j,rows,cols,size,k=0;
  assert(a != 0x0);
  assert(b != 0x0);

  ra = a->r;
  rb = b->r;
  assert(ra!=0x0 || rb!=0x0);

  if(transpose_a==0){
    rows = a->rows;
    size = a->cols;  
  }else{
    rows = a->cols;
    size = a->rows;  
  }

  if(transpose_b==0){
    cols = b->cols;
  }else{
    cols = b->rows;
  }

  if(ra!=0x0 && rb==0x0) rmin = ra;
  if(ra==0x0 && rb!=0x0) rmin = rb;
  if(ra!=0x0 && rb!=0x0){
    if(ra->kt < rb->kt){
      rmin = ra;
    }else{
      rmin = rb;
    }
  }
  k = rmin->kt; 
  rab = new_rkmatrix(k,rows,cols);
  if(rk_arithmetics_adaptive()==1){
    rab->local_eps = rmin->local_eps;
  }
  rab->kt = k;
  if(k==0) return rab;

  if(rmin==ra){
    j=0; 
    for(i=0; i<k; i++){
      if(transpose_a==0){
    if(transpose_b==0){
      evaltrans_supermatrix(b,&ra->b[i*size],&rab->b[j*cols]);
    }else{
      eval_supermatrix(b,&ra->b[i*size],&rab->b[j*cols]);
    }
    dcopy_(&rows,&ra->a[i*rows],eins_,&rab->a[j*rows],eins_);
      }else{
    if(transpose_b==0){
      evaltrans_supermatrix(b,&ra->a[i*size],&rab->b[j*cols]);
    }else{
      eval_supermatrix(b,&ra->a[i*size],&rab->b[j*cols]);
    }
    dcopy_(&rows,&ra->b[i*rows],eins_,&rab->a[j*rows],eins_);
      }


      if((dnrm2_(&rows,&rab->a[j*rows],eins_) + 
      dnrm2_(&cols,&rab->b[j*cols],eins_)) > 1e-32){
    j++;
      }
    }
    rab->kt=j;
  }else{
    j=0; 
    for(i=0; i<k; i++){
      if(transpose_a==0){
    if(transpose_b==0){
      eval_supermatrix(a,&rb->a[i*size],&rab->a[j*rows]);
    }else{
      eval_supermatrix(a,&rb->b[i*size],&rab->a[j*rows]);
    }
      }else{
    if(transpose_b==0){
      evaltrans_supermatrix(a,&rb->a[i*size],&rab->a[j*rows]);
    }else{
      evaltrans_supermatrix(a,&rb->b[i*size],&rab->a[j*rows]);
    }
      }
 
      if(transpose_b==0){
    dcopy_(&cols,&rb->b[i*cols],eins_,&rab->b[j*cols],eins_);
      }else{
    dcopy_(&cols,&rb->a[i*cols],eins_,&rab->b[j*cols],eins_);
      }

      if((dnrm2_(&rows,&rab->a[j*rows],eins_) + 
      dnrm2_(&cols,&rab->b[j*cols],eins_)) > 1e-32){
    j++;
      }
    }
    rab->kt=j;
  }
  return rab;
}

void    
addgeneralprod2_rkmatrix(prkmatrix r, psupermatrix a, psupermatrix b,
             int row_offset, int col_offset, 
             double scale_ab, double transpose_a, double transpose_b){
  int i,j,k,block_rows,block_cols,block_size,rows,cols,size,a_rows,b_cols;
  int* rk_row_offset;
  int* rk_col_offset;
  int* rk_row_offset2;
  int* rk_col_offset2;
  prkmatrix* rk_row;
  prkmatrix* rk_col;
  prkmatrix rtmp;
  psupermatrix stmp,a_row,b_col;
  prkmatrix r2=0x0;
  pfullmatrix f;
  
  assert(r != 0x0);
  assert(a != 0x0);
  assert(b != 0x0);

  if(transpose_a==0){
    rows = a->rows;
    size = a->cols;
    block_rows = a->block_rows;
    block_size = a->block_cols;
  }else{
    rows = a->cols;
    size = a->rows;
    block_rows = a->block_cols;
    block_size = a->block_rows;
  }
  if(transpose_b==0){
    cols = b->cols;
    block_cols = b->block_cols;
  }else{
    cols = b->rows;
    block_cols = b->block_rows;
  }

  if(a->s!=0x0 && b->s!=0x0){
    /* subdivision of a and b, convert subblocks to rk, then add all subblocks */
    assert((transpose_b==0 && b->block_rows == block_size)
       || (transpose_b!=0 && b->block_cols == block_size));
    rk_row_offset = (int*) 
      malloc((block_rows+block_cols+block_cols+block_rows) * sizeof(int));
    assert(rk_row_offset != 0x0);
    rk_row_offset2 = &rk_row_offset[block_rows];
    rk_col_offset  = &rk_row_offset[block_rows+block_cols];
    rk_col_offset2 = &rk_row_offset[block_rows+block_cols+block_cols];
    rk_row = (prkmatrix*) malloc((block_rows+block_cols)*sizeof(prkmatrix));
    assert(rk_row != 0x0);
    rk_col = &rk_row[block_rows];
    
    rk_col_offset[0] = col_offset;
    for(j=0; j<block_cols; j++){
      if(j>0){
    if(transpose_b==0){
      rk_col_offset[j] = rk_col_offset[j-1] + b->s[(j-1)*block_size]->cols;
    }else{
      rk_col_offset[j] = rk_col_offset[j-1] + b->s[j-1]->rows;
    }
      }
      rk_row_offset2[j] = 0;
    }
    for(i=0; i<block_rows; i++){
      if(transpose_a==0){
    rk_row[i] = new_rkmatrix(r->k,a->s[i]->rows,cols);  
      }else{
    rk_row[i] = new_rkmatrix(r->k,a->s[i*block_rows]->cols,cols);
      }
      if(rk_arithmetics_adaptive()==1){
    rk_row[i]->local_eps = r->local_eps;
      }
      for(j=0; j<block_cols; j++){
    if(transpose_a==0){
      a_rows = a->s[i]->rows;
    }else{
      a_rows = a->s[i*block_rows]->cols;
    }
    if(transpose_b==0){
      b_cols = b->s[j*block_size]->cols;
}else{
	  b_cols = b->s[j]->rows;
	}
	
	rk_col[j] = new_rkmatrix(r->k,a_rows,b_cols);
	if(rk_arithmetics_adaptive()==1){
	  rk_col[j]->local_eps = r->local_eps;
	}
	for(k=0; k<block_size; k++){
	  if(transpose_a==0){
	    a_row = a->s[i+k*block_rows];
	  }else{
	    a_row = a->s[k+i*block_rows];
	  }
	  if(transpose_b==0){
	    b_col = b->s[k+j*block_size];
	  }else{
	    b_col = b->s[j+k*block_size];
	  }
	  addgeneralprod2_rkmatrix(rk_col[j],a_row,b_col,0,0,1.0,
				   transpose_a,transpose_b);
	}
      }
      addparts2_rkmatrix(rk_row[i],block_cols,rk_row_offset2,rk_col_offset,rk_col);
      for(j=0; j<block_cols; j++) del_rkmatrix(rk_col[j]);
      scale_rkmatrix(rk_row[i],scale_ab);
    }
    rk_row_offset[0] = row_offset;
    for(i=0; i<block_rows; i++){
      if(i>0){
	if(transpose_a==0){
	  rk_row_offset[i] = rk_row_offset[i-1] + a->s[i-1]->rows;
	}else{
	  rk_row_offset[i] = rk_row_offset[i-1] + a->s[(i-1)*block_rows]->cols;
	}
      }
      rk_col_offset2[i] = col_offset;
    }
    addparts2_rkmatrix(r,block_rows,rk_row_offset,rk_col_offset2,rk_row);
    for(i=0; i<block_rows; i++) del_rkmatrix(rk_row[i]);
    free(rk_row_offset);
    free(rk_row);
  }else{
    if(a->r!=0x0 || b->r!=0x0){
      /* one of a and b is rkmatrix -> trivial */
      r2 = getgeneralmul_rkmatrix(a,b,transpose_a,transpose_b);
      scale_rkmatrix(r2,scale_ab);
      addpart2_rkmatrix(r,row_offset,col_offset,r2,0,0,r2->rows,r2->cols);
      del_rkmatrix(r2);
    }else{
      if(a->f!=0x0 || b->f!=0x0){
	/* both are fullmatrices but a->cols can be very small
	   so that we consider a*b as an rkmatrix factorisation */
	if(size<rows && size<cols){
	  if((rows<cols && a->f!=0x0)|| (b->f==0x0 && a->f!=0x0)){
	    rtmp = new_rkmatrix(size,a->rows,a->cols);
	    stmp = new_supermatrix(1,1,a->rows,a->cols,0x0,rtmp,0x0);
	    addfull2_rkmatrix(a->f,rtmp,0,0);
	    r2 = getgeneralmul_rkmatrix(stmp,b,transpose_a,transpose_b);
	    scale_rkmatrix(r2,scale_ab);
	    del_supermatrix(stmp);
	    addpart2_rkmatrix(r,row_offset,col_offset,r2,0,0,r2->rows,r2->cols);
	    del_rkmatrix(r2);
	  }else{
	    rtmp = new_rkmatrix(size,b->rows,b->cols);
	    stmp = new_supermatrix(1,1,b->rows,b->cols,0x0,rtmp,0x0);
	    addfull2_rkmatrix(b->f,rtmp,0,0);
	    r2 = getgeneralmul_rkmatrix(a,stmp,transpose_a,transpose_b);
	    scale_rkmatrix(r2,scale_ab);
	    del_supermatrix(stmp);
	    addpart2_rkmatrix(r,row_offset,col_offset,r2,0,0,r2->rows,r2->cols);
	    del_rkmatrix(r2);
	  }
	}else{
	  f = getgeneralmul_fullmatrix(a,b,transpose_a,transpose_b);
	  r2 = new_rkmatrix(r->k,rows,cols);
	  if(rk_arithmetics_adaptive()==1){
	    r2->local_eps  = r->local_eps;
	  }
	  addfull2_rkmatrix(f,r2,0,0);
	  scale_rkmatrix(r2,scale_ab);
	  addpart2_rkmatrix(r,row_offset,col_offset,r2,0,0,r2->rows,r2->cols);
	  del_rkmatrix(r2);
	  del_fullmatrix(f);
	}
      }else{
	assert(a->u!=0x0 && b->u!=0x0);
	/* one of the factors is a uniform matrix - what to do ?*/
      }
    } 
  }  
}

void
convertfullhierarchic2_rkmatrix(pfullmatrix f, int rowoffset, int coloffset, prkmatrix r){
  int i,j,rk_ro[4],rk_co[4];
  prkmatrix rk_part[4];
  
  assert(r != 0x0);
  assert(f->rows >= rowoffset+r->rows);
  assert(f->cols >= coloffset+r->cols);
  if(r->rows>42 && r->cols>42){
    rk_part[0] = new_rkmatrix(r->k,r->rows/2,r->cols/2);
    rk_part[1] = new_rkmatrix(r->k,r->rows-r->rows/2,r->cols/2);
    rk_part[2] = new_rkmatrix(r->k,r->rows/2,r->cols-r->cols/2);
    rk_part[3] = new_rkmatrix(r->k,r->rows-r->rows/2,r->cols-r->cols/2);
    for(i=0; i<2; i++){
      for(j=0; j<2; j++){
	rk_ro[i+2*j] = i*r->rows/2;
	rk_co[i+2*j] = j*r->cols/2;
	rk_part[i+2*j]->local_eps = r->local_eps;
	convertfullhierarchic2_rkmatrix(f,rowoffset+rk_ro[i+2*j],
					coloffset+rk_co[i+2*j],rk_part[i+2*j]);
      }
    }
    addparts2_rkmatrix(r,4,rk_ro,rk_co,rk_part);
    for(i=0; i<4; i++) 
      del_rkmatrix(rk_part[i]);
  }else{
    convertfull2_rkmatrix(f,r,rowoffset,coloffset);
  }
}


void
convertsuper2_rkmatrix(psupermatrix a, prkmatrix r){
  int i,j,block_rows,block_cols;
  int* rk_row_offset;
  int* rk_col_offset;
  int* rk_row_offset2;
  int* rk_col_offset2;
  prkmatrix* rk_row;
  prkmatrix* rk_col;
  
  assert(r != 0x0);
  assert(a != 0x0);
  
  block_rows = a->block_rows;
  block_cols = a->block_cols;
  if(a->s!=0x0){
    rk_row_offset = (int*) 
      malloc((block_rows+block_cols+block_cols+block_rows) * sizeof(int));
    assert(rk_row_offset != 0x0);
    rk_row_offset2 = &rk_row_offset[block_rows];
    rk_col_offset = &rk_row_offset[block_rows+block_cols];
    rk_col_offset2 = &rk_row_offset[block_rows+block_cols+block_cols];
    rk_row = (prkmatrix*) malloc((block_rows+block_cols) * sizeof(prkmatrix));
    assert(rk_row != 0x0);
    rk_col = &rk_row[block_rows];
    rk_col_offset[0] = 0;
    for(j=0; j<block_cols; j++){
      if(j>0) rk_col_offset[j] = rk_col_offset[j-1] + a->s[(j-1)*block_rows]->cols;
      rk_row_offset2[j] = 0;
    }
    for(i=0; i<block_rows; i++){
      rk_row[i] = new_rkmatrix(r->k,a->s[i]->rows,a->cols);
      if(rk_arithmetics_adaptive()==1) rk_row[i]->local_eps = r->local_eps;
      for(j=0; j<block_cols; j++){
	rk_col[j] = new_rkmatrix(r->k,a->s[i]->rows,a->s[j*block_rows]->cols);
	if(rk_arithmetics_adaptive()==1){
	  rk_col[j]->local_eps = r->local_eps;
	}
	convertsuper2_rkmatrix(a->s[i+j*block_rows],rk_col[j]);
      }
      addparts2_rkmatrix(rk_row[i],block_cols,rk_row_offset2,rk_col_offset,rk_col);
      for(j=0; j<block_cols; j++) del_rkmatrix(rk_col[j]);
    }
    rk_row_offset[0] = 0;
    for(i=0; i<block_rows; i++){
      if(i>0) rk_row_offset[i] = rk_row_offset[i-1] + a->s[i-1]->rows;
      rk_col_offset2[i] = 0;
    }
    addparts2_rkmatrix(r,block_rows,rk_row_offset,rk_col_offset2,rk_row);
    for(i=0; i<block_rows; i++) del_rkmatrix(rk_row[i]);
    free(rk_row_offset);
    free(rk_row);
  }else{
    if(a->r!=0x0){
      convertrk2_rkmatrix(a->r,r,0,0);
    }else{
      if(a->f != 0x0){
	convertfull2_rkmatrix(a->f,r,0,0);
      }else{
	convertuniform2_rkmatrix(a->u,r);
      }
    }
  }
}

void
convertsparse2_rkmatrix(pcsparsematrix sp, int rowstart, int colstart, 
			prkmatrix r){
  prkmatrix rtmp;
  int i,ii,k,knz,nu;

  r->kt = 0;
  k = 0;
  for(i=0; i<r->rows; i++) {
    ii = rowstart+i;
    knz = 0;
    for(nu=sp->row[ii]; nu<sp->row[ii+1]; nu++)
      if(sp->col[nu] >= colstart && sp->col[nu]<colstart+r->cols &&
		 fabs(sp->coeff[nu])>RK_SVD_EPS2) knz=1;
    if(knz==1) k++;
  }
  if(k==0) return;
  
  rtmp = new_rkmatrix(k,r->rows,r->cols);
  rtmp->kt = k;
  k = 0;
  for(i=0; i<r->rows; i++) {
    ii = rowstart+i;
    knz = 0;
    for(nu=sp->row[ii]; nu<sp->row[ii+1]; nu++)
      if(sp->col[nu] >= colstart && sp->col[nu]<colstart+r->cols &&
		 fabs(sp->coeff[nu])>RK_SVD_EPS2) knz=1;
    if(knz==1){
      for(nu=sp->row[ii]; nu<sp->row[ii+1]; nu++)
		if(sp->col[nu] >= colstart && sp->col[nu]<colstart+r->cols) 
		  rtmp->b[k*r->cols+sp->col[nu]-colstart] = sp->coeff[nu];
	  
      rtmp->a[k*r->rows+i] = 1.0;
      k++;
    }
  }
  addpart2_rkmatrix(r, 0, 0, rtmp, 0, 0, r->rows, r->cols);
  del_rkmatrix(rtmp);
}

double* 
getsigma_rkmatrix(prkmatrix r){
  double *atmp;
  double *btmp;
  int rows,cols,kmax;
  int i,j,k=0;
  double* qr_work;
  double* tau1;
  double* tau2;
  double* sigma;
  double* rarb;
  double* rarbwork;
  double* u=0x0;
  double* v=0x0;
  pfullmatrix f;
  int lwork;
  int info;

  assert(r != 0x0);
  rows = r->rows;
  cols = r->cols;
  kmax=r->kt;

  sigma = (double*) calloc(rows+cols,sizeof(double));
  assert(sigma!=0x0);

  if(kmax==0){
    return sigma;
  }
  if(kmax>=rows || kmax>=cols){
    f = new_fullmatrix(rows,cols);
    convertrk2_fullmatrix(r,f,0,0);
    lwork = 10*rows*cols;
    rarbwork = (double*) calloc(lwork,sizeof(double)); 
    assert(rarbwork!=0x0);
    dgesvd_("N","N",&rows,&cols,f->e,&rows,sigma,u,&rows,v,&cols,
	    rarbwork,&lwork,&info);
    if(info!=0){
      printf("get_sv_rkmatrix: info(dgesvd)=%d\n",info);
      exit(0);
    }
    del_fullmatrix(f);
    free(rarbwork);
    return sigma;
  }
  lwork = rows+cols+10;
  qr_work = (double*) calloc(lwork,sizeof(double)); 
  assert(qr_work!=0x0);
  tau1 = (double*) calloc(rows+cols,sizeof(double));
  assert(tau1!=0x0);
  tau2 = (double*) calloc(rows+cols,sizeof(double));
  assert(tau2!=0x0);
  atmp = (double*) calloc(rows*kmax,sizeof(double));
  assert(atmp!=0x0);
  btmp = (double*) calloc(cols*kmax,sizeof(double));
  assert(btmp!=0x0);
  rarb = (double*) calloc(kmax*kmax,sizeof(double));
  assert(rarb!=0x0);
  for(i=0; i<kmax; i++){
    dcopy_(&rows,&r->a[i*rows],eins_,&atmp[i*rows],eins_);
    dcopy_(&cols,&r->b[i*cols],eins_,&btmp[i*cols],eins_);
  }
  dgeqrf_(&rows,&kmax,atmp,&rows,tau1,qr_work,&lwork,&info);
  if(info!=0){
    printf("get_sv_rkmatrix: info(dgeqrf,a)=%d\n",info);
    exit(0);
  }
  dgeqrf_(&cols,&kmax,btmp,&cols,tau2,qr_work,&lwork,&info);
  if(info!=0){
    printf("get_sv_rkmatrix: info(dgeqrf,b)=%d\n",info);
    exit(0);
  }
  for(i=0; i<kmax; i++){
    for(j=0; j<kmax; j++){
      for(k=0; k<kmax; k++){
	if(k>=i && k>=j) rarb[i+kmax*j] += atmp[i+k*rows]*btmp[j + k*cols];
      }
    }
  }
  lwork = 10*kmax*kmax;
  rarbwork = (double*) calloc(lwork,sizeof(double));
  assert(rarbwork!=0x0);
  dgesvd_("N","N",&kmax,&kmax,rarb,&kmax,sigma,u,&kmax,v,&kmax,
	  rarbwork,&lwork,&info);
  if(info!=0){
    printf("get_sv_kmatrix: info(dgesvd)=%d\n",info);
    exit(0);
  }
  free(rarbwork);
  free(rarb);
  free(tau1);
  free(tau2);
  free(qr_work);
  free(btmp);
  free(atmp);
  return sigma;
}

double
normfrobenius_rkmatrix(prkmatrix r){
  double normf = 0.0;
  double *sigma;
  int i;
  assert(r != 0x0);
  sigma = getsigma_rkmatrix(r);
  for(i=0; i < r->kt; i++)
    normf += sigma[i]*sigma[i];
  normf = sqrt(normf);
  free(sigma);
  return normf;
}

double
norm2_rkmatrix(prkmatrix r){
  double norm2;
  double *sigma;
  assert(r != 0x0);
  sigma = getsigma_rkmatrix(r);
  norm2 = sigma[0]; 
  free(sigma);
  return norm2;
}

void
outputleaf_rkmatrix(prkmatrix r, FILE* datei, pblockcluster bc,  
		    int row_offset, int col_offset, int row_max, int col_max,
		    int rank, int fill, int border, double tol){
  int i,size,rows,cols,k;
  double x1,x2,y,ymax;
  double *sigma;
  double ytol;

  assert(r != 0x0);
  rows = r->rows;
  cols = r->cols;
  k = r->kt;
  sigma = getsigma_rkmatrix(r);
  size = rows; if(cols<rows) size = cols;

  if(k>0){
    if(fill > 1){
      for(i=0; i<k; i++){
	x1 =(0.05+0.9*(i)/k)*cols;
	x2 =(0.05+0.9*(i+1)/k)*cols;
	y = 1.0;
	if(sigma[i]>0){
	  y = log(sigma[i])/log(10);
	  y = 0.05+0.9*(y + 15.0)/16.0;     /* Scale 10^-16 to 10^-0 */
	  if(y>0.95) y = 0.95;
	  if(y>0.07){   
	    fprintf(datei,"%.2f %.2f m\n",col_offset + x1, 
		    row_max-rows-row_offset+0.05*rows);
	    fprintf(datei,"%.2f %.2f l\n",col_offset + x1, 
		    row_max-rows-row_offset+y*rows);
	    fprintf(datei,"%.2f %.2f l\n",col_offset + x2, 
		  row_max-rows-row_offset+y*rows);
	    fprintf(datei,"%.2f %.2f l\n",col_offset + x2, 
		    row_max-rows-row_offset+0.05*rows);
	    fprintf(datei,"%.2f %.2f l\n",col_offset + x1, 
		    row_max-rows-row_offset+0.05*rows);
	    if(fill <= 2){
	      if(bc!=0x0 && bc->adm==0) fprintf(datei,"fi\n");
	      if(bc==0x0 || (bc!=0x0 && bc->adm==1)) fprintf(datei,"fl\n");
	      if(bc!=0x0 && bc->adm==2) fprintf(datei,"fr\n");	      
	    }
	    fprintf(datei,"cs\n");
	  }
	}
      }
    }else{
      if(fill > 0){
	fprintf(datei,"%d %d m\n",col_offset,row_max-row_offset-rows);
	fprintf(datei,"%d %d l\n",col_offset,row_max-row_offset);
	fprintf(datei,"%d %d l\n",col_offset+cols,row_max-row_offset);
	fprintf(datei,"%d %d l\n",col_offset+cols,row_max-row_offset-rows);
	fprintf(datei,"%d %d l\n",col_offset,row_max-row_offset-rows);
	if(bc!=0x0 && bc->adm==0) fprintf(datei,"fi\n");
	if(bc==0x0 || (bc!=0x0 && bc->adm==1)) fprintf(datei,"fl\n");
	if(bc!=0x0 && bc->adm==2) fprintf(datei,"fr\n");	      
	fprintf(datei,"cs\n");
      }
    }
  }
  if(rank > 0 && k>0){
    fprintf(datei,"bl\n");
    if(border<2) fprintf(datei,"hf %.1f sf\n",size*0.5);
    fprintf(datei,"%.1f %.1f m (%d) show\n",col_offset+cols*0.2,
	    row_max-rows-row_offset+rows*0.1,r->kt);
  }
  if(border > 1 && k>0){
    for(i=0; i<6; i++){
      ymax = - (int) log(sigma[0])/log(10);
      y = ymax -3.0*i;
      y = 0.05+0.9*(y + 15.0)/16.0;
      x1 = 0.0*cols;
      x2 = 1.01*cols;
      fprintf(datei,"%.2f %.2f m\n",col_offset + x1, row_max-rows-row_offset+y*rows);
      fprintf(datei,"%.2f %.2f l\n",col_offset + x2, row_max-rows-row_offset+y*rows);
      fprintf(datei,"cs\n");
      x2 = 1.02*cols;
      fprintf(datei,"%.2f %.2f m (10) show\n",col_offset + x2, 
	      row_max-rows-row_offset+y*rows);
      fprintf(datei,"%.2f %.2f m (-%d) show\n",col_offset + x2 + 0.04*rows, 
	      row_max-rows-row_offset+y*rows+0.02*rows,3*i);
    }
  }
  if(border > 0 && k>0){
      fprintf(datei,"bl\n");     
      fprintf(datei,"%d %d m\n",col_offset     ,row_max-rows-row_offset);
      fprintf(datei,"%d %d l\n",col_offset     ,row_max-rows-row_offset+rows);
      fprintf(datei,"%d %d l\n",col_offset+cols,row_max-rows-row_offset+rows);
      fprintf(datei,"%d %d l\n",col_offset+cols,row_max-rows-row_offset);
      fprintf(datei,"%d %d l\n",col_offset     ,row_max-rows-row_offset);
      fprintf(datei,"cs\n");

  }
  if(tol>0 && fill>1){
    ytol = log(tol)/log(10);
    ytol = 0.05+0.9*(ytol + 15.0)/16.0;
    if(ytol>0.07 && ytol<0.95){
      y = ytol;
      x1 = 0.0*cols;
      x2 = 1.01*cols;
      fprintf(datei,"%.2f %.2f m\n",col_offset + x1, row_max-rows-row_offset+y*rows);
      fprintf(datei,"%.2f %.2f l\n",col_offset + x2, row_max-rows-row_offset+y*rows);
      fprintf(datei,"cs\n");
      x2 = 1.02*cols;
      fprintf(datei,"%.2f %.2f m (EPS) show\n",col_offset + x2, 
	      row_max-rows-row_offset+y*rows);
    }
  }
  free(sigma);
}

double
getsv_rkmatrix(prkmatrix r, int k_sv){
  double tmp;
  double *sigma;
  assert(r != 0x0);
  sigma = getsigma_rkmatrix(r);
  if(k_sv>r->rows || k_sv>r->cols){
    tmp = 0.0;
  }else{
    tmp = sigma[k_sv];
  }
  free(sigma);
  return tmp;
}
