
/* ------------------------------------------------------------
   This is the file "solver.c" of the HLib package.

   Copyright 1999 Lars Grasedyck and Steffen Boerm
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#include "basic.h"
#include "solver.h"

#include <math.h>
#include <stdio.h>

double* 
solvepreconditionedcg_supermatrix(psupermatrix s, psupermatrix winv, double* rhs){
  /* preconditioned conjugate gradient method */
  int i,it,n;
  double* x;
  double* r;
  double* p;
  double* ap;
  double* ap_tmp;
  double lambda;
  double app,rap,error;

  n = s->rows;
  x = (double*) malloc(n * sizeof(double));
  r = (double*) malloc(n * sizeof(double));
  p = (double*) malloc(n * sizeof(double));
  ap = (double*) malloc(n * sizeof(double));
  ap_tmp = (double*) malloc(n * sizeof(double));
  lambda = 1.0;

  eval_supermatrix(winv,rhs,r);

  for(i=0; i<n;i++){
    x[i] = 0.0;
  }

  for(i=0; i<n;i++){
    p[i] = r[i];
  }

  for(it=0; it<500; it++){
    eval_supermatrix(s,p,ap_tmp);
    eval_supermatrix(winv,ap_tmp,ap);
    
    error = 0.0;
    for(i=0; i<n;i++) error += r[i]*r[i];
    error = sqrt(error);

    if(error>1e-13){
      lambda = 0.0;
      for(i=0; i<n;i++) lambda += p[i]*r[i];
      app = 0.0;
      for(i=0; i<n;i++) app += p[i]*ap[i];      
      lambda /= app;
      for(i=0; i<n;i++){
	x[i] += lambda*p[i];
	r[i] -= lambda*ap[i];
      }
      rap = 0.0;
      for(i=0; i<n;i++) rap += r[i]*ap[i];      
      for(i=0; i<n;i++){
	p[i] = r[i] -(rap/app)*p[i];
      }
    }else{
      printf("# converged in %d steps.\n",it);
      it = 501;
    }
  }
  if(it==500) printf("# did not converge.\n");
  printf("# |Winv G u - Winv f| = %g\n",error);

  eval_supermatrix(s,x,r);
  error = 0.0;
  for(i=0; i<n; i++)
    error += (r[i] - rhs[i]) * (r[i] - rhs[i]);
  error = sqrt(error);
  printf("# |G u - f| = %g\n", error);

  free(r);
  free(p);
  free(ap);
  free(ap_tmp);
  return x;
}

double* 
solvecholeskypreconditionedcg_supermatrix(psparsematrix s, psupermatrix l, double* rhs, double eps){
  /* preconditioned conjugate gradient method */
  int i,it,n;
  double* x;
  double* r;
  double* p;
  double* ap;
  double* ap_tmp;
  double lambda;
  double app,rap,error,norm;

  n = s->rows;
  x = (double*) malloc(n * sizeof(double));
  r = (double*) malloc(n * sizeof(double));
  p = (double*) malloc(n * sizeof(double));
  ap = (double*) malloc(n * sizeof(double));
  ap_tmp = (double*) malloc(n * sizeof(double));
  lambda = 1.0;

  dcopy_(&s->rows,rhs,eins_,r,eins_);
  solvecholeskyvector_supermatrix(l, r);
  solvecholeskyvectortrans_supermatrix(l, r);

  norm = dnrm2_(&n,rhs,eins_);

  for(i=0; i<n;i++){
    x[i] = 0.0;
  }

  for(i=0; i<n;i++){
    p[i] = r[i];
  }

  error = 1.0;

  for(it=0; it<500; it++){
    if(error>eps){
      eval_sparsematrix(s,p,ap);
      solvecholeskyvector_supermatrix(l, ap);
      solvecholeskyvectortrans_supermatrix(l, ap);
      
      lambda = ddot_(&n,p,eins_,r,eins_);
      app = ddot_(&n,p,eins_,ap,eins_);
      lambda /= app;
      daxpy_(&n,&lambda,p,eins_,x,eins_);
      lambda *= -1.0;
      daxpy_(&n,&lambda,ap,eins_,r,eins_);
      rap = ddot_(&n,r,eins_,ap,eins_);
      lambda = -(rap/app);
      dscal_(&n,&lambda,p,eins_);
      daxpy_(&n,deins_,r,eins_,p,eins_);

      dcopy_(&n,rhs,eins_,ap,eins_);
      dscal_(&n,meins_,ap,eins_);
      addeval_sparsematrix(s,x,ap);
      error = dnrm2_(&n,ap,eins_)/norm;
    }else{
      printf("# converged in %d steps.\n",it);
      it = 501;
    }
  }
  if(it==500) printf("# did not converge.\n");
  printf("# |Winv G u - Winv f| = %g\n",error);

  dcopy_(&n,rhs,eins_,r,eins_);
  dscal_(&n,meins_,r,eins_);
  addeval_sparsematrix(s,x,r);
  error = dnrm2_(&n,r,eins_)/norm;
  printf("# |G u - f|/|f| = %g\n", error);

  free(r);
  free(p);
  free(ap);
  free(ap_tmp);
  return x;
}

double* 
solvecholeskypreconditionedcg2_supermatrix(psupermatrix s, psupermatrix l, double* rhs,
				  double eps){
  /* preconditioned conjugate gradient method */
  int i,it,n;
  double* x;
  double* r;
  double* p;
  double* ap;
  double* ap_tmp;
  double lambda;
  double app,rap,error;

  n = s->rows;
  x = (double*) malloc(n * sizeof(double));
  r = (double*) malloc(n * sizeof(double));
  p = (double*) malloc(n * sizeof(double));
  ap = (double*) malloc(n * sizeof(double));
  ap_tmp = (double*) malloc(n * sizeof(double));
  lambda = 1.0;

  dcopy_(&s->rows,rhs,eins_,r,eins_);
  solvecholeskyvector_supermatrix(l, r);
  solvecholeskyvectortrans_supermatrix(l, r);

  for(i=0; i<n;i++){
    x[i] = 0.0;
  }

  for(i=0; i<n;i++){
    p[i] = r[i];
  }

  for(it=0; it<500; it++){
    eval_supermatrix(s,p,ap);
    solvecholeskyvector_supermatrix(l, ap);
    solvecholeskyvectortrans_supermatrix(l, ap);
    
    error = 0.0;
    for(i=0; i<n;i++) error += r[i]*r[i];
    error = sqrt(error);

    if(error>eps){
      lambda = 0.0;
      for(i=0; i<n;i++) lambda += p[i]*r[i];
      app = 0.0;
      for(i=0; i<n;i++) app += p[i]*ap[i];      
      lambda /= app;
      for(i=0; i<n;i++){
	x[i] += lambda*p[i];
	r[i] -= lambda*ap[i];
      }
      rap = 0.0;
      for(i=0; i<n;i++) rap += r[i]*ap[i];      
      for(i=0; i<n;i++){
	p[i] = r[i] -(rap/app)*p[i];
      }
    }else{
      printf("# converged in %d steps.\n",it);
      it = 501;
    }
  }
  if(it==500) printf("# did not converge.\n");
  printf("# |Winv G u - Winv f| = %g\n",error);

  eval_supermatrix(s,x,r);
  error = 0.0;
  for(i=0; i<n; i++)
    error += (r[i] - rhs[i]) * (r[i] - rhs[i]);
  error = sqrt(error);
  printf("# |G u - f| = %g\n", error);

  free(r);
  free(p);
  free(ap);
  free(ap_tmp);
  return x;
}

double* 
solvelupreconditionedcg2_supermatrix(psupermatrix s, psupermatrix l, double* rhs,
				  double eps){
  /* preconditioned conjugate gradient method */
  int i,it,n;
  double* x;
  double* r;
  double* p;
  double* ap;
  double* ap_tmp;
  double lambda;
  double app,rap,error;

  n = s->rows;
  x = (double*) malloc(n * sizeof(double));
  r = (double*) malloc(n * sizeof(double));
  p = (double*) malloc(n * sizeof(double));
  ap = (double*) malloc(n * sizeof(double));
  ap_tmp = (double*) malloc(n * sizeof(double));
  lambda = 1.0;

  dcopy_(&s->rows,rhs,eins_,r,eins_);
  solvelower_supermatrix(l,r);
  solveupper_supermatrix(l,r);

  for(i=0; i<n;i++){
    x[i] = 0.0;
  }

  for(i=0; i<n;i++){
    p[i] = r[i];
  }

  for(it=0; it<500; it++){
    eval_supermatrix(s,p,ap);
    solvelower_supermatrix(l,ap);
    solveupper_supermatrix(l,ap);
    
    error = 0.0;
    for(i=0; i<n;i++) error += r[i]*r[i];
    error = sqrt(error);
    printf(" err = %e\n",error);

    if(error>eps){
      lambda = 0.0;
      for(i=0; i<n;i++) lambda += p[i]*r[i];
      app = 0.0;
      for(i=0; i<n;i++) app += p[i]*ap[i];      
      lambda /= app;
      for(i=0; i<n;i++){
	x[i] += lambda*p[i];
	r[i] -= lambda*ap[i];
      }
      rap = 0.0;
      for(i=0; i<n;i++) rap += r[i]*ap[i];      
      for(i=0; i<n;i++){
	p[i] = r[i] -(rap/app)*p[i];
      }
    }else{
      printf("# converged in %d steps.\n",it);
      it = 501;
    }
  }
  if(it==500) printf("# did not converge.\n");
  printf("# |Winv G u - Winv f| = %g\n",error);

  eval_supermatrix(s,x,r);
  error = 0.0;
  for(i=0; i<n; i++)
    error += (r[i] - rhs[i]) * (r[i] - rhs[i]);
  error = sqrt(error);
  printf("# |G u - f| = %g\n", error);

  free(r);
  free(p);
  free(ap);
  free(ap_tmp);
  return x;
}

double* 
solvelupreconditionedcg3_supermatrix(psupermatrix s, prkmatrix rk, 
				     psupermatrix l, double* rhs,
				     double eps){
  /* preconditioned conjugate gradient method */
  int i,it,n;
  double* x;
  double* r;
  double* p;
  double* ap;
  double* ap_tmp;
  double lambda;
  double app,rap,error;

  n = s->rows;
  x = (double*) malloc(n * sizeof(double));
  r = (double*) malloc(n * sizeof(double));
  p = (double*) malloc(n * sizeof(double));
  ap = (double*) malloc(n * sizeof(double));
  ap_tmp = (double*) malloc(n * sizeof(double));
  lambda = 1.0;

  dcopy_(&s->rows,rhs,eins_,r,eins_);
  solvelower_supermatrix(l,r);
  solveupper_supermatrix(l,r);

  for(i=0; i<n;i++){
    x[i] = 0.0;
  }

  for(i=0; i<n;i++){
    p[i] = r[i];
  }

  for(it=0; it<100; it++){
    eval_supermatrix(s,p,ap);
    addeval_rkmatrix(rk,p,ap);
    solvelower_supermatrix(l,ap);
    solveupper_supermatrix(l,ap);
    
    error = 0.0;
    for(i=0; i<n;i++) error += r[i]*r[i];
    error = sqrt(error);
    if(it<5 || it%10==0) printf(" err = %e\n",error);

    if(error>eps){
      lambda = 0.0;
      for(i=0; i<n;i++) lambda += p[i]*r[i];
      app = 0.0;
      for(i=0; i<n;i++) app += p[i]*ap[i];      
      lambda /= app;
      for(i=0; i<n;i++){
	x[i] += lambda*p[i];
	r[i] -= lambda*ap[i];
      }
      rap = 0.0;
      for(i=0; i<n;i++) rap += r[i]*ap[i];      
      for(i=0; i<n;i++){
	p[i] = r[i] -(rap/app)*p[i];
      }
    }else{
      printf("# converged in %d steps.\n",it);
      it = 501;
    }
  }
  if(it==500) printf("# did not converge.\n");
  printf("# |Winv G u - Winv f| = %g\n",error);

  eval_supermatrix(s,x,r);
  addeval_rkmatrix(rk,x,r);
  error = 0.0;
  for(i=0; i<n; i++)
    error += (r[i] - rhs[i]) * (r[i] - rhs[i]);
  error = sqrt(error);
  printf("# |G u - f| = %g\n", error);

  free(r);
  free(p);
  free(ap);
  free(ap_tmp);
  return x;
}

double* 
solvecg_supermatrix(psupermatrix s, double* rhs, double eps){
  /* conjugate gradient method */
  int i,it,n;
  double* x;
  double* r;
  double* p;
  double* ap;
  double lambda;
  double app,rap,error;

  n = s->rows;
  x = (double*) malloc(n * sizeof(double));
  r = (double*) malloc(n * sizeof(double));
  p = (double*) malloc(n * sizeof(double));
  ap = (double*) malloc(n * sizeof(double));
  lambda = 1.0;

  for(i=0; i<n;i++){
    x[i] = 0.0;
    r[i] = rhs[i];
    p[i] = r[i];
  }

  for(it=0; it<50; it++){
    eval_supermatrix(s,p,ap);
    
    error = 0.0;
    for(i=0; i<n;i++) error += r[i]*r[i];
    error = sqrt(error);

    if(error>eps){
      printf("# step %d, residual = %e\n",it+1,error);
      lambda = 0.0;
      for(i=0; i<n;i++) lambda += p[i]*r[i];
      app = 0.0;
      for(i=0; i<n;i++) app += p[i]*ap[i];      
      lambda /= app;
      for(i=0; i<n;i++){
	x[i] += lambda*p[i];
	r[i] -= lambda*ap[i];
      }
      rap = 0.0;
      for(i=0; i<n;i++) rap += r[i]*ap[i];      
      for(i=0; i<n;i++){
	p[i] = r[i] -(rap/app)*p[i];
      }
    }else{
      printf("# converged in %d steps.\n",it);
      it = 501;
    }
  }
  if(it==500) printf("# did not converge.\n");
  printf("# |Gu-f| = %g\n",error);
  free(r);
  free(p);
  free(ap);
  return x;
}

double* 
solvecgkaczmarz_supermatrix(psupermatrix s, double* rhs, double eps){
  /* conjugate gradient method */
  int i,it,n;
  double* x;
  double* r;
  double* p;
  double *ap, *ap2;
  double lambda;
  double app,rap,error;

  n = s->rows;
  x = (double*) malloc(n * sizeof(double));
  r = (double*) malloc(n * sizeof(double));
  p = (double*) malloc(n * sizeof(double));
  ap2 = (double*) malloc(n * sizeof(double));
  ap = (double*) malloc(n * sizeof(double));
  lambda = 1.0;

  evaltrans_supermatrix(s,rhs,r);
  for(i=0; i<n;i++){
    x[i] = 0.0;
    p[i] = r[i];
  }

  for(it=0; it<500; it++){
    eval_supermatrix(s,p,ap2);
    evaltrans_supermatrix(s,ap2,ap);
    
    error = 0.0;
    for(i=0; i<n;i++) error += r[i]*r[i];
    error = sqrt(error);

    if(error>eps){
      lambda = 0.0;
      for(i=0; i<n;i++) lambda += p[i]*r[i];
      app = 0.0;
      for(i=0; i<n;i++) app += p[i]*ap[i];      
      lambda /= app;
      for(i=0; i<n;i++){
	x[i] += lambda*p[i];
	r[i] -= lambda*ap[i];
      }
      rap = 0.0;
      for(i=0; i<n;i++) rap += r[i]*ap[i];      
      for(i=0; i<n;i++){
	p[i] = r[i] -(rap/app)*p[i];
      }
    }else{
      printf("# converged in %d steps.\n",it);
      it = 501;
    }
  }
  if(it==500) printf("# did not converge.\n");
  printf("# |Gu-f| = %g\n",error);
  free(r);
  free(p);
  free(ap);
  return x;
}
