
/* ------------------------------------------------------------
   This is the file "sparsematrix.h" of the HLib package.

   Copyright 1999 Steffen Boerm and Lars Grasedyck
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#ifndef SparseMatrix
#define SparseMatrix

typedef struct _sparsefactory sparsefactory;
typedef sparsefactory *psparsefactory;
typedef const sparsefactory *pcsparsefactory;

typedef struct _sparsematrix sparsematrix;
typedef sparsematrix *psparsematrix;
typedef const sparsematrix *pcsparsematrix;

#include "basic.h"
#include "supermatrix.h"
#include "blockcluster.h"

struct _sparsefactory {
  int rows;			/* Number of rows */
  int cols;			/* Number of columns */

  struct _smat_entry **row;	/* List of non-zero entries */

  long int *nonzeros;		/* Number of non-zero entries per row */
};

struct _sparsematrix {
  int rows;			/* Number of rows*/
  int cols;			/* Number of columns */
  int nz;			/* Number of non-zero entries */

  int *row;			/* Pointer to non-zero entries for each row */
  int *col;			/* Column indices of non-zero entries */
  double *coeff;		/* Coefficients of non-zero entries */
};

/* A sparsefactory is used to store the structure of sparse matrices,
   namely the pattern of non-zero entries.

   For each row, we have a list of non-zero matrix entries.
   These lists are filled by calls to SparseFactory_AddNonzero.
   As soon as the pattern is established, a call to
   SparseFactory_BuildSparseMatrix creates a corresponding
   SparseMatrix object. */

/* Create an empty SparseFactory object
   rows: Number of rows of the sparse matrix 
   cols: Number of columns of the sparse matrix */
psparsefactory
new_sparsefactory(int rows, int cols);

/* Free the memory used by a SparseFactory object */
void
del_sparsefactory(psparsefactory fac);

/* Remove all non-zero entries from a sparsefactory object */
void
clear_sparsefactory(psparsefactory fac);

/* Add a non-zero entry to a sparsefactory object,
   if it does not already exist
   row: Row of the non-zero entry
   col: Column of the non-zero entry */
void
addnz_sparsefactory(psparsefactory fac, int row, int col);

/* Print the contents of a sparsefactory object */
void
print_sparsefactory(pcsparsefactory fac);

/* A sparsematrix object stores a matrix in the compressed row sparse
   format: For each row i, row[i] gives the index of first non-zero entry
   in the arrays col and coeff. row[i+1]-1 is the index of the last
   non-zero entry. */

/* Create a SparseMatrix object using an array denoting the number
   of non-zero entries per row. The arrays col and coeff are _not_
   initialized, this has to be done by the caller. */   
psparsematrix
new_sparsematrix(psparsefactory fac);

/* Free the memory used by a SparseMatrix object */
void
del_sparsematrix(psparsematrix sp);

/* Set all non-zero entries to zero (without removing them from the
   pattern, of course) */
void
clear_sparsematrix(psparsematrix sp);

/* Add a value to a non-zero entry
   row: Row of the non-zero entry
   col: Column of the non-zero entry
   val: Value to be added */
void
addcoeff_sparsematrix(psparsematrix sp, int row, int col, double val);

/* Set a non-zero entry to a certain value
   row: Row of the non-zero entry
   col: Column of the non-zero entry
   val: Value to be set */
void
setcoeff_sparsematrix(psparsematrix sp, int row, int col, double val);

/* Print the contents of a SparseMatrix object */
void
print_sparsematrix(pcsparsematrix sp);

/* Multiply a sparsematrix by a vector v and write the result to w. */
void 
eval_sparsematrix(pcsparsematrix sp, const double *v, double *w);

/* Multiply a sparsematrix by a vector v and add the result to w. */
void 
addeval_sparsematrix(pcsparsematrix sp, const double *v, double *w);

/* Multiply the transposed of a sparsematrix by a vector v and write
   the result to w. */
void 
evaltrans_sparsematrix(pcsparsematrix sp, const double *v, double *w);

/* Multiply the transposed of a sparsematrix by a vector v and add
   the result to w. */
void 
addevaltrans_sparsematrix(pcsparsematrix sp, const double *v, double *w);

/* Convert a sparsematrix to a supermatrix. Only fullmatrix blocks will
   be different from zero. */
void
convertsparse2_supermatrix(pcsparsematrix sp, psupermatrix s);

/* Convert a sparsematrix to a supermatrix and allow all offdiagonal blocks 
   corresponding to inadmissible leaves to be rkmatrices. */
void
convertsparse3_supermatrix(pcsparsematrix sp, psupermatrix s, 
			   pblockcluster bc, int nmin);
void
convertsparse3sym_supermatrix(pcsparsematrix sp, psupermatrix s, 
			      pblockcluster bc, int nmin);

#endif
