
/* ------------------------------------------------------------
   This is the file "supermatrix.c" of the HLib package.

   Copyright 1999 Lars Grasedyck and Steffen Boerm
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#include "basic.h"
#include "lapack.h"
#include "supermatrix.h"

#include <assert.h>
#include <math.h>
#include <string.h>
#include <sys/times.h>
#include <time.h>

static long int allocated_supermatrices;
static long int current_supermatrices;

void
init_supermatrices(int verbosity){
  if(verbosity>0)
    printf("Initializing SuperMatrices.\n");
  allocated_supermatrices=0;
  current_supermatrices=0;
}

void
finish_supermatrices(int verbosity){
  if(verbosity>0){
    printf("Finishing SuperMatrices is ");
    if(current_supermatrices!=0){
      printf("not successful:\n");
      printf(" Still In Use : %ld\n",current_supermatrices);
    }else{
      printf("successful:\n");
    }
    printf(" Allocated : %ld\n",allocated_supermatrices);
  }
}

psupermatrix
new_supermatrix(int block_rows, int block_cols, int rows, int cols,
		puniformmatrix u_new, prkmatrix r_new, pfullmatrix f_new)
{
  psupermatrix s;
  int i;

  s = (psupermatrix) malloc(sizeof(supermatrix));
  assert(s!=0x0);
  s->block_rows = block_rows;
  s->block_cols = block_cols;
  s->s = 0x0;
  s->u = 0x0;
  s->r = 0x0;
  s->f = 0x0;
  s->col = 0x0;
  s->row = 0x0;
  if(u_new==0x0 && r_new==0x0 && f_new==0x0){ 
    s->s = (psupermatrix*) malloc(sizeof(psupermatrix)*block_rows*block_cols);
    assert(s->s!=0x0);
    for(i=0; i<block_rows*block_cols; i++)
      s->s[i] = NULL;
    s->rows = rows;
    s->cols = cols;
  }else{
    if(u_new!=0x0){
      assert(r_new==0x0 && f_new==0x0);
      assert(u_new->row == NULL || u_new->rows == rows);
      assert(u_new->col == NULL || u_new->cols == cols);
      s->u = u_new;
      s->rows = rows;
      s->cols = cols;
    }
    if(r_new!=0x0){
      assert(u_new==0x0 && f_new==0x0);
      assert(r_new->rows == rows);
      assert(r_new->cols == cols);
      s->r = r_new;
      s->rows = rows;
      s->cols = cols;
    }
    if(f_new!=0x0){
      assert(u_new==0x0 && r_new==0x0);
      assert(f_new->rows == rows);
      assert(f_new->cols == cols);
      s->f = f_new;
      s->rows = rows;
      s->cols = cols;
    }
  }
  s->l = 0;
  allocated_supermatrices++;
  current_supermatrices++;
  return s;
}

void
initrandom_supermatrix(psupermatrix s){
  int i,j;
  int rows,cols,k,block_rows,block_cols;
  assert(s != 0x0);
  block_rows = s->block_rows;
  block_cols = s->block_cols;
  rows = s->rows;
  cols = s->cols;
  if(s->s != 0x0){
    for(i=0; i<block_rows*block_cols; i++){
      initrandom_supermatrix(s->s[i]);
    }
  }else{
    if(s->r!=0x0){
      k = s->r->k;
      s->r->kt = k;
      for(i=0; i<k; i++){
	for(j=0; j<rows; j++){
	  s->r->a[i*rows+j] = ((double)(random()%10000))/10000.0;
	}
	for(j=0; j<cols; j++){
	  s->r->b[i*cols+j] = ((double)(random()%10000))/10000.0;
	}
      }
    }else{
      for(i=0; i<rows*cols; i++) 
	s->f->e[i] = ((double)(random()%10000))/10000.0;
    }
  }
}

void
init0_supermatrix(psupermatrix s){
  int i;
  int rows,cols,block_rows,block_cols;
  assert(s != 0x0);
  block_rows = s->block_rows;
  block_cols = s->block_cols;
  rows = s->rows;
  cols = s->cols;
  if(s->s != 0x0){
    for(i=0; i<block_rows*block_cols; i++){
      init0_supermatrix(s->s[i]);
    }
  }else{
    if(s->r!=0x0){
      s->r->kt=0;
    }else{
      for(i=0; i<rows*cols; i++) s->f->e[i] = 0.0;
    }
  }
}

size_t
getsize_supermatrix(psupermatrix s){
  size_t size = 0;
  int i;
  assert(s != 0x0);
  size += sizeof(supermatrix);
  size += sizeof(psupermatrix)*s->block_rows*s->block_cols;
  if(s->s!=0x0)
    for(i=0; i<s->block_rows*s->block_cols; i++)
      size += getsize_supermatrix(s->s[i]);
  if(s->f!=0x0) size += getsize_fullmatrix(s->f);
  if(s->r!=0x0) size += getsize_rkmatrix(s->r);
  if(s->u!=0x0) size += getsize_uniformmatrix(s->u);
  return size;
}

size_t
getsizefull_supermatrix(psupermatrix s){
  size_t size = 0;
  int i;
  assert(s != 0x0);
  if(s->s!=0x0)
    for(i=0; i<s->block_rows*s->block_cols; i++)
      size += getsizefull_supermatrix(s->s[i]);
  if(s->f!=0x0) size += getsize_fullmatrix(s->f);
  return size;
}

size_t 
getsizerk_supermatrix(psupermatrix s){
  size_t size = 0;
  int i;
  assert(s != 0x0);
  if(s->s!=0x0)
    for(i=0; i<s->block_rows*s->block_cols; i++)
      size += getsizerk_supermatrix(s->s[i]);
  if(s->r!=0x0) size += getsize_rkmatrix(s->r);
  return size;
}

psupermatrix
copy_supermatrix(psupermatrix s){
  int i;
  int block_rows,block_cols;
  psupermatrix sc;
  psupermatrix s_el;
  puniformmatrix u;
  prkmatrix r;
  pfullmatrix f;
  assert(s != 0x0);
  block_rows = s->block_rows;
  block_cols = s->block_cols;
  f = s->f;
  r = s->r;
  u = s->u;
  sc = new_supermatrix(block_rows,block_cols,s->rows,s->cols,u,r,f);
  if(s->s != 0x0){
    for(i=0; i<block_rows*block_cols; i++){
      s_el = s->s[i];
      sc->s[i] = copy_supermatrix(s_el);
    }
  }else{
    if(s->u!=0x0){
      assert(s->r==0x0 && s->f==0x0);
      sc->u = copy_uniformmatrix(u);
    }
    else if(s->r!=0x0){
      assert(s->f==0x0);
      sc->r = copy_rkmatrix(r);
    }else{
      assert(s->f!=0x0);
      sc->f = copy_fullmatrix(f);
    }
  }
  sc->rows = s->rows;
  sc->cols = s->cols;
  sc->l = s->l;
  return sc;
}

psupermatrix
extendedcopy_supermatrix(psupermatrix s, int newrank){
  int i;
  int block_rows,block_cols;
  psupermatrix sc;
  psupermatrix s_el;
  prkmatrix r;
  pfullmatrix f;
  assert(s != 0x0);
  assert(s->u == 0x0);
  block_rows = s->block_rows;
  block_cols = s->block_cols;
  f = s->f;
  r = s->r;
  sc = new_supermatrix(block_rows,block_cols,s->rows,s->cols,0x0,r,f);
  if(s->s != 0x0){
    for(i=0; i<block_rows*block_cols; i++){
      s_el = s->s[i];
      sc->s[i] = extendedcopy_supermatrix(s_el,newrank);
    }
  }else{
    if(s->r!=0x0){
      assert(s->f==0x0);
      sc->r = extendedcopy_rkmatrix(r,newrank);
    }else{
      assert(s->f!=0x0);
      sc->f = copy_fullmatrix(f);
    }
  }
  sc->rows = s->rows;
  sc->cols = s->cols;
  sc->l = s->l;
  return sc;
}

psupermatrix
copystructure_supermatrix(psupermatrix s){
  int i;
  int block_rows,block_cols;
  psupermatrix sc;
  psupermatrix s_el;
  puniformmatrix u;
  prkmatrix r;
  pfullmatrix f;
  assert(s != 0x0);
  block_rows = s->block_rows;
  block_cols = s->block_cols;
  f = s->f;
  r = s->r;
  u = s->u;
  sc = new_supermatrix(block_rows,block_cols,s->rows,s->cols,u,r,f);
  sc->row = s->row;
  sc->col = s->col;
  if(s->s != 0x0){
    for(i=0; i<block_rows*block_cols; i++){
      s_el = s->s[i];
      sc->s[i] = copystructure_supermatrix(s_el);
    }
  }else{
    if(s->u!=0x0){
      assert(s->r==0x0 && s->f==0x0);
      sc->u = copystructure_uniformmatrix(u);
    }
    else if(s->r!=0x0){
      assert(s->f==0x0);
      sc->r = copystructure_rkmatrix(r);
    }else{
      assert(s->f!=0x0);
      sc->f = copystructure_fullmatrix(f);
    }
  }
  sc->rows = s->rows;
  sc->cols = s->cols;
  sc->l = s->l;
  return sc;
}

void
reallocstructure_supermatrix(psupermatrix s, int new_rank){
  int i;
  assert(s != 0x0);
  assert(s->u == 0x0);
  if(s->f==0x0){
    if(s->r!=0x0){
      reallocstructure_rkmatrix(s->r,new_rank);
    }else{
      for(i=0; i<s->block_rows*s->block_cols; i++)
	reallocstructure_supermatrix(s->s[i],new_rank);
    }
  }
}

void
copycontent_supermatrix(psupermatrix ssource, psupermatrix starget){
  int i;
  int block_rows,block_cols;
  puniformmatrix u;
  prkmatrix r;
  pfullmatrix f;
  assert(ssource != 0x0);
  assert(starget != 0x0);
  block_rows = ssource->block_rows;
  block_cols = ssource->block_cols;
  f = ssource->f;
  r = ssource->r;
  u = ssource->u;
  if(ssource->s != 0x0){
    assert(starget->s!=0x0);
    for(i=0; i<block_rows*block_cols; i++){
      copycontent_supermatrix(ssource->s[i],starget->s[i]);
    }
  }else{
    if(ssource->u!=0x0){
      assert(starget->u!=0x0);
      copycontent_uniformmatrix(u,starget->u);
    }
    else if(ssource->r!=0x0){ 
      assert(starget->r!=0x0);
      copycontent_rkmatrix(r,starget->r);
    }else{
      assert(starget->f!=0x0);
      copycontent_fullmatrix(f,starget->f);
    }
  }
}

void
del_supermatrix(psupermatrix s){
  int i;
  assert(s != 0x0);
  if(s->u != 0x0)
    del_uniformmatrix(s->u);
  if(s->r != 0x0)
    del_rkmatrix(s->r);
  if(s->f != 0x0)
    del_fullmatrix(s->f);
  if(s->s != 0x0){
    for(i=0; i<s->block_rows * s->block_cols; i++)
      del_supermatrix(s->s[i]);
    free(s->s);
  }
  free(s);
  current_supermatrices--;
}

int
getp0_supermatrix(psupermatrix s, int thislvl){
  int i,nl;
  int maxlvl = thislvl;
  assert(s != 0x0);
  if(s->s!=0x0)
    for(i=0; i<(s->block_rows)*(s->block_cols); i++){
      nl = getp0_supermatrix(s->s[i],thislvl+1);
      if(nl>maxlvl) maxlvl = nl;
    }
  return maxlvl;
}

int
depth_supermatrix(psupermatrix s){
  assert(s != 0x0);
  return getp0_supermatrix(s,1);
}

void
seterror_supermatrix(psupermatrix s, double eps){
  int i;
  assert(s != 0x0);
  if(s->s!=0x0){
    for(i=0; i<(s->block_rows)*(s->block_cols); i++)
      seterror_supermatrix(s->s[i],eps);
  }else{
    if(rk_arithmetics_adaptive()==1){
      if(s->r!=0x0){
	s->r->local_eps=eps;
      }
    }
  }
}

static void
fastaddscaleeval_supermatrix(psupermatrix s, double alpha, double *v, double *w) {
  int i,j;
  int vindex;
  int windex;
  int block_rows,block_cols;
  psupermatrix *s_el;

  assert(s != 0x0);
  assert(v != 0x0);
  assert(w != 0x0);

  block_rows = s->block_rows;
  block_cols = s->block_cols;
  s_el = s->s;

  if(s->u == 0x0 && s->r == 0x0 && s->f == 0x0) {
    assert(s_el != 0x0);

    vindex = 0;
    for(j=0; j<block_cols; j++) {
      windex = 0;
      for(i=0; i<block_rows; i++) {
	fastaddscaleeval_supermatrix(s_el[i+j*block_rows], alpha, 
				     v+vindex, w+windex);
	windex += s_el[i+j*block_rows]->rows;
      }
      assert(windex == s->rows);
      vindex += s_el[j*block_rows]->cols;
    }
    assert(vindex == s->cols);
  }
  else {
    if(s->u != 0x0) {
      fastaddscaleeval_uniformmatrix(s->u, alpha);
    }
    else if(s->r != 0x0) {
      addscaleeval_rkmatrix(s->r, alpha, v, w);
    }
    else {
      addscaleeval_fullmatrix(s->f, alpha, v, w);
    }
  }
}

void
addeval_supermatrix(psupermatrix s, double *v, double *w) {
  assert(s != 0x0);
  assert(v != 0x0);
  assert(w != 0x0);

  if(s->row || s->col) {
    assert(s->row && s->col);

    clear_clusterbasis(s->row);
    clear_clusterbasis(s->col);

    forward_clusterbasis(s->col, v);
  }

  fastaddscaleeval_supermatrix(s, 1.0, v, w);

  if(s->row || s->col) {
    assert(s->row && s->col);

    backward_clusterbasis(s->row, w);
  }
}

void
eval_supermatrix(psupermatrix s, double *v, double *w) {
  int i, rows;

  rows = s->rows;
  for(i=0; i<rows; i++)
    w[i] = 0.0;

  addeval_supermatrix(s, v, w);
}

void
fastaddevaltrans_supermatrix(psupermatrix s, double *v, double *w) {
  int i, j;
  int vindex;
  int windex;
  int block_rows,block_cols;
  psupermatrix *s_el;

  assert(s != 0x0);
  assert(v != 0x0);
  assert(w != 0x0);

  block_rows = s->block_rows;
  block_cols = s->block_cols;
  s_el = s->s;

  if(s->u == 0x0 && s->r == 0x0 && s->f == 0x0) {
    assert(s_el != 0x0);

    vindex = 0;
    for(i=0; i<block_rows; i++){
      windex = 0;
      for(j=0; j<block_cols; j++){
	fastaddevaltrans_supermatrix(s_el[i+j*block_rows], v+vindex, w+windex);
	windex += s_el[i+j*block_rows]->cols;
      }
      assert(windex == s->cols);
      vindex += s_el[i]->rows;
    }
    assert(vindex == s->rows);
  }
  else {
    if(s->u != 0x0) {
      fastaddevaltrans_uniformmatrix(s->u);
    }
    else if(s->r != 0x0) {
      addevaltrans_rkmatrix(s->r, v, w);
    }
    else {
      addevaltrans_fullmatrix(s->f, v, w);
    }
  }
}

void
addevaltrans_supermatrix(psupermatrix s, double* v, double* w) {
  if(s->row || s->col) {
    assert(s->row && s->col);

    clear_clusterbasis(s->row);
    clear_clusterbasis(s->col);

    forward_clusterbasis(s->row, v);
  }

  fastaddevaltrans_supermatrix(s, v, w);

  if(s->row || s->col) {
    assert(s->row && s->col);

    backward_clusterbasis(s->col, w);
  }
}

void
evaltrans_supermatrix(psupermatrix s, double *v, double *w) {
  int i, cols;

  cols = s->cols;
  for(i=0; i<cols; i++)
    w[i] = 0.0;

  addevaltrans_supermatrix(s, v, w);
}

static void
lowersolverec_supermatrix(psupermatrix s, double *v) {
  psupermatrix *s_el;
  int blocks;
  int rindex, cindex;
  int r, c;

  assert(s->rows == s->cols);

  if(s->u == 0x0 && s->r == 0x0 && s->f == 0x0) {
    s_el = s->s;
    blocks = s->block_rows;

    assert(s_el != 0x0);
    assert(blocks == s->block_cols);

    if(s->row)
      clear_vector(s->row->kt, s->row->xt);

    rindex = 0;
    for(r=0; r<blocks; r++) {
      cindex = 0;
      for(c=0; c<r; c++) {
	fastaddscaleeval_supermatrix(s_el[r+c*blocks], -1.0, v+cindex, v+rindex);
	cindex += s_el[r+c*blocks]->cols;
      }
      assert(cindex == rindex);
      rindex += s_el[r]->rows;
      
      if(s->row) {
	assert(s->row->sons == s->block_rows);
	addeval_lapack(s->row->son[r]->kt, s->row->kt,
		       s->row->T[r], s->row->yt, s->row->son[r]->yt);
      }

      lowersolverec_supermatrix(s_el[r+r*blocks], v+rindex);

      if(s->row)
	addevaltrans_lapack(s->row->son[r]->kt, s->row->kt,
			    s->row->T[r], s->row->son[r]->xt, s->row->xt);
    }
    assert(rindex == s->rows);
  }
  else {
    assert(s->u == 0x0);
    assert(s->r == 0x0);

    if(s->row) {
      assert(s->row->sons == 0);
      addeval_lapack(s->rows, s->row->kt, s->row->V, s->row->yt, v);
    }
    
    lowersolve_fullmatrix(s->f, v);

    if(s->row)
      evaltrans_lapack(s->rows, s->row->kt, s->row->V, v, s->row->xt);
  }
}

void
lowersolve_supermatrix(psupermatrix s, double *v)
{
  if(s->row)
    cleary_clusterbasis(s->row);

  lowersolverec_supermatrix(s, v);
}

static void
uppersolverec_supermatrix(psupermatrix s, double *v) {
  psupermatrix *s_el;
  int blocks;
  int rindex, cindex;
  int r, c;

  assert(s->rows == s->cols);

  if(s->u == 0x0 && s->r == 0x0 && s->f == 0x0) {
    s_el = s->s;
    blocks = s->block_rows;

    assert(s_el != 0x0);
    assert(blocks == s->block_cols);

    if(s->row)
      clear_vector(s->row->kt, s->row->xt);

    rindex = s->rows;
    for(r=blocks-1; r>=0; r--) {
      cindex = rindex;
      rindex -= s_el[r]->rows;
      for(c=r+1; c<blocks; c++) {
	fastaddscaleeval_supermatrix(s_el[r+c*blocks], -1.0, v+cindex, v+rindex);
	cindex += s_el[r+c*blocks]->cols;
      }
      assert(cindex == s->cols);
      
      if(s->row) {
	assert(s->row->sons == s->block_rows);
	addeval_lapack(s->row->son[r]->kt, s->row->kt,
		       s->row->T[r], s->row->yt, s->row->son[r]->yt);
      }

      uppersolverec_supermatrix(s_el[r+r*blocks], v+rindex);

      if(s->row)
	addevaltrans_lapack(s->row->son[r]->kt, s->row->kt,
			    s->row->T[r], s->row->son[r]->xt, s->row->xt);
    }
    assert(rindex == s->rows);
  }
  else {
    assert(s->u == 0x0);
    assert(s->r == 0x0);

    if(s->row) {
      assert(s->row->sons == 0);
      addeval_lapack(s->rows, s->row->kt, s->row->V, s->row->yt, v);
    }
    
    uppersolve_fullmatrix(s->f, v);

    if(s->row)
      evaltrans_lapack(s->rows, s->row->kt, s->row->V, v, s->row->xt);
  }
}

void
uppersolve_supermatrix(psupermatrix s, double *v)
{
  if(s->row)
    cleary_clusterbasis(s->row);

  uppersolverec_supermatrix(s, v);
}

psupermatrix
read2_supermatrix(FILE* datei, int lvl){
  int type;
  int i,j; 
  int row_offset,col_offset,block_rows,block_cols,k,kt,rows,cols;
  psupermatrix s = 0x0;
  pfullmatrix f = 0x0;
  prkmatrix r = 0x0;
  char* buf = (char*) malloc(120*sizeof(char));
  assert(datei != 0x0);
  assert(buf != 0x0);
  /* reading part of the matrix */
  if(feof(datei)){printf("Unexpected end of file.\n");return 0x0;}
  fgets(buf,120,datei);
  sscanf(buf,"Type %d\n",&type);
  if(type==3){
    fgets(buf,120,datei);
    sscanf(buf,"row_offset = %d\n",&row_offset);
    fgets(buf,120,datei);
    sscanf(buf,"col_offset = %d\n",&col_offset);
    fgets(buf,120,datei);
    sscanf(buf,"rows = %d\n",&rows);
    fgets(buf,120,datei);
    sscanf(buf,"cols = %d\n",&cols);
    fgets(buf,120,datei);
    sscanf(buf,"block_rows = %d\n",&block_rows);
    fgets(buf,120,datei);
    sscanf(buf,"block_cols = %d\n",&block_cols);   
    s = new_supermatrix(block_rows,block_cols,rows,cols,0x0,0x0,0x0);
    s->rows = rows;
    s->cols = cols;
    s->l = lvl;
    lvl++;
    for(j=0; j<block_cols; j++)
      for(i=0; i<block_rows; i++)
	(s->s)[i+j*block_rows] = read2_supermatrix(datei,lvl);
  }else{
    if(type==2){
      fgets(buf,120,datei);
      sscanf(buf,"row_offset = %d\n",&row_offset);
      fgets(buf,120,datei);
      sscanf(buf,"col_offset = %d\n",&col_offset);
      fgets(buf,120,datei);
      sscanf(buf,"rows = %d\n",&rows);
      fgets(buf,120,datei);
      sscanf(buf,"cols = %d\n",&cols);
      f = new_fullmatrix(rows,cols);
      s = new_supermatrix(1,1,rows,cols,0x0,0x0,f);
      s->rows = rows;
      s->cols = cols;
      s->l = lvl;
      for(i=0; i<rows*cols; i++){
	fgets(buf,120,datei);
	sscanf(buf,"%lf\n",&(f->e[i]));
      }
    }else{
      if(type==1){
	fgets(buf,120,datei);
	sscanf(buf,"row_offset = %d\n",&row_offset);
	fgets(buf,120,datei);
	sscanf(buf,"col_offset = %d\n",&col_offset);
	fgets(buf,120,datei);
	sscanf(buf,"rows = %d\n",&rows);
	fgets(buf,120,datei);
	sscanf(buf,"cols = %d\n",&cols);
	fgets(buf,120,datei);
	sscanf(buf,"k = %d\n",&k);
	fgets(buf,120,datei);
	sscanf(buf,"kt = %d\n",&kt);
	r = new_rkmatrix(k,rows,cols);
	s = new_supermatrix(1,1,rows,cols,0x0,r,0x0);
	s->rows = rows;
	s->cols = cols;
	s->l = lvl;
	r->kt = kt;
	for(i=0; i<kt*rows; i++){
	  fgets(buf,120,datei);
	  sscanf(buf,"%lf\n",&(r->a[i]));
	}
	for(i=0; i<kt*cols; i++){
	  fgets(buf,120,datei);
	  sscanf(buf,"%lf\n",&(r->b[i]));
	}
      }else{
	printf("Unknown type of matrix: %d.\n",type);
      }
    }
  }
  free(buf);
  return s;
}

psupermatrix
read_supermatrix(char* dateiname){
  FILE* datei;
  int type;
  psupermatrix s;
  char* buf = (char*) malloc(120*sizeof(char));
  assert(buf != 0x0);
  datei = fopen(dateiname,"r");
  if(datei){
    if(feof(datei)){
      printf("Unexpected end of file (empty?).\n");
      return 0x0;
    }
    fgets(buf,120,datei); 
    type = strcmp(buf,"Beginn der Matrix\n");
    if(type!=0){
      printf("The matrix-file does not have the correct format.");
      printf("The first row is:\n%s\n",buf);
      fclose(datei);
      return 0x0;
    }
    s = read2_supermatrix(datei,0);
    if(feof(datei)){
      printf("Unexpected end of file (last row missing?).\n");
      return 0x0;
     }
    fgets(buf,120,datei);
    type = strcmp(buf,"Ende der Matrix\n");
    if(type!=0){
      printf("Reading the matrix was a failure. The last row was:\n%s\n",buf);
    }
    fclose(datei);
  }else{
    printf("Could not open file %s.\n",dateiname);
    return 0x0;
  }
  return s;
}

void
writepart_supermatrix(FILE* datei, psupermatrix s, 
		      int row_offset2, int col_offset2){
  int i,j,block_rows,block_cols,k,kt,row_offset,col_offset,rows=0,cols=0;
  assert(s->u==0x0);

  /* write part of the matrix */
  if(s->r==0x0 && s->f==0x0){
    fprintf(datei,"Type 3\n");
    fprintf(datei,"row_offset = %d\n",row_offset2);
    fprintf(datei,"col_offset = %d\n",col_offset2);
    fprintf(datei,"rows = %d\n",s->rows);
    fprintf(datei,"cols = %d\n",s->cols);
    fprintf(datei,"block_rows = %d\n",s->block_rows);
    fprintf(datei,"block_cols = %d\n",s->block_cols);
    block_rows = s->block_rows;
    block_cols = s->block_cols;
    col_offset = col_offset2;
    for(j=0; j<block_cols; j++){
      row_offset = row_offset2;
      for(i=0; i<block_rows; i++){
	writepart_supermatrix(datei,(s->s)[i+j*block_rows],
			      row_offset,col_offset);
	row_offset += s->s[i]->rows;
      }
      col_offset += s->s[j*rows]->cols;
    }
  }else{
    if(s->f != 0x0){
      fprintf(datei,"Type 2\n");
      fprintf(datei,"row_offset = %d\n",row_offset2);
      fprintf(datei,"col_offset = %d\n",col_offset2);
      fprintf(datei,"rows = %d\n",s->rows);
      fprintf(datei,"cols = %d\n",s->cols);
      rows = s->rows;
      cols = s->cols;
      for(i=0; i<rows*cols; i++){
	fprintf(datei,"%.16e\n",s->f->e[i]);
      }
    }else{
      if(s->r != 0x0){
	fprintf(datei,"Type 1\n");
	fprintf(datei,"row_offset = %d\n",row_offset2);
	fprintf(datei,"col_offset = %d\n",col_offset2);
	fprintf(datei,"rows = %d\n",s->rows);
	fprintf(datei,"cols = %d\n",s->cols);
	rows = s->rows;
	cols = s->cols;
	k = s->r->k;
	kt = s->r->kt;
	fprintf(datei,"k = %d\n",k);
	fprintf(datei,"kt = %d\n",kt);
	for(i=0; i<kt*rows; i++){
	  fprintf(datei,"%.16e\n",s->r->a[i]);
	}
	for(i=0; i<kt*cols; i++){
	  fprintf(datei,"%.16e\n",s->r->b[i]);
	}
      }
    }
  }
}

void
write_supermatrix(char* dateiname, psupermatrix s){
  FILE* datei;
  /* open file and write first row */
  datei = fopen(dateiname,"w");
  if(datei){
    fprintf(datei,"Beginn der Matrix\n");
    writepart_supermatrix(datei,s,0,0);
    fprintf(datei,"Ende der Matrix\n");
    fclose(datei);
  }else{
    printf("Unable to open file %s.\n",dateiname);
  }
}

int
getnruniformmatricesin_supermatrix(psupermatrix s){
  int nr=0;
  int i;
  if(s->s!=0x0){
    for(i=0; i<s->block_rows*s->block_cols; i++)
      nr += getnruniformmatricesin_supermatrix(s->s[i]);
  }else{
    if(s->u!=0x0)
      nr = 1;
  }
  return nr;
}

int 
getnrrkmatricesin_supermatrix(psupermatrix s){
  int nr=0;
  int i;
  if(s->s!=0x0){
    for(i=0; i<s->block_rows*s->block_cols; i++)
      nr += getnrrkmatricesin_supermatrix(s->s[i]);
  }else{
    if(s->r!=0x0)
      nr = 1;
  }
  return nr;
}

int 
getnrfullmatricesin_supermatrix(psupermatrix s){
  int nr=0;
  int i;
  if(s->s!=0x0){
    for(i=0; i<s->block_rows*s->block_cols; i++)
      nr += getnrfullmatricesin_supermatrix(s->s[i]);
  }else{
    if(s->f!=0x0)
      nr = 1;
  }
  return nr;
}

void
convertsuper2_fullmatrix(pfullmatrix f, psupermatrix s){
  double* w;
  int i,rows,cols,size;
  double* v;
  assert(f != 0x0);
  assert(s != 0x0);
  assert(s->u == 0x0);
  rows = f->rows;
  cols = f->cols;
  size = rows*cols;
  if(s->f!=0x0){
    dcopy_(&size,s->f->e,eins_,f->e,eins_);
  }else{
    v = (double*) malloc(f->cols * sizeof(double));
    for(i=0; i<cols; i++) v[i] = 0.0;
    assert(v!=0x0);
    for(i=0; i<cols; i++){
      w = &f->e[i*rows];
      v[i] = 1.0;
      eval_supermatrix(s,v,w);
      v[i] = 0.0;
    }
    free(v);
  }
}

void
convertfull2_rkmatrix(pfullmatrix f, prkmatrix r, 
		      int row_offset, int col_offset){
  assert(f != 0x0);
  assert(r != 0x0);
  r->kt = 0;
  addfull2_rkmatrix(f,r,row_offset,col_offset);
}

void
convertfull2_fullmatrix(pfullmatrix f, pfullmatrix ft, 
			int row_offset, int col_offset){
  int rows,cols,size;
  assert(f != 0x0);
  assert(ft != 0x0);
  rows = ft->rows;
  cols = ft->cols;
  assert(row_offset+rows <= f->rows && row_offset>=0);
  assert(col_offset+cols <= f->cols && col_offset>=0) ;
  if(row_offset==0 && col_offset==0 && f->rows==rows && f->cols==cols){
    size = rows*cols;
    dcopy_(&size,f->e,eins_,ft->e,eins_);
  }else{
    for(size=0; size<cols; size++)
      dcopy_(&rows,&f->e[row_offset+(size+col_offset)*f->rows],
	     eins_,&ft->e[size*rows],eins_);
  }
}

void
convertfull2_supermatrix(pfullmatrix f, psupermatrix c, 
			 int row_offset, int col_offset){
  int i,j;
  int block_rows,block_cols;
  int no,mo;
  assert(c != 0x0);
  assert(f != 0x0);
  assert(c->u==0x0);
  block_rows = c->block_rows;
  block_cols = c->block_cols;
  if(c->s!=0x0){
    no = row_offset;
    for(i=0; i<block_rows; i++){
      mo = col_offset;
      for(j=0 ;j<block_cols; j++){
	convertfull2_supermatrix(f,c->s[i+j*block_rows],no,mo);
	mo += c->s[j*block_rows]->cols; 
      }
      no += c->s[i]->rows; 
    }
  }else{
    if(c->r != 0x0){
      convertfull2_rkmatrix(f,c->r,row_offset,col_offset);
    }else{
      convertfull2_fullmatrix(f,c->f,row_offset,col_offset);
    }
  }
}

void
convertrk2_rkmatrix(prkmatrix r, prkmatrix rt, int row_offset, int col_offset){
  int i,k,rows,cols,rowst,colst;
  double *a_el,*b_el,*rta,*rtb;
  assert(r != 0x0);
  assert(rt != 0x0);
  a_el = r->a;
  b_el = r->b;
  rta = rt->a;
  rtb = rt->b;
  rows = r->rows;
  cols = r->cols;
  rowst = rt->rows;
  colst = rt->cols;
  k = r->kt;
  assert(row_offset+rowst <= rows && row_offset>=0);
  assert(col_offset+colst <= cols && col_offset>=0);
  if(rt->k >= k){
    rt->kt = k;
    for(i=0; i<k; i++)
      dcopy_(&rowst,&a_el[row_offset+i*rows],eins_,&rta[i*rowst],eins_);
    for(i=0; i<k; i++)
      dcopy_(&colst,&b_el[col_offset+i*cols],eins_,&rtb[i*colst],eins_);
  }else{
    rt->kt = 0;
    addpart2_rkmatrix(rt,0,0,r,row_offset,col_offset,rt->rows,rt->cols);
  }
}

void
convertrk2_fullmatrix(prkmatrix r, pfullmatrix ft, 
		      int row_offset, int col_offset){
  int size;
  assert(r != 0x0);
  assert(ft != 0x0);
  assert(row_offset+ft->rows <= r->rows && row_offset>=0); 
  assert(col_offset+ft->cols <= r->cols && col_offset>=0); 
  size = ft->rows * ft->cols;
  if(r->kt>0){
    dgemm_(ntrans,ttrans,&ft->rows,&ft->cols,&r->kt,deins_,
	   &r->a[row_offset],&r->rows,&r->b[col_offset],&r->cols,dnull_,
	   ft->e,&ft->rows);
  }else{
    dscal_(&size,dnull_,ft->e,eins_);
  }
}

void
convertrk2_supermatrix(prkmatrix r, psupermatrix c, 
		       int row_offset, int col_offset){
  int i,j,block_rows,block_cols;
  int no,mo;
  assert(r != 0x0);
  assert(c != 0x0);
  assert(c->u == 0x0);
  block_rows = c->block_rows;
  block_cols = c->block_cols;
  assert(row_offset+c->rows <= r->rows && row_offset>=0); 
  assert(col_offset+c->cols <= r->cols && col_offset>=0); 
  if(c->s!=0x0){
    no = row_offset;
    for(i=0; i<block_rows; i++){
      mo = col_offset;
      for(j=0 ;j<block_cols; j++){
	convertrk2_supermatrix(r,c->s[i+j*block_rows],no,mo);
	mo += c->s[j*block_rows]->cols; 
      }
      no += c->s[i]->rows; 
    }
  }else{
    if(c->r != 0x0){
      convertrk2_rkmatrix(r,c->r,row_offset,col_offset);
    }else{
      convertrk2_fullmatrix(r,c->f,row_offset,col_offset);
    }
  }
}

void
addfullpart2_supermatrix(pfullmatrix f, psupermatrix c, 
			 int row_offset, int col_offset){
  int i,j;
  int no,mo;
  assert(f != 0x0);
  assert(c != 0x0);
  assert(c->u == 0x0);
  
  if(c->s!=0x0){
    no = row_offset;
    for(i=0; i<c->block_rows; i++){
      mo = col_offset;
      for(j=0 ;j<c->block_cols; j++){
	addfullpart2_supermatrix(f,c->s[i+j*c->block_rows],no,mo);
	mo += c->s[j*c->block_rows]->cols; 
      }
      no += c->s[i]->rows; 
    }
  }else{
    assert(row_offset>=0 && row_offset+c->rows <= f->rows);
    assert(col_offset>=0 && col_offset+c->cols <= f->cols);
    if(c->r != 0x0){
      addfullpart2_rkmatrix(f,c->r,row_offset,col_offset);
    }else{
      for(i=0; i<c->f->cols; i++)
	daxpy_(&c->f->rows,deins_,&f->e[row_offset+(i+col_offset)*f->rows],
	       eins_,&c->f->e[i*c->f->rows],eins_);
    }
  }
}

void
addfull2_supermatrix(pfullmatrix f, psupermatrix c, 
		     int row_offset, int col_offset){
  assert(f != 0x0);
  assert(c->u == 0x0);
  addfullpart2_supermatrix(f,c,row_offset,col_offset);
}

void
generaladdfullpart2_supermatrix(pfullmatrix f, psupermatrix c, 
				int row_offset, int col_offset, 
				int uplow_or_all){
  int i,j;
  int no,mo;
  assert(f != 0x0);
  assert(c != 0x0);
  assert(c->u == 0x0);

  if(c->s!=0x0){
    no = row_offset;
    for(i=0; i<c->block_rows; i++){
      mo = col_offset;
      for(j=0 ;j<c->block_cols; j++){
	if((i<j && (uplow_or_all != 2)) ||
	   (i>j && (uplow_or_all != 1)) ||
	   uplow_or_all == 0){
	  generaladdfullpart2_supermatrix(f,c->s[i+j*c->block_rows],no,mo,0);
	}else{
	  if(i==j)
	    generaladdfullpart2_supermatrix(f,c->s[i+j*c->block_rows],no,mo,
				      uplow_or_all);
	}
	mo += c->s[j*c->block_rows]->cols; 
      }
      no += c->s[i]->rows; 
    }
  }else{
    assert(row_offset>=0 && row_offset+c->rows <= f->rows);
    assert(col_offset>=0 && col_offset+c->cols <= f->cols);
    if(c->r != 0x0){
      assert(uplow_or_all==0);
      addfullpart2_rkmatrix(f,c->r,row_offset,col_offset);
    }else{
      for(i=0; i<c->rows; i++){
	for(j=0; j<c->cols; j++){
	  if((i<j && uplow_or_all==1) || (i>j && uplow_or_all==2) ||
	     i==j || uplow_or_all==0)
	    c->f->e[i+j*c->rows] 
	      += f->e[i+row_offset + (j+col_offset)*f->rows]; 
	}
      }      
    }
  }
}

void
generaladdfull2_supermatrix(pfullmatrix f, psupermatrix c, 
			    int row_offset, int col_offset, int uplow_or_all){
  assert(f != 0x0);
  assert(c->u == 0x0);
  generaladdfullpart2_supermatrix(f,c,row_offset,col_offset,uplow_or_all);
}

void
addrk2_fullmatrix(prkmatrix r, pfullmatrix ft, 
		  int  row_offset, int col_offset){
  assert(r != 0x0);
  assert(ft != 0x0);
  assert(row_offset>=0 && row_offset+ft->rows <= r->rows);
  assert(col_offset>=0 && col_offset+ft->cols <= r->cols);
  if(r->kt>0){
    dgemm_(ntrans,ttrans,&ft->rows,&ft->cols,&r->kt,deins_,
	   &r->a[row_offset],&r->rows,&r->b[col_offset],&r->cols,deins_,
	   ft->e,&ft->rows);
  }
}

void
addrk2_supermatrix(prkmatrix r, psupermatrix c, int row_offset, int col_offset){
  int i,j,block_rows,block_cols;
  int no,mo;
  assert(c != 0x0);
  assert(r != 0x0);
  assert(c->u == 0x0);
  block_rows = c->block_rows;
  block_cols = c->block_cols;
  if(c->s!=0x0){
    no = row_offset;
    for(i=0; i<block_rows; i++){
      mo = col_offset;
      for(j=0 ;j<block_cols; j++){
	addrk2_supermatrix(r,c->s[i+j*block_rows],no,mo);
	mo += c->s[j*block_rows]->cols; 
      }
      no += c->s[i]->rows; 
    }
  }else{
    if(c->r != 0x0){
      addpart2_rkmatrix(c->r,0,0,r,row_offset,col_offset,c->rows,c->cols);
    }else{
      addrk2_fullmatrix(r,c->f,row_offset,col_offset);
    }
  }
}

void
generaladdrk2_fullmatrix(prkmatrix r, pfullmatrix ft, 
			 int  row_offset, int col_offset, int uplow_or_all){
  int i,j;
  assert(r != 0x0);
  assert(ft != 0x0);
  assert(row_offset>=0 && row_offset+ft->rows <= r->rows);
  assert(col_offset>=0 && col_offset+ft->cols <= r->cols);
  if(r->kt>0){
    if(uplow_or_all==0){
      dgemm_(ntrans,ttrans,&ft->rows,&ft->cols,&r->kt,deins_,
      &r->a[row_offset],&r->rows,&r->b[col_offset],&r->cols,deins_,
      ft->e,&ft->rows);
    }else{
      for(i=0; i<ft->rows; i++){
	for(j=0; j<ft->cols; j++){
	  if((i<j && uplow_or_all==1) || (i>j && uplow_or_all==2) ||
	     i==j)
	    ft->e[i+j*ft->rows] += ddot_(&r->kt,&r->a[i+row_offset],&r->rows,
					 &r->b[j+col_offset],&r->cols); 
	}
      }      
    }
  }
}

void
generaladdrk2_supermatrix(prkmatrix r, psupermatrix c, int row_offset, int col_offset, int uplow_or_all){
  int i,j,block_rows,block_cols;
  int no,mo;
  assert(c != 0x0);
  assert(r != 0x0);
  assert(c->u == 0x0);
  block_rows = c->block_rows;
  block_cols = c->block_cols;
  if(c->s!=0x0){
    no = row_offset;
    for(i=0; i<block_rows; i++){
      mo = col_offset;
      for(j=0 ;j<block_cols; j++){
	if((i<j && (uplow_or_all != 2)) ||
	   (i>j && (uplow_or_all != 1)) ||
	   uplow_or_all == 0){
	  generaladdrk2_supermatrix(r,c->s[i+j*block_rows],no,mo,0);
	}else{
	  if(i==j)
	    generaladdrk2_supermatrix(r,c->s[i+j*block_rows],no,mo,
				      uplow_or_all);
	}
	mo += c->s[j*block_rows]->cols; 
      }
      no += c->s[i]->rows; 
    }
  }else{
    if(c->r != 0x0){
      assert(uplow_or_all==0);
      addpart2_rkmatrix(c->r,0,0,r,row_offset,col_offset,c->rows,c->cols);
    }else{
      generaladdrk2_fullmatrix(r,c->f,row_offset,col_offset,uplow_or_all);
    }
  }
}

void 
add_supermatrix(psupermatrix c, psupermatrix a, psupermatrix b){
  int i,size;
  assert(c != 0x0);
  assert(a != 0x0);
  assert(b != 0x0);
  assert(c->u == 0x0);
  assert(a->u == 0x0);
  assert(b->u == 0x0);
  if(c->s!=0x0){
    /* addition of supermatrices stored in a supermatrix */
    assert(a->block_rows==c->block_rows && 
	   a->block_cols==b->block_rows &&
	   b->block_cols==c->block_cols);
    size = (c->block_rows) * (c->block_cols);
    for(i=0; i<size; i++)
      add_supermatrix(c->s[i],a->s[i],b->s[i]);
  }else{
    if(c->r!=0x0){
      /* addition of rkmatrices stored in an rkmatrix */
      add_rkmatrix(c->r,a->r,b->r);
    }else{
      /* addition of fullmatrices stored in a fullmatrix */
      add_fullmatrix(c->f,a->f,b->f);
    }  
  } 
}

void 
addto_supermatrix(psupermatrix c, psupermatrix a){
  int i,size;
  prkmatrix r;
  pfullmatrix f;
  psupermatrix s;
  assert(c != 0x0);
  assert(a != 0x0);
  assert(c->u == 0x0);
  assert(a->u == 0x0);
  assert(c->rows == a->rows);
  assert(c->cols == a->cols);
  if(a->s!=0x0){
    /* addition of supermatrices stored in a supermatrix */
    if(c->s!=0x0){
      assert(c->block_rows==a->block_rows);
      assert(c->block_cols==a->block_cols);
      size = (c->block_rows) * (c->block_cols);
      for(i=0; i<size; i++)
	addto_supermatrix(c->s[i],a->s[i]);
    }else{
      if(c->r!=0x0){
	/* supermatrix to rkmatrix */
	r = new_rkmatrix(c->r->kt,c->rows,c->cols);
	convertsuper2_rkmatrix(a, r);
	addpart2_rkmatrix(c->r, 0, 0, r, 0, 0,c->rows, c->cols);
	del_rkmatrix(r);
      }else{
	/* supermatrix to fullmatrix */
	f = new_fullmatrix(c->rows,c->cols);
	convertsuper2_fullmatrix(f,a);
	addfull2_fullmatrix(c->f,f);
	del_fullmatrix(f);
      }
    }
  }else{
    if(a->r!=0x0){
      /* rkmatrix addition to supermatrix */
      addrk2_supermatrix(a->r,c,0,0);
    }else{
      if(c->f != 0x0){
	addfull2_fullmatrix(c->f,a->f);
      }else{
	s = copy_supermatrix(c);
	convertfull2_supermatrix(a->f, s, 0, 0);
	addto_supermatrix(c,s);
	del_supermatrix(s);
      }
    }  
  } 
}

void
muladd_supermatrix(psupermatrix c, psupermatrix a, psupermatrix b){
  int i,j,block_rows,block_cols,k,block_size,rows,cols,size;
  psupermatrix c_el;
  psupermatrix a_el;
  psupermatrix b_el;
  prkmatrix r,rtmp;
  psupermatrix stmp;
  pfullmatrix f;
  assert(c != 0x0);
  assert(a != 0x0);
  assert(b != 0x0);
  block_rows = c->block_rows;
  block_cols = c->block_cols;
  block_size = a->block_cols;
  assert(a->cols==b->rows);
  assert(a->rows==c->rows);
  assert(b->cols==c->cols);
  assert(a->u==0x0);
  assert(b->u==0x0);
  assert(c->u==0x0);

  if(c->s!=0x0){
    if(a->s!=0x0 && b->s!=0x0){
      /* only supermatrices -> recursion */ 
      for(i=0; i<block_rows; i++)
	for(j=0; j<block_cols; j++){
	  c_el = c->s[i+j*block_rows];
	  a_el = a->s[i];
	  b_el = b->s[j*block_size];
	  muladd_supermatrix(c_el,a_el,b_el);
	  for(k=1; k<block_size; k++){
	    a_el = a->s[i+k*block_rows];
	    b_el = b->s[k+j*block_size];
	    muladd_supermatrix(c_el,a_el,b_el);
	  }
	}
    }else{
      /* a or b is rk or fullmatrix */
      if(a->r!=0x0 || b->r!=0x0){
	/* rkmul */
	r = getmul_rkmatrix(a,b);
	addrk2_supermatrix(r,c,0,0);
	del_rkmatrix(r);
      }else{
	/* fullmul */
	rows = c->rows;
	cols = c->cols;
	size = a->cols;      
	if(size<rows && size<cols){
	  if(rows<cols || (b->f==0x0 && a->f!=0x0)){
	    rtmp = new_rkmatrix(size,rows,size);
	    stmp = new_supermatrix(1,1,rows,size,0x0,rtmp,0x0);
	    addfull2_rkmatrix(a->f,rtmp,0,0);
	    r = getmul_rkmatrix(stmp,b);
	    del_supermatrix(stmp);
	    addrk2_supermatrix(r,c,0,0);
	    del_rkmatrix(r);
	  }else{
	    rtmp = new_rkmatrix(size,size,cols);
	    stmp = new_supermatrix(1,1,size,cols,0x0,rtmp,0x0);
	    addfull2_rkmatrix(b->f,rtmp,0,0);
	    r = getmul_rkmatrix(a,stmp);
	    del_supermatrix(stmp);
	    addrk2_supermatrix(r,c,0,0);
	    del_rkmatrix(r);
	  }
	}else{
	  f = getmul_fullmatrix(a,b);
	  addfull2_supermatrix(f,c,0,0);
	  del_fullmatrix(f);
	}
      }
    }
  }else{
    if(c->r!=0x0){
      /* a or b is rk or fullmatrix */
      if(a->r!=0x0 || b->r!=0x0){
	/* rkmul */
	r = getmul_rkmatrix(a,b);
	addpart2_rkmatrix(c->r,0,0,r,0,0,r->rows,r->cols);
	del_rkmatrix(r);
      }else{
	/* product of 2 supermatrices to be stored in a rkmatrix*/
	addprod2_rkmatrix(c->r,a,b,0,0); 
      }
    }else{
      /* a or b is rk or fullmatrix */
      if(a->r!=0x0 || b->r!=0x0){
	/* rkmul */
	r = getmul_rkmatrix(a,b);
	addrk2_fullmatrix(r,c->f,0,0);
	del_rkmatrix(r);
      }else{
	/* fullmul */
	if(a->f!=0x0 && b->f!=0x0){
	  dgemm_(ntrans,ntrans,&c->rows,&c->cols,&a->cols,deins_,
		 a->f->e,&a->rows,b->f->e,&b->rows,deins_,c->f->e,&c->rows);
	}else{
	  f = getmul_fullmatrix(a,b);
	  addfull2_fullmatrix(c->f,f);
	  del_fullmatrix(f);
	}
      }
    }  
  } 
}

void
mul_supermatrix(psupermatrix c, psupermatrix a, psupermatrix b){
  assert(c != 0x0);
  assert(a != 0x0);
  assert(b != 0x0);
  init0_supermatrix(c);
  muladd_supermatrix(c,a,b);
}

int 
getnrnodes_supermatrix(psupermatrix c){
  int block_size,i,nr_total=0;
  assert(c != 0x0);
  block_size = c->block_rows * c->block_cols;
  if(c->s == 0x0){
    nr_total = 1;
  }else{
    for(i=0; i<block_size; i++)
      nr_total += getnrnodes_supermatrix(c->s[i]);
  }
  return nr_total;
}

void
copydata_supermatrix(psupermatrix s, psupermatrix sc){
  int i,block_rows,block_cols;
  psupermatrix s_el;
  prkmatrix r;
  pfullmatrix f;
  assert(s != 0x0);
  assert(sc != 0x0);
  assert(s->u == 0x0);
  assert(sc->u == 0x0);
  block_rows = s->block_rows;
  block_cols = s->block_cols;
  f = s->f;
  r = s->r;
  if(s->s != 0x0){
    assert(sc->s != 0x0 && sc->block_rows==block_rows 
	   && sc->block_cols==block_cols);
    for(i=0; i<block_rows*block_cols; i++){
      s_el = s->s[i];
      copydata_supermatrix(s_el,sc->s[i]);
    }
  }else{
    if(s->r!=0x0){
      assert(sc->r != 0x0); 
      convertrk2_rkmatrix(r,sc->r,0,0);
    }else{
      assert(sc->f != 0x0); 
      convertfull2_fullmatrix(f,sc->f,0,0);
    }
  }
}

void
scale_supermatrix(psupermatrix s, double value){
  int i,block_rows,block_cols,nk;
  psupermatrix s_el;
  prkmatrix r;
  pfullmatrix f;
  assert(s != 0x0);
  assert(s->u == 0x0);
  block_rows = s->block_rows;
  block_cols = s->block_cols;
  f = s->f;
  r = s->r;
  if(s->s != 0x0){
    for(i=0; i<block_rows*block_cols; i++){
      s_el = s->s[i];
      scale_supermatrix(s_el,value);
    }
  }else{
    if(r!=0x0){
      if(r->kt>0){
	nk = r->rows * r->kt; 
	dscal_(&nk,&value,r->a,eins_);
      }
    }else{
      scale_fullmatrix(f,value);
    }
  }
}

void
subid_supermatrix(psupermatrix s){
  subcid_supermatrix(s,1.0);
}

void
subcid_supermatrix(psupermatrix s, double c){
  int i,block_rows,block_cols,rows,cols;
  psupermatrix s_el;
  prkmatrix r;
  pfullmatrix f;
  assert(s != 0x0);
  assert(s->u == 0x0);
  block_rows = s->block_rows;
  block_cols = s->block_cols;
  f = s->f;
  r = s->r;
  if(s->s != 0x0){
    for(i=0; i<block_rows && i<block_cols; i++){
      s_el = s->s[i+block_rows*i];
      subcid_supermatrix(s_el,c);
    }
  }else{
    if(r!=0x0){
      rows = r->rows;
      cols = r->cols;
      f = new_fullmatrix(rows,cols);
      for(i=0; i<rows && i<cols; i++) f->e[i+i*rows] = -c;
      addfull2_rkmatrix(f,r,0,0);
      del_fullmatrix(f);
    }else{
      rows = f->rows;
      cols = f->cols;
      for(i=0; i<rows && i<cols; i++) f->e[i+i*rows] -= c;
    }
  }
}

double 
getentry_supermatrix(psupermatrix s, int row, int col){
  double entry = 0.0;
  int br,bc;
  int ro, co;

  assert(s!=0x0);
  assert(row>=0 && row<s->rows && col>=0 && col<s->cols);
  
  if(s->s != 0x0){
    ro = 0;
    for(br=0; br<s->block_rows; br++){
      co = 0;
      for(bc=0; bc<s->block_cols; bc++){
	if(row-ro>=0 && row-ro<s->s[br + bc*s->block_rows]->rows &&
	   col-co>=0 && col-co<s->s[br + bc*s->block_rows]->cols){
	  entry = getentry_supermatrix(s->s[br + bc*s->block_rows],
				       row-ro,col-co);
	}
	co += s->s[br + bc*s->block_rows]->cols;
      }
      ro += s->s[br]->rows;
    }
  }else{
    if(s->f != 0x0){
      entry = getentry_fullmatrix(s->f,row,col);
    }else{
      if(s->r != 0x0){
	entry = getentry_rkmatrix(s->r,row,col);
      }else{
	entry = getentry_uniformmatrix(s->u,row,col);
      }
    }
  }
  return entry;
}

void 
setentry_supermatrix(psupermatrix s, int row, int col, double value){
  int br,bc;
  int ro, co;

  assert(s!=0x0);
  assert(row>=0 && row<s->rows && col>=0 && col<s->cols);
  
  if(s->s != 0x0){
    ro = 0;
    for(br=0; br<s->block_rows; br++){
      co = 0;
      for(bc=0; bc<s->block_cols; bc++){
	if(row-ro>=0 && row-ro<s->s[br + bc*s->block_rows]->rows &&
	   col-co>=0 && col-co<s->s[br + bc*s->block_rows]->cols){
	  setentry_supermatrix(s->s[br + bc*s->block_rows],
			       row-ro,col-co,value);
	}
	co += s->s[br + bc*s->block_rows]->cols;
      }
      ro += s->s[br]->rows;
    }
  }else{
    if(s->f != 0x0){
      setentry_fullmatrix(s->f,row,col,value);
    }else{
      if(s->r != 0x0){
	setentry_rkmatrix(s->r,row,col,value);
      }else{
	printf("setentry_uniformmatrix not implemented\n");
      }
    }
  }
}

double
trace_supermatrix(psupermatrix s){
  int row;
  double trace = 0.0;
  for(row=0; row<s->rows && row<s->cols; row++){
    trace += getentry_supermatrix(s,row,row);
  }
  return trace;
}

double 
normfrobenius_supermatrix(psupermatrix s){
  double normf = 0.0;
  double tmp = 0.0;
  int i,size;
  assert(s != 0x0);
  assert(s->u == 0x0);
  if(s->s!=0x0){
    for(i=0; i<((s->block_rows)*(s->block_cols));i++){
      tmp = normfrobenius_supermatrix(s->s[i]);
      normf += tmp*tmp;
    }
    normf = sqrt(fabs(normf));
  }else{
    if(s->r!=0x0){
      normf = normfrobenius_rkmatrix(s->r);
    }else{
      size = s->f->rows * s->f->cols;
      for(i=0; i<size; i++) normf += (s->f->e[i])*(s->f->e[i]);
      normf = sqrt(fabs(normf));
    }
  }
  return normf;
}

double 
norm2_supermatrix(psupermatrix s){
  return norm2nr_supermatrix(s,10);
}

double 
norm2nr_supermatrix(psupermatrix s, int nr){
  double norm2 = 0.0;
  double tmp = 0.0;
  double *v, *w, *v2;
  int i,rows,cols;
  assert(s != 0x0);
  assert(s->u == 0x0);
  rows = s->rows;
  cols = s->cols;
  v = (double*) malloc(cols*sizeof(double));
  w = (double*) malloc(rows*sizeof(double));
  v2 = (double*) malloc(cols*sizeof(double));
  assert(v!=0x0 && w!=0x0 && v2!=0x0);
  for(i=0; i<cols; i++){
    v[i] = rand();
  }
  tmp = 1.0/dnrm2_(&cols,v,eins_);
  dscal_(&cols,&tmp,v,eins_);
  for(i=0; i<nr; i++){
    eval_supermatrix(s,v,w);
    evaltrans_supermatrix(s,w,v2);
    norm2 = dnrm2_(&cols,v2,eins_);
    if(tmp>1e-32){
      tmp=1.0/norm2;
      dscal_(&cols,&tmp,v2,eins_);
      dcopy_(&cols,v2,eins_,v,eins_);
    }else{
      i=nr;
    }
  }
  free(v);
  free(v2);
  free(w);
  return sqrt(norm2);
}

double 
norm2prodminusid_supermatrix(psupermatrix si, psupermatrix s){
  return norm2nrprodminusid_supermatrix(si,s,10);
}

double 
norm2nrprodminusid_supermatrix(psupermatrix si, psupermatrix s, 
			       int nr){
  double norm2 = 0.0;
  double tmp = 0.0;
  double *v, *w, *v2, *v3;
  int i,rows,cols;
  assert(s != 0x0);
  assert(si != 0x0);
  assert(s->u == 0x0);
  assert(si->u == 0x0);
  rows = si->rows;
  cols = si->cols;
  assert(s->rows == cols);
  assert(s->cols == rows);

  v = (double*) malloc(rows*sizeof(double));
  w = (double*) malloc(cols*sizeof(double));
  v2 = (double*) malloc(rows*sizeof(double));
  v3 = (double*) malloc(rows*sizeof(double));
  assert(v!=0x0 && w!=0x0 && v2!=0x0 && v3!=0x0);
  for(i=0; i<rows; i++){
    v[i] = rand();
  }
  tmp = 1.0/dnrm2_(&rows,v,eins_);
  dscal_(&rows,&tmp,v,eins_);
  for(i=0; i<nr; i++){
    evaltrans_supermatrix(si,v,w);
    evaltrans_supermatrix(s,w,v3);

    dcopy_(&rows,v,eins_,v2,eins_);
    daxpy_(&rows,meins_,v3,eins_,v2,eins_);

    daxpy_(&rows,meins_,v,eins_,v3,eins_);

    eval_supermatrix(s,v3,w);
    eval_supermatrix(si,w,v3);
    daxpy_(&rows,deins_,v3,eins_,v2,eins_);
 
    norm2 = sqrt(fabs(ddot_(&cols,v2,eins_,v,eins_)));
    tmp = dnrm2_(&cols,v2,eins_);
    if(tmp>1e-32){
      tmp = 1.0/tmp;
      dscal_(&cols,&tmp,v2,eins_);
      dcopy_(&cols,v2,eins_,v,eins_);
    }else{
      i=nr;
    }
  }
  free(v);
  free(v2);
  free(w);
  free(v3);
  return norm2;
}

double 
norm2nrprodminusid2_supermatrix(psupermatrix si, psparsematrix s, 
			       int nr){
  double norm2 = 0.0;
  double tmp = 0.0;
  double *v, *w, *v2, *v3;
  int i,rows,cols;
  assert(s != 0x0);
  assert(si != 0x0);
  assert(si->u == 0x0);
  rows = si->rows;
  cols = si->cols;
  assert(s->rows == cols);
  assert(s->cols == rows);

  v = (double*) malloc(rows*sizeof(double));
  w = (double*) malloc(cols*sizeof(double));
  v2 = (double*) malloc(rows*sizeof(double));
  v3 = (double*) malloc(rows*sizeof(double));
  assert(v!=0x0 && w!=0x0 && v2!=0x0 && v3!=0x0);
  for(i=0; i<rows; i++){
    v[i] = rand();
  }
  tmp = 1.0/dnrm2_(&rows,v,eins_);
  dscal_(&rows,&tmp,v,eins_);
  for(i=0; i<nr; i++){
    evaltrans_supermatrix(si,v,w);
    evaltrans_sparsematrix(s,w,v3);

    dcopy_(&rows,v,eins_,v2,eins_);
    daxpy_(&rows,meins_,v3,eins_,v2,eins_);

    daxpy_(&rows,meins_,v,eins_,v3,eins_);

    eval_sparsematrix(s,v3,w);
    eval_supermatrix(si,w,v3);
    daxpy_(&rows,deins_,v3,eins_,v2,eins_);
 
    norm2 = sqrt(fabs(ddot_(&cols,v2,eins_,v,eins_)));
    tmp = dnrm2_(&cols,v2,eins_);
    if(tmp>1e-32){
      tmp = 1.0/tmp;
      dscal_(&cols,&tmp,v2,eins_);
      dcopy_(&cols,v2,eins_,v,eins_);
    }else{
      i=nr;
    }
  }
  free(v);
  free(v2);
  free(w);
  free(v3);
  return norm2;
}

double 
norm2prodminusid2_supermatrix(psupermatrix si, psparsematrix s){
  return norm2nrprodminusid2_supermatrix(si,s,10);
}

double 
norm2diff_supermatrix(psupermatrix s1, psupermatrix s2){
  return norm2nrdiff_supermatrix(s1,s2,10);
}

double 
norm2nrdiff_supermatrix(psupermatrix s1, psupermatrix s2, int nr){
  double norm2 = 0.0;
  double tmp = 0.0;
  double *v, *w, *v2;
  int i,rows,cols;
  assert(s1 != 0x0);
  assert(s2 != 0x0);
  assert(s1->u == 0x0);
  assert(s2->u == 0x0);
  rows = s1->rows;
  cols = s1->cols;
  v = (double*) malloc(cols*sizeof(double));
  w = (double*) malloc(rows*sizeof(double));
  v2 = (double*) malloc(cols*sizeof(double));
  assert(v!=0x0 && w!=0x0 && v2!=0x0);
  for(i=0; i<cols; i++){
    v[i] = rand();
  }
  tmp = 1.0/dnrm2_(&cols,v,eins_);
  dscal_(&cols,&tmp,v,eins_);
  for(i=0; i<nr; i++){
    eval_supermatrix(s1,v,w);
    dscal_(&rows,meins_,w,eins_);
    addeval_supermatrix(s2,v,w);
    evaltrans_supermatrix(s1,w,v2);
    dscal_(&cols,meins_,v2,eins_);
    addevaltrans_supermatrix(s2,w,v2);

    norm2 = dnrm2_(&cols, v2, eins_);
    if(norm2>1e-32){
      tmp = 1.0/norm2;
      dscal_(&cols,&tmp,v2,eins_);
      dcopy_(&cols,v2,eins_,v,eins_);
    }else{
      i=nr;
    }
  }
  free(v);
  free(v2);
  free(w);
  return sqrt(norm2);
}

double 
norm2ludiff_supermatrix(psupermatrix lu, psupermatrix s){
  return norm2nrludiff_supermatrix(lu,s,10);
}

double 
norm2nrludiff_supermatrix(psupermatrix lu, psupermatrix s, int nr){
  double norm2 = 0.0;
  double tmp = 0.0;
  double *v, *w, *v2;
  int i,rows,cols;
  assert(lu != 0x0);
  assert(s != 0x0);
  assert(lu->u == 0x0);
  assert(s->u == 0x0);
  rows = lu->rows;
  cols = lu->cols;
  /*
  printf(" rows = %d, cols = %d\n",rows,cols);
  */
  v = (double*) malloc(cols*sizeof(double));
  w = (double*) malloc(rows*sizeof(double));
  v2 = (double*) malloc(cols*sizeof(double));
  assert(v!=0x0 && w!=0x0 && v2!=0x0);
  for(i=0; i<cols; i++){
    v[i] = rand();
    w[i] = 0.0;
    v2[i] = 0.0;
  }
  tmp = 1.0/dnrm2_(&cols,v,eins_);
  dscal_(&cols,&tmp,v,eins_);
  for(i=0; i<nr; i++){
    generaleval_supermatrix(lu,2,0,0,0,1.0,v,v2);
    generaleval_supermatrix(lu,1,0,0,1,1.0,v2,w);

    dscal_(&rows,meins_,w,eins_);
    addeval_supermatrix(s,v,w);

    generaleval_supermatrix(lu,1,1,0,1,1.0,w,v);
    generaleval_supermatrix(lu,2,1,0,0,1.0,v,v2);

    dscal_(&cols,meins_,v2,eins_);
    addevaltrans_supermatrix(s,w,v2);

    norm2 = dnrm2_(&cols, v2, eins_);
    if(norm2>1e-32){
      tmp = 1.0/norm2;
      dscal_(&cols,&tmp,v2,eins_);
      dcopy_(&cols,v2,eins_,v,eins_);
    }else{
      i=nr;
    }
  }
  free(v);
  free(v2);
  free(w);
  return sqrt(norm2);
}

double 
norm2inverse_supermatrix(psupermatrix s){
  double norm2 = 0.0;
  double tmp = 0.0;
  double *v, *w, *v2, normold,theta,rho;
  int i,rows;
  assert(s != 0x0);
  assert(s->u == 0x0);
  rows = s->rows;
  assert(rows == s->cols);
  v = (double*) malloc(rows*sizeof(double));
  w = (double*) malloc(rows*sizeof(double));
  v2 = (double*) malloc(rows*sizeof(double));
  assert(v!=0x0 && w!=0x0 && v2!=0x0);
  for(i=0; i<rows; i++){
    v[i] = rand();
  }
  tmp = 1.0/dnrm2_(&rows,v,eins_);
  theta = norm2_supermatrix(s);
  theta = -1.0/(theta*theta);
  dscal_(&rows,&tmp,v,eins_);
  normold = 1.0;
  for(i=0; i<33; i++){
    eval_supermatrix(s,v,w);
    evaltrans_supermatrix(s,w,v2);
    dscal_(&rows,&theta,v2,eins_);
    daxpy_(&rows,deins_,v,eins_,v2,eins_);

    norm2 = ddot_(&rows,v2,eins_,v,eins_);
    tmp = dnrm2_(&rows,v2,eins_);
    if(tmp>1e-32){
      tmp = 1.0/tmp;
      dscal_(&rows,&tmp,v2,eins_);
      dcopy_(&rows,v2,eins_,v,eins_);
    }else{
      i=33;
    }
  }
  free(v);
  free(w);
  free(v2);
  rho = norm2;
  return sqrt(fabs(1.0/(1-norm2)*theta));
}

double 
norm2luinverse_supermatrix(psupermatrix l){
  double norm2 = 0.0;
  double tmp = 0.0;
  double *v, *w, *v2;
  int i,rows,cols;
  assert(l != 0x0);
  assert(l->u == 0x0);
  rows = l->rows;
  cols = l->cols;
  v = (double*) malloc(cols*sizeof(double));
  w = (double*) malloc(rows*sizeof(double));
  v2 = (double*) malloc(cols*sizeof(double));
  assert(v!=0x0 && w!=0x0 && v2!=0x0);
  for(i=0; i<cols; i++){
    v[i] = rand();
  }
  tmp = 1.0/dnrm2_(&cols,v,eins_);
  dscal_(&cols,&tmp,v,eins_);
  for(i=0; i<10; i++){
    
    dcopy_(&rows,v,eins_,v2,eins_);
    solvelower_supermatrix(l,v2);
    solveupper_supermatrix(l,v2);
    solveuppertrans_supermatrix(l,v2);
    solvelowertrans_supermatrix(l,v2);

    norm2 = dnrm2_(&cols,v2,eins_);
    if(tmp>1e-32){
      tmp=1.0/norm2;
      dscal_(&cols,&tmp,v2,eins_);
      dcopy_(&cols,v2,eins_,v,eins_);
    }else{
      i=100;
    }
  }
  free(v);
  free(v2);
  free(w);
  return sqrt(norm2);
}

double 
norm2choleskyinverse_supermatrix(psupermatrix l){
  double norm2 = 0.0;
  double tmp = 0.0;
  double *v, *w, *v2;
  int i,rows,cols;
  assert(l != 0x0);
  assert(l->u == 0x0);
  rows = l->rows;
  cols = l->cols;
  v = (double*) malloc(cols*sizeof(double));
  w = (double*) malloc(rows*sizeof(double));
  v2 = (double*) malloc(cols*sizeof(double));
  assert(v!=0x0 && w!=0x0 && v2!=0x0);
  for(i=0; i<cols; i++){
    v[i] = rand();
  }
  tmp = 1.0/dnrm2_(&cols,v,eins_);
  dscal_(&cols,&tmp,v,eins_);
  dcopy_(&rows,v,eins_,v2,eins_);
  for(i=0; i<10; i++){
    
    solvecholeskyvector_supermatrix(l,v2);
    solvecholeskyvectortrans_supermatrix(l,v2);

    norm2 = dnrm2_(&cols,v2,eins_);

    if(norm2>1e-32){
      tmp=1.0/norm2;
      dscal_(&cols,&tmp,v2,eins_);
      dcopy_(&cols,v2,eins_,v,eins_);
    }else{
      i=10;
    }
  }
  free(v);
  free(v2);
  free(w);
  return norm2;
}

void
invert_supermatrix(psupermatrix si, psupermatrix s, psupermatrix swork){
  int block_rows,block_cols,i,j,l;
  psupermatrix *sw_e, *s_e, *si_e;
  assert(si!=0x0);
  assert(si->u == 0x0);
  assert(s!=0x0);
  assert(s->u == 0x0);
  assert(swork!=0x0);
  assert(swork->u == 0x0);
  sw_e = swork->s;
  s_e = s->s;
  si_e = si->s;
  block_rows = s->block_rows;
  block_cols = s->block_cols;
  if(si->s==0x0 || s->s==0x0 || swork->s==0x0){
    assert(si->f!=0x0 && s->f!=0x0 && swork->f!=0x0);
    /* inversion in the space of fullmatrices */
    invert_fullmatrix(si->f,s->f,swork->f);
  }else{
    assert(s->block_rows  == si->block_rows    && 
	   si->block_rows == swork->block_rows && 
	   si->block_cols == s->block_cols     && 
	   s->block_cols  == swork->block_cols);
    /* lower triangular */
    for(l=0; l<block_rows; l++){
      /* invert diagonal */
      invert_supermatrix(si_e[l+block_rows*l],s_e[l+block_rows*l],
			 sw_e[l+block_rows*l]);
      /* scale inverse with inverted diagonal */
      for(j=0; j<l; j++){
	mul_supermatrix(sw_e[l+block_rows*j],
			si_e[l+block_rows*l],si_e[l+block_rows*j]);
	copydata_supermatrix(sw_e[l+block_rows*j],si_e[l+block_rows*j]);
      }
      /* scale matrix with inverted diagonal */
      for(j=l+1; j<block_cols; j++){
	mul_supermatrix(sw_e[l+block_rows*j],
			si_e[l+block_rows*l],s_e[l+block_rows*j]);
	copydata_supermatrix(sw_e[l+block_rows*j],s_e[l+block_rows*j]);
      }
      /* eliminate lower part */
      for(i=l+1; i<block_rows; i++){
	/* subtract from inverse */
	for(j=0; j<=l; j++){
	  mul_supermatrix(sw_e[i+block_rows*j],
			  s_e[i+block_rows*l],si_e[l+block_rows*j]);
	  scale_supermatrix(sw_e[i+block_rows*j],-1.0);
	  addto_supermatrix(si_e[i+block_rows*j],sw_e[i+block_rows*j]);
	}
	/* subtract from matrix */
	for(j=l+1; j<block_cols; j++){
	  mul_supermatrix(sw_e[i+block_rows*j],
			  s_e[i+block_rows*l],s_e[l+block_rows*j]);
	  scale_supermatrix(sw_e[i+block_rows*j],-1.0);
	  addto_supermatrix(s_e[i+block_rows*j],sw_e[i+block_rows*j]);
	}
      }
    }
    /* upper triangular */
    for(l=block_rows-1; l>=0; l--){
      for(i=l-1; i>=0; i--){
	for(j=0; j<block_cols; j++){
	  mul_supermatrix(sw_e[i+block_rows*j],
			  s_e[i+block_rows*l],si_e[l+block_rows*j]);
	  scale_supermatrix(sw_e[i+block_rows*j],-1.0);
	  addto_supermatrix(si_e[i+block_rows*j],sw_e[i+block_rows*j]);
	}
      }
    }
  }
}

void
outputleaf_supermatrix(psupermatrix s, FILE* datei, pblockcluster bc, 
		       int row_offset, int col_offset,int row_max, int col_max, 
		       int rank, int ufill, int rfill, int ffill, 
		       int border, double tol){
  int i,j,no,mo,block_rows,block_cols;
  pblockcluster bc2;

  if(s->s!=0x0){
  block_rows = s->block_rows;
  block_cols = s->block_cols;
    no = row_offset;
    for(i=0; i<block_rows; i++){
      mo = col_offset;
      for(j=0; j<block_cols; j++){
	bc2 = 0x0;
	if(bc!=0x0) bc2 = bc->son[i+j*block_rows];
	outputleaf_supermatrix(s->s[i+j*block_rows],datei,bc2,no,mo,
			       row_max,col_max,
			       rank,ufill,rfill,ffill,border,tol);
	mo += s->s[j*block_rows]->cols; 
      }
      no += s->s[i]->rows;
    }
  }else{
    if(s->u!=0x0){
      outputleaf_uniformmatrix(s->u, datei, bc, row_offset, col_offset,
			       row_max, col_max, rank, ufill, border, tol);
    }
    else if(s->r!=0x0){
      outputleaf_rkmatrix(s->r, datei, bc, row_offset, col_offset,
			  row_max, col_max, rank, rfill, border, tol);
    }else{
      outputleaf_fullmatrix(s->f, datei, bc, row_offset, col_offset, 
			    row_max, col_max, rank, ffill, border, tol);
    }
  }
}

void
output_supermatrix(psupermatrix s, char* dateiname, pblockcluster bc, 
		   int rank, int ufill, int rfill, int ffill, int border, 
		   double tol, int extraoffset){
  double scale = s->rows;
  FILE* datei;
  if(s->cols>scale) scale = s->cols;
  datei = fopen(dateiname,"w");
  assert(datei!=0);
  scale = 500.0/scale;
  fprintf(datei,"%%!PS-Adobe-2.0-2.0 EPSF-2.0\n");
  fprintf(datei,"%%%%BoundingBox: 0 0 %f %f\n",
	  (s->cols)*scale+20+extraoffset,(s->rows)*scale+20+extraoffset);
  fprintf(datei,"%f dup translate\n",10.0+extraoffset);
  fprintf(datei,"%f dup scale\n",scale);
  if(scale<5){
    fprintf(datei,"%f setlinewidth\n",0.02*scale);
  }else{
    fprintf(datei,"%f setlinewidth\n",0.1);
  }
  fprintf(datei,"/Helvetica findfont %f scalefont setfont\n",12.0/scale);
  fprintf(datei,"0 setgray\n");
  fprintf(datei,"/bl{0 setgray} def\n");
  fprintf(datei,"/fb{0.95 setgray fill} def\n");
  fprintf(datei,"/fr{1.0 0.6 0.6  setrgbcolor fill} def\n");
  fprintf(datei,"/fg{0.5 setgray fill} def\n");
  fprintf(datei,"/fl{0.6 1.0 0.6  setrgbcolor fill} def\n");
  fprintf(datei,"/fi{0.8 0.0 0.0  setrgbcolor fill} def\n");
  fprintf(datei,"/fu{0.3 0.3 1.0  setrgbcolor fill} def\n");
  fprintf(datei,"/cs{closepath stroke} def\n");
  fprintf(datei,"/m{moveto} def\n");
  fprintf(datei,"/l{lineto} def\n");
  fprintf(datei,"/hf{/Helvetica findfont} def\n");
  fprintf(datei,"/sf{scalefont setfont} def\n");

  outputleaf_supermatrix(s,datei,bc,0,0,s->rows,s->cols,
			  rank,ufill,rfill,ffill,border,tol);
  fprintf(datei,"showpage\n");
  fclose(datei);
}

void
outputrank_supermatrix(psupermatrix s, char* dateiname,  
		       int rank, int extraoffset){
  output_supermatrix(s, dateiname, 0x0, rank, 0, 0, 0, 1, 0,extraoffset);
}

void
outputsvd_supermatrix(psupermatrix s, char* dateiname, int extraoffset){
  output_supermatrix(s, dateiname, 0x0, 1, 0, 2, 1, 1, 0,extraoffset);
}

void
determineworstblock_supermatrix(psupermatrix s, double* maxerror, 
				psupermatrix* rs, int k){
  int i;
  double error;
  if(s->s!=0x0){
    for(i=0; i<s->block_rows*s->block_cols; i++)
      determineworstblock_supermatrix(s->s[i],maxerror,rs,k);
  }else{
    if(s->r!=0x0){
      if(s->r->kt>k){
	error = getsv_rkmatrix(s->r,k);
	if(error>=*maxerror){
	  *maxerror =error;
	  *rs = s;
	}
      }
    }
  }
  
}

void
outputworstblock_supermatrix(psupermatrix s, char* dateiname, 
			     int k, double tolerance){
  FILE* datei;
  psupermatrix* r;
  double* maxerror;
  assert(s != 0x0);
  r = (psupermatrix*) malloc(sizeof(psupermatrix));
  maxerror = (double*) malloc(sizeof(double));
  assert(r!=0x0 && maxerror!=0x0);
  datei = fopen(dateiname,"w");
  if(datei){
    determineworstblock_supermatrix(s,maxerror,r,k);
    printf("worstblock-error = %g\n",*maxerror);
    fprintf(datei,"%%!PS-Adobe-2.0-2.0 EPSF-2.0\n");
    fprintf(datei,"%%%%BoundingBox: 0 0 %f %f\n",580.0,520.0);
    fprintf(datei,"%f setlinewidth\n",0.1);
    fprintf(datei,"%f dup translate\n",10.0);
    fprintf(datei,"%f dup scale \n",500.0/((double)(*r)->rows));
    
    fprintf(datei,"/Helvetica findfont %f scalefont setfont\n",
	    (*r)->cols/32.0);
    fprintf(datei,"0 setgray\n");
    fprintf(datei,"/bl{0 setgray} def\n");
    fprintf(datei,"/fb{0 setgray fill} def\n");
    fprintf(datei,"/fr{1.0 0.6 0.6  setrgbcolor fill} def\n");
    fprintf(datei,"/fg{0.5 setgray fill} def\n");
    fprintf(datei,"/fl{0.6 1.0 0.6  setrgbcolor fill} def\n");
    fprintf(datei,"/fi{0.8 0.0 0.0  setrgbcolor fill} def\n");
    fprintf(datei,"/cs{closepath stroke} def\n");
    fprintf(datei,"/m{moveto} def\n");
    fprintf(datei,"/l{lineto} def\n");
    fprintf(datei,"/hf{/Helvetica findfont} def\n");
    fprintf(datei,"/sf{scalefont setfont} def\n");
    
    outputleaf_rkmatrix((*r)->r,datei,0x0, 0,0,(*r)->rows,(*r)->cols,
			 1,2,2,tolerance);
    
    fprintf(datei,"showpage\n");
    fclose(datei);
  }else{
    printf("could not open file %s.\n",dateiname);
  }
  free(maxerror);
  free(r);
}

static double
poweriteration(psupermatrix s, pfullmatrix f,
	       double *x, double *y,
	       int rstart, int cstart,
	       int maxiter)
{
  double norm;
  int i;
  int iter;

  for(i=0; i<s->cols; i++)
    x[i] = (double) rand() / RAND_MAX;

  norm = norm2_lapack(s->cols, x);
  iter = 0;
  while(iter < maxiter && norm > 0.0) {
    scale_lapack(s->cols, x, 1.0 / norm);

    dgemv_("Not Transposed",
	   &s->rows, &s->cols,
	   deins_,
	   f->e + rstart + cstart * f->rows, &f->rows,
	   x, eins_,
	   dnull_,
	   y, eins_);
    scale_lapack(s->rows, y, -1.0);
    addeval_supermatrix(s, x, y);

    dgemv_("Transposed",
	   &s->rows, &s->cols,
	   deins_,
	   f->e + rstart + cstart * f->rows, &f->rows,
	   y, eins_,
	   dnull_,
	   x, eins_);
    scale_lapack(s->cols, x, -1.0);
    addevaltrans_supermatrix(s, y, x);

    norm = norm2_lapack(s->cols, x);
    iter++;
  }

  return sqrt(norm);
}

static void
drawerrors(psupermatrix s, pfullmatrix f,
	   double *x, double *y,
	   int rstart, int cstart,
	   int computeonly, double *maxerror, FILE *out)
{
  double error;
  int roff, coff;
  int i, j;

  if(s->s) {
    roff = rstart;
    for(i=0; i<s->block_rows; i++) {
      coff = cstart;
      for(j=0; j<s->block_cols; j++) {
	drawerrors(s->s[i+j*s->block_rows], f, x, y, roff, coff,
		   computeonly, maxerror, out);
	coff += s->s[i+j*s->block_rows]->cols;
      }
      assert(coff == cstart + s->cols);
      roff += s->s[i]->rows;
    }
    assert(roff == rstart + s->rows);
  }
  else {
    error = poweriteration(s, f, x, y, rstart, cstart, 10);
    if(computeonly) {
      if(*maxerror < error)
	*maxerror = error;
    }
    else
      (void) fprintf(out,
		     "%f %d %d %d %d box\n",
		     dmin(error / *maxerror, 1.0),
		     rstart, cstart, s->rows, s->cols);
  }
}

void
outputerror_supermatrix(psupermatrix s, pfullmatrix f, char *dateiname)
{
  FILE *out;
  double *x;
  double *y;
  double sc;
  double maxerror;

  assert(s->rows == f->rows);
  assert(s->cols == f->cols);
  assert(dateiname != NULL);

  out = fopen(dateiname, "w");
  if(out) {
    x = (double *) malloc((size_t) sizeof(double) * s->rows);
    assert(x != NULL);

    y = (double *) malloc((size_t) sizeof(double) * s->cols);
    assert(y != NULL);

    maxerror = 0.0;
    drawerrors(s, f, x, y, 0, 0, 1, &maxerror, out);

    sc = 400.0 / imax(s->rows, s->cols);

    (void) fprintf(out,
		   "%%!PS-Adobe-2.0 EPSF-2.0\n"
		   "%%%%BoundingBox: 0 0 %d %d\n"
		   "%% Maximal local error %g\n"
		   "0.1 setlinewidth\n"
		   "%f dup scale\n"
		   "/box {\n"
		   "  4 3 roll %d exch sub 4 -1 roll exch moveto\n"
		   "  1 index neg 0 exch rlineto\n"
		   "  0 rlineto\n"
		   "  0 exch rlineto\n"
		   "  closepath gsave\n"
		   "  dup 1 exch sub 0 exch setrgbcolor\n"
		   "  fill grestore stroke } def\n\n",
		   (int)(sc * s->cols + 0.5),
		   (int)(sc * s->rows + 0.5),
		   maxerror,
		   sc,
		   s->rows);

    drawerrors(s, f, x, y, 0, 0, 0, &maxerror, out);

    (void) fprintf(out,
		   "\n"
		   "showpage\n");

    (void) fclose(out);
  }
  else
    (void) fprintf(stderr,
		   "Could not write to file \"%s\"\n",
		   dateiname);
}

void
distribute_csp(psupermatrix s, int* csp_row, int* csp_col, int level, 
	       int row_offset, int col_offset, int maxrows, int maxcols){
  int block_row, block_col;
  int ro, co;

  if(s->s != 0x0){
    ro = row_offset;
    for(block_row=0; block_row<s->block_rows; block_row++){
      co = col_offset;
      for(block_col=0; block_col<s->block_cols; block_col++){
	csp_row[ro + level*maxrows]++;
	csp_col[co + level*maxcols]++;
	distribute_csp(s->s[block_row + s->block_rows*block_col],
		       csp_row,csp_col,level+1,ro,co,maxrows,maxcols);
	co += s->s[block_row + s->block_rows*block_col]->cols;
      }
      ro += s->s[block_row]->rows;
    }  
  }
}

int
csp_supermatrix(psupermatrix s){
  int i,j,p,csp;
  int* csp_row;
  int* csp_col;
  int rows = s->rows;
  int cols = s->cols;
  p = depth_supermatrix(s);
  csp_row = (int*) malloc(p*rows*sizeof(int));
  csp_col = (int*) malloc(p*cols*sizeof(int));
  for(i=0; i<p*rows; i++) csp_row[i] = 0;
  for(i=0; i<p*cols; i++) csp_col[i] = 0;
  csp_row[0] = 1;
  csp_col[0] = 1;
  distribute_csp(s,csp_row,csp_col,1,0,0,rows,cols);
  
  csp = 0;
  for(i=0; i<rows; i++){
    for(j=0; j<p; j++){
      if(csp_row[i + j*rows]>csp) csp = csp_row[i + j*rows];
    }
  }
  for(i=0; i<cols; i++){
    for(j=0; j<p; j++){
      if(csp_col[i + j*cols]>csp) csp = csp_col[i + j*cols];
    }
  }

  free(csp_row);
  free(csp_col);
  return csp;
}

typedef struct _cid_le cid_le;

struct _cid_le {
  int level;
  int row;
  int col;
  cid_le *next;
};

cid_le*
add_cid_le(cid_le *first, int row, int col, int level){
  cid_le *next;
  for(next = first; 
      next!=0x0 && (next->row!=row || next->col!=col || next->level!=level); 
      next=next->next);
  if(next==0x0){
    next = (cid_le*) malloc(sizeof(cid_le));
    next->next = first;
    next->row = row;
    next->col = col;
    next->level = level;
    return next;
  }
  return first;
}

void
del_cid_le(cid_le *first){
  cid_le* next;
  next = first->next;
  free(first);
  if(next != 0x0)
    del_cid_le(next);
}

void
updatecidlistprod2_rkmatrix(psupermatrix c, psupermatrix a, psupermatrix b,
			    int c_row, int row_offset, 
			    int c_col, int col_offset, 
			    int level, cid_le **blocks, int *blockindex, 
			    int *blocklevel, int maxsize, int rowtotal){
  int i,j,k,n,m,am,ro,co;
  n = a->block_rows;
  am = a->block_cols;
  m = b->block_cols;
  assert(b->rows==a->cols);
  if(a->s!=0x0 && b->s!=0x0){
    ro = row_offset;
    for(i=0; i<n; i++){
      co = col_offset;
      for(j=0; j<m; j++){
	for(k=0; k<am; k++){
	  updatecidlistprod2_rkmatrix(c, a->s[i+k*n], b->s[k+j*am], 
				      c_row, ro, c_col, co, 
				      level+1,
				      blocks,blockindex,blocklevel,
				      maxsize,rowtotal);
	}
	co += b->s[j*am]->cols;
      }
      ro += a->s[i]->rows;
    }
  }else{
    for(i=0; i<maxsize && blocks[c_row + i*rowtotal] != 0x0 && 
	  (blockindex[c_row + i*rowtotal] != c_col ||
	   blocklevel[c_row + i*rowtotal] != level); i++);
    assert(i<maxsize);
    blocks[c_row + i*rowtotal] = 
      add_cid_le(blocks[c_row + i*rowtotal],row_offset,col_offset,level);
    blockindex[c_row + i*rowtotal] = c_col;
    blocklevel[c_row + i*rowtotal] = level;
  }
}


void
computecidlist_supermatrix(psupermatrix c, psupermatrix a, psupermatrix b,
			    int row_offset, int col_offset, int level,
			    cid_le **blocks, int *blockindex, int *blocklevel,
			    int maxsize, int rowtotal){
  int i,j,n,m,k,nm,ro,co;
  psupermatrix c_el;
  psupermatrix a_el;
  psupermatrix b_el;
  n = c->block_rows;
  m = c->block_cols;
  nm = a->block_cols;
  assert((a->cols==b->rows) && 
	 (a->rows==c->rows) && 
	 (b->cols==c->cols));

  if(c->s!=0x0){
    if(a->s!=0x0 && b->s!=0x0){
      /* only supermatrices, not a leaf -> recursion */ 
      ro = row_offset;
      for(i=0; i<n; i++){
	co = col_offset;
	for(j=0; j<m; j++){
	  c_el = c->s[i+j*n];
	  for(k=0; k<nm; k++){
	    a_el = a->s[i+k*n];
	    b_el = b->s[k+j*nm];
	    computecidlist_supermatrix(c_el,a_el,b_el, ro, co, level+1,
			    blocks, blockindex, blocklevel, maxsize, rowtotal);
	  }
	  co += c->s[i+j*n]->cols;
	}
	ro += c->s[i]->rows;
      }
    }else{
      /* a or b is rk or fullmatrix or uniformmatrix */
    }
  }else{
    if(c->r != 0x0 || c->u != 0x0){
      if(a->s != 0x0 && b->s != 0x0){
	/* product of 2 supermatrices to be stored in an 
	   rkmatrix or uniformmatrix */
	updatecidlistprod2_rkmatrix(c,a,b,row_offset,row_offset,
				    col_offset,col_offset,
				    level, blocks, blockindex, 
				    blocklevel, maxsize, rowtotal);
      }
    }  
  } 
}

/* no use trying to understand */
int
cid_supermatrix(psupermatrix s){
  int cid=0;
  int csp,depth;
  cid_le **blocks;
  int* blockindex;
  int* blocklevel;
  int i,j,maxsize;
  cid_le *thisblock;
  
  csp = csp_supermatrix(s);
  depth = depth_supermatrix(s);
  maxsize = csp*depth;
  blocks = (cid_le**) malloc(s->rows*maxsize*sizeof(cid_le*));
  blockindex = (int*) malloc(s->rows*maxsize*sizeof(int));
  blocklevel = (int*) malloc(s->rows*maxsize*sizeof(int));
  for(i=0; i<s->rows*maxsize; i++){
    blocks[i] = 0x0;
    blockindex[i] = -1;
    blocklevel[i] = -1;
  }
  computecidlist_supermatrix(s,s,s,0,0,0,blocks,blockindex,
			     blocklevel,maxsize,s->rows);
  cid = 1;
  for(i=0; i<s->rows*maxsize; i++){
    if(blocks[i] !=0x0){
      for(thisblock=blocks[i], j=2; thisblock->next!=0x0; 
	  thisblock=thisblock->next, j++);
      if(j>cid) cid = j;
    }
  }
  
  for(i=0; i<s->rows*maxsize; i++){
    if(blocks[i] != 0x0) del_cid_le(blocks[i]);
  }
  free(blocks);
  free(blockindex);
  free(blocklevel);
  return cid;
}

/* stuff for cholesky and lu */

void
generalmuladd_supermatrix(psupermatrix c, psupermatrix a, psupermatrix b,
		       double scale_ab, int transpose_a, int transpose_b, 
		       int uplow_or_all){
  int i,j,block_rows,block_cols,k,block_size,rows,cols,size;
  psupermatrix c_el;
  psupermatrix a_el;
  psupermatrix b_el;
  prkmatrix r;
  pfullmatrix f;
  assert(c != 0x0);
  assert(a != 0x0);
  assert(b != 0x0);
  block_rows = c->block_rows;
  block_cols = c->block_cols;
  rows = c->rows;
  cols = c->cols;

  if(transpose_a==0){
    block_size = a->block_cols;
    size = a->cols;
    assert(a->rows==c->rows);
    if(transpose_b==0){
      assert(a->cols==b->rows);
      assert(b->cols==c->cols);
    }else{
      assert(a->cols==b->cols);
      assert(b->rows==c->cols);
    }
  }else{
    block_size = a->block_rows;
    size = a->rows;
    assert(a->cols==c->rows);
    if(transpose_b==0){
      assert(a->rows==b->rows);
      assert(b->cols==c->cols);
    }else{
      assert(a->rows==b->cols);
      assert(b->rows==c->cols);
    }
  }
  assert(a->u==0x0);
  assert(b->u==0x0);
  assert(c->u==0x0);

  if(c->s!=0x0){
    if(a->s!=0x0 && b->s!=0x0){
      /* only supermatrices -> recursion */ 
      assert(((transpose_a==0 && a->block_rows==c->block_rows)
	      || (transpose_a!=0 && a->block_cols==c->block_rows))
	     && 
	     ((transpose_b==0 && b->block_cols==c->block_cols)
	      || (transpose_b!=0 && c->block_rows==c->block_cols))
	     );
      for(i=0; i<block_rows; i++)
	for(j=0; j<block_cols; j++){
	  if((i<j && (uplow_or_all != 2)) ||
	     (i>j && (uplow_or_all != 1)) ||
	     uplow_or_all == 0){
	    c_el = c->s[i+j*block_rows];
	    for(k=0; k<block_size; k++){
	      if(transpose_a==0){
		a_el = a->s[i+k*block_rows];
	      }else{
		a_el = a->s[k+i*block_size];
	      }
	      if(transpose_b==0){
		b_el = b->s[k+j*block_size];
	      }else{
		b_el = b->s[j+k*block_cols];
	      }
	      generalmuladd_supermatrix(c_el,a_el,b_el,
					scale_ab,transpose_a,transpose_b,0);
	    }
	  }else{
	    if(i==j){
	      c_el = c->s[i+j*block_rows];
	      for(k=0; k<block_size; k++){
		if(transpose_a==0){
		  a_el = a->s[i+k*block_rows];
		}else{
		  a_el = a->s[k+i*block_size];
		}
		if(transpose_b==0){
		  b_el = b->s[k+j*block_size];
		}else{
		  b_el = b->s[j+k*block_cols];
		}
		generalmuladd_supermatrix(c_el,a_el,b_el,scale_ab,
					  transpose_a,transpose_b,
					  uplow_or_all);    
	      }
	    }
	  }
	}
    }else{
      /* a or b is rk or fullmatrix */
      if(a->r!=0x0 || b->r!=0x0){
	/* rkmul */
	r = getgeneralmul_rkmatrix(a,b,transpose_a,transpose_b);
	if(scale_ab != 1.0) 
	  scale_rkmatrix(r,scale_ab);
	generaladdrk2_supermatrix(r,c,0,0,uplow_or_all);
	del_rkmatrix(r);
      }else{
	/* fullmul */
	rows = c->rows;
	cols = c->cols;
	if(transpose_a==0){
	  size = a->cols;    
	}else{
	  size = a->rows;    
	}
	f = getgeneralmul_fullmatrix(a,b,transpose_a,transpose_b);
	if(scale_ab != 1.0) 
	  scale_fullmatrix(f,scale_ab);
	
	generaladdfull2_supermatrix(f,c,0,0,uplow_or_all);
	del_fullmatrix(f);
      }
    }
  }else{
    if(c->r!=0x0){
      assert(uplow_or_all==0);
      if(a->r!=0x0 || b->r!=0x0){
	/* a or b is rk or fullmatrix */
	/* rkmul */
	r = getgeneralmul_rkmatrix(a,b,transpose_a,transpose_b);
	if(scale_ab != 1.0) 
	  scale_rkmatrix(r,scale_ab);
	addpart2_rkmatrix(c->r,0,0,r,0,0,r->rows,r->cols);
	del_rkmatrix(r);
      }else{
	/* product of 2 supermatrices to be stored in an rkmatrix*/
	addgeneralprod2_rkmatrix(c->r,a,b,0,0,
				 scale_ab,transpose_a,transpose_b); 
      }
    }else{
      if(a->r!=0x0 || b->r!=0x0){
	/* a or b is rk or fullmatrix */
	/* rkmul */
	f = getgeneralmul_fullmatrix(a,b,transpose_a,transpose_b);
	if(scale_ab != 1.0) 
	  scale_fullmatrix(f,scale_ab);
	generaladdfull2_supermatrix(f,c,0,0,uplow_or_all);
	del_fullmatrix(f);
      }else{
	/* fullmul */
	f = getgeneralmul_fullmatrix(a,b,transpose_a,transpose_b);
	if(scale_ab != 1.0) 
	  scale_fullmatrix(f,scale_ab);
	generaladdfull2_supermatrix(f,c,0,0,uplow_or_all);
	del_fullmatrix(f);
      }
    }
  } 
}

void
generalmul_supermatrix(psupermatrix c, psupermatrix a, psupermatrix b,
		       double scale_ab, int transpose_a, int transpose_b, 
		       int uplow_or_all){
  init0_supermatrix(c);
  generalmuladd_supermatrix(c,a,b,scale_ab,
			    transpose_a,transpose_b,uplow_or_all);
}

void
subeval_supermatrix(psupermatrix s, double *v, double *w){
  int i,j;
  int vindex=0;
  int windex=0;
  int sindex=0;
  int n,m;
  psupermatrix *s_el;
  assert(s != 0x0);
  assert(v != 0x0);
  assert(w != 0x0);
  n = s->block_rows;
  m = s->block_cols;
  s_el = s->s;
  if(s->r==0x0 && s->f==0x0){
    /* 1.-m. column */
    for(j=0; j<m; j++){
      windex = 0;
      for(i=0; i<n; i++){
	subeval_supermatrix(s_el[sindex],&v[vindex],&w[windex]);
	windex += s_el[sindex]->rows;
	sindex++;
      }
      vindex += s_el[sindex-1]->cols;
    }
  }else{
    if(s->r!=0x0){
      subeval_rkmatrix(s->r,v,w);
    }else{
      subeval_fullmatrix(s->f,v,w);
    }
  }  
}

void
subevaltrans_supermatrix(psupermatrix s, double* v, double* w){
  int i,j;
  int vindex=0;
  int windex=0;
  int sindex=0;
  int n,m;
  psupermatrix *s_el;
  assert(s != 0x0);
  assert(v != 0x0);
  assert(w != 0x0);
  n = s->block_rows;
  m = s->block_cols;
  s_el = s->s;
  if(s->r==0x0 && s->f==0x0){
    /* 1.-n. row */
    for(i=0; i<n; i++){
      windex = 0;
      sindex = i;
      for(j=0; j<m; j++){
	subevaltrans_supermatrix(s_el[sindex],&v[vindex],&w[windex]);
	windex += s_el[sindex]->cols;
	sindex+=n;
      }
      vindex += s_el[sindex-n]->rows;
    }
  }else{
    if(s->r!=0x0){
      subevaltrans_rkmatrix(s->r,v,w);
    }else{
      subevaltrans_fullmatrix(s->f,v,w);
    }
  }
}

void
solvecholeskyvector_supermatrix(psupermatrix l, double *v){
  int i,j;
  int row_of, col_of;

  if(l->s!=0x0){
    assert(l->block_rows == l->block_cols);
    assert(l->rows == l->cols);
    row_of = 0;
    for(i=0; i<l->block_rows; i++){
      col_of = 0;
      for(j=0; j<i; j++){
	subeval_supermatrix(l->s[i+j*l->block_rows],&v[col_of],&v[row_of]);
	col_of += l->s[j*l->block_rows]->cols;
      }
      solvecholeskyvector_supermatrix(l->s[i+i*l->block_rows],&v[row_of]);
      row_of += l->s[i]->rows;
    }
  }else{
    assert(l->f != 0x0);
    assert(l->rows == l->cols);
    dtrsv_("L","N","N",&l->rows,l->f->e,&l->rows,v,eins_);
  }
}

void
generaleval_supermatrix(psupermatrix s, int uplo, int trans, int add,
			int diag, double alpha, double *v, double *w){
  /* w = (w) +alpha A(trans,uplo)*v */
  /* uplo=0 -> all, uplo=1 -> lower, uplo=2 -> upper */
  /* trans=1 -> transposed */
  /* diag=1 -> unit diagonal */
  /* add = 1 w+=... */
  int i,j;
  int vindex;
  int windex;
  int block_rows,block_cols;
  psupermatrix *s_el;

  assert(s != 0x0);
  assert(v != 0x0);
  assert(w != 0x0);

  block_rows = s->block_rows;
  block_cols = s->block_cols;
  s_el = s->s;

  if(add==0){
    if(trans==0) scale_lapack(s->rows,w,0.0);
    if(trans==1) scale_lapack(s->cols,w,0.0);
    generaleval_supermatrix(s,uplo,trans,1,diag,alpha,v,w);
    return;
  }

  if(s->s != 0x0){
    assert(s_el != 0x0);
    
    if(trans==0){
      vindex = 0;
      for(j=0; j<block_cols; j++) {
	windex = 0;
	for(i=0; i<block_rows; i++) {
	  if(i==j){
	    generaleval_supermatrix(s_el[i+j*block_rows], uplo, trans, add,
				    diag, alpha, v+vindex, w+windex);
	  }else{
	    if((uplo==1 && i>j) || (uplo==2 && i<j) || uplo==0){
	      generaleval_supermatrix(s_el[i+j*block_rows], 0, trans, add,
				      0, alpha, v+vindex, w+windex);
	    }
	  }
	  windex += s_el[i+j*block_rows]->rows;
	}
	assert(windex == s->rows);
	vindex += s_el[j*block_rows]->cols;
      }
      assert(vindex == s->cols);
    }else{
      vindex = 0;
      for(i=0; i<block_rows; i++){
	windex = 0;
	for(j=0; j<block_cols; j++){
	  if(i==j){
	    generaleval_supermatrix(s_el[i+j*block_rows], uplo, trans, add,
				    diag, alpha, v+vindex, w+windex);
	  }else{
	    if((uplo==1 && i>j) || (uplo==2 && i<j) || uplo==0){
	      generaleval_supermatrix(s_el[i+j*block_rows], 0, trans, add,
				      0, alpha, v+vindex, w+windex);
	    }
	  }
	  windex += s_el[i+j*block_rows]->cols;
	}
	assert(windex == s->cols);
	vindex += s_el[i]->rows;
      }
      assert(vindex == s->rows);
    }
  }else{
    if(s->r != 0x0){
      generaleval_rkmatrix(s->r, uplo, trans, add, alpha, v, w);
    }else{
      generaleval_fullmatrix(s->f, uplo, trans, add, diag, alpha, v, w);
    }
  }
}

void
solvecholeskyvectortrans_supermatrix(psupermatrix l, double *v){
  int i,j;
  int n = l->block_rows;
  int nof, mof;
  assert(l->block_rows == l->block_cols);
  assert(l->rows == l->cols);
  if(l->s!=0x0){
    nof = l->rows;
    for(i=n-1; i>=0; i--){
      mof = nof;
      nof -= l->s[i]->rows;
      for(j=i+1; j<n; j++){
	subevaltrans_supermatrix(l->s[j+i*n],&v[mof],&v[nof]);
	mof += l->s[j+i*n]->rows;
      }
      solvecholeskyvectortrans_supermatrix(l->s[i+i*n],&v[nof]);
    }
  }else{
    assert(l->f!=0x0);
    dtrsv_("L","T","N",&l->rows,l->f->e,&l->rows,v,eins_);
  }
}

/* X L^T = A, store X in A */
void
solveleftcholesky_supermatrix(psupermatrix l, psupermatrix a){
  int i,j,nu;
  double *v;
  pfullmatrix f;
  assert(a!=0x0 && l!=0x0);
  if(l->s!=0x0){
    if(a->s!=0x0){
      assert(l->block_rows == l->block_cols);
      assert(l->block_rows == a->block_cols);
      for(j=0; j<a->block_cols; j++){
	for(i=0; i<a->block_rows; i++){
	  for(nu=0; nu<j; nu++){
	    generalmuladd_supermatrix(a->s[i+j*a->block_rows], 
				      a->s[i+nu*a->block_rows], 
				      l->s[j+nu*l->block_rows],
				      -1.0, 0, 1, 0);
	  }
	  solveleftcholesky_supermatrix(l->s[j+j*l->block_rows],
					a->s[i+j*a->block_rows]);
	  
	}
      }
    }else{
      if(a->f!=0x0){
	v = (double*) malloc(a->cols * sizeof(double));
	for(i=0; i<a->rows; i++){
	  dcopy_(&a->cols,&a->f->e[i],&a->rows,v,eins_);
	  solvecholeskyvector_supermatrix(l,v);
	  dcopy_(&a->cols,v,eins_,&a->f->e[i],&a->rows);
	}
	free(v);
      }else{
	assert(a->r!=0x0);
	for(i=0; i<a->r->kt; i++){
	  solvecholeskyvector_supermatrix(l,&a->r->b[i*a->cols]);
	}
      }
    }
    }else{
    if(a->f!=0x0){
      v = (double*) malloc(a->cols * sizeof(double));
      for(i=0; i<a->rows; i++){
	dcopy_(&a->cols,&a->f->e[i],&a->rows,v,eins_);
	solvecholeskyvector_supermatrix(l,v);
	dcopy_(&a->cols,v,eins_,&a->f->e[i],&a->rows);
      }
      free(v);
    }else{
      if(a->r!=0x0){
	for(i=0; i<a->r->kt; i++){
	  solvecholeskyvector_supermatrix(l,&a->r->b[i*a->cols]);
	}
      }else{
	f = new_fullmatrix(a->rows,a->cols);
	convertsuper2_fullmatrix(f,a);
	v = (double*) malloc(a->cols * sizeof(double));
	for(i=0; i<a->rows; i++){
	  dcopy_(&a->cols,&f->e[i],&a->rows,v,eins_);
	  solvecholeskyvector_supermatrix(l,v);
	  dcopy_(&a->cols,v,eins_,&f->e[i],&a->rows);
	}
	free(v);
	convertfull2_supermatrix(f,a,0,0);
	del_fullmatrix(f);
      }
    } 
  }
}

/* X L^T = A, store X in A 
void
solveleftcholesky_supermatrix(psupermatrix l, psupermatrix a){
  int i,j,nu;
  double *v;
  pfullmatrix fa;

  assert(a!=0x0 && l!=0x0);
  if(l->s!=0x0){
    if(a->s!=0x0){
      assert(l->block_rows == l->block_cols);
      assert(l->block_cols == a->block_cols);
      for(j=0; j<a->block_cols; j++){
	for(i=0; i<a->block_rows; i++){
	  for(nu=0; nu<j; nu++){
	    generalmuladd_supermatrix(a->s[i+j*a->block_rows], 
				      a->s[i+nu*a->block_rows], 
				      l->s[j+nu*l->block_rows],
				      -1.0, 0, 1, 0);
	  }
	  solveleftcholesky_supermatrix(l->s[j+j*l->block_rows],
					a->s[i+j*a->block_rows]);
	  
	}
      }
    }else{
      if(a->f!=0x0){
	v = (double*) malloc(a->cols * sizeof(double));
	for(i=0; i<a->rows; i++){
	  dcopy_(&a->cols,&a->f->e[i],&a->rows,v,eins_);
	  solvecholeskyvector_supermatrix(l,v);
	  dcopy_(&a->cols,v,eins_,&a->f->e[i],&a->rows);
	}
	free(v);
      }else{
	assert(a->r!=0x0);
	for(i=0; i<a->r->kt; i++){
	  solvecholeskyvector_supermatrix(l,&a->r->b[i*a->cols]);
	}
      }
    }
  }else{
    assert(l->f!=0x0);
    if(a->f!=0x0){
      v = (double*) malloc(a->cols * sizeof(double));
      for(i=0; i<a->rows; i++){
	dcopy_(&a->cols,&a->f->e[i],&a->rows,v,eins_);
	solvecholeskyvector_supermatrix(l,v);
	dcopy_(&a->cols,v,eins_,&a->f->e[i],&a->rows);
      }
      free(v);
    }else{
      if(a->r!=0x0){
	  for(i=0; i<a->r->kt; i++){
	      solvecholeskyvector_supermatrix(l,&a->r->b[i*a->cols]);
	  }
      }else{
	  fa = new_fullmatrix(a->rows,a->cols);
	  convertsuper2_fullmatrix(fa,a);
	  v = (double*) malloc(a->cols*sizeof(double));
	  for(i=0; i<fa->rows; i++){
	      dcopy_(&a->cols,&fa->e[i],&a->rows,v,eins_);
	      solvecholeskyvector_supermatrix(l,v);
	      dcopy_(&a->cols,v,eins_,&fa->e[i],&a->rows);
	  } 
	  convertfull2_supermatrix(fa,a,0,0);
	  free(v);
	  del_fullmatrix(fa);
      }
    } 
  }
}
*/

/* L^{T}X = A, store X in A */
void
solverightcholesky_supermatrix(psupermatrix l, psupermatrix a){
  int i,j,nu;
  pfullmatrix fa;

  assert(a!=0x0 && l!=0x0);
  if(l->s!=0x0){
    if(a->s!=0x0){
      assert(l->block_rows == l->block_cols);
      assert(l->block_rows == a->block_rows);
      for(i=a->block_rows-1; i>=0; i--){
	for(j=0; j<a->block_cols; j++){
	  for(nu=a->block_rows-1; nu>i; nu--){
	    generalmuladd_supermatrix(a->s[i+j*a->block_rows], 
				      l->s[nu+i*l->block_rows], 
				      a->s[nu+j*a->block_rows],
				      -1.0, 0, 1, 0);
	  }
	  solverightcholesky_supermatrix(l->s[i+i*l->block_rows],
					 a->s[i+j*a->block_rows]);
	  
	}
      }
    }else{
      if(a->f!=0x0){
	for(i=0; i<a->cols; i++){
	  solvecholeskyvectortrans_supermatrix(l,&a->f->e[i*a->rows]);
	}
      }else{
	assert(a->r!=0x0);
	for(i=0; i<a->r->kt; i++){
	  solvecholeskyvectortrans_supermatrix(l,&a->r->a[i*a->rows]);
	}
      }
    }
  }else{
    assert(l->f!=0x0);
    if(a->f!=0x0){
      for(i=0; i<a->cols; i++){
	solvecholeskyvectortrans_supermatrix(l,&a->f->e[i*a->rows]);
      }
    }else{
      if(a->r!=0x0){
	for(i=0; i<a->r->kt; i++){
	  solvecholeskyvectortrans_supermatrix(l,&a->r->a[i*a->rows]);
	}
      }else{
	  fa = new_fullmatrix(a->rows,a->cols);
	  convertsuper2_fullmatrix(fa,a);
	  for(i=0; i<fa->cols; i++){
	    solvecholeskyvectortrans_supermatrix(l,&fa->e[i*a->rows]);
	  }
	  convertfull2_supermatrix(fa,a,0,0);
	  del_fullmatrix(fa);
      }
    } 
  }
}

void
setzeroupper_supermatrix(psupermatrix s, int rof, int cof){
  int i,j;
  int ro,co;

  if(s->s!=0x0){
    ro = rof;
    for(i=0; i<s->block_rows; i++){
      co = cof;
      for(j=0; j<s->block_cols; j++){
	setzeroupper_supermatrix(s->s[i+j*s->block_rows],ro,co);
	co += s->s[j*s->block_rows]->cols;
      }
      ro += s->s[i]->rows;
    }
  }else{
    if(rof<=cof){
      if(s->r!=0x0) s->r->kt=0;
      if(s->f!=0x0){
	for(i=0; i<s->rows; i++)
	  for(j=0; j<s->cols; j++)
	    if(rof+i < cof+j) s->f->e[i+j*s->rows] = 0.0;
      }
    }
  }
}

void
setzerolower_supermatrix(psupermatrix s, int rof, int cof){
  int i,j;
  int ro,co;

  if(s->s!=0x0){
    ro = rof;
    for(i=0; i<s->block_rows; i++){
      co = cof;
      for(j=0; j<s->block_cols; j++){
	setzerolower_supermatrix(s->s[i+j*s->block_rows],ro,co);
	co += s->s[j*s->block_rows]->cols;
      }
      ro += s->s[i]->rows;
    }
  }else{
    if(rof>=cof){
      if(s->r!=0x0) s->r->kt=0;
      if(s->f!=0x0){
	for(i=0; i<s->rows; i++)
	  for(j=0; j<s->cols; j++)
	    if(rof+i > cof+j) s->f->e[i+j*s->rows] = 0.0;
      }
    }
  }
}

void
choleskydecomposition_supermatrix(psupermatrix a){
  int info=0,i,j,nu;

  assert(a!=0x0);
  if(a->s!=0x0){
    for(j=0; j<a->block_cols; j++){
      for(i=j; i<a->block_rows; i++){
	for(nu=0; nu<j; nu++){
	  if(i==j){
	    generalmuladd_supermatrix(a->s[i+i*a->block_rows], 
				      a->s[i+nu*a->block_rows],
				      a->s[i+nu*a->block_rows], 
				      -1.0, 0, 1, 2);	  
	  }else{
	    generalmuladd_supermatrix(a->s[i+j*a->block_rows], 
				      a->s[i+nu*a->block_rows],
				      a->s[j+nu*a->block_rows], 
				      -1.0, 0, 1, 0);	  
	  }
	}
	if(i==j){
	  choleskydecomposition_supermatrix(a->s[i+i*a->block_rows]);
	}else{
	  solveleftcholesky_supermatrix(a->s[j+j*a->block_rows],
					a->s[i+j*a->block_rows]);	
	}
      }
    }
  }else{
    assert(a->f!=0x0);
    dpotrf_("L",&a->rows,a->f->e,&a->cols,&info);
    if(info!=0){
      printf("In choleskydecomposition_supermatrix: info=%d\n",info);
      exit(0);
    }
  }
}

void
solvecholesky_supermatrix(psupermatrix l, double *v){
  solvecholeskyvector_supermatrix(l, v);
  solvecholeskyvectortrans_supermatrix(l, v);
}

/******/
/* LU */
/******/

void
solvelower_supermatrix(psupermatrix a, double *v){
  int i,j;
  int row_of, col_of;

  assert(a!=0x0);
  if(a->s!=0x0){
    assert(a->block_rows == a->block_cols);
    row_of = 0;
    for(i=0; i<a->block_rows; i++){
      col_of = 0;
      for(j=0; j<i; j++){
	subeval_supermatrix(a->s[i+j*a->block_rows],&v[col_of],&v[row_of]);
	col_of += a->s[j*a->block_rows]->cols;
      }
      solvelower_supermatrix(a->s[i+i*a->block_rows],&v[row_of]);
      row_of += a->s[i]->rows;
    }
  }else{
    assert(a->f!=0x0 && a->rows==a->cols);
    dtrsv_("L","N","U",&a->rows,a->f->e,&a->rows,v,eins_);
  }
}

void
solvelowertrans_supermatrix(psupermatrix a, double *v){
  int i,j;
  int n = a->block_rows;
  int nof, mof;
  assert(a!=0x0);
  assert(a->block_rows == a->block_cols);
  if(a->s!=0x0){
    nof = a->rows;
    for(i=n-1; i>=0; i--){
      mof = nof;
      nof -= a->s[i]->rows;
      for(j=i+1; j<n; j++){
	subevaltrans_supermatrix(a->s[j+i*n],&v[mof],&v[nof]);
	mof += a->s[i+j*n]->cols;
      }
      solvelowertrans_supermatrix(a->s[i+i*n],&v[nof]);
    }
  }else{
    assert(a->f!=0x0 && a->rows==a->cols);
    dtrsv_("L","T","U",&a->rows,a->f->e,&a->rows,v,eins_);
  }
}

void
solveupper_supermatrix(psupermatrix a, double *v){
  int i,j;
  int n = a->block_rows;
  int nof, mof;
  assert(a!=0x0);
  assert(a->block_rows == a->block_cols);
  if(a->s!=0x0){
    nof = a->cols;
    for(i=n-1; i>=0; i--){
      mof = nof;
      nof -= a->s[i]->rows;
      for(j=i+1; j<n; j++){
	subeval_supermatrix(a->s[i+j*n],&v[mof],&v[nof]);
	mof += a->s[i+j*n]->cols;
      }
      solveupper_supermatrix(a->s[i+i*n],&v[nof]);
    }
  }else{
    assert(a->f!=0x0 && a->rows==a->cols);
    dtrsv_("U","N","N",&a->rows,a->f->e,&a->rows,v,eins_);
  }
}

void
solveuppertrans_supermatrix(psupermatrix a, double *v){
  int i,j;
  int row_of, col_of;
   assert(a!=0x0);
  if(a->s!=0x0){
    assert(a->block_rows == a->block_cols);
    row_of = 0;
    for(i=0; i<a->block_rows; i++){
      col_of = 0;
      for(j=0; j<i; j++){
	subevaltrans_supermatrix(a->s[j+i*a->block_rows],&v[col_of],&v[row_of]);
	col_of += a->s[j*a->block_rows]->cols;
      }
      solveuppertrans_supermatrix(a->s[i+i*a->block_rows],&v[row_of]);
      row_of += a->s[i]->rows;
    }
  }else{
    assert(a->f!=0x0 && a->rows==a->cols);
    dtrsv_("U","T","N",&a->rows,a->f->e,&a->rows,v,eins_);
  }
}

/* X U = A, store X in A */
void
solveleftlu_supermatrix(psupermatrix u, psupermatrix a){
  int i,j,nu;
  double *v;
  pfullmatrix f;
  assert(a!=0x0 && u!=0x0);
  if(u->s!=0x0){
    if(a->s!=0x0){
      assert(u->block_rows == u->block_cols);
      assert(u->block_rows == a->block_cols);
      for(j=0; j<a->block_cols; j++){
	for(i=0; i<a->block_rows; i++){
	  for(nu=0; nu<j; nu++){
	    generalmuladd_supermatrix(a->s[i+j*a->block_rows], 
				      a->s[i+nu*a->block_rows], 
				      u->s[nu+j*u->block_rows],
				      -1.0, 0, 0, 0);
	  }
	  solveleftlu_supermatrix(u->s[j+j*u->block_rows],
				  a->s[i+j*a->block_rows]);
	  
	}
      }
    }else{
      if(a->f!=0x0){
	v = (double*) malloc(a->cols * sizeof(double));
	for(i=0; i<a->rows; i++){
	  dcopy_(&a->cols,&a->f->e[i],&a->rows,v,eins_);
	  solveuppertrans_supermatrix(u,v);
	  dcopy_(&a->cols,v,eins_,&a->f->e[i],&a->rows);
	}
	free(v);
      }else{
	assert(a->r!=0x0);
	for(i=0; i<a->r->kt; i++){
	  solveuppertrans_supermatrix(u,&a->r->b[i*a->cols]);
	}
      }
    }
  }else{
    if(a->f!=0x0){
      v = (double*) malloc(a->cols * sizeof(double));
      for(i=0; i<a->rows; i++){
	dcopy_(&a->cols,&a->f->e[i],&a->rows,v,eins_);
	solveuppertrans_supermatrix(u,v);
	dcopy_(&a->cols,v,eins_,&a->f->e[i],&a->rows);
      }
      free(v);
    }else{
      if(a->r!=0x0){
	for(i=0; i<a->r->kt; i++){
	  solveuppertrans_supermatrix(u,&a->r->b[i*a->cols]);
	}
      }else{
	f = new_fullmatrix(a->rows,a->cols);
	convertsuper2_fullmatrix(f,a);
	v = (double*) malloc(a->cols * sizeof(double));
	for(i=0; i<a->rows; i++){
	  dcopy_(&a->cols,&f->e[i],&a->rows,v,eins_);
	  solveuppertrans_supermatrix(u,v);
	  dcopy_(&a->cols,v,eins_,&f->e[i],&a->rows);
	}
	free(v);
	convertfull2_supermatrix(f,a,0,0);
	del_fullmatrix(f);
      }
    } 
  }
}

/* LX = A, store X in A */
void
solverightlu_supermatrix(psupermatrix l, psupermatrix a){
  int i,j,nu;
  pfullmatrix f;
  assert(a!=0x0 && l!=0x0);
  if(l->s!=0x0){
    if(a->s!=0x0){
      assert(l->block_rows == l->block_cols);
      assert(l->block_rows == a->block_rows);
      
      for(j=0; j<a->block_cols; j++){
	for(i=0; i<a->block_rows; i++){
	  for(nu=0; nu<i; nu++){
	    generalmuladd_supermatrix(a->s[i+j*a->block_rows], 
				      l->s[i+nu*l->block_rows], 
				      a->s[nu+j*a->block_rows],
				      -1.0, 0, 0, 0);
	  }
	  solverightlu_supermatrix(l->s[i+i*l->block_rows],
				   a->s[i+j*a->block_rows]);
	  
	}
      }
    }else{
      if(a->f!=0x0){
	for(i=0; i<a->cols; i++){
	  solvelower_supermatrix(l,&a->f->e[i*a->rows]);
	}
      }else{
	assert(a->r!=0x0);
	for(i=0; i<a->r->kt; i++){
	  solvelower_supermatrix(l,&a->r->a[i*a->rows]);
	}
      }
    }
  }else{
    if(a->f!=0x0){
      for(i=0; i<a->cols; i++){
	solvelower_supermatrix(l,&a->f->e[i*a->rows]);
      }
    }else{
      if(a->r!=0x0){
	for(i=0; i<a->r->kt; i++){
	  solvelower_supermatrix(l,&a->r->a[i*a->rows]);
	}
      }else{
	f = new_fullmatrix(a->rows,a->cols);
	convertsuper2_fullmatrix(f,a);
	for(i=0; i<a->cols; i++){
	  solvelower_supermatrix(l,&f->e[i*a->rows]);
	}
	convertfull2_supermatrix(f,a,0,0);
	del_fullmatrix(f);
      }
    } 
  }
}

void
ludecomposition_supermatrix(psupermatrix a){
  int i,j,k,nu;
  double *m;
  int rows,cols;

  assert(a!=0x0);
  if(a->s!=0x0){
    for(nu=0; nu<a->block_rows && nu<a->block_cols; nu++){
      /* diagonal part */
      for(i=0; i<nu; i++){
	generalmuladd_supermatrix(a->s[nu+nu*a->block_rows], 
				  a->s[nu+i*a->block_rows],
				  a->s[i+nu*a->block_rows], -1.0, 0, 0, 0);
      }
      ludecomposition_supermatrix(a->s[nu+nu*a->block_rows]);   
      /* strictly lower part */
      for(i=nu+1; i<a->block_rows; i++){
	for(j=0; i<nu; j++){
	  generalmuladd_supermatrix(a->s[i+nu*a->block_rows], 
				    a->s[i+j*a->block_rows],
				    a->s[j+nu*a->block_rows], -1.0, 0, 0, 0);
	}
	solveleftlu_supermatrix(a->s[nu+nu*a->block_rows], 
				a->s[i+nu*a->block_rows]);
      }
      /* strictly upper part */
      for(j=nu+1; j<a->block_cols; j++){
	for(i=0; i<nu; i++){
	  generalmuladd_supermatrix(a->s[nu+j*a->block_rows], 
				    a->s[nu+i*a->block_rows],
				    a->s[i+j*a->block_rows], -1.0, 0, 0, 0);
	}
	solverightlu_supermatrix(a->s[nu+nu*a->block_rows], 
				 a->s[nu+j*a->block_rows]);
      }
    }
  }else{
    assert(a->f!=0x0);
    assert(a->rows==a->cols);
    /* lu decomposition for fullmatrix without pivoting,
       l_ii = 1, l(ower) without diagonal(=1) and u(pper) stored in a */
    m = a->f->e;
    rows = a->rows;
    cols = a->cols;
    
    for(k=0; k<rows; k++){
      /* u_kk = a_kk - sum_i<k l_ki*u_ik */
      for(i=0; i<k; i++)
	m[k+k*rows] -= m[k+i*rows]*m[i+k*rows];
      /* l_ik = (a_ik - sum_j<k l_ij*u_jk)/u_kk */
      for(i=k+1; i<rows; i++){
	for(j=0; j<k; j++){
	  m[i+k*rows] -= m[i+j*rows]*m[j+k*rows];
	}
	if(fabs(m[k+k*rows])<1e-16)
	  printf("warning: diag_element(LU) = %e\n",m[k+k*rows]);
	m[i+k*rows] /= m[k+k*rows];
      }
      /* u_kj = a_kj - sum_i<k l_ki*u_ij */
      for(j=k+1; j<cols; j++){
	for(i=0; i<k; i++){
	  m[k+j*rows] -= m[k+i*rows]*m[i+j*rows];
	}
      }
    }
  }
}

void
solvelu_supermatrix(psupermatrix lu, double *v){
  solvelower_supermatrix(lu,v);
  solveupper_supermatrix(lu,v);
}

double 
norm2cholesky2_supermatrix(psupermatrix l, psupermatrix s){
  double norm2 = 0.0;
  double tmp = 0.0;
  double *v, *w, *v2, *v3;
  int i,n,m;
  int nr = 10;
  assert(s != 0x0);
  assert(l != 0x0);
  n = l->rows;
  m = l->cols;
  assert(s->rows == m);
  assert(s->cols == n);

  v = (double*) malloc(n*sizeof(double));
  w = (double*) malloc(m*sizeof(double));
  v2 = (double*) malloc(n*sizeof(double));
  v3 = (double*) malloc(n*sizeof(double));
  assert(v!=0x0 && w!=0x0 && v2!=0x0 && v3!=0x0);
  for(i=0; i<n; i++){
    v[i] = ((double) (random()%1024))/1024.0;
  }
  tmp = 1.0/dnrm2_(&n,v,eins_);
  dscal_(&n,&tmp,v,eins_);
  for(i=0; i<nr; i++){

    dcopy_(&n,v,eins_,v2,eins_);
    dcopy_(&n,v,eins_,w,eins_);
    solvecholeskyvector_supermatrix(l,w);    
    solvecholeskyvectortrans_supermatrix(l,w);    

    evaltrans_supermatrix(s,w,v3);

    dcopy_(&n,v,eins_,v2,eins_);
    daxpy_(&n,meins_,v3,eins_,v2,eins_);

    daxpy_(&n,meins_,v,eins_,v3,eins_);

    eval_supermatrix(s,v3,w);

    dcopy_(&m,w,eins_,v3,eins_);
    solvecholeskyvector_supermatrix(l,v3);    
    solvecholeskyvectortrans_supermatrix(l,v3);    

    daxpy_(&n,deins_,v3,eins_,v2,eins_);
 
    norm2 = sqrt(fabs(ddot_(&m,v2,eins_,v,eins_)));
    tmp = dnrm2_(&m,v2,eins_);
    if(tmp>1e-64){
      tmp = 1.0/tmp;
      dscal_(&m,&tmp,v2,eins_);
      dcopy_(&m,v2,eins_,v,eins_);
    }else{
      i=nr;
    }
  }
  free(v);
  free(v2);
  free(w);
  free(v3);
  return norm2;
}

double 
norm2nrcholeskydiff_supermatrix(psupermatrix l, psupermatrix s, int nr){
  double norm2 = 0.0;
  double tmp = 0.0;
  double *v, *w, *v2;
  int i,rows,cols;
  assert(l != 0x0);
  assert(s != 0x0);
  assert(l->u == 0x0);
  assert(s->u == 0x0);
  rows = l->rows;
  cols = l->cols;

  v = (double*) malloc(cols*sizeof(double));
  w = (double*) malloc(rows*sizeof(double));
  v2 = (double*) malloc(cols*sizeof(double));
  assert(v!=0x0 && w!=0x0 && v2!=0x0);
  for(i=0; i<cols; i++){
    v[i] = rand();
    w[i] = 0.0;
    v2[i] = 0.0;
  }
  tmp = 1.0/dnrm2_(&cols,v,eins_);
  dscal_(&cols,&tmp,v,eins_);
  for(i=0; i<nr; i++){
    generaleval_supermatrix(l,1,1,0,0,1.0,v,v2);
    generaleval_supermatrix(l,1,0,0,0,1.0,v2,w);

    dscal_(&rows,meins_,w,eins_);
    addeval_supermatrix(s,v,w);

    generaleval_supermatrix(l,1,1,0,0,1.0,w,v);
    generaleval_supermatrix(l,1,0,0,0,1.0,v,v2);

    dscal_(&cols,meins_,v2,eins_);
    addevaltrans_supermatrix(s,w,v2);

    norm2 = dnrm2_(&cols, v2, eins_);
    if(norm2>1e-32){
      tmp = 1.0/norm2;
      dscal_(&cols,&tmp,v2,eins_);
      dcopy_(&cols,v2,eins_,v,eins_);
    }else{
      i=nr;
    }
  }
  free(v);
  free(v2);
  free(w);
  return sqrt(norm2);
}

double 
norm2choleskydiff_supermatrix(psupermatrix l, psupermatrix s){
  return norm2nrcholeskydiff_supermatrix(l,s,10);
}

double 
norm2choleskyh2_supermatrix(psupermatrix l, psupermatrix s){
  double norm2 = 0.0;
  double tmp = 0.0;
  double *v, *w, *v2, *v3;
  int i,n,m;
  int nr = 10;
  assert(s != 0x0);
  assert(l != 0x0);
  n = l->rows;
  m = l->cols;
  assert(s->rows == m);
  assert(s->cols == n);

  v = (double*) malloc(n*sizeof(double));
  w = (double*) malloc(m*sizeof(double));
  v2 = (double*) malloc(n*sizeof(double));
  v3 = (double*) malloc(n*sizeof(double));
  assert(v!=0x0 && w!=0x0 && v2!=0x0 && v3!=0x0);
  for(i=0; i<n; i++){
    v[i] = ((double) (random()%1024))/1024.0;
  }
  tmp = 1.0/dnrm2_(&n,v,eins_);
  dscal_(&n,&tmp,v,eins_);
  for(i=0; i<nr; i++){

    dcopy_(&n,v,eins_,v2,eins_);
    dcopy_(&n,v,eins_,w,eins_);
    lowersolve_supermatrix(l,w);    
    uppersolve_supermatrix(l,w);    

    evaltrans_supermatrix(s,w,v3);

    dcopy_(&n,v,eins_,v2,eins_);
    daxpy_(&n,meins_,v3,eins_,v2,eins_);

    daxpy_(&n,meins_,v,eins_,v3,eins_);

    eval_supermatrix(s,v3,w);

    dcopy_(&m,w,eins_,v3,eins_);
    lowersolve_supermatrix(l,v3);    
    uppersolve_supermatrix(l,v3);    

    daxpy_(&n,deins_,v3,eins_,v2,eins_);
 
    norm2 = sqrt(fabs(ddot_(&m,v2,eins_,v,eins_)));
    tmp = dnrm2_(&m,v2,eins_);
    if(tmp>1e-64){
      tmp = 1.0/tmp;
      dscal_(&m,&tmp,v2,eins_);
      dcopy_(&m,v2,eins_,v,eins_);
    }else{
      i=nr;
    }
  }
  free(v);
  free(v2);
  free(w);
  free(v3);
  return norm2;
}

double 
norm2cholesky_sparsematrix(psupermatrix l, psparsematrix s){
  double norm2 = 0.0;
  double tmp = 0.0;
  double *v, *w, *v2, *v3;
  int i,n,m;
  int nr = 10;
  assert(s != 0x0);
  assert(l != 0x0);
  n = l->rows;
  m = l->cols;
  assert(s->rows == m);
  assert(s->cols == n);

  v = (double*) malloc(n*sizeof(double));
  w = (double*) malloc(m*sizeof(double));
  v2 = (double*) malloc(n*sizeof(double));
  v3 = (double*) malloc(n*sizeof(double));
  assert(v!=0x0 && w!=0x0 && v2!=0x0 && v3!=0x0);
  for(i=0; i<n; i++){
    v[i] = ((double) (random()%1024))/1024.0;
  }
  tmp = 1.0/dnrm2_(&n,v,eins_);
  dscal_(&n,&tmp,v,eins_);
  for(i=0; i<nr; i++){

    dcopy_(&n,v,eins_,v2,eins_);
    dcopy_(&n,v,eins_,w,eins_);
    solvecholeskyvector_supermatrix(l,w);    
    solvecholeskyvectortrans_supermatrix(l,w);    

    evaltrans_sparsematrix(s,w,v3);

    dcopy_(&n,v,eins_,v2,eins_);
    daxpy_(&n,meins_,v3,eins_,v2,eins_);

    daxpy_(&n,meins_,v,eins_,v3,eins_);

    eval_sparsematrix(s,v3,w);

    dcopy_(&m,w,eins_,v3,eins_);
    solvecholeskyvector_supermatrix(l,v3);    
    solvecholeskyvectortrans_supermatrix(l,v3);    

    daxpy_(&n,deins_,v3,eins_,v2,eins_);
 
    norm2 = sqrt(fabs(ddot_(&m,v2,eins_,v,eins_)));
    tmp = dnrm2_(&m,v2,eins_);
    if(tmp>1e-64){
      tmp = 1.0/tmp;
      dscal_(&m,&tmp,v2,eins_);
      dcopy_(&m,v2,eins_,v,eins_);
    }else{
      i=nr;
    }
  }
  free(v);
  free(v2);
  free(w);
  free(v3);
  return norm2;
}

double 
norm2lu_supermatrix(psupermatrix lu, psparsematrix s){
  double norm2 = 0.0;
  double tmp = 0.0;
  double *v, *w, *v2, *v3;
  int i,n,m;
  int nr = 10;
  assert(s != 0x0);
  assert(lu != 0x0);
  n = lu->rows;
  m = lu->cols;
  assert(s->rows == m);
  assert(s->cols == n);

  v = (double*) malloc(n*sizeof(double));
  w = (double*) malloc(m*sizeof(double));
  v2 = (double*) malloc(n*sizeof(double));
  v3 = (double*) malloc(n*sizeof(double));
  assert(v!=0x0 && w!=0x0 && v2!=0x0 && v3!=0x0);
  for(i=0; i<n; i++){
    v[i] = ((double) (random()%1024))/1024.0;
  }
  tmp = 1.0/dnrm2_(&n,v,eins_);
  dscal_(&n,&tmp,v,eins_);
  for(i=0; i<nr; i++){

    dcopy_(&n,v,eins_,v2,eins_);
    dcopy_(&n,v,eins_,w,eins_);
    solvelower_supermatrix(lu,w);    
    solveupper_supermatrix(lu,w);    

    eval_sparsematrix(s,w,v3);

    dcopy_(&n,v,eins_,v2,eins_);
    daxpy_(&n,meins_,v3,eins_,v2,eins_);

    daxpy_(&n,meins_,v,eins_,v3,eins_);

    evaltrans_sparsematrix(s,v3,w);

    dcopy_(&m,w,eins_,v3,eins_);
    solveuppertrans_supermatrix(lu,v3);    
    solvelowertrans_supermatrix(lu,v3);    

    daxpy_(&n,deins_,v3,eins_,v2,eins_);
 
    norm2 = sqrt(fabs(ddot_(&m,v2,eins_,v,eins_)));
    tmp = dnrm2_(&m,v2,eins_);
    if(tmp>1e-64){
      tmp = 1.0/tmp;
      dscal_(&m,&tmp,v2,eins_);
      dcopy_(&m,v2,eins_,v,eins_);
    }else{
      i=nr;
    }
  }
  free(v);
  free(v2);
  free(w);
  free(v3);
  return norm2;
}

double 
norm2lu2_supermatrix(psupermatrix lu, psupermatrix s){
  double norm2 = 0.0;
  double tmp = 0.0;
  double *v, *w, *v2, *v3;
  int i,n,m;
  int nr = 10;
  assert(s != 0x0);
  assert(lu != 0x0);
  n = lu->rows;
  m = lu->cols;
  assert(s->rows == m);
  assert(s->cols == n);

  v = (double*) malloc(n*sizeof(double));
  w = (double*) malloc(m*sizeof(double));
  v2 = (double*) malloc(n*sizeof(double));
  v3 = (double*) malloc(n*sizeof(double));
  assert(v!=0x0 && w!=0x0 && v2!=0x0 && v3!=0x0);
  for(i=0; i<n; i++){
    v[i] = ((double) (random()%1024))/1024.0;
  }
  tmp = 1.0/dnrm2_(&n,v,eins_);
  dscal_(&n,&tmp,v,eins_);
  for(i=0; i<nr; i++){

    dcopy_(&n,v,eins_,v2,eins_);
    dcopy_(&n,v,eins_,w,eins_);
    solvelower_supermatrix(lu,w);    
    solveupper_supermatrix(lu,w);    

    eval_supermatrix(s,w,v3);

    dcopy_(&n,v,eins_,v2,eins_);
    daxpy_(&n,meins_,v3,eins_,v2,eins_);

    daxpy_(&n,meins_,v,eins_,v3,eins_);

    evaltrans_supermatrix(s,v3,w);

    dcopy_(&m,w,eins_,v3,eins_);
    solveuppertrans_supermatrix(lu,v3);    
    solvelowertrans_supermatrix(lu,v3);    

    daxpy_(&n,deins_,v3,eins_,v2,eins_);
 
    norm2 = sqrt(fabs(ddot_(&m,v2,eins_,v,eins_)));
    tmp = dnrm2_(&m,v2,eins_);
    if(tmp>1e-64){
      tmp = 1.0/tmp;
      dscal_(&m,&tmp,v2,eins_);
      dcopy_(&m,v2,eins_,v,eins_);
    }else{
      i=nr;
    }
  }
  free(v);
  free(v2);
  free(w);
  free(v3);
  return norm2;
}

void
transpose_supermatrix(psupermatrix s, psupermatrix st, int ro, int co){
  int n = s->block_rows;
  int m = s->block_cols;
  int i,j,rof,cof;
  double tmp;
  if(s->s!=0x0){
    assert(st->s!=0x0);
    rof=ro;
    for(i=0; i<n; i++){
      cof=co;
      for(j=0; j<m; j++){
	if(rof<=cof)
	  transpose_supermatrix(s->s[i+j*n],st->s[j+i*m],rof,cof);
	cof += s->s[i+j*n]->cols;
      }
      rof += s->s[i]->rows;
    }
  }else{
    if(s->f!=0x0){
      assert(st->f!=0x0);
      n = s->rows;
      m = s->cols;
      for(i=0; i<n; i++){
	for(j=0; j<m && (ro<co || j<i); j++){
	  tmp = st->f->e[j+i*m];
	  st->f->e[j+i*m] = s->f->e[i+j*n];
	  s->f->e[i+j*n] = tmp;
	}
      }
    }else{
      assert(st->r!=0x0);
      n = s->rows * s->r->kt;
      m = s->cols * s->r->kt;
      if(st->r->k<s->r->kt){
	reallocstructure_rkmatrix(st->r,s->r->kt);
      }else{
	if(s->r->k<st->r->kt) reallocstructure_rkmatrix(s->r,st->r->kt);
      }
      if(s->r->kt>st->r->kt){
	st->r->kt = s->r->kt;
      }else{
	s->r->kt = st->r->kt;
      }
      dswap_(&n,s->r->a,eins_,st->r->b,eins_);
      if(ro!=co)
	dswap_(&m,s->r->b,eins_,st->r->a,eins_);
    }
  }
}

void
copyup_supermatrix(psupermatrix s, psupermatrix st, int ro, int co){
  int n = s->block_rows;
  int m = s->block_cols;
  int i,j,rof,cof;
  if(s->s!=0x0){
    assert(st->s!=0x0);
    rof=ro;
    for(i=0; i<n; i++){
      cof=co;
      for(j=0; j<m; j++){
	if(rof>=cof)
	  copyup_supermatrix(s->s[i+j*n],st->s[j+i*m],rof,cof);
	cof += s->s[i+j*n]->cols;
      }
      rof += s->s[i]->rows;
    }
  }else{
    if(s->f!=0x0){
      assert(st->f!=0x0);
      n = s->rows;
      m = s->cols;
      for(i=0; i<n; i++){
	for(j=0; j<m && (ro<co || j<i); j++){
	  st->f->e[j+i*m] = s->f->e[i+j*n];
	}
      }
    }else{
      assert(st->r!=0x0);
      n = s->rows * s->r->kt;
      m = s->cols * s->r->kt;
      if(st->r->k<s->r->kt){
	reallocstructure_rkmatrix(st->r,s->r->kt);
      }else{
	if(s->r->k<st->r->kt) reallocstructure_rkmatrix(s->r,st->r->kt);
      }
      if(s->r->kt>st->r->kt){
	st->r->kt = s->r->kt;
      }else{
	s->r->kt = st->r->kt;
      }
      dswap_(&n,s->r->a,eins_,st->r->b,eins_);
      if(ro!=co)
	dswap_(&m,s->r->b,eins_,st->r->a,eins_);
    }
  }
}

void
eigenbasis_supermatrix(psupermatrix s, pfullmatrix ev, double* ew){
  pfullmatrix f;
  double* ewi;
  int info;
  int lwork;
  double* work;
  assert(s != 0x0); 
  assert(ew != 0x0);
  assert(ev != 0x0);
  f = new_fullmatrix(s->rows,s->cols);
  ewi = (double*) malloc(s->rows * sizeof(double));
  assert(ewi != 0x0);

  convertsuper2_fullmatrix(f,s);
  lwork = f->rows*f->cols*8;
  work = (double*) malloc(lwork * sizeof(double));
  assert(work != 0x0);
  dgeev_("N","V",&f->rows,f->e,&f->rows,ew,ewi,ev->e,&f->rows,
	 ev->e,&f->rows,work,&lwork,&info);
  if(info!=0){
    printf("Error in eigenbasis_supermatrix: info=%d\n",info);
  }
  del_fullmatrix(f);
  free(work);
  free(ewi);
}
