
/* ------------------------------------------------------------
   This is the file "supermatrix.h" of the HLib package.

   Copyright 1999 Lars Grasedyck and Steffen Boerm
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#ifndef SuperMatrix
#define SuperMatrix

typedef struct _supermatrix supermatrix;
typedef supermatrix *psupermatrix;
typedef const supermatrix *pcsupermatrix;

#include "basic.h"
#include "fullmatrix.h"
#include "rkmatrix.h"
#include "uniformmatrix.h"
#include "sparsematrix.h"

#include <stdio.h>
#include <stdlib.h>

/* a supermatrix is a rows times cols matrix consisting of
   block_rows times block_cols submatrices. In some hierarchy
   l is the level. The array s contains the pointers to the
   submatrices which are again of supermatrix structure.
   If there is only one submatrix, then
   this may be a fullmatrix, rkmatrix or uniformmatrix instead
   of a supermatrix and s is a nullpointer */

struct _supermatrix {
  int rows;              
  int cols;
  int block_rows;         
  int block_cols;
  int l;              
  pclusterbasis row;
  pclusterbasis col;
  puniformmatrix u;   /* if u is not 0 this is a uniform matrix */
  prkmatrix r;        /* if r is not 0 this is a rank-k matrix */
  pfullmatrix f;      /* if f is not 0 this is a full matrix */
  psupermatrix* s;    /* usually s is not 0 and this is indeed a supermatrix */
};

/* Summary :
   1. General subroutines for the initialisation
   2. Constructor, Destructor, Copyconstructor
   3. Evaluate, Add, Scale, Multiply, Invert
   4. Depth, sparsity and idempotency constant
   5. Output to Postscript
*/

/*************************************************/
/*                                               */
/* 1. General subroutines for the initialisation */
/*                                               */
/*************************************************/

/* before using supermatrices, initialize them, 
   verbosity=0 => no output */
void
init_supermatrices(int verbosity);

/* when done using them call the destructor */
void
finish_supermatrices(int verbosity);

/* set the local error for the adaptive rkmatrix
   arithmetics to eps in each rkmatrix block of s.
   By default eps = machine_eps. In order to use
   adaptive arithmetics instead of fixed rank 
   (formatted) arithmetics one has to call
   rk_arithmetics_set_adaptive().
*/
void
seterror_supermatrix(psupermatrix s, double eps);

/*************************************************/
/*                                               */
/* 2. Constructor, Destructor, Copyconstructor   */
/*                                               */
/*************************************************/

/* create a new supermatrix of block_rows_new times block_cols_new 
   submatrices, if r_new/f_new is not 0 then there are no 
   submatrices and u/r/f is set to u_new/r_new/f_new */
psupermatrix
new_supermatrix(int block_rows, int block_cols, int rows, int cols,
		puniformmatrix u_new, prkmatrix r_new, pfullmatrix f_new);

/* return the size of the supermatrix in bytes */
size_t
getsize_supermatrix(psupermatrix s);

/* return the size of the fullmatrices in a supermatrix in bytes */
size_t
getsizefull_supermatrix(psupermatrix s);

/* get the size of the rkmatrices in a supermatrix in bytes */
size_t
getsizerk_supermatrix(psupermatrix s);

/* create a new supermatrix that has the same content as s */
psupermatrix
copy_supermatrix(psupermatrix s);

/* create a new supermatrix that has the same content/structure as s 
   but higher rank in the rkmatrix blocks */
psupermatrix
extendedcopy_supermatrix(psupermatrix s, int newrank);

/* create a new supermatrix that has the same structure as s */
psupermatrix
copystructure_supermatrix(psupermatrix s);

/* realloc for larger rank */
void
reallocstructure_supermatrix(psupermatrix s, int new_rank);

/* copy the content of ssource to starget */
void
copycontent_supermatrix(psupermatrix ssource, psupermatrix starget);

/* st := transpose(s), s:=transpose(s) is allowed, call with ro=co=0 */
void
transpose_supermatrix(psupermatrix s, psupermatrix st, int ro, int co);

/* upper(st) := transpose(lower(s)), call with ro=co=0 */
void
copyup_supermatrix(psupermatrix s, psupermatrix st, int ro, int co);

/* the destructor for a supermatrix */
void
del_supermatrix(psupermatrix s);

/*************************************************/
/*                                               */
/* 3. Evaluate, Add, Scale, Multiply, Invert     */
/*                                               */
/*************************************************/

/* fill the supermatrix s with random numbers, in 
   rkmatrix subblocks the rank r->kt is set to r->k */
void
initrandom_supermatrix(psupermatrix s);

/* initialise s by zero */
void
init0_supermatrix(psupermatrix s);

void
setzerolower_supermatrix(psupermatrix s, int rof, int cof);

void
setzeroupper_supermatrix(psupermatrix s, int rof, int cof);

/* w = s*v */
void 
eval_supermatrix(psupermatrix s, double* v, double* w);

/* w += s*v */
void 
addeval_supermatrix(psupermatrix s, double* v, double* w);

/* w = s'*v */
void 
evaltrans_supermatrix(psupermatrix s, double* v, double* w);

/* w += s'*v */
void 
addevaltrans_supermatrix(psupermatrix s, double* v, double* w);

/* w -= s*v */
void
subeval_supermatrix(psupermatrix s, double *v, double *w);

/* w -= s'*v */
void
subevaltrans_supermatrix(psupermatrix s, double* v, double* w);

/* w = alpha * s * v, where
 trans = 1 means the transpose of s is evaluated, 0 = default,
 add   = 1 means w = w + s*v, 0 = default,
 diag  = 1 means the diagonal is constant 1, 0 = default,
 uplo  = 1 means only the upper triangular part is evaluated,
         2 for the lower triangluar part,
	 0 means all of the matrix is evaluated,
*/
void
generaleval_supermatrix(psupermatrix s, int uplo, int trans, int add,
			int diag, double alpha, double *v, double *w);

/* Solve lower-triangular system of equations */
void
lowersolve_supermatrix(psupermatrix s, double *v);

/* Solve upper-triangular system of equations */
void
uppersolve_supermatrix(psupermatrix s, double *v);

/* dateiname is the desired file to read the matrix from */
psupermatrix
read_supermatrix(char* dateiname);

/* print s to the file dateiname */
void
write_supermatrix(char* dateiname, psupermatrix s);
void
writepart_supermatrix(FILE* datei, psupermatrix s, 
		      int row_offset2, int col_offset2);

/* convert s to f */
void
convertsuper2_fullmatrix(pfullmatrix f, psupermatrix s);

/* r = f(row_offset..,col_offset..) */
void
convertfull2_rkmatrix(pfullmatrix f, prkmatrix r, 
		      int row_offset, int col_offset);

/* ft = f(row_offset..,col_offset..) */
void
convertfull2_fullmatrix(pfullmatrix f, pfullmatrix ft, 
			int row_offset, int col_offset);

/* c = f(row_offset..,col_offset..) */
void
convertfull2_supermatrix(pfullmatrix f, psupermatrix c, 
			 int row_offset, int col_offset);

/* rt = r(row_offset..,col_offset..) */
void
convertrk2_rkmatrix(prkmatrix r, prkmatrix rt, int 
		    row_offset, int col_offset);

/* ft = r(row_offset..,col_offset..) */
void
convertrk2_fullmatrix(prkmatrix r, pfullmatrix ft, 
		      int row_offset, int col_offset);

/* c = r(row_offset..,col_offset..) */
void
convertrk2_supermatrix(prkmatrix r, psupermatrix c, 
		       int row_offset, int col_offset);

/* ft += r(row_offset..,col_offset..) */
void
addrk2_fullmatrix(prkmatrix r, pfullmatrix ft, 
		  int  row_offset, int col_offset);

/* c += r(row_offset..,col_offset..) */
void
addrk2_supermatrix(prkmatrix r, psupermatrix c, 
		   int row_offset, int col_offset);
void
addfull2_supermatrix(pfullmatrix f, psupermatrix c, int row_offset, int col_offset);

/* c = a+b where a,b,c are of the same (supermatrix)structure */
void 
add_supermatrix(psupermatrix c, psupermatrix a, psupermatrix b);

/* c += a where a,c are of the same (supermatrix)structure */
void 
addto_supermatrix(psupermatrix c, psupermatrix a);

/* c += a*b where a*b is of compatible structure to c */
void
muladd_supermatrix(psupermatrix c, psupermatrix a, psupermatrix b);

/* c = a*b where a*b is of compatible structure to c */
void 
mul_supermatrix(psupermatrix c, psupermatrix a, psupermatrix b);

/* c += scale_ab * (transpose)a * (transpose)b, compute only 
 upper(1)/lower(2) or both(0) parts of c*/
void
generalmuladd_supermatrix(psupermatrix c, psupermatrix a, psupermatrix b,
			  double scale_ab, int transpose_a, int transpose_b,
			  int uplow_or_all);
void
generalmul_supermatrix(psupermatrix c, psupermatrix a, psupermatrix b,
		       double scale_ab, int transpose_a, int transpose_b,
		       int uplow_or_all);

/* copy content of s into sc and truncate where necessary */
void
copydata_supermatrix(psupermatrix s, psupermatrix sc);

/* s *= value */
void
scale_supermatrix(psupermatrix s, double value);

/* s = s - Identity */
void
subid_supermatrix(psupermatrix s);

/* s = s - c*Identity */
void
subcid_supermatrix(psupermatrix s, double c);

/* returnvalue = s[row,col] */
double 
getentry_supermatrix(psupermatrix s, int row, int col);

/* s[row,col] = value */
void 
setentry_supermatrix(psupermatrix s, int row, int col, double value);

/* trace (=Sum(s_ii)) of the supermatrix */
double
trace_supermatrix(psupermatrix s);

/* si = inverse(s), s and si and swork are used as 
   workspace (same size), si and swork are assumed 
   to be initialized to zero */
void
invert_supermatrix(psupermatrix si, psupermatrix s, psupermatrix swork);

/* returnvalue = |s|_{F} */
double 
normfrobenius_supermatrix(psupermatrix s);

/* returnvalue = estimate for |s|_{2} after nr steps of the poweriteration */
double 
norm2nr_supermatrix(psupermatrix s, int nr);
/* same as above with nr=10 */
double 
norm2_supermatrix(psupermatrix s);

/* returnvalue = estimate for |si*s - I|_{2}, nr steps of poweriteration */
double 
norm2nrprodminusid_supermatrix(psupermatrix si,psupermatrix s, int nr);
/* same as above with nr=10 */
double 
norm2prodminusid_supermatrix(psupermatrix si,psupermatrix s);

/*returnvalue = estimate for |si*s - I|_{2}, nr steps of poweriteration */
double 
norm2prodminusid2_supermatrix(psupermatrix si, psparsematrix s);
/* same as above with nr=10 */
double 
norm2nrprodminusid2_supermatrix(psupermatrix si, psparsematrix s, int nr);

/* returnvalue = estimate for |s1-s2|_{2}, nr steps of poweriteration */
double 
norm2nrdiff_supermatrix(psupermatrix s1, psupermatrix s2, int nr);
/* same as above with nr=10 */
double 
norm2diff_supermatrix(psupermatrix s1, psupermatrix s2);
double 
norm2nrludiff_supermatrix(psupermatrix lu, psupermatrix s, int nr);
/* same as above with nr=10 */
double 
norm2ludiff_supermatrix(psupermatrix lu, psupermatrix s);

/* returnvalue = estimate for |s^-1|_{2}, unusable for illcond. matrices */
double 
norm2inverse_supermatrix(psupermatrix s);

/* returnvalue = estimate for |(l * u)^-1|_{2}, 10 steps of poweriteration,
 where the two factors of an LU-decomposition are stored in l */
double 
norm2luinverse_supermatrix(psupermatrix l);

/* returnvalue = estimate for |(l * l^T)^-1|_{2}, 10 steps of poweriteration,
 where l is regular lower triangular */
double 
norm2choleskyinverse_supermatrix(psupermatrix l);


/********************************************/
/*  Subroutines for Cholesky Decomposition  */
/********************************************/

/* compute the Cholesky decomposition */
void
choleskydecomposition_supermatrix(psupermatrix a);

/* solve for a decomposed matrix */
void
solvecholesky_supermatrix(psupermatrix l, double *v);

/* service subroutines */
void
solvecholeskyvector_supermatrix(psupermatrix l, double *v);
void
solvecholeskyvectortrans_supermatrix(psupermatrix l, double *v);
void
solveleftcholesky_supermatrix(psupermatrix l, psupermatrix a);
double 
norm2cholesky2_supermatrix(psupermatrix l, psupermatrix s);
double 
norm2choleskydiff_supermatrix(psupermatrix l, psupermatrix s);
double 
norm2choleskyh2_supermatrix(psupermatrix l, psupermatrix s);
double 
norm2cholesky_sparsematrix(psupermatrix l, psparsematrix s);


/******************************************/
/*    Subroutines for LU Decomposition    */
/******************************************/

/* compute the LU decomposition */
void
ludecomposition_supermatrix(psupermatrix a);

/* solve for a decomposed matrix */
void
solvelu_supermatrix(psupermatrix lu, double *v);

/* service subroutines */
void
solvelower_supermatrix(psupermatrix a, double *v);
void
solvelowertrans_supermatrix(psupermatrix a, double *v);
void
solveupper_supermatrix(psupermatrix a, double *v);
void
solveuppertrans_supermatrix(psupermatrix a, double *v);
void
solveleftlu_supermatrix(psupermatrix lu, psupermatrix a);
void
solverightlu_supermatrix(psupermatrix lu, psupermatrix a);
double 
norm2lu_supermatrix(psupermatrix lu, psparsematrix s);
double 
norm2lu2_supermatrix(psupermatrix lu, psupermatrix s);

/* eigenvectors and ~values */
void
eigenbasis_supermatrix(psupermatrix s, pfullmatrix ev, double* ew);

/*************************************************/
/*                                               */
/* 4. Depth, sparsity and idempotency constant   */
/*                                               */
/*************************************************/

/* returns the depth of the matrix-structure */
int 
depth_supermatrix(psupermatrix s);

/* returns the number of nodes in the tree-structure */
int 
getnrnodes_supermatrix(psupermatrix c);

/* get the sparsity constant of the supermatrix */
int csp_supermatrix(psupermatrix s);

/* get the idempotency constant of the supermatrix */
int cid_supermatrix(psupermatrix s);

/*************************************************/
/*                                               */
/* 5. Output to Postscript                       */
/*                                               */
/*************************************************/

/* postscript ouput of a blockwise svd of s to file
   'dateiname' with extraoffset pixels offset in
   the lower left corner 
   (for printing extraoffset = 50) */
void
outputsvd_supermatrix(psupermatrix s, char* dateiname, int extraoffset);

/* postscript ouput of the block-structure of s to file
   'dateiname' with extraoffset pixels offset in
   the lower left corner (for printing extraoffset = 50),
   if showrank>0 then r->kt is printed in each rkmatrix block
*/
void
outputrank_supermatrix(psupermatrix s, char* dateiname, 
		       int showrank, int extraoffset);

/* print the supermatrix to a postscript file.
   s          : supermatrix to be printed
   datei      : filename
   rank       : if rank > 0 then output r->kt
   [u,r,f]fill: 0-> filled square, 1-> output svd, 
                2-> output detailed svd with axis description
   border     : if border > 0 then draw a block border around the block
   tol        : if fill>1 then mark 'tol' at the y-axis
   extraoffset: offset in the lower left corner (for printing 50)
*/
void
output_supermatrix(psupermatrix s, char* dateiname, pblockcluster bc,
		   int rank, int ufill, int rfill, int ffill, 
		   int border, double tol, int extraoffset);

/* determine the rank-k-block with the largest k-th singular value,
   stores the singular value in maxerror and the block in r */
void
determineworstblock_supermatrix(psupermatrix s, double* maxerror, 
				psupermatrix* r, int k);

/* detailed postscript output of the rank-k-block with the largest 
   k-th singular value, at the y-axis the value tolerance is marked.
   The filename is given in 'dateiname'. */
void
outputworstblock_supermatrix(psupermatrix s, char* dateiname, 
			     int k, double tolerance);

/* Print the block partition of a supermatrix and color the blocks
   according to the error compared to a full matrix
   s: supermatrix to be printed
   f: fullmatrix to be used in the computation of the error
   dateiname: filename */
void
outputerror_supermatrix(psupermatrix s, pfullmatrix f, char *dateiname);

#endif
