
/* ------------------------------------------------------------
   This is the file "surfacebem.c" of the HLib package.

   Copyright 1999 Steffen Boerm and Lars Grasedyck
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#include "aca.h"
#include "basic.h"
#include "lapack.h"
#include "surfacebem.h"

#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

/* Are we going to use Sauter's transforming quadrature rules? */
#define SAUTER

static const int not_in_list = -2;
static const int end_of_list = -1;

static double
slp_kernel(double x0, double x1, double x2,
	   double y0, double y1, double y2)
{
  double dist = sqrt(dsqr(x0-y0) + dsqr(x1-y1) + dsqr(x2-y2));

  return 1.0 / (4.0 * M_PI * dist);
}

static double
dlp_kernel(double x0, double x1, double x2,
	   double y0, double y1, double y2,
	   double n0, double n1, double n2)
{
  double dx0, dx1, dx2, dist2, dist;

  dx0 = x0 - y0;
  dx1 = x1 - y1;
  dx2 = x2 - y2;
  dist2 = dsqr(dx0) + dsqr(dx1) + dsqr(dx2);
  dist = sqrt(dist2);

  return (dx0 * n0 + dx1 * n1 + dx2 * n2) / (4.0 * M_PI * dist2 * dist);
}

static void
integrate_nearfield_slp(pccluster row, pccluster col,
			psurfacebemfactory bfactory, double *X, int ldX)
{
  const double (*x)[3] = (const double (*)[3]) bfactory->gr->x;
  const int (*t)[3] = (const int (*)[3]) bfactory->gr->t;
  const double *gdet = bfactory->gr->gdet;
  double *qbuf = bfactory->qbuf;
  const int *dof2tri = bfactory->dof2tri;
  const int *triangle = bfactory->triangle;
  const int *tri2dof = bfactory->tri2dof;
  const int *dof = bfactory->dof;
  int ld_rows = bfactory->ld_rows;
  int ld_cols = bfactory->ld_cols;
  int *arow = bfactory->arow;
  int *acol = bfactory->acol;
  int *row_dof = bfactory->row_dof;
  int *col_dof = bfactory->col_dof;
  int first_row, first_col;
  const double (*xq)[2], (*yq)[2], *wq;
  int ei[3], ej[3], pi[3], pj[3], fi[3], fj[3];
  int nq;
  double xx[3], yy[3];
  double a0, a1, a2, b0, b1, b2;
  int i, j, i1, j1, i2, j2;
  int type;
  int q;

  assert(ldX >= row->size);

  for(j=0; j<col->size; j++)
    for(i=0; i<row->size; i++)
      X[i+j*ldX] = 0.0;

  first_row = end_of_list;
  for(i=0; i<row->size; i++) {
    i1 = i+row->start;
    for(j=dof2tri[i1]; j<dof2tri[i1+1]; j++) {
      j1 = triangle[j];
      if(arow[j1] == not_in_list) {
	arow[j1] = first_row;
	first_row = j1;
      }
    }
  }

  first_col = end_of_list;
  for(i=0; i<col->size; i++) {
    i1 = i+col->start;
    for(j=dof2tri[i1]; j<dof2tri[i1+1]; j++) {
      j1 = triangle[j];
      if(acol[j1] == not_in_list) {
	acol[j1] = first_col;
	first_col = j1;
      }
    }
  }

  for(i=first_row; i>=0; i=arow[i]) {
    ei[0] = t[i][0];
    ei[1] = t[i][1];
    ei[2] = t[i][2];

    for(j=first_col; j>=0; j=acol[j]) {
      ej[0] = t[j][0];
      ej[1] = t[j][1];
      ej[2] = t[j][2];

      type = select_quadrature2d(bfactory->q2, ei, ej, pi, pj,
				 &xq, &yq, &wq, &nq);

      fi[0] = ei[pi[0]];
      fi[1] = ei[pi[1]];
      fi[2] = ei[pi[2]];

      fj[0] = ej[pj[0]];
      fj[1] = ej[pj[1]];
      fj[2] = ej[pj[2]];

      for(q=0; q<nq; q++) {
	a2 = xq[q][1];
	a1 = xq[q][0];
	a0 = 1.0 - a1 - a2;

	b2 = yq[q][1];
	b1 = yq[q][0];
	b0 = 1.0 - b1 - b2;

	xx[0] = a0 * x[fi[0]][0] + a1 * x[fi[1]][0] + a2 * x[fi[2]][0];
	xx[1] = a0 * x[fi[0]][1] + a1 * x[fi[1]][1] + a2 * x[fi[2]][1];
	xx[2] = a0 * x[fi[0]][2] + a1 * x[fi[1]][2] + a2 * x[fi[2]][2];

	yy[0] = b0 * x[fj[0]][0] + b1 * x[fj[1]][0] + b2 * x[fj[2]][0];
	yy[1] = b0 * x[fj[0]][1] + b1 * x[fj[1]][1] + b2 * x[fj[2]][1];
	yy[2] = b0 * x[fj[0]][2] + b1 * x[fj[1]][2] + b2 * x[fj[2]][2];

	qbuf[q] = slp_kernel(xx[0], xx[1], xx[2], yy[0], yy[1], yy[2]);
      }

      assert(tri2dof[i+1] - tri2dof[i] == ld_rows);
      assert(tri2dof[j+1] - tri2dof[j] == ld_cols);

      if(bfactory->permutation) {
	bfactory->permutation(dof + tri2dof[i], pi, row_dof);
	bfactory->permutation(dof + tri2dof[j], pj, col_dof);
      }
      else {
	(void) memcpy(row_dof, dof + tri2dof[i], (size_t) sizeof(int) * ld_rows);
	(void) memcpy(col_dof, dof + tri2dof[j], (size_t) sizeof(int) * ld_cols);
      }

      for(j1=0; j1<ld_cols; j1++) {
	j2 = col_dof[j1] - col->start;
	for(i1=0; i1<ld_rows; i1++) {
	  i2 = row_dof[i1] - row->start;
	  if(0 <= i2 && i2 < row->size && 0 <= j2 && j2 < col->size)
	    X[i2 + j2*ldX] += gdet[i] * gdet[j] * ddot_(&nq, qbuf, eins_,
							wq, eins_);
	  wq += nq;
	}
      }
    }
  }

  while(first_col != end_of_list) {
    i = acol[first_col];
    acol[first_col] = not_in_list;
    first_col = i;
  }

  while(first_row != end_of_list) {
    i = arow[first_row];
    arow[first_row] = not_in_list;
    first_row = i;
  }
}

static void
integrate_nearfield_dlp(pccluster row, pccluster col,
			psurfacebemfactory bfactory, double *X, int ldX)
{
  const double (*x)[3] = (const double (*)[3]) bfactory->gr->x;
  const int (*t)[3] = (const int (*)[3]) bfactory->gr->t;
  const double (*n)[3] = (const double (*)[3]) bfactory->gr->n;
  const double *gdet = bfactory->gr->gdet;
  double *qbuf = bfactory->qbuf;
  const int *dof2tri = bfactory->dof2tri;
  const int *triangle = bfactory->triangle;
  const int *tri2dof = bfactory->tri2dof;
  const int *dof = bfactory->dof;
  int ld_rows = bfactory->ld_rows;
  int ld_cols = bfactory->ld_cols;
  int *arow = bfactory->arow;
  int *acol = bfactory->acol;
  int *row_dof = bfactory->row_dof;
  int *col_dof = bfactory->col_dof;
  int first_row, first_col;
  const double (*xq)[2], (*yq)[2], *wq, *m;
  int ei[3], ej[3], pi[3], pj[3], fi[3], fj[3];
  int nq;
  double xx[3], yy[3], nn[3];
  double a0, a1, a2, b0, b1, b2;
  int i, j, i1, j1, i2, j2;
  int type;
  int q;

  assert(ldX >= row->size);

  for(j=0; j<col->size; j++)
    for(i=0; i<row->size; i++)
      X[i+j*ldX] = 0.0;

  first_row = end_of_list;
  for(i=0; i<row->size; i++) {
    i1 = i+row->start;
    for(j=dof2tri[i1]; j<dof2tri[i1+1]; j++) {
      j1 = triangle[j];
      if(arow[j1] == not_in_list) {
	arow[j1] = first_row;
	first_row = j1;
      }
    }
  }

  first_col = end_of_list;
  for(i=0; i<col->size; i++) {
    i1 = i+col->start;
    for(j=dof2tri[i1]; j<dof2tri[i1+1]; j++) {
      j1 = triangle[j];
      if(acol[j1] == not_in_list) {
	acol[j1] = first_col;
	first_col = j1;
      }
    }
  }

  for(i=first_row; i>=0; i=arow[i]) {
    ei[0] = t[i][0];
    ei[1] = t[i][1];
    ei[2] = t[i][2];

    for(j=first_col; j>=0; j=acol[j]) {
      ej[0] = t[j][0];
      ej[1] = t[j][1];
      ej[2] = t[j][2];

      nn[0] = n[j][0];
      nn[1] = n[j][1];
      nn[2] = n[j][2];

      type = select_quadrature2d(bfactory->q2, ei, ej, pi, pj,
				 &xq, &yq, &wq, &nq);

      fi[0] = ei[pi[0]];
      fi[1] = ei[pi[1]];
      fi[2] = ei[pi[2]];

      fj[0] = ej[pj[0]];
      fj[1] = ej[pj[1]];
      fj[2] = ej[pj[2]];

      for(q=0; q<nq; q++) {
	a2 = xq[q][1];
	a1 = xq[q][0];
	a0 = 1.0 - a1 - a2;

	b2 = yq[q][1];
	b1 = yq[q][0];
	b0 = 1.0 - b1 - b2;

	xx[0] = a0 * x[fi[0]][0] + a1 * x[fi[1]][0] + a2 * x[fi[2]][0];
	xx[1] = a0 * x[fi[0]][1] + a1 * x[fi[1]][1] + a2 * x[fi[2]][1];
	xx[2] = a0 * x[fi[0]][2] + a1 * x[fi[1]][2] + a2 * x[fi[2]][2];

	yy[0] = b0 * x[fj[0]][0] + b1 * x[fj[1]][0] + b2 * x[fj[2]][0];
	yy[1] = b0 * x[fj[0]][1] + b1 * x[fj[1]][1] + b2 * x[fj[2]][1];
	yy[2] = b0 * x[fj[0]][2] + b1 * x[fj[1]][2] + b2 * x[fj[2]][2];

	qbuf[q] = dlp_kernel(xx[0], xx[1], xx[2],
			     yy[0], yy[1], yy[2],
			     nn[0], nn[1], nn[2]);
      }

      m = bfactory->mass;

      assert(tri2dof[i+1] - tri2dof[i] == ld_rows);
      assert(tri2dof[j+1] - tri2dof[j] == ld_cols);

      if(bfactory->permutation) {
	bfactory->permutation(dof + tri2dof[i], pi, row_dof);
	bfactory->permutation(dof + tri2dof[j], pj, col_dof);
      }
      else {
	(void) memcpy(row_dof, dof + tri2dof[i], (size_t) sizeof(int) * ld_rows);
	(void) memcpy(col_dof, dof + tri2dof[j], (size_t) sizeof(int) * ld_cols);
      }

      for(j1=0; j1<ld_cols; j1++) {
	j2 = col_dof[j1] - col->start;
	for(i1=0; i1<ld_rows; i1++) {
	  i2 = row_dof[i1] - row->start;
	  if(0 <= i2 && i2 < row->size && 0 <= j2 && j2 < col->size) {
	    X[i2 + j2*ldX] += gdet[i] * gdet[j] * ddot_(&nq, qbuf, eins_,
							wq, eins_);
	    if(i == j)
	      X[i2 + j2*ldX] += bfactory->lambda * gdet[i] * (*m);
	  }
	  wq += nq;
	  m++;
	}
      }
    }
  }

  while(first_col != end_of_list) {
    i = acol[first_col];
    acol[first_col] = not_in_list;
    first_col = i;
  }

  while(first_row != end_of_list) {
    i = arow[first_row];
    arow[first_row] = not_in_list;
    first_row = i;
  }
}

static void
transform_interpolation_points(pccluster tau, psurfacebemfactory bfactory,
			       double (*l)[3])
{
  const double *xp = bfactory->xp;
  int p = bfactory->p;
  double mx, my, mz, dx, dy, dz;
  int i;

  mx = 0.5 * (tau->bmax[0] + tau->bmin[0]);
  my = 0.5 * (tau->bmax[1] + tau->bmin[1]);
  mz = 0.5 * (tau->bmax[2] + tau->bmin[2]);
  dx = 0.5 * (tau->bmax[0] - tau->bmin[0]);
  dy = 0.5 * (tau->bmax[1] - tau->bmin[1]);
  dz = 0.5 * (tau->bmax[2] - tau->bmin[2]);

  for(i=0; i<p; i++) {
    l[i][0] = mx + xp[i] * dx;
    l[i][1] = my + xp[i] * dy;
    l[i][2] = mz + xp[i] * dz;
  }
}

static void
integrate_slp(pccluster tau, pccluster sigma, psurfacebemfactory bfactory,
	      double *X, int ldX)
{
  const double (*x)[3] = (const double (*)[3]) bfactory->gr->x;
  const int (*t)[3] = (const int (*)[3]) bfactory->gr->t;
  const double *gdet = bfactory->gr->gdet;
  double *qbuf = bfactory->qbuf;
  const int *dof2tri = bfactory->dof2tri;
  const int *triangle = bfactory->triangle;
  const int *tri2dof = bfactory->tri2dof;
  const int *dof = bfactory->dof;
  int *atau = bfactory->arow;
  int first_tau;
  double (*l)[3] = bfactory->l;
  int p = bfactory->p;
  const double (*xq)[2] = (const double (*)[2]) bfactory->xq;
  const double *wq;
  int q = bfactory->q;
  double xx[3];
  double a0, a1, a2;
  int i, j, i1, i2, jx, jy, jz, k;

  assert(ldX >= tau->size);

  for(j=0; j<p*p*p; j++)
    for(i=0; i<tau->size; i++)
      X[i+j*ldX] = 0.0;
  
  first_tau = end_of_list;
  for(i=0; i<tau->size; i++) {
    i1 = i+tau->start;
    for(i2=dof2tri[i1]; i2<dof2tri[i1+1]; i2++) {
      k = triangle[i2];
      if(atau[k] == not_in_list) {
	atau[k] = first_tau;
	first_tau = k;
      }
    }
  }

  transform_interpolation_points(sigma, bfactory, l);

  for(i=first_tau; i>=0; i=atau[i]) {
    for(jx=0; jx<p; jx++)
      for(jy=0; jy<p; jy++)
	for(jz=0; jz<p; jz++) {
	  j = jx + p*(jy + p*jz);

	  for(k=0; k<q; k++) {
	    a2 = xq[k][1];
	    a1 = xq[k][0];
	    a0 = 1.0 - a1 - a2;

	    xx[0] = a0 * x[t[i][0]][0] + a1 * x[t[i][1]][0] + a2 * x[t[i][2]][0];
	    xx[1] = a0 * x[t[i][0]][1] + a1 * x[t[i][1]][1] + a2 * x[t[i][2]][1];
	    xx[2] = a0 * x[t[i][0]][2] + a1 * x[t[i][1]][2] + a2 * x[t[i][2]][2];

	    qbuf[k] = slp_kernel(l[jx][0], l[jy][1], l[jz][2],
				 xx[0], xx[1], xx[2]);
	  }

	  wq = bfactory->wq;
	  for(i1=tri2dof[i]; i1<tri2dof[i+1]; i1++) {
	    i2 = dof[i1] - tau->start;
	    if(0 <= i2 && i2 < tau->size)
	      X[i2 + j*ldX] += gdet[i] * ddot_(&q, qbuf, eins_,
					       wq, eins_);
	    wq += q;
	  }
	}
  }

  while(first_tau != end_of_list) {
    i = atau[first_tau];
    atau[first_tau] = not_in_list;
    first_tau = i;
  }
}

static void
integrate_dlp(pccluster tau, pccluster sigma, psurfacebemfactory bfactory,
	      double *X, int ldX)
{
  const double (*x)[3] = (const double (*)[3]) bfactory->gr->x;
  const int (*t)[3] = (const int (*)[3]) bfactory->gr->t;
  const double (*n)[3] = (const double (*)[3]) bfactory->gr->n;
  const double *gdet = bfactory->gr->gdet;
  double *qbuf = bfactory->qbuf;
  const int *dof2tri = bfactory->dof2tri;
  const int *triangle = bfactory->triangle;
  const int *tri2dof = bfactory->tri2dof;
  const int *dof = bfactory->dof;
  int *atau = bfactory->arow;
  int first_tau;
  double (*l)[3] = bfactory->l;
  int p = bfactory->p;
  const double (*xq)[2] = (const double (*)[2]) bfactory->xq;
  const double *wq;
  int q = bfactory->q;
  int size = tau->size;
  double xx[3], nn[3];
  double a0, a1, a2;
  int i, j, i1, i2, jx, jy, jz, k;

  assert(ldX == size);
  
  for(j=0; j<p*p*p; j++)
    for(i=0; i<size; i++)
      X[i+j*ldX] = 0.0;
  
  first_tau = end_of_list;
  for(i=0; i<size; i++) {
    i1 = i+tau->start;
    for(i2=dof2tri[i1]; i2<dof2tri[i1+1]; i2++) {
      k = triangle[i2];
      if(atau[k] == not_in_list) {
	atau[k] = first_tau;
	first_tau = k;
      }
    }
  }

  transform_interpolation_points(sigma, bfactory, l);

  for(i=first_tau; i>=0; i=atau[i]) {
    nn[0] = n[i][0];
    nn[1] = n[i][1];
    nn[2] = n[i][2];

    for(jx=0; jx<p; jx++)
      for(jy=0; jy<p; jy++)
	for(jz=0; jz<p; jz++) {
	  j = jx + p*(jy + p*jz);

	  for(k=0; k<q; k++) {
	    a2 = xq[k][1];
	    a1 = xq[k][0];
	    a0 = 1.0 - a1 - a2;

	    xx[0] = a0 * x[t[i][0]][0] + a1 * x[t[i][1]][0] + a2 * x[t[i][2]][0];
	    xx[1] = a0 * x[t[i][0]][1] + a1 * x[t[i][1]][1] + a2 * x[t[i][2]][1];
	    xx[2] = a0 * x[t[i][0]][2] + a1 * x[t[i][1]][2] + a2 * x[t[i][2]][2];

	    qbuf[k] = dlp_kernel(l[jx][0], l[jy][1], l[jz][2],
				 xx[0], xx[1], xx[2],
				 nn[0], nn[1], nn[2]);
	  }

	  wq = bfactory->wq;
	  for(i1=tri2dof[i]; i1<tri2dof[i+1]; i1++) {
	    i2 = dof[i1] - tau->start;
	    if(0 <= i2 && i2 < size)
	      X[i2 + j*ldX] += gdet[i] * ddot_(&q, qbuf, eins_,
					       wq, eins_);
	    wq += q;
	  }
	}
  }

  while(first_tau != end_of_list) {
    i = atau[first_tau];
    atau[first_tau] = not_in_list;
    first_tau = i;
  }
}

static void
integrate_polynomial(pccluster tau, psurfacebemfactory bfactory,
		     double *X, int ldX)
{
  const double (*x)[3] = (const double (*)[3]) bfactory->gr->x;
  const int (*t)[3] = (const int (*)[3]) bfactory->gr->t;
  const double *gdet = bfactory->gr->gdet;
  double *qbuf = bfactory->qbuf;
  const int *dof2tri = bfactory->dof2tri;
  const int *triangle = bfactory->triangle;
  const int *tri2dof = bfactory->tri2dof;
  const int *dof = bfactory->dof;
  int *atau = bfactory->arow;
  int first_tau;
  double (*l)[3] = bfactory->l;
  int p = bfactory->p;
  const double (*xq)[2] = (const double (*)[2]) bfactory->xq;
  const double *wq;
  int q = bfactory->q;
  int size = tau->size;
  double xx[3];
  double a0, a1, a2, poly;
  int i, j, i1, i2, jx, jy, jz, k, m;

  assert(ldX == size);

  for(j=0; j<p*p*p; j++)
    for(i=0; i<tau->size; i++)
      X[i+j*ldX] = 0.0;
  
  first_tau = end_of_list;
  for(i=0; i<size; i++) {
    i1 = i+tau->start;
    for(i2=dof2tri[i1]; i2<dof2tri[i1+1]; i2++) {
      k = triangle[i2];
      if(atau[k] == not_in_list) {
	atau[k] = first_tau;
	first_tau = k;
      }
    }
  }

  transform_interpolation_points(tau, bfactory, l);

  for(i=first_tau; i>=0; i=atau[i]) {
    for(jx=0; jx<p; jx++)
      for(jy=0; jy<p; jy++)
	for(jz=0; jz<p; jz++) {
	  j = jx + p*(jy + p*jz);

	  for(k=0; k<q; k++) {
	    a2 = xq[k][1];
	    a1 = xq[k][0];
	    a0 = 1.0 - a1 - a2;

	    xx[0] = a0 * x[t[i][0]][0] + a1 * x[t[i][1]][0] + a2 * x[t[i][2]][0];
	    xx[1] = a0 * x[t[i][0]][1] + a1 * x[t[i][1]][1] + a2 * x[t[i][2]][1];
	    xx[2] = a0 * x[t[i][0]][2] + a1 * x[t[i][1]][2] + a2 * x[t[i][2]][2];

	    poly = 1.0;
	    for(m=0; m<p; m++)
	      if(m != jx)
		poly *= (xx[0] - l[m][0]) / (l[jx][0] - l[m][0]);
	    for(m=0; m<p; m++)
	      if(m != jy)
		poly *= (xx[1] - l[m][1]) / (l[jy][1] - l[m][1]);
	    for(m=0; m<p; m++)
	      if(m != jz)
		poly *= (xx[2] - l[m][2]) / (l[jz][2] - l[m][2]);

	    qbuf[k] = poly;
	  }

	  wq = bfactory->wq;
	  for(i1=tri2dof[i]; i1<tri2dof[i+1]; i1++) {
	    i2 = dof[i1] - tau->start;
	    if(0 <= i2 && i2 < size)
	      X[i2 + j*ldX] += gdet[i] * ddot_(&q, qbuf, eins_,
					       wq, eins_);
	    wq += q;
	  }
	}
  }

  while(first_tau != end_of_list) {
    i = atau[first_tau];
    atau[first_tau] = not_in_list;
    first_tau = i;
  }
}

static void
integrate_polynomial_dn(pccluster tau, psurfacebemfactory bfactory,
			double *X, int ldX)
{
  const double (*x)[3] = (const double (*)[3]) bfactory->gr->x;
  const int (*t)[3] = (const int (*)[3]) bfactory->gr->t;
  const double (*n)[3] = (const double (*)[3]) bfactory->gr->n;
  const double *gdet = bfactory->gr->gdet;
  double *qbuf = bfactory->qbuf;
  const int *dof2tri = bfactory->dof2tri;
  const int *triangle = bfactory->triangle;
  const int *tri2dof = bfactory->tri2dof;
  const int *dof = bfactory->dof;
  int *atau = bfactory->arow;
  int first_tau;
  double (*l)[3] = bfactory->l;
  int p = bfactory->p;
  const double (*xq)[2] = (const double (*)[2]) bfactory->xq;
  const double *wq = bfactory->wq;
  int q = bfactory->q;
  int size = tau->size;
  double xx[3], nn[3];
  double a0, a1, a2, monomial, dmonomial, poly, dpoly, poly3, dpoly3;
  int i, j, i1, i2, jx, jy, jz, k, m;

  assert(ldX == size);
  
  for(j=0; j<p*p*p; j++)
    for(i=0; i<size; i++)
      X[i+j*ldX] = 0.0;
  
  first_tau = end_of_list;
  for(i=0; i<size; i++) {
    i1 = i+tau->start;
    for(i2=dof2tri[i1]; i2<dof2tri[i1+1]; i2++) {
      k = triangle[i2];
      if(atau[k] == not_in_list) {
	atau[k] = first_tau;
	first_tau = k;
      }
    }
  }

  transform_interpolation_points(tau, bfactory, l);

  for(i=first_tau; i>=0; i=atau[i]) {
    nn[0] = n[i][0];
    nn[1] = n[i][1];
    nn[2] = n[i][2];

    for(jx=0; jx<p; jx++)
      for(jy=0; jy<p; jy++)
	for(jz=0; jz<p; jz++) {
	  j = jx + p*(jy + p*jz);

	  for(k=0; k<q; k++) {
	    a2 = xq[k][1];
	    a1 = xq[k][0];
	    a0 = 1.0 - a1 - a2;

	    xx[0] = a0 * x[t[i][0]][0] + a1 * x[t[i][1]][0] + a2 * x[t[i][2]][0];
	    xx[1] = a0 * x[t[i][0]][1] + a1 * x[t[i][1]][1] + a2 * x[t[i][2]][1];
	    xx[2] = a0 * x[t[i][0]][2] + a1 * x[t[i][1]][2] + a2 * x[t[i][2]][2];

	    poly3 = 1.0;
	    dpoly3 = 0.0;

	    poly = 1.0;
	    dpoly = 0.0;
	    for(m=0; m<p; m++)
	      if(m != jx) {
		dmonomial = 1.0 / (l[jx][0] - l[m][0]);
		monomial = (xx[0] - l[m][0]) * dmonomial;
		dpoly = dpoly * monomial + poly * dmonomial;
		poly *= monomial;
	      }
	    dpoly3 = dpoly3 * poly + poly3 * dpoly * nn[0];
	    poly3 *= poly;

	    poly = 1.0;
	    dpoly = 0.0;
	    for(m=0; m<p; m++)
	      if(m != jy) {
		dmonomial = 1.0 / (l[jy][1] - l[m][1]);
		monomial = (xx[1] - l[m][1]) * dmonomial;
		dpoly = dpoly * monomial + poly * dmonomial;
		poly *= monomial;
	      }
	    dpoly3 = dpoly3 * poly + poly3 * dpoly * nn[1];
	    poly3 *= poly;
	    
	    poly = 1.0;
	    dpoly = 0.0;
	    for(m=0; m<p; m++)
	      if(m != jz) {
		dmonomial = 1.0 / (l[jz][2] - l[m][2]);
		monomial = (xx[2] - l[m][2]) * dmonomial;
		dpoly = dpoly * monomial + poly * dmonomial;
		poly *= monomial;
	      }
	    dpoly3 = dpoly3 * poly + poly3 * dpoly * nn[2];
	    poly3 *= poly;

	    qbuf[k] = dpoly3;
	  }

	  wq = bfactory->wq;
	  for(i1=tri2dof[i]; i1<tri2dof[i+1]; i1++) {
	    i2 = dof[i1] - tau->start;
	    if(0 <= i2 && i2 < size)
	      X[i2 + j*ldX] += gdet[i] * ddot_(&q, qbuf, eins_,
					       wq, eins_);
	    wq += q;
	  }
	}
  }

  while(first_tau != end_of_list) {
    i = atau[first_tau];
    atau[first_tau] = not_in_list;
    first_tau = i;
  }
}

static void
transfer_lagrange(pccluster trg, pccluster src,
		  psurfacebemfactory bfactory,
		  double *T, int ldT)
{
  double (*l)[3] = bfactory->l;
  double (*l2)[3] = bfactory->l2;
  int p = bfactory->p;
  double polyx, polyy, polyz;
  int ix, iy, iz, i;
  int jx, jy, jz, j;
  int r;

  assert(p*p*p <= ldT);
  
  transform_interpolation_points(trg, bfactory, l);
  transform_interpolation_points(src, bfactory, l2);

  for(ix=0; ix<p; ix++)
    for(jx=0; jx<p; jx++) {
      polyx = 1.0;
      for(r=0; r<p; r++)
	if(r != ix)
	  polyx *= (l[jx][0] - l2[r][0]) / (l2[ix][0] - l2[r][0]);
      for(iy=0; iy<p; iy++)
	for(jy=0; jy<p; jy++) {
	  polyy = polyx;
	  for(r=0; r<p; r++)
	    if(r != iy)
	      polyy *= (l[jy][1] - l2[r][1]) / (l2[iy][1] - l2[r][1]);
	  for(iz=0; iz<p; iz++)
	    for(jz=0; jz<p; jz++) {
	      polyz = polyy;
	      for(r=0; r<p; r++)
		if(r != iz)
		  polyz *= (l[jz][2] - l2[r][2]) / (l2[iz][2] - l2[r][2]);

	      i = ix + p * (iy + p * iz);
	      j = jx + p * (jy + p * jz);
	      T[j+i*ldT] = polyz;
	    }
	}
    }
}

static void
interpolate_kernel(pccluster row, pccluster col,
		   psurfacebemfactory bfactory,
		   double *S, int ldS)
{
  double (*l)[3] = bfactory->l;
  double (*l2)[3] = bfactory->l2;
  int p = bfactory->p;
  int ix, iy, iz, i;
  int jx, jy, jz, j;

  assert(p*p*p <= ldS);

  transform_interpolation_points(row, bfactory, l);
  transform_interpolation_points(col, bfactory, l2);

  for(ix=0; ix<p; ix++)
    for(iy=0; iy<p; iy++)
      for(iz=0; iz<p; iz++) {
	i = ix + p*(iy + p*iz);
	for(jx=0; jx<p; jx++)
	  for(jy=0; jy<p; jy++)
	    for(jz=0; jz<p; jz++) {
	      j = jx + p*(jy + p*jz);

	      S[i+j*ldS] = slp_kernel(l[ix][0], l[iy][1], l[iz][2],
				      l2[jx][0], l2[jy][1], l2[jz][2]);
	    }
      }
}

static double
entry_slp(int row, int col, void *data) {
  psurfacebemfactory bfactory = (psurfacebemfactory) data;
  const int *index = bfactory->rct->dof2idx;
  const double (*x)[3] = (const double (*)[3]) bfactory->gr->x;
  const int (*t)[3] = (const int (*)[3]) bfactory->gr->t;
  const double *gdet = bfactory->gr->gdet;
  const double (*xq)[2], (*yq)[2], *wq;
  int ei[3], ej[3], pi[3], pj[3], fi[3], fj[3];
  int nq;
  double xx[3], yy[3];
  double sum, a0, a1, a2, b0, b1, b2;
  int ii, jj;
  int type;
  int q;

  assert(0 <= row);
  assert(row < bfactory->gr->triangles);
  assert(0 <= col);
  assert(col < bfactory->gr->triangles);

  ii = index[row];

  ei[0] = t[ii][0];
  ei[1] = t[ii][1];
  ei[2] = t[ii][2];

  jj = index[col];
  
  ej[0] = t[jj][0];
  ej[1] = t[jj][1];
  ej[2] = t[jj][2];
  
  type = select_quadrature2d(bfactory->q2, ei, ej, pi, pj,
			     &xq, &yq, &wq, &nq);
  
  fi[0] = ei[pi[0]];
  fi[1] = ei[pi[1]];
  fi[2] = ei[pi[2]];
  
  fj[0] = ej[pj[0]];
  fj[1] = ej[pj[1]];
  fj[2] = ej[pj[2]];
  
  sum = 0.0;
  for(q=0; q<nq; q++) {
    a2 = xq[q][1];
    a1 = xq[q][0];
    a0 = 1.0 - a1 - a2;
    
    b2 = yq[q][1];
    b1 = yq[q][0];
    b0 = 1.0 - b1 - b2;
    
    xx[0] = a0 * x[fi[0]][0] + a1 * x[fi[1]][0] + a2 * x[fi[2]][0];
    xx[1] = a0 * x[fi[0]][1] + a1 * x[fi[1]][1] + a2 * x[fi[2]][1];
    xx[2] = a0 * x[fi[0]][2] + a1 * x[fi[1]][2] + a2 * x[fi[2]][2];
    
    yy[0] = b0 * x[fj[0]][0] + b1 * x[fj[1]][0] + b2 * x[fj[2]][0];
    yy[1] = b0 * x[fj[0]][1] + b1 * x[fj[1]][1] + b2 * x[fj[2]][1];
    yy[2] = b0 * x[fj[0]][2] + b1 * x[fj[1]][2] + b2 * x[fj[2]][2];
    
    sum += wq[q] * slp_kernel(xx[0], xx[1], xx[2],
			      yy[0], yy[1], yy[2]);
  }

  return sum * gdet[ii] * gdet[jj];
}

static double
entry_dlp(int row, int col, void *data) {
  psurfacebemfactory bfactory = (psurfacebemfactory) data;
  const int *index = bfactory->rct->dof2idx;
  const double (*x)[3] = (const double (*)[3]) bfactory->gr->x;
  const int (*t)[3] = (const int (*)[3]) bfactory->gr->t;
  const double (*n)[3] = (const double (*)[3]) bfactory->gr->n;
  const double *gdet = bfactory->gr->gdet;
  const double (*xq)[2], (*yq)[2], *wq;
  int ei[3], ej[3], pi[3], pj[3], fi[3], fj[3];
  int nq;
  double xx[3], yy[3], nn[3];
  double sum, a0, a1, a2, b0, b1, b2;
  int ii, jj;
  int type;
  int q;

  assert(0 <= row);
  assert(row < bfactory->gr->triangles);
  assert(0 <= col);
  assert(col < bfactory->gr->triangles);

  ii = index[row];
  
  ei[0] = t[ii][0];
  ei[1] = t[ii][1];
  ei[2] = t[ii][2];

  jj = index[col];

  ej[0] = t[jj][0];
  ej[1] = t[jj][1];
  ej[2] = t[jj][2];
  
  nn[0] = n[jj][0];
  nn[1] = n[jj][1];
  nn[2] = n[jj][2];
  
  type = select_quadrature2d(bfactory->q2, ei, ej, pi, pj,
			     &xq, &yq, &wq, &nq);
  
  fi[0] = ei[pi[0]];
  fi[1] = ei[pi[1]];
  fi[2] = ei[pi[2]];
  
  fj[0] = ej[pj[0]];
  fj[1] = ej[pj[1]];
  fj[2] = ej[pj[2]];
  
  sum = 0.0;
  for(q=0; q<nq; q++) {
    a2 = xq[q][1];
    a1 = xq[q][0];
    a0 = 1.0 - a1 - a2;
    
    b2 = yq[q][1];
    b1 = yq[q][0];
    b0 = 1.0 - b1 - b2;
    
    xx[0] = a0 * x[fi[0]][0] + a1 * x[fi[1]][0] + a2 * x[fi[2]][0];
    xx[1] = a0 * x[fi[0]][1] + a1 * x[fi[1]][1] + a2 * x[fi[2]][1];
    xx[2] = a0 * x[fi[0]][2] + a1 * x[fi[1]][2] + a2 * x[fi[2]][2];
    
    yy[0] = b0 * x[fj[0]][0] + b1 * x[fj[1]][0] + b2 * x[fj[2]][0];
    yy[1] = b0 * x[fj[0]][1] + b1 * x[fj[1]][1] + b2 * x[fj[2]][1];
    yy[2] = b0 * x[fj[0]][2] + b1 * x[fj[1]][2] + b2 * x[fj[2]][2];
    
    sum += wq[q] * dlp_kernel(xx[0], xx[1], xx[2],
			      yy[0], yy[1], yy[2],
			      nn[0], nn[1], nn[2]);
  }

  return gdet[ii] * (sum * gdet[jj] +
		     (ii == jj ? bfactory->lambda * 0.5 : 0.0));
}

static void
build_invgraph(int n_src, int n_trg, int nn,
	       const int *s_src, const int *src,
	       int *s_trg, int *trg)
{
  int i, j, k;

  for(i=0; i<n_trg; i++)
    s_trg[i] = 0;

  for(i=0; i<n_src; i++)
    for(j=s_src[i]; j<s_src[i+1]; j++)
      s_trg[src[j]]++;

  for(i=1; i<n_trg; i++)
    s_trg[i] += s_trg[i-1];
  s_trg[n_trg] = s_trg[n_trg-1];
  assert(s_trg[n_trg] <= nn);

  for(i=0; i<n_src; i++)
    for(j=s_src[i]; j<s_src[i+1]; j++) {
      k = --s_trg[src[j]];
      assert(0 <= k && k < nn);
      trg[k] = i;
    }
}

static psurfacebemfactory
common_initialization(pbemgrid3d gr, pcclustertree rct, pcclustertree cct,
		      int q, int p, int n, int ld_rows, int ld_cols)
{
  psurfacebemfactory bfactory;
  int triangles = gr->triangles;
  int i;

  bfactory = (psurfacebemfactory) malloc(sizeof(surfacebemfactory));
  assert(bfactory != NULL);

  bfactory->xp = (double *) malloc((size_t) sizeof(double) * p);
  assert(bfactory->xp != NULL);

  bfactory->l = (double (*)[3]) malloc((size_t) sizeof(double [3]) * p);
  assert(bfactory->l != NULL);

  bfactory->l2 = (double (*)[3]) malloc((size_t) sizeof(double [3]) * p);
  assert(bfactory->l2 != NULL);

  for(i=0; i<p; i++)
    bfactory->xp[i] = cos(M_PI*(2.0*i+1.0)/(2.0*p));
  bfactory->p = p;
  bfactory->lambda = 0.0;

  bfactory->row_dof = (int *) malloc((size_t) sizeof(int) * ld_rows);
  assert(bfactory->row_dof != NULL);
  bfactory->col_dof = (int *) malloc((size_t) sizeof(int) * ld_cols);
  assert(bfactory->col_dof != NULL);

  bfactory->gr = gr;
  bfactory->rct = rct;
  bfactory->cct = cct;
  bfactory->ld_rows = ld_rows;
  bfactory->ld_cols = ld_cols;

  bfactory->mass = (double *) malloc((size_t) sizeof(double) * ld_rows * ld_cols);
  assert(bfactory->mass != NULL);

  bfactory->dof2tri = (int *) malloc((size_t) sizeof(int) * (n+1));
  assert(bfactory->dof2tri != NULL);

  bfactory->triangle = (int *) malloc((size_t) sizeof(int) * ld_rows * ld_cols * triangles);
  assert(bfactory->triangle != NULL);

  bfactory->tri2dof = (int *) malloc((size_t) sizeof(int) * (triangles+1));
  assert(bfactory->triangle != NULL);

  bfactory->dof = (int *) malloc((size_t) sizeof(int) * ld_rows * ld_cols * triangles);
  assert(bfactory->dof != NULL);

  bfactory->arow = (int *) malloc((size_t) sizeof(int) * triangles);
  assert(bfactory->arow != NULL);

  bfactory->acol = (int *) malloc((size_t) sizeof(int) * triangles);
  assert(bfactory->acol != NULL);

  for(i=0; i<triangles; i++)
    bfactory->arow[i] = bfactory->acol[i] = not_in_list;

  bfactory->transfer_row = transfer_lagrange;
  bfactory->transfer_col = transfer_lagrange;
  bfactory->kernel_coefficients = interpolate_kernel;

  return bfactory;
}

psurfacebemfactory
new_surfacebemfactory_slp(pbemgrid3d gr,
			  pcclustertree rct, pcclustertree cct,
			  int q, int p)
{
  psurfacebemfactory bfactory;
  int triangles = gr->triangles;
  int nqmax;
  int i;

  bfactory = common_initialization(gr, rct, cct, q, p, triangles, 1, 1);

  build_triangle_gauss(q, &bfactory->q, &bfactory->xq, &bfactory->wq);

#ifdef SAUTER
  bfactory->q2 = new_quadrature2d_sauter(q);
#else
  bfactory->q2 = new_quadrature2d(q, -1.0);
#endif

  nqmax = imax(imax(bfactory->q2->n_identical, bfactory->q2->n_edge),
	       imax(bfactory->q2->n_vertex, bfactory->q2->n_distant));

  bfactory->qbuf = (double *) malloc((size_t) sizeof(double) * nqmax);
  assert(bfactory->qbuf != NULL);

  for(i=0; i<=triangles; i++)
    bfactory->tri2dof[i] = i;
  (void) memcpy(bfactory->dof, rct->idx2dof, (size_t) sizeof(int) * triangles);
  build_invgraph(triangles, triangles, triangles,
		 bfactory->tri2dof, bfactory->dof,
		 bfactory->dof2tri, bfactory->triangle);

  bfactory->integrate_nearfield = integrate_nearfield_slp;
  bfactory->nearfield_entry = entry_slp;
  bfactory->integrate_row_kernel = integrate_slp;
  bfactory->integrate_row_lagrange = integrate_polynomial;
  bfactory->integrate_col_kernel = integrate_slp;
  bfactory->integrate_col_lagrange = integrate_polynomial;
  bfactory->permutation = NULL;

  bfactory->mass[0] = 0.5;
  bfactory->lambda = 0.0;

  return bfactory;
}

psurfacebemfactory
new_surfacebemfactory_dlp(pbemgrid3d gr,
			  pcclustertree rct, pcclustertree cct,
			  int q, int p, double lambda)
{
  psurfacebemfactory bfactory;
  int triangles = gr->triangles;
  int nqmax;
  int i;

  bfactory = common_initialization(gr, rct, cct, q, p, triangles, 1, 1);

  build_triangle_gauss(q, &bfactory->q, &bfactory->xq, &bfactory->wq);

#ifdef SAUTER
  bfactory->q2 = new_quadrature2d_sauter(q);
#else
  bfactory->q2 = new_quadrature2d_antisymm(q, -2.0);
#endif

  nqmax = imax(imax(bfactory->q2->n_identical, bfactory->q2->n_edge),
	       imax(bfactory->q2->n_vertex, bfactory->q2->n_distant));

  bfactory->qbuf = (double *) malloc((size_t) sizeof(double) * nqmax);
  assert(bfactory->qbuf != NULL);

  for(i=0; i<=triangles; i++)
    bfactory->tri2dof[i] = i;
  (void) memcpy(bfactory->dof, rct->idx2dof, (size_t) sizeof(int) * triangles);
  build_invgraph(triangles, triangles, triangles,
		 bfactory->tri2dof, bfactory->dof,
		 bfactory->dof2tri, bfactory->triangle);

  bfactory->integrate_nearfield = integrate_nearfield_dlp;
  bfactory->nearfield_entry = entry_dlp;
  bfactory->integrate_row_kernel = integrate_dlp;
  bfactory->integrate_row_lagrange = integrate_polynomial;
  bfactory->integrate_col_kernel = integrate_slp;
  bfactory->integrate_col_lagrange = integrate_polynomial_dn;
  bfactory->permutation = NULL;

  bfactory->mass[0] = 0.5;
  bfactory->lambda = lambda;

  return bfactory;
}

static void
permute_l(const int *dof, const int *lperm, int *pdof)
{
  pdof[0] = dof[lperm[0]];
  pdof[1] = dof[lperm[1]];
  pdof[2] = dof[lperm[2]];
}

psurfacebemfactory
new_surfacebemfactory_slp_ll(pbemgrid3d gr,
			     pcclustertree rct, pcclustertree cct,
			     int q, int p)
{
  psurfacebemfactory bfactory;
  int vertices = gr->vertices;
  int triangles = gr->triangles;
  int nqmax;
  int i, j;

  bfactory = common_initialization(gr, rct, cct, q, p, vertices, 3, 3);

  build_triangle_gauss_l(q, &bfactory->q, &bfactory->xq, &bfactory->wq);

  bfactory->q2 = new_quadrature2d_sauter_ll(q);

  nqmax = imax(imax(bfactory->q2->n_identical, bfactory->q2->n_edge),
	       imax(bfactory->q2->n_vertex, bfactory->q2->n_distant));

  bfactory->qbuf = (double *) malloc((size_t) sizeof(double) * nqmax);
  assert(bfactory->qbuf != NULL);

  for(i=0; i<triangles; i++) {
    bfactory->tri2dof[i] = 3*i;
    for(j=0; j<3; j++) {
      bfactory->dof[j+3*i] = rct->idx2dof[gr->t[i][j]];
      assert(0 <= bfactory->dof[j+3*i] && bfactory->dof[j+3*i] < vertices);
    }
  }
  bfactory->tri2dof[triangles] = 3*triangles;
  build_invgraph(triangles, vertices, 3*triangles,
		 bfactory->tri2dof, bfactory->dof,
		 bfactory->dof2tri, bfactory->triangle);

  bfactory->integrate_nearfield = integrate_nearfield_slp;
  bfactory->nearfield_entry = NULL;
  bfactory->integrate_row_kernel = integrate_slp;
  bfactory->integrate_row_lagrange = integrate_polynomial;
  bfactory->integrate_col_kernel = integrate_slp;
  bfactory->integrate_col_lagrange = integrate_polynomial;
  bfactory->permutation = permute_l;

  bfactory->mass[0] = 1.0 / 12.0;
  bfactory->mass[1] = 1.0 / 24.0;
  bfactory->mass[2] = 1.0 / 24.0;
  bfactory->mass[3] = 1.0 / 24.0;
  bfactory->mass[4] = 1.0 / 12.0;
  bfactory->mass[5] = 1.0 / 24.0;
  bfactory->mass[6] = 1.0 / 24.0;
  bfactory->mass[7] = 1.0 / 24.0;
  bfactory->mass[8] = 1.0 / 12.0;
  bfactory->lambda = 0.0;

  return bfactory;
}

psurfacebemfactory
new_surfacebemfactory_dlp_ll(pbemgrid3d gr,
			     pcclustertree rct, pcclustertree cct,
			     int q, int p, double lambda)
{
  psurfacebemfactory bfactory;
  int vertices = gr->vertices;
  int triangles = gr->triangles;
  int nqmax;
  int i, j;

  bfactory = common_initialization(gr, rct, cct, q, p, vertices, 3, 3);

  build_triangle_gauss_l(q, &bfactory->q, &bfactory->xq, &bfactory->wq);

  bfactory->q2 = new_quadrature2d_sauter_ll(q);

  nqmax = imax(imax(bfactory->q2->n_identical, bfactory->q2->n_edge),
	       imax(bfactory->q2->n_vertex, bfactory->q2->n_distant));

  bfactory->qbuf = (double *) malloc((size_t) sizeof(double) * nqmax);
  assert(bfactory->qbuf != NULL);

  for(i=0; i<triangles; i++) {
    bfactory->tri2dof[i] = 3*i;
    for(j=0; j<3; j++) {
      bfactory->dof[j+3*i] = rct->idx2dof[gr->t[i][j]];
      assert(0 <= bfactory->dof[j+3*i] && bfactory->dof[j+3*i] < vertices);
    }
  }
  bfactory->tri2dof[triangles] = 3*triangles;
  build_invgraph(triangles, vertices, 3*triangles,
		 bfactory->tri2dof, bfactory->dof,
		 bfactory->dof2tri, bfactory->triangle);

  bfactory->integrate_nearfield = integrate_nearfield_dlp;
  bfactory->nearfield_entry = NULL;
  bfactory->integrate_row_kernel = integrate_dlp;
  bfactory->integrate_row_lagrange = integrate_polynomial;
  bfactory->integrate_col_kernel = integrate_slp;
  bfactory->integrate_col_lagrange = integrate_polynomial_dn;
  bfactory->permutation = permute_l;

  bfactory->mass[0] = 1.0 / 12.0;
  bfactory->mass[1] = 1.0 / 24.0;
  bfactory->mass[2] = 1.0 / 24.0;
  bfactory->mass[3] = 1.0 / 24.0;
  bfactory->mass[4] = 1.0 / 12.0;
  bfactory->mass[5] = 1.0 / 24.0;
  bfactory->mass[6] = 1.0 / 24.0;
  bfactory->mass[7] = 1.0 / 24.0;
  bfactory->mass[8] = 1.0 / 12.0;
  bfactory->lambda = lambda;

  return bfactory;
}

void
del_surfacebemfactory(psurfacebemfactory bfactory)
{
  assert(bfactory != NULL);
  assert(bfactory->dof2tri != NULL);
  assert(bfactory->triangle != NULL);
  assert(bfactory->tri2dof != NULL);
  assert(bfactory->dof != NULL);
  assert(bfactory->xq != NULL);
  assert(bfactory->wq != NULL);
  assert(bfactory->q2 != NULL);
  assert(bfactory->qbuf != NULL);
  assert(bfactory->xp != NULL);
  assert(bfactory->l != NULL);
  assert(bfactory->l2 != NULL);
  assert(bfactory->arow != NULL);
  assert(bfactory->acol != NULL);

  free(bfactory->acol);
  free(bfactory->arow);
  free(bfactory->l2);
  free(bfactory->l);
  free(bfactory->xp);
  free(bfactory->qbuf);
  del_quadrature2d(bfactory->q2);
  free(bfactory->wq);
  free(bfactory->xq);
  free(bfactory->dof);
  free(bfactory->tri2dof);
  free(bfactory->triangle);
  free(bfactory->dof2tri);
  free(bfactory);
}

void
fill_surfacebem_supermatrix(psupermatrix M, pccluster row, pccluster col,
			    psurfacebemfactory bfactory)
{
  prkmatrix r = M->r;
  pfullmatrix f = M->f;
  int p3 = bfactory->p * bfactory->p * bfactory->p;
  int i, j;

  assert(M->u == NULL);

  if(r != NULL) {
    assert(r->rows == row->size);
    assert(r->cols == col->size);

    reallocstructure_rkmatrix(r, p3);
    r->kt = p3;
    
    if(diameter_cluster(row) < diameter_cluster(col)) {
      bfactory->integrate_row_lagrange(row, bfactory, r->a, r->rows);
      bfactory->integrate_row_kernel(col, row, bfactory, r->b, r->cols);
    }
    else {
      bfactory->integrate_col_kernel(row, col, bfactory, r->a, r->rows);
      bfactory->integrate_col_lagrange(col, bfactory, r->b, r->cols);
    }
  }
  else {
    if(f != NULL)
      bfactory->integrate_nearfield(row, col, bfactory, f->e, f->rows);
    else
      if(row->sons == 0) {
	if(col->sons == 0)
	  fill_surfacebem_supermatrix(M->s[0],
				      row, col,
				      bfactory);
	else
	  for(j=0; j<col->sons; j++)
	    fill_surfacebem_supermatrix(M->s[j],
					row, col->son[j],
					bfactory);
      }
      else {
	if(col->sons == 0)
	  for(i=0; i<row->sons; i++)
	    fill_surfacebem_supermatrix(M->s[i],
					row->son[i], col,
					bfactory);
	else
	  for(i=0; i<col->sons; i++)
	    for(j=0; j<row->sons; j++)
	      fill_surfacebem_supermatrix(M->s[i+j*row->sons],
					  row->son[i], col->son[j],
					  bfactory);
      }
  }
}

void
acafill_surfacebem_supermatrix(psupermatrix M, pccluster row, pccluster col,
			       psurfacebemfactory bfactory,
			       double eps, int k)
{
  prkmatrix r = M->r;
  pfullmatrix f = M->f;
  int i, j;

  assert(M->u == NULL);

  if(r != NULL) {
    assert(r->rows == row->size);
    assert(r->cols == col->size);

    reallocstructure_rkmatrix(r, k);
    aca2_rkmatrix(bfactory->nearfield_entry, bfactory,
		  eps, row->start, col->start, r, HLIB_ACA_DEFAULT);
  }
  else {
    if(f != NULL)
      bfactory->integrate_nearfield(row, col, bfactory, f->e, f->rows);
    else
      if(row->sons == 0) {
	if(col->sons == 0)
	  acafill_surfacebem_supermatrix(M->s[0], row, col,
					 bfactory, eps, k);
	else
	  for(j=0; j<col->sons; j++)
	    acafill_surfacebem_supermatrix(M->s[j], row, col->son[j],
					   bfactory, eps, k);
      }
      else {
	if(col->sons == 0)
	  for(i=0; i<row->sons; i++)
	    acafill_surfacebem_supermatrix(M->s[i], row->son[i], col,
					   bfactory, eps, k);
	else
	  for(i=0; i<col->sons; i++)
	    for(j=0; j<row->sons; j++)
	      acafill_surfacebem_supermatrix(M->s[i+j*row->sons], row->son[i], col->son[j],
					     bfactory, eps, k);
      }
  }
}

void
bufacafill_surfacebem_supermatrix(psupermatrix M, pccluster row, pccluster col,
				  psurfacebemfactory bfactory,
				  double eps, int kmax,
				  double **buf, int *bufsize)
{
  prkmatrix r = M->r;
  pfullmatrix f = M->f;
  int i, j;

  assert(M->u == NULL);

  if(r != NULL) {
    assert(r->rows == row->size);
    assert(r->cols == col->size);

    bufaca2_rkmatrix(bfactory->nearfield_entry, bfactory,
		     eps, kmax, row->start, col->start, buf, bufsize,
		     r, HLIB_ACA_DEFAULT);
  }
  else {
    if(f != NULL)
      bfactory->integrate_nearfield(row, col, bfactory, f->e, f->rows);
    else
      if(row->sons == 0) {
	if(col->sons == 0)
	  bufacafill_surfacebem_supermatrix(M->s[0],
					    row, col,
					    bfactory, eps, kmax,
					    buf, bufsize);
	else
	  for(j=0; j<col->sons; j++)
	    bufacafill_surfacebem_supermatrix(M->s[j],
					      row, col->son[j],
					      bfactory, eps, kmax,
					      buf, bufsize);
      }
      else {
	if(col->sons == 0)
	  for(i=0; i<row->sons; i++)
	    bufacafill_surfacebem_supermatrix(M->s[i],
					      row->son[i], col,
					      bfactory, eps, kmax,
					      buf, bufsize);
	else
	  for(i=0; i<col->sons; i++)
	    for(j=0; j<row->sons; j++)
	      bufacafill_surfacebem_supermatrix(M->s[i+j*row->sons],
						row->son[i], col->son[j],
						bfactory, eps, kmax,
						buf, bufsize);
      }
  }
}

pclusterbasis
build_surfacebem_rowbasis(pccluster t, psurfacebemfactory bfactory)
{
  pclusterbasis cb;
  int p = bfactory->p;
  int s;

  cb = new_clusterbasis(t, 0, 0);
  if(t->sons > 0) {
    for(s=0; s<t->sons; s++)
      cb->son[s] = build_surfacebem_rowbasis(t->son[s], bfactory);

    reallocstructure_clusterbasis(cb, p*p*p);

    for(s=0; s<t->sons; s++) {
      assert(cb->T[s] != NULL);
      bfactory->transfer_row(t->son[s], t, bfactory, cb->T[s], p*p*p);
    }

    cb->kt = p*p*p;
  }
  else {
    reallocstructure_clusterbasis(cb, p*p*p);

    assert(cb->V != NULL);
    bfactory->integrate_row_lagrange(t, bfactory, cb->V, t->size);

    cb->kt = p*p*p;
  }

  return cb;
}

pclusterbasis
build_surfacebem_colbasis(pccluster t, psurfacebemfactory bfactory)
{
  pclusterbasis cb;
  int p = bfactory->p;
  int s;

  cb = new_clusterbasis(t, 0, 0);
  if(t->sons > 0) {
    for(s=0; s<t->sons; s++)
      cb->son[s] = build_surfacebem_colbasis(t->son[s], bfactory);

    reallocstructure_clusterbasis(cb, p*p*p);

    for(s=0; s<t->sons; s++) {
      assert(cb->T[s] != NULL);
      bfactory->transfer_col(t->son[s], t, bfactory, cb->T[s], p*p*p);
    }

    cb->kt = p*p*p;
  }
  else {
    reallocstructure_clusterbasis(cb, p*p*p);

    assert(cb->V != NULL);
    bfactory->integrate_col_lagrange(t, bfactory, cb->V, t->size);

    cb->kt = p*p*p;
  }

  return cb;
}

void
fill_surfacebem_supermatrix2(psupermatrix M, pccluster row, pccluster col,
			     psurfacebemfactory bfactory)
{
  puniformmatrix u = M->u;
  pfullmatrix f = M->f;
  int i, j;

  assert(M->r == NULL);
  assert((!M->row) || M->row->t == row);
  assert((!M->col) || M->col->t == col);

  if(u != NULL) {
    assert(u->row->t == row);
    assert(u->col->t == col);

    bfactory->kernel_coefficients(u->row->t, u->col->t, bfactory, u->S, u->row->kt);

    u->ktr = u->row->kt;
    u->ktc = u->col->kt;
  }
  else if(f != NULL)
    bfactory->integrate_nearfield(row, col, bfactory, f->e, f->rows);
  else
    if(row->sons > 0)
      if(col->sons > 0) 
	for(i=0; i<row->sons; i++)
	  for(j=0; j<col->sons; j++)
	    fill_surfacebem_supermatrix2(M->s[i+j*row->sons],
					 row->son[i], col->son[j],
					 bfactory);
      else
	for(i=0; i<row->sons; i++)
	  fill_surfacebem_supermatrix2(M->s[i],
				       row->son[i], col,
				       bfactory);
    else
      if(col->sons > 0)
	for(j=0; j<col->sons; j++)
	  fill_surfacebem_supermatrix2(M->s[j],
				       row, col->son[j],
				       bfactory);
}

void
buildrhs_surfacebem(double *b,
		    double (*rhs)(const double *x, const double *n),
		    psurfacebemfactory bfactory)
{
  const int *index = bfactory->rct->dof2idx;
  const double (*x)[3] = (const double (*)[3]) bfactory->gr->x;
  const int (*t)[3] = (const int (*)[3]) bfactory->gr->t;
  const double (*n)[3] = (const double (*)[3]) bfactory->gr->n;
  const double *gdet = bfactory->gr->gdet;
  const double (*xq)[2] = (const double (*)[2]) bfactory->xq;
  const double *wq = bfactory->wq;
  int q = bfactory->q;
  int triangles = bfactory->gr->triangles;
  double xx[3], nn[3];
  double sum, a0, a1, a2;
  int i, j, ii;

  for(i=0; i<triangles; i++) {
    ii = index[i];

    nn[0] = n[ii][0];
    nn[1] = n[ii][1];
    nn[2] = n[ii][2];

    sum = 0.0;
    for(j=0; j<q; j++) {
      a2 = xq[j][1];
      a1 = xq[j][0];
      a0 = 1.0 - a1 - a2;

      xx[0] = a0 * x[t[ii][0]][0] + a1 * x[t[ii][1]][0] + a2 * x[t[ii][2]][0];
      xx[1] = a0 * x[t[ii][0]][1] + a1 * x[t[ii][1]][1] + a2 * x[t[ii][2]][1];
      xx[2] = a0 * x[t[ii][0]][2] + a1 * x[t[ii][1]][2] + a2 * x[t[ii][2]][2];
      
      sum += wq[j] * rhs(xx, nn);
    }
    b[i] = sum * gdet[ii];
  }
}

void
buildapprox_surfacebem(double *u,
		       double (*func)(const double *x, const double *n),
		       psurfacebemfactory bfactory)
{
  const int *index = bfactory->rct->dof2idx;
  const double (*x)[3] = (const double (*)[3]) bfactory->gr->x;
  const int (*t)[3] = (const int (*)[3]) bfactory->gr->t;
  const double (*n)[3] = (const double (*)[3]) bfactory->gr->n;
  const double (*xq)[2] = (const double (*)[2]) bfactory->xq;
  const double *wq = bfactory->wq;
  int q = bfactory->q;
  int triangles = bfactory->gr->triangles;
  double xx[3], nn[3];
  double sum, a0, a1, a2;
  int i, j, ii;

  for(i=0; i<triangles; i++) {
    ii = index[i];

    nn[0] = n[ii][0];
    nn[1] = n[ii][1];
    nn[2] = n[ii][2];

    sum = 0.0;
    for(j=0; j<q; j++) {
      a2 = xq[j][1];
      a1 = xq[j][0];
      a0 = 1.0 - a1 - a2;

      xx[0] = a0 * x[t[ii][0]][0] + a1 * x[t[ii][1]][0] + a2 * x[t[ii][2]][0];
      xx[1] = a0 * x[t[ii][0]][1] + a1 * x[t[ii][1]][1] + a2 * x[t[ii][2]][1];
      xx[2] = a0 * x[t[ii][0]][2] + a1 * x[t[ii][1]][2] + a2 * x[t[ii][2]][2];
      
      sum += wq[j] * func(xx, nn);
    }
    u[i] = sum * 2.0;
  }
}

double
l2norm_surfacebem(double *u,
		  psurfacebemfactory bfactory)
{
  const int *index = bfactory->rct->dof2idx;
  const double *gdet = bfactory->gr->gdet;
  int triangles = bfactory->gr->triangles;
  double sum;
  int i, ii;

  sum = 0.0;
  for(i=0; i<triangles; i++) {
    ii = index[i];

    sum += gdet[ii] * dsqr(u[ii]);
  }

  return sqrt(sum);
}

double
l2diff_surfacebem(double *u,
		  double (*func)(const double *x, const double *n),
		  psurfacebemfactory bfactory)
{
  const int *index = bfactory->rct->dof2idx;
  const double (*x)[3] = (const double (*)[3]) bfactory->gr->x;
  const int (*t)[3] = (const int (*)[3]) bfactory->gr->t;
  const double (*n)[3] = (const double (*)[3]) bfactory->gr->n;
  const double *gdet = bfactory->gr->gdet;
  const double (*xq)[2] = (const double (*)[2]) bfactory->xq;
  const double *wq = bfactory->wq;
  int q = bfactory->q;
  int triangles = bfactory->gr->triangles;
  double xx[3], nn[3];
  double norm, sum, a0, a1, a2;
  int i, j, ii;

  norm = 0.0;
  for(i=0; i<triangles; i++) {
    ii = index[i];

    nn[0] = n[ii][0];
    nn[1] = n[ii][1];
    nn[2] = n[ii][2];

    sum = 0.0;
    for(j=0; j<q; j++) {
      a2 = xq[j][1];
      a1 = xq[j][0];
      a0 = 1.0 - a1 - a2;

      xx[0] = a0 * x[t[ii][0]][0] + a1 * x[t[ii][1]][0] + a2 * x[t[ii][2]][0];
      xx[1] = a0 * x[t[ii][0]][1] + a1 * x[t[ii][1]][1] + a2 * x[t[ii][2]][1];
      xx[2] = a0 * x[t[ii][0]][2] + a1 * x[t[ii][1]][2] + a2 * x[t[ii][2]][2];
      
      sum += wq[j] * dsqr(func(xx, nn) - u[i]);
    }
    norm += gdet[ii] * sum;
  }

  return sqrt(norm);
}

static void
h2r_buildN(pccluster row, pccluster col, void *data, double *N, int ldN)
{
  psurfacebemfactory bfactory = (psurfacebemfactory) data;

  bfactory->integrate_nearfield(row, col, bfactory, N, ldN);
}

static void
h2r_buildS(pccluster row, pccluster col, void *data, double *S, int ldS)
{
  psurfacebemfactory bfactory = (psurfacebemfactory) data;

  bfactory->kernel_coefficients(row, col, bfactory, S, ldS);
}

static int
h2r_getrankrow(pccluster row, void *data)
{
  psurfacebemfactory bfactory = (psurfacebemfactory) data;

  return bfactory->p * bfactory->p * bfactory->p;
}

static void
h2r_buildVrow(pccluster tau, void *data, double *V, int ldV)
{
  psurfacebemfactory bfactory = (psurfacebemfactory) data;

  bfactory->integrate_row_lagrange(tau, bfactory, V, ldV);
}

static void
h2r_buildTrow(pccluster trg, pccluster src, void *data, double *T, int ldT)
{
  psurfacebemfactory bfactory = (psurfacebemfactory) data;

  bfactory->transfer_row(trg, src, bfactory, T, ldT);
}

static int
h2r_getrankcol(pccluster row, void *data)
{
  psurfacebemfactory bfactory = (psurfacebemfactory) data;

  return bfactory->p * bfactory->p * bfactory->p;
}

static void
h2r_buildVcol(pccluster tau, void *data, double *V, int ldV)
{
  psurfacebemfactory bfactory = (psurfacebemfactory) data;

  bfactory->integrate_col_lagrange(tau, bfactory, V, ldV);
}

static void
h2r_buildTcol(pccluster trg, pccluster src, void *data, double *T, int ldT)
{
  psurfacebemfactory bfactory = (psurfacebemfactory) data;

  bfactory->transfer_col(trg, src, bfactory, T, ldT);
}

ph2recompression
newrecompression_surfacebem(psurfacebemfactory bfactory,
			    double eps, int kmax, TruncationStrategy strategy)
{
  ph2recompression h2r;

  h2r = new_h2recompression();

  h2r->buildN = h2r_buildN;
  h2r->buildS = h2r_buildS;
  h2r->getrankrow = h2r_getrankrow;
  h2r->buildVrow = h2r_buildVrow;
  h2r->buildTrow = h2r_buildTrow;
  h2r->getrankcol = h2r_getrankcol;
  h2r->buildVcol = h2r_buildVcol;
  h2r->buildTcol = h2r_buildTcol;
  h2r->data = bfactory;

  h2r->eps = eps;
  h2r->kmax = kmax;
  h2r->strategy = strategy;

  return h2r;
}
