
/* ------------------------------------------------------------
   This is the file "surfacebem.h" of the HLib package.

   Copyright 1999 Steffen Boerm and Lars Grasedyck
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#ifndef SurfaceBEM
#define SurfaceBEM

typedef struct _surfacebemfactory surfacebemfactory;
typedef surfacebemfactory *psurfacebemfactory;
typedef const surfacebemfactory *pcsurfacebemfactory;

#include "basic.h"
#include "bem3d.h"
#include "cluster.h"
#include "clusterbasis.h"
#include "h2virtual.h"
#include "h2conversion.h"
#include "quadrature2d.h"
#include "supermatrix.h"

/* ============================================================
   A surfacebemfactory contains all the information required for the
   computation of an H- oder H^2-matrix approximation of the discretized
   single layer potential on a curve in 3D: An array of points describing
   the polygonal approximation of the curve, quadrature rules for singular
   and regular integrals as well as interpolation points used for the
   approximation of the kernel function.
   ============================================================ */

struct _surfacebemfactory {
  pbemgrid3d gr;		/* BEM grid */
  pcclustertree rct;		/* Row cluster tree */
  pcclustertree cct;		/* Column cluster tree */

  int *dof2tri;			/* List of triangles per degree of freedom */
  int *triangle;

  int *tri2dof;			/* List of degrees of freedom per triangle */
  int *dof;

  double (*xq)[2];		/* Quadrature rule for regular integral */
  double *wq;
  int q;

  int *row_dof;			/* Local degrees of freedom for row */
  int *col_dof;			/* Local degrees of freedom for column */
  int ld_rows;			/* Number of row basis functions */
  int ld_cols;			/* Number of column basis functions */

  pquadrature2d q2;		/* Quadrature rules for singular integrals */
  double *mass;			/* Element mass matrix */

  double *qbuf;			/* Quadrature buffer */

  double *xp;			/* Points for Chebyshev interpolation */
  int p;			/* Number of interpolation points */

  double lambda;		/* Mass matrix scaling in double layer potential */

  void (*integrate_nearfield)(pccluster row, pccluster col,
			      psurfacebemfactory bfactory, double *X, int ldX);
  double (*nearfield_entry)(int row, int col, void *data);
  void (*kernel_coefficients)(pccluster row, pccluster col,
			      psurfacebemfactory bfactory, double *S, int ldS);
  void (*integrate_row_kernel)(pccluster tau, pccluster sigma,
			       psurfacebemfactory bfactory, double *X, int ldX);
  void (*integrate_row_lagrange)(pccluster tau,
				 psurfacebemfactory bfactory, double *X, int ldX);
  void (*transfer_row)(pccluster trg, pccluster src,
		       psurfacebemfactory bfactory, double *T, int ldT);
  void (*integrate_col_kernel)(pccluster tau, pccluster sigma,
			       psurfacebemfactory bfactory, double *X, int ldX);
  void (*integrate_col_lagrange)(pccluster tau,
				 psurfacebemfactory bfactory, double *X, int ldX);
  void (*transfer_col)(pccluster trg, pccluster src,
		       psurfacebemfactory bfactory, double *T, int ldT);
  void (*permutation)(const int *dof, const int *lperm, int *pdof);

  double (*l)[3];		/* Transformed 3D interpolation points */
  double (*l2)[3];

  int *arow;			/* List of active triangles for a row cluster */
  int *acol;			/* List of active triangles for a column cluster */
};

/* Create a surfacebemfactory object
   gr: Surface grid
   rct: Row cluster tree
   cct: Column cluster tree
   q: Order of the quadrature
   p: Order of the polynomial approximation */
psurfacebemfactory
new_surfacebemfactory_slp(pbemgrid3d gr,
			  pcclustertree rct, pcclustertree cct,
			  int q, int p);

psurfacebemfactory
new_surfacebemfactory_dlp(pbemgrid3d gr,
			  pcclustertree rct, pcclustertree cct,
			  int q, int p, double lambda);

psurfacebemfactory
new_surfacebemfactory_slp_ll(pbemgrid3d gr,
			     pcclustertree rct, pcclustertree cct,
			     int q, int p);

psurfacebemfactory
new_surfacebemfactory_dlp_ll(pbemgrid3d gr,
			     pcclustertree rct, pcclustertree cct,
			     int q, int p, double lambda);

/* Free the memory used by a curvebemfactory object */
void
del_surfacebemfactory(psurfacebemfactory bfactory);

/* Compute the entries of an H-matrix approximation by interpolation
   M: H-matrix
   row: Cluster corresponding to the row of the matrix
   col: Cluster corresponding to the column of the matrix
   bfactory: Factory */
void
fill_surfacebem_supermatrix(psupermatrix M, pccluster row, pccluster col,
			    psurfacebemfactory bfactory);

/* Compute the entries of an H-matrix adaptive cross approximation
   M: H-Matrix
   row: Cluster corresponding to the row of the matrix
   col: Cluster corresponding to the column of the matrix
   bfactory: Factory
   eps: Precision*/
void
acafill_surfacebem_supermatrix(psupermatrix M, pccluster row, pccluster col,
			       psurfacebemfactory bfactory, double eps, int k);

/* Compute the entries of an H-matrix adaptive cross approximation
   M: H-Matrix
   row: Cluster corresponding to the row of the matrix
   col: Cluster corresponding to the column of the matrix
   bfactory: Factory
   eps: Precision
   kmax: Maximal rank
   buf: Pointer to pointer to buffer
   bufsize: Pointer to size of buffer */
void
bufacafill_surfacebem_supermatrix(psupermatrix M, pccluster row, pccluster col,
				  psurfacebemfactory bfactory,
				  double eps, int kmax,
				  double **buf, int *bufsize);

/* Create a nested row cluster basis by integrating Lagrange polynomials
   t: Root cluster
   bfactory: Factory */
pclusterbasis
build_surfacebem_rowbasis(pccluster t, psurfacebemfactory bfactory);

/* Create a nested column cluster basis by integrating Lagrange polynomials
   t: Root cluster
   bfactory: Factory */
pclusterbasis
build_surfacebem_colbasis(pccluster t, psurfacebemfactory bfactory);

/* Compute the entries of an H^2-matrix approximation of the
   single layer potential
   M: H^2-matrix
   row: Cluster corresponding to the row of the matrix
   col: Cluster corresponding to the column of the matrix
   bfactory: Factory */
void
fill_surfacebem_supermatrix2(psupermatrix M, pccluster row, pccluster col,
			     psurfacebemfactory bfactory);

/* Discretize the right hand side of the integral equation
   FOR PIECEWISE CONSTANT BASIS FUNCTIONS ONLY
   b: Output vector, of size bfactory->gr->triangles
   rhs: Right hand side, depending on position and normal vector
   bfactory: Factory */
void
buildrhs_surfacebem(double *b,
		    double (*rhs)(const double *x, const double *n),
		    psurfacebemfactory bfactory);

/* Approximate a function in the trial space
   FOR PIECEWISE CONSTANT BASIS FUNCTIONS ONLY
   u: Output vector, of size bfactory->gr->triangles
   func: Function, depending on position and normal vector
   bfactory: Factory */
void
buildapprox_surfacebem(double *u,
		       double (*func)(const double *x, const double *n),
		       psurfacebemfactory bfactory);

/* Compute L^2 norm of a function in the trial space
   FOR PIECEWISE CONSTANT BASIS FUNCTIONS ONLY
   u: Output vector, of size bfactory->gr->triangles
   bfactory: Factory */
double
l2norm_surfacebem(double *u,
		  psurfacebemfactory bfactory);

/* Compute L^2 norm of the difference of a function and a discrete function
   FOR PIECEWISE CONSTANT BASIS FUNCTIONS ONLY
   u: Coefficients of the discrete function
   func: Function, depending on position and normal vector
   bfactory: Factory */
double
l2diff_surfacebem(double *u,
		  double (*func)(const double *x, const double *n),
		  psurfacebemfactory bfactory);

/* Create an h2recompression object that uses a surfacebemfactory in
   order to perform a virtual H^2 recompression
   bfactory: Factory
   eps: Tolerance for H^2 recompression
   kmax: Maximal rank of adaptive cluster bases
   strategy: Truncation strategy, see h2conversion.h */
ph2recompression
newrecompression_surfacebem(psurfacebemfactory bfactory,
			    double eps, int kmax, TruncationStrategy strategy);

#endif
