
/* ------------------------------------------------------------
   This is the file "uniformmatrix.h" of the HLib package.

   Copyright 1999 Steffen Boerm and Lars Grasedyck
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#ifndef UniformMatrix
#define UniformMatrix

typedef struct _uniformmatrix uniformmatrix;
typedef uniformmatrix *puniformmatrix;
typedef const uniformmatrix *pcuniformmatrix;

#include "basic.h"
#include "blockcluster.h"
#include "clusterbasis.h"
#include "rkmatrix.h"

#include <stdio.h>

struct _uniformmatrix {
  pclusterbasis row;		/* Cluster basis for the row */
  pclusterbasis col;		/* Cluster basis for the column */

  int rows;			/* Number of matrix rows */
  int cols;			/* Number of matrix columns */
  
  int kr;			/* Maximal rank of row basis */
  int kc;			/* Maximal rank of column basis */
  int ktr;			/* Current rank of row basis */
  int ktc;			/* Current rank of column basis */

  double *S;			/* Coefficient matrix, ktr x ktc */
};

/* A uniformmatrix stores a matrix in the form V S W^T, where V and W
   are given by clusterbasis objects, i.e., V depends only on the
   row cluster and W depends only on the column cluster. Due to this
   uniformity, we can split the matrix-vector multiplication into
   three steps: First xt = W^T x is computed, this is called the
   forward transformation, then we perform the multiplication yt = S xt
   by the coefficent matrix S, and finally we apply the backward
   transformation y += V yt in order to find the result.
   If a number of uniformmatrix objects is contained in a supermatrix
   hierarchy, the forward and backward transformations have to be
   performed only once for each cluster, not for each block as in the
   case of rkmatrix objects. This leads to improved performance. */

/* Create a uniformmatrix object
   row: Corresponding row cluster basis
   col: Corresponding column cluster basis */
puniformmatrix
new_uniformmatrix(pclusterbasis row, pclusterbasis col);

/* Free the memory used by a uniformmatrix object */
void
del_uniformmatrix(puniformmatrix u);

/* Return the size of a uniformmatrix (without basis) */
size_t
getsize_uniformmatrix(puniformmatrix u);

/* Change uniformmatrix to match new row and column basis */
void
reallocstructure_uniformmatrix(puniformmatrix u,
			       pclusterbasis row, pclusterbasis col);

/* Create a new uniformmatrix object that is an exact copy of the
   original one */
puniformmatrix
copy_uniformmatrix(pcuniformmatrix u);

/* Create a new uniformmatrix object that has the same structure as
   the original one, but whose S matrix is not yet initialized */
puniformmatrix
copystructure_uniformmatrix(pcuniformmatrix u);

/* Copy the matrix S from a given uniformmatrix object to another one
   with the correct structure
   usource: Source uniformmatrix object
   utarget: Target uniformmatrix object, only S will be overwritten */
void
copycontent_uniformmatrix(pcuniformmatrix usource, puniformmatrix utarget);

/* Switch a uniformmatrix to different cluster bases
   u: Original uniformmatrix
   rb: New row basis
   rproj: Row basis projection
   cb: New column basis
   cproj: Column basis projection */
puniformmatrix
project_uniformmatrix(pcuniformmatrix u,
		      pclusterbasis rb, pcclusteroperator rproj,
		      pclusterbasis cb, pcclusteroperator cproj);

/* Create a PostScript(TM) output of a uniformmatrix object */
void
outputleaf_uniformmatrix(pcuniformmatrix u, FILE *file, pblockcluster bc,
			 int row_offset, int col_offset,
			 int row_max, int col_max, 
			 int rank, int ufill, int border, double tol);

/* Convert a uniformmatrix object to an rkmatrix object.
   u: Source uniformmatrix
   r: Target rkmatrix */
void
convertuniform2_rkmatrix(puniformmatrix u, prkmatrix r);

/* Convert an rkmatrix to a uniformmatrix by projecting onto the
   associated cluster bases. WORKS ONLY FOR ORTHOGONAL CLUSTER BASES.
   r: Source rkmatrix
   u: Target uniformmatrix */
void
convertrk2_uniformmatrix(prkmatrix r, puniformmatrix u);

/* Approximate a uniformmatrix by an rkmatrix by performing an SVD
   of the coefficient matrix. WORKS ONLY FOR ORTHOGONAL CLUSTER BASES.
   u: Original uniformmatrix, u->row and u->col have to be orthogonal
   r: Target rkmatrix
   eps: Desired precision
   kmax: Upper bound for the rank
   strategy: Truncation strategy */
void
approxuniform2_rkmatrix(puniformmatrix u, prkmatrix r,
			double eps, int kmax, TruncationStrategy strategy);

/* Multiply a uniformmatrix by a vector v and write the result to w. */
void 
eval_uniformmatrix(puniformmatrix u, const double *v, double *w);

/* Multiply a uniformmatrix by a vector v and add the result to w. */
void 
addeval_uniformmatrix(puniformmatrix u, const double *v, double *w);

/* Multiply the transposed of a uniformmatrix by a vector v and write
   the result to w. */
void 
evaltrans_uniformmatrix(puniformmatrix u, const double *v, double *w);

/* Multiply the transposed of a uniformmatrix by a vector v and add
   the result to w. */
void 
addevaltrans_uniformmatrix(puniformmatrix u, const double *v, double *w);

/* Multiply the coefficient vector xt by S and write the result to yt */
void 
fasteval_uniformmatrix(puniformmatrix u);

/* Multiply the coefficient vector xt by S and add the result to yt */
void 
fastaddeval_uniformmatrix(puniformmatrix u);

/* Multiply the coefficient vector xt by alpha*S and add the result to yt */
void
fastaddscaleeval_uniformmatrix(puniformmatrix u, double alpha);

/* Multiply the coefficient vector xt by S^T and write the result to yt */
void 
fastevaltrans_uniformmatrix(puniformmatrix u);

/* Multiply the coefficient vector xt by S^T and add the result to yt */
void 
fastaddevaltrans_uniformmatrix(puniformmatrix u);

/* Compute an entry of a uniformmatrix object */
double
getentry_uniformmatrix(puniformmatrix u, int row, int col);

/* Return the exact product a*b in rkmatrix representation,
   where at least one of the factors is a uniformmatrix */
prkmatrix
getmul_uniformmatrix(psupermatrix a, psupermatrix b);

#endif
